/**
 * A an object for drawing svg/vml graphics:
 *
 * @class MapChat.Canvas
 *
 * Depends:
 *	kamap/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 */

MapChat.Canvas = function(options)
{
	var t = this,
		map = (t.map = options.map),
		name = (t.name = "MapChat.Canvas"),
		zIndex = (t.zIndex = options.zIndex||300),
		canvas = (t.canvas = null),
		svg = (t.svg = F),
		geoms = (t.geoms = []),
		active = (t.active = F),
		mapCanvas = (t.mapCanvas = $(t.map.createDrawingCanvas(t.zIndex)));
	
	kaTool.apply( t, [map] );
	
	t.bInfoTool = T;
	
	/** get all the prototype functions from the kaTool object */
	for (var p in kaTool.prototype) {
		if (!MapChat.Canvas.prototype[p])
			MapChat.Canvas.prototype[p]= kaTool.prototype[p];
	}
	
	// Initialize the vector canvas:
	if (document.createElementNS)
	{
		t.initSVG();
	}
	else if(document.createStyleSheet)
	{
		t.initVML();
	}
	
	t.canvas.onkepress = map.onkeypress;
	t.canvas.onmousedown = function(){return F;}; // prevents svg canvas from being highlighted in FF3
	
	return t;
}

/**
 * activate this tool.  Activating the tool causes any existing tools to be
 * deactivated.
 */
MapChat.Canvas.prototype.activate = function()
{
	var t = this;
	
	if (t.active) return;
	
	if (t.svg)
	{
		// FF needs the svg canvas to have a viewBox applied in order to display vectors outside the bounds
		// of the DIV tag that it is created within:
		t.canvas.setAttributeNS(null, "viewBox", (-1*t.mapCanvas.offset().left-500)+" "+(-1*t.mapCanvas.offset().top-500)+" "+
			parseInt(t.mapCanvas.css("width"))+" "+parseInt(t.mapCanvas.css("height")));
		$(t.canvas).css({left:-1*t.mapCanvas.offset().left-500,top:-1*t.mapCanvas.offset().top-500});
	}
	
	// Shapes need to be redrawn if the scale is changed...
	t.map.registerForEvent( KAMAP_SCALE_CHANGED, t, t.scaleChanged );
	
	// The SVG canvas needs to be shifted if extents are changes.
	if (t.svg) t.map.registerForEvent( KAMAP_EXTENTS_CHANGED, t, t.extentsChanged );
	
	// Redraw geometries, since this canvas may have been deactivated and any existing geometries
	// on the canvas may not be in sync with the map
	t.scaleChanged();
	
	t.canvas.style.visibility = 'visible';
	
	kaTool.prototype.activate.apply(t);
	
	t.active = T;
	
};

/**
 * deactivate this tool.
 */
MapChat.Canvas.prototype.deactivate = function()
{
	var t = this;
	
	if (!t.active) return;
	
	t.canvas.style.visibility = 'hidden';
	
	t.map.deregisterForEvent( KAMAP_SCALE_CHANGED, t, t.scaleChanged );
	
	if (t.svg) t.map.deregisterForEvent( KAMAP_EXTENTS_CHANGED, t, t.extentsChanged );
	
	kaTool.prototype.deactivate.apply(t);
	
	t.active = F;
};

/**
 * Initialize a canvas using SVG
 */
MapChat.Canvas.prototype.initSVG = function()
{
	var t = this;
	t.svg = T;
	t.svgNS = "http://www.w3.org/2000/svg";
	t.def = document.createElementNS(t.svgNS, "def");
	t.canvas = document.createElementNS(t.svgNS, "svg");
	t.canvas.setAttribute("width",parseInt($(t.mapCanvas).css("width")));
	t.canvas.setAttribute("height",parseInt($(t.mapCanvas).css("height")));
	t.canvas.style.position="absolute";
	t.canvas.style.overflow="visible";
	t.mapCanvas[0].appendChild(t.def);
	t.mapCanvas[0].appendChild(t.canvas);
}

/**
 * Initialize a canvas using VML
 */
MapChat.Canvas.prototype.initVML = function()
{
	if(!document.namespaces["g_vml_"])
	{
		document.namespaces.add("g_vml_", "urn:schemas-microsoft-com:vml");
		var style = document.createStyleSheet();
		style.addRule('g_vml_\\:*', "behavior: url(#default#VML); position: absolute; display: inline-block;");
	}
	this.canvas = this.mapCanvas[0];
}

/**
 * MapChat.Canvas.scaleChanged()
 *
 * When the scale changes, shapes need to be redrawn at the correct scale.
 *
 */
MapChat.Canvas.prototype.scaleChanged = function()
{
	this.extentsChanged();
	
	$.each(this.geoms,function(){
		this.draw();
	});
}

/**
 * MapChat.Canvas.extentsChanged()
 *
 * update the viewBox for svg canvases:
 */
MapChat.Canvas.prototype.extentsChanged = function()
{
	var t = this;
	if (t.svg)
	{
		t.canvas.setAttributeNS(null, "viewBox", (-1*t.mapCanvas.offset().left-500)+" "+(-1*t.mapCanvas.offset().top-500)+" "+
			parseInt(t.mapCanvas.css("width"))+" "+parseInt(t.mapCanvas.css("height")));
		$(t.canvas).css({left:-1*t.mapCanvas.offset().left-500,top:-1*t.mapCanvas.offset().top-500});
	}
}

/**
 * MapChat.Canvas.addGeom(svg,options)
 *
 * Adds vector geometries based on SVG vector coordinates (just line/polygon paths and 
 * points/circles for now - assumes georeferenced coordinates)
 *
 * svg - string, the svg-formated vector coordinates (path or point)
 * options - object, determines the styling (e.g., fill/stroke colors, point radius, etc)
 */
MapChat.Canvas.prototype.addGeom = function(svg,options)
{
	var t = this, geom;
	svg = svg.toUpperCase();
	if (svg.indexOf("M")>=0)
	{
		geom = t.newGeom("path",options,(svg.indexOf("Z")>=0));
		var subPaths = svg.split(" Z").join("").split(" L").join("").split("M ");
		$.each(subPaths,function(subPath,v){
			if (v.length>0)
			{
				var coords = $.trim(v).split(" "), addPath = [];
				for (var i=0;i<coords.length;i=i+2)
				{
					coords[i+1] = coords[i+1].replace(/;/g,"");
					addPath.push([1.0*coords[i],-1.0*coords[i+1]]);
				}
				geom.xy.push(addPath);
			}
		});
	}
	else
	{
		geom = t.newGeom("point",options);
		var points = svg.split(",");
		$.each(points,function(p,coord){
			var cxy = coord.split(" ");
			cx = cxy[0].substring(cxy[0].indexOf('"')+1,cxy[0].length-1);
			cy = cxy[1].substring(cxy[1].indexOf('"')+1,cxy[1].length-1);
			geom.addPoint([1.0*cx,-1.0*cy]);
		});
	}
	
	t.geoms.push(geom);
	geom.draw();
	
	return geom;
}

/**
 * MapChat.Canvas.newGeom(geomtype,options)
 *
 * Creates an empty vector path or point geometry that draws on the current instance of the 
 * Canvas object.  It will not be drawn by default - the geom.xy property must be populated first,
 * then the geom.draw() method called to draw the geometry.
 *
 * geomtype - string, either 'path' or 'point' - the type of geometry to be drawn
 * options - object, determines the styling (e.g., fill/stroke colors, point radius, etc)
 * closePath - boolean, determines whether polylines will be closed as polygons
 */
MapChat.Canvas.prototype.newGeom = function(geomtype,options,closePath)
{
	var t = this, geom = {xy:[],type:geomtype,opts:{}};
	if (options) $.extend(geom.opts,options);
	if (geomtype=="path")
	{
		geom.type = "path";
		geom.closePath = !!closePath;
		
		if (t.svg)
		{
			geom.o = t.canvas.appendChild(document.createElementNS(t.svgNS,"path"));
		}
		else
		{
			geom.o = t.canvas.appendChild(document.createElement("g_vml_:shape"));
			geom.o.style.width=t.mapCanvas.css("width");
			geom.o.style.height=t.mapCanvas.css("height");
			geom.o.coordsize = parseInt(t.mapCanvas.css("width"))+","+parseInt(t.mapCanvas.css("height"));
			geom.o.f = geom.o.appendChild(document.createElement("g_vml_:fill"));
			geom.o.s = geom.o.appendChild(document.createElement("g_vml_:stroke"));
		}
		
		/**
		 * geom.addPoint(xy,g)
		 *
		 * Adds an xy coordinate to this object.
		 *
		 * xy - array, the geographic x/y coordinates of the point.
		 * g - integer, the index of the shape to add this to (e.g., if supporting multipolygons/multilines). Defaults to last geometry.
		 */
		geom.addPoint = function(xy,g)
		{
			if (geom.xy.length==0) geom.xy.push([]);
			if (g==null) g = geom.xy.length-1;
			if (!geom.xy[g]) geom.xy[g] = [];
			geom.xy[g].push(xy);
		}
		
		/**
		 * geom.addPart()
		 *
		 * Adds a new array to the xy coordinates, starting a new part in a multi-part geometry.
		 */
		geom.addPart = function()
		{
			if (geom.xy[geom.xy.length-1] && geom.xy[geom.xy.length-1].length>0) geom.xy.push([]);
		}
		
		/**
		 * geom.dropPart()
		 *
		 * Drops an array from the xy coordinate in a multi-part geometry.
		 */
		geom.dropPart = function()
		{
			if (geom.xy.length>1) geom.xy.pop();
			else geom.xy = [];
		}
		
		/**
		 * geom.dropPoint()
		 *
		 * Removes the last point in the geometry.  If the current part of a multipart geometry is empty, it is also dropped.
		 */
		geom.dropPoint = function()
		{
			if (geom.xy.length>0)
			{
				while (geom.xy.length>1 && geom.xy[geom.xy.length-1].length==0)
				{
					geom.dropPart();
				}
				if (geom.xy[geom.xy.length-1].length>0) return geom.xy[geom.xy.length-1].pop();
			}
		}
		
		/**
		 * geom.lastPoint()
		 *
		 * Returns the last xy coordinate.
		 */
		geom.lastPoint = function()
		{
			if (geom.xy.length==1)
				var curXY = geom.xy[0];
			else if (geom.xy.length>1)
				var curXY = (geom.xy[geom.xy.length-1].length==0?geom.xy[geom.xy.length-2]:geom.xy[geom.xy.length-1]);
			else return N;
			
			if (curXY.length>=1)
				return curXY[curXY.length-1];
		}
		
		/**
		 * geom.numPoints();
		 *
		 * Returns the number of points.
		 */
		geom.numPoints = function()
		{
			var p = 0;
			$.each(geom.xy,function(){
				p += this.length;
			});
			return p;
		}
		
		/**
		 * geom.allPoints();
		 *
		 * Returns the all the points in a single array:
		 */
		geom.allPoints = function()
		{
			var xy = [];
			$.each(geom.xy,function(){
				xy = xy.concat(this);
			});
			return xy;
		}
	}
	else
	{
		geom.type = "point";
		geom.o = [];
		
		/**
		 * geom.addPoint(xy)
		 *
		 * Adds an xy coordinate to a point object, and creates a corresponding vector object that will be used
		 * when geom.draw() is called.
		 *
		 * xy - array, the geographic x/y coordinates of the point.
		 */
		geom.addPoint = function(xy)
		{
			geom.xy.push(xy);
			if (t.svg)
			{
				newPt = t.canvas.appendChild(document.createElementNS(t.svgNS, "circle"));
			}
			else
			{
				newPt = t.canvas.appendChild(document.createElement("g_vml_:oval"));
				newPt.f = newPt.appendChild(document.createElement("g_vml_:fill"));
				newPt.s = newPt.appendChild(document.createElement("g_vml_:stroke"));
			}
			geom.setStyle(null,newPt);
			geom.o.push(newPt);
		}
		
		/**
		 * geom.dropPoint()
		 *
		 * Removes the last point in the geometry.
		 */
		geom.dropPoint = function()
		{
			if (geom.xy.length>0)
			{
				var o = geom.o.pop();
				$(o).remove();
				return geom.xy.pop();
			}
		}
		
		/**
		 * geom.lastPoint()
		 *
		 * Returns the last xy coordinate.
		 */
		geom.lastPoint = function()
		{
			if (geom.xy.length>=1)
				return geom.xy[geom.xy.length-1];
		}
		
		/**
		 * geom.numPoints();
		 *
		 * Returns the number of points.
		 */
		geom.numPoints = function()
		{
			return geom.xy.length;
		}
		
		/**
		 * geom.allPoints();
		 *
		 * Returns the all the points (just the xy attribute in this case):
		 */
		geom.allPoints = function()
		{
			return geom.xy;
		}
	}
	
	/**
	 * geom.setStyle(options,o)
	 *
	 * Adds an xy coordinate to a point object, and creates a corresponding vector object that will be used
	 * when geom.draw() is called.
	 *
	 * options - object, list of options to apply/override the style of the vector object
	 * o - object, if specified, only apply the style to this vector object
	 */
	geom.setStyle = function(options,o)
	{
		var vectorObjects = geom.o;
		if (options) $.extend(geom.opts,options);
		if (o) vectorObjects = o;
		$(vectorObjects).each(function(){
			var v = this;
			
			if (t.svg)
			{
				if (geom.opts.display=="none") $(v).hide();
				else if (geom.opts.display!=N) $(v).show();
				
				if (geom.type=="point") v.setAttribute("r",geom.opts.r);
				
				if (geom.opts.fill) $(v).attr(geom.opts.fill);
				else $(v).attr({fill:"none"});
				
				if (geom.opts.stroke) $(v).attr(geom.opts.stroke);
			}
			else
			{
				if (geom.opts.display=="none") $(v).css("visibility","hidden");
				else if (geom.opts.display!=N) $(v).css("visibility","visible");
				
				if (geom.type=="point") $(v).css({width:geom.opts.r*2,height:geom.opts.r*2});
				if (geom.opts.fill)
				{
					v.f.color = geom.opts.fill.fill||"white";
					v.f.opacity = geom.opts.fill["fill-opacity"]||1.0;
				}
				else
				{
					v.f.color = "false";
					v.filled = "false";
				}
				if (geom.opts.stroke)
				{
					v.s.weight = geom.opts.stroke["stroke-width"]||1;
					v.s.color = geom.opts.stroke.stroke||"black";
					v.s.opacity = geom.opts.stroke["stroke-opacity"]||1.0;
				}
				else
				{
					v.s.weight="0";
					v.stroked="false";
				}
			}
		});
	}
	
	geom.setStyle();
	
	geom.draw = function(){ t.unDropGeom(this); t.drawGeom(this); };
	geom.drop = function(){ t.dropGeom(this); };
	geom.unDrop = function(){ t.unDropGeom(this); };
	
	return geom;
}

/**
 * MapChat.Canvas.drawGeom(geom)
 *
 * Draws vector geometries that have been added to the canvas by converting the
 * geographic coordinates, and updating the vector properties on the canvas:
 *
 * geom - object, previously created by the addGeom function...
 */
MapChat.Canvas.prototype.drawGeom = function(geom)
{
	var t = this;
	if ($.inArray(geom,t.geoms)==-1)
	{
		t.geoms.push(geom);
	}
	switch (geom.type)
	{
		case "path":
			var sPath = "";
			// the xy should include one or more arrays of coordinate pairs:
			$.each(geom.xy,function(subPath,c){
				if (c.length>0)
				{
					if (subPath>0)
					{
						sPath += " ";
					}
					sPath += "M ";
					$.each(c,function(ci,xy){
						if (ci>0)
						{
							sPath += " ";
						}
						if (ci==1)
						{
							sPath += "L "
						}
						var pixPt = t.map.geoToPix(xy[0],xy[1]);
						sPath += (pixPt[0]-t.map.xOrigin)+" "+(pixPt[1]-t.map.yOrigin);
					});
					if (geom.closePath)
					{
						sPath += (t.svg?" Z":" X");
					}
				}
			});
			if (t.svg)
			{
				$(geom.o).attr("d",sPath);
			}
			else
			{
				geom.o.path.value = sPath;
			}
			break;
		
		case "point":
			$.each(geom.xy,function(p,xy){
				var pixPt = t.map.geoToPix(xy[0],xy[1]);
				if (t.svg)
				{
					geom.o[p].setAttribute("cx",(pixPt[0]-t.map.xOrigin)+"px");
					geom.o[p].setAttribute("cy",(pixPt[1]-t.map.yOrigin)+"px");
				}
				else
				{
					$(geom.o[p]).css({left:(pixPt[0]-t.map.xOrigin-geom.opts.r)+"px",
						top:(pixPt[1]-t.map.yOrigin-geom.opts.r)+"px"});
				}
			});
			break;
	}
}

/**
 * MapChat.Canvas.dropGeom()
 *
 * Drops all references of a geometry object from the canvas.
 *
 */
MapChat.Canvas.prototype.dropGeom = function(geom)
{
	if (geom.type=="path")
		$(geom.o).remove();
	else
		$.each(geom.o,function(){$(this).remove();});
	
	if ($.inArray(geom,this.geoms)!=-1) this.geoms.splice($.inArray(geom,this.geoms),1);
}

/**
 * MapChat.Canvas.dropGeom()
 *
 * un-Drops the a geometry object from the canvas (doesn't require recreating the geometry).
 *
 */
MapChat.Canvas.prototype.unDropGeom = function(geom)
{
	var t = this;
	if ($.inArray(geom,t.geoms)==-1)
	{
		if (geom.type=="path")
			t.canvas.appendChild(geom.o);
		else
			$.each(geom.o,function(){t.canvas.appendChild(this);});
		
		t.geoms.push(geom);
	}
}

/**
 * MapChat.Canvas.clear()
 *
 * Removes all child-nodes from the canvas.
 *
 */
MapChat.Canvas.prototype.clear = function()
{
	$(this.canvas).children().remove();
	this.geoms = [];
}


/**
 * MapChat.Canvas.validateGeom(geom)
 *
 * check for self-intersecting polygons before calculating the area - only
 * check to see if the last (or closing) line segments intersect any previous
 * line segments.
 *
 * @param object geom - the geometry to validate.  If the geom type is point, it will return true without checking.
 * @param boolean ignoreClose - ignore the closing-line between the first/last points of the 
 */
MapChat.Canvas.prototype.validateGeom = function(geom,ignoreClose)
{
	// Not a polygon - nothing to validate.
	if (!(geom.type=="path" && geom.closePath)) return T;
	
	var t = this, r = T;
	
	$.each(geom.xy,function(i,xy){
		
		var pts = xy.length;
		
		if (pts>0 && pts<3) return (r = F);
		
		if (pts>3)
		{
			for (var p0=0;p0<(pts-2);p0++)
			{
				p1 = p0+1;
				for (var p2=(p0+2);p2<(pts);p2++)
				{
					var p3 = (p2<=(pts-2)?p2+1:0);
					if ((i!=(geom.xy.length-1) || (p3!=0 || !ignoreClose))
						&& (p3!=p0) && t.lineIntersect(xy[p0][0],xy[p0][1],xy[p1][0],xy[p1][1],xy[p2][0],xy[p2][1],xy[p3][0],xy[p3][1]))
					{
						return (r = F);
					}
				}
			}
		}
	});
	return r;
};

/**
 * MapChat.Canvas.lineIntersect(x0,y0,x1,y1,x2,y2,x3,y3)
 *
 * returns true if the two lines defined by the four pairs of coordinates intersect.
 *
 * x0,y0,x1,y1,x2,y2,x3,y3 - integers that correspond to the endpoints of the two lines
 */
MapChat.Canvas.prototype.lineIntersect = function(x0,y0,x1,y1,x2,y2,x3,y3)
{
	//codi extret de http://www.whisqu.se/per/docs/math28.htm
	var a1,b1,c1,a2,b2,c2,det_inv,m1,m2,x,y;
	
	if ((x1-x0)!=0)
		m1 = (y1-y0)/(x1-x0);
	else
		m1 = 999999999.999999;   // close enough to infinity
	
	if ((x3-x2)!=0)
		m2 = (y3-y2)/(x3-x2);
	else
		m2 = 999999999.999999;   // close enough to infinity
	
	a1 = m1;
	a2 = m2;
	b1 = -1;
	b2 = -1;
	c1 = (y0-m1*x0);
	c2 = (y2-m2*x2);
	
	// inverse of the determinate
	det_inv = 1/(a1*b2 - a2*b1);
	
	// use Kramers rule to compute xi and yi
	x=((b1*c2 - b2*c1)*det_inv);
	y=((a2*c1 - a1*c2)*det_inv);

	if ( (x>=Math.min(x0,x1) && x<=Math.max(x0,x1) 
		&& y>=Math.min(y0,y1) && y<=Math.max(y0,y1)) 
		&& (x>=Math.min(x2,x3) && x<=Math.max(x2,x3) 
		&& y>=Math.min(y2,y3) && y<=Math.max(y2,y3)) ) return T;
	
	else return F;
}
