/**
 * An object that presents a Chat interface for private messages.
 *
 * @class MapChat.Chat.PrivateDialog
 *
 * @param map - a kaMap class instance. If not defined, it will default to MapChat.map.
 * @param chat - a ChatLib class instance, which includes the methods required for
 *               keeping chat messages synchronized with the server (e.g., Get(),
 *               Send(), etc.).  If not defined, it will default to MapChat.chat.
 * @param options - an associative array of options that configure the behaviour of the dialog.
 *   object targets - an associative array of target users/groups that messages will be sent to/from
 *
 * Depends:
 *	kamap/htdocs/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 *	MapChat/MapLayers.js
 *	jquery/ui/ui.dialog.js
 *  MapChat/SelectLib.js
 *  MapChat/DrawLib.js
 */

// Get an alias for the the MapChat.Chat namespace (to make the code shorter).
var mcc = MapChat.UseNS("MapChat.Chat"),

	// Alias the disable/enable button functions and the QueueEvent function:
	dbtn = MapChat.DisableButton, ebtn = MapChat.EnableButton, mcqe = MapChat.QueueEvent;


mcc.PrivateDialog = function(map,chat,options)
{
	var t = this,
		options = (t.options = options||{}),
		name = (t.name = "MapChat.Chat.PrivateDialog"),
		targets = (t.targets = options.targets),
		tid = (t.tid = $.toJSON(targets)),
		tgNames = (t.tgNames = MapChat.ChatLib.TargetNames(targets)),
		zIndex = (t.zIndex = options.zIndex||300),
		map = (t.map = map||MapChat.map),
		mapLayers = (t.mapLayers = MapChat.MapSingleton("MapChat.MapLayers",t.map)),
		ChatLib = (t.ChatLib = chat||MapChat.chat),
		
		SelectLib = (t.SelectLib = MapChat.MapSingleton("MapChat.SelectLib",t.map)),
		DrawLib = (t.DrawLib = MapChat.MapSingleton("MapChat.DrawLib",t.map)),
		
		content = (t.content =
			$('<div class="mccd" style="font-size:.8em;line-height:16px">'+
				'<div class="topPanel">'+
					(t.tgNames.target_users?'Users in this private discussion: '+t.tgNames.target_users.join(', ')+'<br />':'')+
					(t.tgNames.target_groups?'Groups in this private discussion: '+t.tgNames.target_groups.join(', ')+'<br />':'')+
					'Display mode: <input type="radio" name="thread_view" checked="true" /> Threaded &nbsp;&nbsp;&nbsp; '+
					'<input type="radio" name="sequential_view" /> Sequential'+
				'</div>'+
				'<div class="chat_messages autoscroll"></div>'+
				'<div class="bottomPanel"><button class="ui-corner-all ui-state-default" name="new_topic">New Topic</button></div>'+
			'</div>')
			.dialog({title:"Private Discussion",
				width:350,height:300,modal:false,autoOpen:false,closeOnEscape:false,
				resizeStop:function()
				{
					t.content.trigger("resizeDiv");
				},
				open:function()
				{
					mcqe(9,"ChatPrivateDialog",["Opened",t.user,tid]);
					MapChat.PushQueue(); // Makes sure that this event is recorded asap
					t.content.trigger("resizeDiv");
				},
				close:function()
				{
					mcqe(9,"ChatPrivateDialog",["Closed",t.user,tid]);
					MapChat.PushQueue(); // Makes sure that this event is recorded asap
				}
			})
		).bind("resizeDiv",function(){
			// Make the continer is fit to the dialog's height (so we get scrolling):
			t.content.css({height:$(".ui-resizable-s", t.dialog).offset().top-t.content.offset().top});
			t.content.children('.chat_messages').css({height:t.content.height()-
				t.content.children('.topPanel').height()-
				t.content.children('.bottomPanel').height()-10});
		}),
		
		dialog = (t.dialog = content.parents(".ui-dialog:first")),
		
		displayCanvas = (t.displayCanvas = map.defaultCanvas||MapChat.Class("MapChat.Canvas",{map:t.map,zIndex:600})),
		
		// Styles for selection/draw shapes:
		selectOpts = (t.selectOpts = options.selectOpts||{stroke:{stroke:"#00F",'stroke-width':2,'stroke-opacity':0.60},r:5}),
		selectFill = (t.selectFill = $.extend(options.selectFill||{fill:{fill:"#00F",'fill-opacity':0.35}},t.selectOpts)),
		extentOpts = (t.extentOpts = options.extentOpts||{stroke:{stroke:"#FCA",'stroke-width':2,'stroke-opacity':0.60},r:5}),
		extentFill = (t.extentFill = $.extend(options.extentFill||{fill:{fill:"#FCA",'fill-opacity':0.35}},t.extentOpts)),
		drawOpts = (t.drawOpts = options.drawOpts||{stroke:{stroke:"#77F",'stroke-width':2,'stroke-opacity':0.60},r:5}),
		drawFill = (t.drawFill = $.extend(options.drawFill||{fill:{fill:"#77F",'fill-opacity':0.35}},t.drawOpts));
	
	t.displayCanvas.activate();
	
	t.ChatLib.AddContext(t.content.children('.chat_messages'),t,'private_listener');
	
	// Bind events to the chat library.
	t.content.bind("MessagesReceived",function(e,chat_ids,tid,scroll){t.UpdateChat(chat_ids,tid,scroll);});
	
	// Setup the threaded/sequential switch:
	t.content.children('.topPanel').children('input[type="radio"]').click(function(){t.SwitchView($(this));});
	
	$('button[name="new_topic"]',t.content).click(function(){
		// Display an input area above the new topic button:
		nt = $(this);
		if (!nt.data("NewMessage"))
		{
			nt.data("NewMessage",
			t.ChatLib.NewMessage({
				obj:nt,
				chat_opts:{private:t.targets},
				append:'sibling',
				map:t.map,
				SelectLib:t.SelectLib,
				DrawLib:t.DrawLib,
				above:T,
				close:function(e,msg){
					nt.removeData("NewMessage");
					t.content.trigger("resizeDiv");
					//if (msg) MapChat.ChatLib.ScrollTo(t.content.children('.chat_messages'),{msgID:msg.data('md').id,stick:F});
				}
			}));
			t.content.trigger("resizeDiv");
		}
	});
	
	// Now, call t.SwitchView, which will get all the private messages that are currently 
	// in the cache, and match the current user and selected targets, then display them in
	// whichever mode is currently checked.
	t.SwitchView(t.content.children('.topPanel').children('input[type="radio"]:checked'));
}

var mccpd = mcc.PrivateDialog;

mccpd.prototype.Open = function()
{
	$.blockUI();
	
	var t = this;
	
	t.content.dialog("open").dialog("moveToTop");
	
	$.unblockUI();
	
	return t;
	
}

/**
 * This function should trigger whenever new messages are received.
 */
mccpd.prototype.UpdateChat = function(chat_ids,tid,scroll)
{
	
	var t = this, threaded = t.content.children('.topPanel').children('input[name="thread_view"]').attr("checked");
	
	if (tid!=t.tid) return;
	
	if (t.content.dialog("isOpen"))
	{
		mcqe(9,"ChatPrivateDialog",["MessageRead",t.user,tid].concat(chat_ids));
		MapChat.PushQueue(); // Makes sure that this event is recorded asap
	}
	
	var pmsgs = t.ChatLib.msgData.children().filter('.chat_private').filter(function(){
			var msg = $(this), md = msg.data('md');
			return ($.inArray(md.id,chat_ids)!=-1);
		});
	
	pmsgs.each(function(){
		var msg = $(this);
		if (threaded)
			var dm = t.ChatLib.DisplayMessage(msg,{showReplies:T});
		else
			var dm = t.ChatLib.DisplayMessage(msg);
		
		MapChat.ChatLib[threaded?"AppendThread":"AppendList"](dm,t.content.children('.chat_messages'));
	});
	
	t.ChatLib.BindMessageElems('chat_new_elem',T,{map:t.map,SelectLib:t.SelectLib,DrawLib:t.DrawLib});
	MapChat.ChatLib.ScrollTo(t.content.children('.chat_messages'),{msgID:pmsgs.filter(':first').data('md').id,threaded:threaded});
}

/**
 * Switch between threaded/sequential views
 */
mccpd.prototype.SwitchView = function(cur)
{
	var t = this,
		thr = t.content.children('.topPanel').children('input[name="thread_view"]'),
		seq = t.content.children('.topPanel').children('input[name="sequential_view"]');
	if (cur.attr("name")=="thread_view") seq.attr("checked",F);
	else thr.attr("checked",F);
	
	var pmsgs = t.ChatLib.msgData.children('.chat_private').filter(function(){
		var msg = $(this), md = msg.data('md');
		//needs cleanup...should parse target_users when messages are recieved anyway.
		return (((!t.targets.target_groups && !md.target_groups)||$.toJSON(t.targets.target_groups)==$.toJSON((typeof(md.target_groups)=="string"?$.parseJSON(md.target_groups):md.target_groups)))
			&& ((!t.targets.target_users && !md.target_users)||$.toJSON(t.targets.target_users)==$.toJSON((typeof(md.target_users)=="string"?$.parseJSON(md.target_users):md.target_users))));
	});
	
	t.content.children('.chat_messages').children().remove();
	t.content.children('.chat_messages').append(t.ChatLib.DisplayAllMessages(thr.attr("checked"),T,pmsgs));
	t.ChatLib.BindMessageElems('chat_new_elem',T,{map:t.map,SelectLib:t.SelectLib,DrawLib:t.DrawLib});
	MapChat.ChatLib.ScrollTo(t.content.children('.chat_messages'),{msgID:T,stick:T,threaded:thr.attr("checked")});
}

