/**
 * A library for handling the sending/receiving of chat messages
 *
 * @class MapChat.Chat
 *
 * @param conf - a jQuery object containing the chat node from the discussion's xml config
 *
 * Depends:
 *	kamap/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 *  jquery/jquery.scrollTo.js
 *  jquery/jquery.notice.js
 */

MapChat.Using("jquery.scrollTo");
MapChat.Using("jquery.notice");

MapChat.ChatLib = function(conf)
{
	var t = this,
		conf = (t.conf = conf),
		maps = (t.maps = []), // Any kaMap objects added to this object will have events triggered (e.g., new drawings) by this instance of ChatLib
		defaultMap = (t.defaultMap = N), // If this is set, it will be the default map object for some contexts (e.g., when private messages are opened from a jquery.notice element)
		serverTime = (t.serverTime = N), // The timestamp returned by the last poll to the server
		lastTime = (t.lastTime = F), /* ((new Date()).getTime()+MapChat.timeDelta), */
		minID = (t.minID = -2), // the lowest message ID that has been retrieved.
		maxID = (t.maxID = -2), // the maximum message ID that has been retrieved.
		lastMinID = (t.lastMinID = N), // the lowest message ID that was retrieved from the last poll.
		lastMaxID = (t.lastMaxID = N), // the maximum message ID that was retrieved from the last poll.
		historyDepth = (t.historyDepth = 0), // this will determine how deep into the thread tree to retrieve messages when loading the history
		depthMin = (t.depthMin = []), // this will keep track of the minID retrieved for a given thread depth
		depthMax = (t.depthMax = []), // this will keep track of the minID retrieved for a given thread depth
		firstPoll = (t.firstPoll = F), // Tells us whether the first setTimeout poll has executed
		uInfo = (t.uInfo = {}), // Stores information for users in the *current* discussion
		pollTime = (t.pollTime = (t.conf.children("pollTime:first").val()*1||5000)), // The time in milliseconds between polls for new messages
		pollInterval = (t.pollInterval = -1), // this will be the ID for the instance of pollInterval() that polls the server for messages.
		pollInProgress = (t.pollInProgress = F), // This should be set to true when a poll is in progress...to prevent multiple polls if they take longer than pollTime to complete
		globalFlags = (t.globalFlags = []), // Any flags added to this array will be appended to any messages sent using this class until the flag is removed from the array.
		privateDialogs = (t.privateDialogs = {}), // Keeps track of open private message dialogs.
		
		msgData = (t.msgData = $("<div />")), // a jQuery object will store markup containing message data - this will allow the use of jQuery's selectors for searching 
		userData = (t.userData = {}), // stores user data as it is retreived from the server.
		
		context = (t.context = $([])), // This will be the contexts that are searched for elements when triggering events:
		
		intialized = (t.initialized = F); // This should be set to true after the pollInterval has been set.
		
	MapChat.events.bind("DroppedUsers",function(e,du){
		$.each(du,function(i,u){
			MapChat.Debug(['info','Dropped',u]);
			MapChat.UserInfo(u).data("u").online = F;
			$(".chat_user_icon[alt='"+u.replace(/'/g,"\\'")+"'],.chat_user_icon[name='"+u.replace(/'/g,"\\'")+"']",t.context).trigger("UserOffline");
		});
	});
	MapChat.events.bind("AddedUsers",function(e,au){
		$.each(au,function(i,u){
			MapChat.Debug(['info','Added',u]);
			MapChat.UserInfo(u).data("u").online = T;
			$(".chat_user_icon[alt='"+u.replace(/'/g,"\\'")+"'],.chat_user_icon[name='"+u.replace(/'/g,"\\'")+"']",t.context).trigger("UserOnline");
		});
	});
	MapChat.events.bind("EditedMessages",function(e,em){
		var aRefresh = [];
		$.each(em,function(i,id){
			MapChat.Debug(['info','Edited',id]);
			if ($(".chat_id-"+id,t.msgData).length>0) aRefresh.push(id);
		});
		if (aRefresh.length>0)
		{
			MapChat.Debug(['info','Refresh',aRefresh]);
			t.Get({describe:aRefresh},{ignore:T});
		}
		t.context.filter('.edit-listener').trigger("EditedMessages",[em]);
	});
	
	MapChat.events.bind("NewUserDrawings",function(e,exts){
		$.each(t.maps,function(mi,map){
			if (exts[map.currentMap])
			{
				var xy = F;
				$.each(exts[map.currentMap],function(i,ux){
					if (!xy) xy = [ux.xmin,ux.ymin,ux.xmax,ux.ymax];
					else xy = [Math.min(xy[0],ux.xmin),Math.min(xy[1],ux.ymin),Math.min(xy[2],ux.xmax),Math.min(xy[3],ux.ymax)];
				});
				if (xy)
				{
						$.each(MapChat.FindTilesByExtent(
								map.unadjustPixPosition(xy[0],xy[1],T),
								map.unadjustPixPosition(xy[2],xy[3],T),
								map,"USER_LAYERS",5),
							function(){ map.resetTile(""+this,F,T);}
						);
						
						// Also need to update the timestamp on the user layer, otherwise the browser may still use cached layers
						var ts = new Date();
						map.getCurrentMap().getLayer("USER_LAYERS").timeStamp = Math.round(ts.getTime()/1000) + ts.getTimezoneOffset() * 60;
				}
			}
		});
	});
}

//shortcut to the MapChat.Chat class:
var mccl = MapChat.ChatLib,
	
	// shortcut for the queue event function:
	mcqe = MapChat.QueueEvent;

/**
 * Initialze the chat poll interval.
 */
mccl.prototype.Init = function()
{
	var t = this;
	t.Get({sinceID:-2},{after:function(xhr){
		if (parseInt($("chat_messages:first",xhr.responseXML).attr("length"))<25) t.historyDepth++;
		t.pollInterval = setInterval(function(){if (!t.polling) t.Get({sinceID:t.maxID},{after:function(){t.firstPoll=T;}}); },t.pollTime);
	}});
	t.intialized = T;
}

/**
 * Get messages from the server.
 *
 * @param opts - configure what this function returns from the server.
 * @param getOpts - extra configuration that alters the behaviour of this function, but are not sent to
 *                  the server in the AJAX API request.
 *
 * Possible values that can be included in the opts parameter:
 *  sinceTime: if specified, only messages recorded after this number of milliseconds since the UNIX 'epoch' time will be returned.
 *  sinceID: if specified, -2 gets the most recent batch of messages; greater than or equal to -1 gets the next batch
 *           of in sequential order messages with IDs greater than sinceID.
 *  limit: defaults to 25; 0 will retrieve all messages; values greater than zero will limit the number of messages returned to
 *         the specified value.
 *  from: if set, only return messages from the specified username
 *  reverse: if true, then get messages *before* sinceID
 *  depth: the thread-depth to get messages at; -1 is any depth, 0 is the top-level message that starts thread, 1, etc...
 *  flags: an array of text flags - if defined, only messages that satisfy these flags will be returned.
 *  selection: an integer that references the id of a set of features selected by the user - only messages explicitly linked
 *             to these features will be returned.
 *  intersects: the same as 'selection', but will return any messages that are referenced to geographic features that intersect
 *              the selection set identified.
 *  extent: an array of xmin,ymin,xmax,ymax coordinates to explicitly define a geographical extent for this message - if
 *          the min/max coordinates are the same (or max is missing), the extent will be treated as a point.
 *  distance: the distance to search for nearby messages (works with intersects or extent).
 *  list: if true, only IDs will be returned for matching messages
 *  listgeo: same as list, except geographic information is included in the results.
 *  private: an object containing target_users or target_groups, which are must be sorted arrays of user/group names.
 *  describe: if an array of IDs is supplied, details will be returned for the matching messages - NOTE: all above parameters
 *            will be ignored in the request if this is used
 *  children: if true, then the messages returned will be replies to sinceID or IDs listed in 'describe' (ignored if sinceID or 
 *            describe are not set)
 *
 * Possible values that can be included in the getOpts parameter:
 *  ignore: if true, the pollInProgress flag will be ignored
 *  after: if set, this function will be executed after the ParseMessages method has been executed.
 *  updateDepth: if set, the depthMin/depthMax values will be set.
 *  delayScroll: if set, individual messages will not trigger a ScrollTo in the dialog...
 *  noParse: if true, t.ParseMessages will not be executed when the results are returned.
 */
mccl.prototype.Get = function(opts,getOpts)
{
	var t = this,
		opts = opts||{},
		getOpts = getOpts||{},
		maps = [];
	
	// Add the list of maps that will have events triggered on them if changes to user drawings occur.
	$.each(t.maps,function(mi,map){
		maps.push(map.currentMap);
	});
	opts.maps = maps;
	opts.timeDelta = MapChat.timeDelta;
	
	if (!getOpts.ignore && t.pollInProgress)
	{
		MapChat.Debug(["warn","Chat_Get: poll already in progress - skipping the current request."]);
		return;
	}
	
	// Set the flag indicating we are currently polling the server...
	if (!getOpts.ignore) t.pollInProgress = T;
	
	// Fetch the messages from the server in asynchronous mode
	MapChat.RunFunction("Chat_Get",{args:{opts:$.toJSON(opts)}},{
		async:T,
		complete: function(xhr){
			try
			{
				if (!opts.list && !opts.listgeo && !getOpts.noParse) t.ParseMessages(xhr,getOpts.updateDepth,!getOpts.delayScroll);
				if (getOpts.after) getOpts.after(xhr);
			}
			catch (e)
			{
				MapChat.Debug(["warn","Chat_Get: Error parsing AJAX response: ",e,xhr.responseText]);
			}
			if (!getOpts.ignore) t.pollInProgress = F;
		},
		error: function(xhr,sStatus,nError){
			MapChat.Debug(["warn","Chat_Get: Error on AJAX request. Status: %s Error: %s ",sStatus,nError]);
			if (!getOpts.ignore) t.pollInProgress = F;
		}
	});
}

/**
 * Handles the receipt of new messages from the server.
 */
mccl.prototype.ParseMessages = function(xhr,updateDepth,delayScroll)
{
	var t = this,
		cm = $("chat_messages:first",$(xhr.responseXML)),
		newIDs = [],
		newPrivateIDs = {},
		updatePIDs = {},
		new_count = cm.attr("count");
	
	t.lastMinID = N;
	t.lastMaxID = N;
		
	if (new_count!=N)
		t.msgData.data("count",1*new_count);
	
	t.serverTime = cm.attr("time");
	
	$.each($.parseJSON(cm.text()),function(i,md){
		var msg = t.msgData.children(".chat_id-"+md.id),
			flags = "chat_id-"+md.id+" chat_pid-"+(md.parent_id||'');
		
		// Parse any remaining elements (i.e., text that should be converted to arrays):
		mccl.FixMsgExtents(md);
		
		// Setup the flags and replies properties:
		if (md.flags) flags += " "+md.flags;
		
		// Format the display date
		md.display_date = dateFormat(md.epoch_ms-MapChat.timeDelta,"mmm dd h:MMt");
		
		// Flag private vs. public messages:
		if (md.private) flags += " chat_private";
		else flags += " chat_public";
		
		// Sort target user/groupnames:
		var targets = {};
		if (md.target_users)
		{
			md.target_users = $.parseJSON(md.target_users).sort();
			targets.target_users = md.target_users;
		}
		if (md.target_groups)
		{
			md.target_users = $.parseJSON(md.target_groups).sort();
			targets.target_groups = md.target_groups;
		}
		
		// Make sure the replies data attribute is an array of integers:
		if (!md.replies||md.replies=="") md.replies = [];
		else md.replies = $.parseJSON(md.replies);
		$.each(md.replies,function(i){md.replies[i] = parseInt(md.replies[i]);});
		
		// If the message has not been received before.
		if (msg.length==0)
		{
			// Create the message element.
			msg = $('<p class="'+flags+'" name="'+md.username+'" />');
			t.msgData.append(msg);
			msg.data("md",md);
			
			// Keep a list of new messages IDs:
			if (!md.private)
			{
				newIDs.push(md.id);
			}
			else
			{
				var tid = $.toJSON(targets);
				if (!newPrivateIDs[tid]) newPrivateIDs[tid] = [md.id];
				else newPrivateIDs[tid].push(md.id);
			}
			
			// Update any parents for new messages:
			if (md.parent_id!=N && md.parent_id>=0)
			{
				var p = $("p.chat_id-"+md.parent_id+":first",t.msgData),
					pmd = p.data("md");
				if (p.length==1 && md.id!=N && $.inArray(md.id,pmd.replies)==-1)
				{
					pmd.replies.push(md.id);
					if (updatePIDs[md.parent_id]) updatePIDs[md.parent_id].push(md.id);
					else updatePIDs[md.parent_id] = [md.id];
				}
			}
			
			if (new_count==N && t.firstPoll && md.id > t.maxID) t.msgData.data("count",t.msgData.data("count")+1);
			
		}
		
		// If the message *has* been received before, then update its properties and
		// trigger the MessageChanged event:
		else
		{
			msg.attr("class",flags);
			md.test = T;
			$.extend(msg.data("md"),md);
			
			msg.data("select_svg_new",T).data("draw_svg_new",T);
			
			$("div.chat_message.chat_id-"+md.id,t.context).trigger("MessageChanged");
		}
		
		// Increment minID/maxID properties:
		if (md.id<t.minID||t.minID<0) t.minID = md.id;
		if (md.id>t.maxID) t.maxID = md.id;
		if (t.lastMinID==N||md.id<t.lastMinID) t.lastMinID = md.id;
		if (t.lastMaxID==N||md.id>t.lastMaxID) t.lastMaxID = md.id;
		if (updateDepth)
		{
			if (md.id<t.depthMin[md.depth]||t.depthMin[md.depth]==N) t.depthMin[md.depth] = md.id;
			if (md.id>t.depthMax[md.depth]||t.depthMax[md.depth]==N) t.depthMax[md.depth] = md.id;
		}
	});
	
	// If any acutal new messages have been received, 
	if (newIDs.length>0) t.context.filter('.message-listener:not(.private_listener)').trigger("MessagesReceived",[MapChat.UniqueArray(newIDs),delayScroll]);
	$.each(newPrivateIDs,function(tid,pIDs){
		if (t.privateDialogs[tid])
		{
			t.context.filter('.private_listener').trigger("MessagesReceived",[MapChat.UniqueArray(pIDs),tid,delayScroll]);
		}
		
		if (!t.privateDialogs[tid] || !t.privateDialogs[tid].content.dialog('isOpen'))
		{
			var targets = $.parseJSON(tid),
				tgNames = MapChat.ChatLib.TargetNames(targets);
			$.noticeAdd({stay:true,content:$('<span style="cursor:pointer;">New privates messages from: '+
				(tgNames.target_users?tgNames.target_users.join(', '):'')+
				((tgNames.target_users&&tgNames.target_groups)?' | ':'')+
				(tgNames.target_groups?tgNames.target_groups.join(', '):'')+'</span>').click(function(){
					t.GetPrivateDialog(targets).Open();
					$.noticeRemove($(this).parents('.notice-item:first'));
				}),stayTime:5000});
		}
	});
	
}

/**
 * Send new messages to the server.
 *
 * @param opts - configures what information is included with this message
 * @param libs - links to map, SelectLib, and DrawLib objects, if applicable.
 *
 * Possible values that can be included in the opts parameter:
 *  clientTime: the time reported by the client, adjusted for the time delta relative to the server (should be pretty much equal).
 *  parentID: if specified, this message is in reply to a previous message.
 *  flags: an array of text flags - if defined, these will be saved in a space-separated string in the chat_log
 *  selection: a list of integers that references a set of features selected by the user - if specified, these will be included
 *             in the spatial links for the chat message.
 *  drawing: an integer that references the id of a user-drawn feature - if specified, this will be included
 *           in the spatial links for the chat message.
 *  extent: a map object, whose current extent will be added to the spatial data for this message.
 *  distance: (not yet implemented) if set, the geometry for the chat_log will be a buffer around the extent and/or selected
 *            features included above
 *  type: if set, this will determine the 'type' of message (e.g., if some level of argumentation structure is to be used in a discussion)
 *  private: if set, this should be an array containing lists of usernames and/or groupnames, to whom the messages are targeted.
 *  text: the text of the message
 */
mccl.prototype.Send = function(opts,libs)
{
	// No empty messages allowed:
	if (!opts || !opts.text) return F;
	
	var t = this,
		md = $.extend({},opts); //break the reference to the object supplied as the opts argument.
	
	// Record the local time that this message is sent
	md.clientTime = ((new Date()).getTime()+MapChat.timeDelta);
	if (md.flags && t.globalFlags.length>0) md.flags = md.flags.concat(t.globalFlags);
	else if (t.globalFlags.length>0) md.flags = [].concat(t.globalFlags);
	
	// If set, the 'extent' should be a pointer to a map object.
	if (md.extent) md.extent = md.extent.getGeoExtents();
	
	// Flag current selection/drawing as having been linked.
	if (opts.selection && libs.SelectLib) libs.SelectLib.hasLinked = T;
	if (opts.drawing && libs.DrawLib) libs.DrawLib.hasLinked = T;
	
	msg = $("chat_send:first", MapChat.RunFunction("Chat_Send",{args:{opts:$.toJSON(md)}}));
	
	// The above is rendered moot by this, but keeping selections/drawings after linking them can
	// be problematic, where people don't notice that they already have a selection or drawing, and
	// they keep adding to the selection/drawing without paying attention.
	if (opts.selection && libs.SelectLib) libs.SelectLib.GetSelection({mode:'clear'});
	if (opts.drawing && libs.DrawLib) libs.DrawLib.Clear();
	
	if (msg.length>0)
	{
		var md = $.extend(md,$.parseJSON(msg.text()));
			
		// Get the message info:
		md.flags = "chat_msg chat_id-"+md.id+" chat_pid-"+md.parentID+(md.flags?((typeof(md.flags)=="string")?md.flags:" "+md.flags.join(" ")):"");
		md.replies = [];
		
		// Make sure fields match what they would coming from the database:
		if (md.parentID)
		{
			md.parent_id = md.parentID;
			delete md.parentID;
		}
		md.epoch_ms = md.clientTime;
		md.display_date = dateFormat(md.epoch_ms-MapChat.timeDelta,"mmm dd h:MMt");
		delete md.clientTime;
		
		// Add the username to the message's properties:
		md.username = MapChat.session.user_identity;
		
		var msg = $('<p class="'+md.flags+(md.private?' chat_private':'')+'" />').appendTo(t.msgData);
		msg.data("md",md);
		
		t.msgData.data("count",t.msgData.data("count")+1);
		
		if (md.parent_id) $("div.chat_id-"+md.parent_id,t.context).trigger("NewReplies",[[md.id]]);
		
		// Parse any remaining elements (i.e., text that should be converted to arrays):
		mccl.FixMsgExtents(md);
		
		if (!md.private) t.context.filter('.message-listener:not(.private_listener)').trigger("MessagesReceived",[[md.id]]);
		else
		{
			var private = {},
				tu = md.target_users,
				tg = md.target_groups;
			
			// Build the targetid for the dialog, then trigger the private listener event (so this message appears in the dialog)
			if (tu) private.target_users = (typeof(tu)=="string"?$.parseJSON(tu):tu);
			if (tg) private.target_groups = (typeof(tg)=="string"?$.parseJSON(tg):tg);
			t.context.filter('.private_listener').trigger("MessagesReceived",[[md.id],$.toJSON(private)]);
		}
		
		return msg;
		
	}
	else
	{
		MapChat.Debug(["error","Unable to send message: ",opts]);
		return false;
	}
}


/**
 * Creates an input area for chat messages.
 *
 * @param object opts - defines how to present the input area.
 *
 *  Possible values that can be included in opts:
 *   obj: the jQuery object that refers to the html element that
 *        the input area will be displayed with
 *   append: if this is equal to 'sibling', the input will be appended
 *           as a sibling relative to the obj element.  If it is equal
 *           to 'child', then the element is appended as a child element
 *           within the obj element.  Otherwise, the default is false,
 *           which will add the element seperately within the page.
 *   above: if true and append='sibling', the input area is displayed
 *          above the obj element, otherwise it is added below.  If
 *          append is not set or false, then the position of the input
 *          area will be calculated relative to obj, and will be placed
 *          above/below.
 *   css: an object-array that contains css properties that will be applied
 *        to the input area.
 *   cls: if set, this class is added to the input area's div element.
 *   chat_opts: if set, this will preset any opts that should be submitted
 *              to the t.chat.Send() method.
 *   close: if set, this should be a callback function that is executed
 *          when the input area is closed.  The first argument to this
 *          function will be a boolean value indicating whether the
 *          message was sent to the server (true) or not (false).
 *   eflags: an array of text flags that will be added to the message.
 */
mccl.prototype.NewMessage = function(opts)
{
	var t = this,
		obj = opts.obj,
		append = opts.append||'none',
		above = opts.above||F,
		
		map = opts.map||F,
		SelectLib = opts.SelectLib||F,
		DrawLib = opts.DrawLib||F,
		
		eflags = opts.eflags||[],
		css = opts.css,
		chat_opts = opts.chat_opts||{},
		link_opts = {},
		
		oInput = $('<div class="chat_input"><p><textarea /></p>'+
			'<p class="alignright">'+
				(map?
					'<img class="icon png24 extent_link bordered" svg_mode="extent" title="Link to Current Extent" src="images/icons/link_extent.png" />'
				:'')+
				(SelectLib?
					'<img class="icon png24 select_link bordered select-listener'+(SelectLib.count>0?'':' disabled')+'" title="Link to Selected Features" src="'+
						(SelectLib.count>0?
							'images/icons/link_selection.png':'images/icons/link_selection_grayed.png'
						)+'" />'
				:'')+
				(DrawLib?
					'<img class="icon png24 draw_link bordered draw-listener'+(DrawLib.geomID!=N?'':' disabled')+'" title="Link to Current Drawing" src="'+
						(DrawLib.geomID!=N?
							'images/icons/link_drawing.png':'images/icons/link_drawing_grayed.png'
						)+'" />'
				:'')+
			'</p>'+
		'</div>');
	
	$("img.icon",oInput).tooltip({track:T,delay:0,showURL:F,extraClass:"tooltip"})
		/*.hover(mccl.MsgIconHoverIn,mccl.MsgIconHoverOut)*/.data("link_opts",link_opts)
		.data("ta",$('textarea',oInput));
	
	// Disabled for now...to prevent confusion:
// 	if (SelectLib && SelectLib.count>0 && !SelectLib.hasLinked) $("img.icon.select_link",oInput).click();
// 	if (DrawLib && DrawLib.geomID!=N && !DrawLib.hasLinked) $("img.icon.draw_link",oInput).click();
	
	$("img.icon.extent_link",oInput).bind("click",{map:map},mccl.ExtentLinkClick);
	$("img.icon.select_link",oInput).bind("click",{SelectLib:SelectLib},mccl.SelectLinkClick)
		.bind("UpdateSelection",{SelectLib:SelectLib},mccl.SelectLinkUpdate).trigger("UpdateSelection");
	$("img.icon.draw_link",oInput).bind("click",{DrawLib:DrawLib},mccl.DrawLinkClick)
		.bind("DrawUpdate",{DrawLib:DrawLib},mccl.DrawLinkUpdate)
		.bind("DrawingCleared",{DrawLib:DrawLib},mccl.DrawLinkUpdate).trigger("DrawUpdate");
	
	MapChat.KeyProtect($("textarea",oInput));
	
	if (opts.cls) oInput.addClass(opts.cls);
	if (opts.css) oInput.css(opts.css);
	
	// Append the input area in the appropriate location:
	switch (append)
	{
		case F:
			alert("Not Implemented Yet");
			break;
		
		case "child":
			obj.append(oInput);
			break;
		
		default:
			// This will call either the 'before' or 'after' jQuery 
			// methods on the obj element:
			obj[(above?'before':'after')](oInput);
			break;	
	}
	
	return $('textarea',oInput).blur(function(){
		var ta = $(this);
		ta.data("leaveTO",setTimeout(function(){
			if (ta.val()!="") return; // Don't hide the dialog if there's text in it.
			oInput.remove();
			if (opts.close) opts.close(F);
		},200));
	}).focus(function(){
		var ta = $(this);
		if (ta.data("leaveTO"))
		{
			clearTimeout(ta.data("leaveTO"));
			ta.removeData("leaveTO");
		}
	}).keypress(function(e){
		switch(e.keyCode)
		{
			case 13:
				if (e.ctrlKey||e.metaKey||e.shiftKey) return;
				// TODO: send the message;
				
				var msg = t.Send($.extend(chat_opts,{ text:$("textarea",oInput).val(), flags:eflags },link_opts),{map:map,SelectLib:SelectLib,DrawLib:DrawLib});
				
				oInput.remove();
				if (opts.close) opts.close(T,msg);
				return F;
				break;
			
			case 27:
				mcqe(9,"ChatLib",["CancelNewMessage",chat_opts.parentID]);
				oInput.remove();
				if (opts.close) opts.close(F);
				break;
		}
		e.stopPropagation();
	}).focus();
}

/**
 * Places an editable input area in place of the text for a chat message.
 *
 * @param object opts - defines how to present the input area.
 *
 *  Possible values that can be included in opts:
 *   obj: the jQuery object that refers to the chat_text html element that
 *        the input area will be displayed for editing
 *   md: the data for the existing message
 *   map: the map object to get extents from
 *   SelectLib: the SelectLib to get selections from
 *   DrawLib: the SelectLib to get drawings from
 *   css: an object-array that contains css properties that will be applied
 *        to the input area.
 *   cls: if set, this class is added to the input area's div element.
 *   close: if set, this should be a callback function that is executed
 *          when the input area is closed.  The first argument to this
 *          function will be a boolean value indicating whether the
 *          message was sent to the server (true) or not (false).
 */
mccl.prototype.EditMessage = function(opts)
{
	var t = this,
		obj = opts.obj,
		md = opts.md,
		map = opts.map||F,
		SelectLib = opts.SelectLib||F,
		DrawLib = opts.DrawLib||F,
		msg_opts = $.extend({},md),
		link_opts = {},
		
		oInput = $('<div class="chat_edit_input"><p><textarea>'+md.text+'</textarea></p>'+
			'<p style="float:left;">'+
				((md.extent && md.extent.length>0)?
					'<img class="icon png24 extent_unlink bordered" title="Unlink Extent" src="images/icons/unlink_extent.png" />'
				:'')+
				((md.selection && md.selection.extent && md.selection.extent.length>0)?
					'<img class="icon png24 select_unlink bordered" title="Unlink Selected Features" src="images/icons/unlink_selection.png" />'
				:'')+
				((md.drawing && md.drawing.extent && md.drawing.extent.length>0)?
					'<img class="icon png24 draw_unlink bordered" title="Unlink Drawing" src="images/icons/unlink_drawing.png" />'
				:'')+
			'</p>'+
			'<p class="alignright">&nbsp;'+
				(map?
					'<img class="icon png24 extent_link bordered" title="Link to Current Extent" src="images/icons/link_extent.png" />'
				:'')+
				(SelectLib?
					'<img class="icon png24 select_link bordered select-listener'+(SelectLib.count>0?'':' disabled')+'" title="Link to Selected Features" src="'+
						(SelectLib.count>0?
							'images/icons/link_selection.png':'images/icons/link_selection_grayed.png'
						)+'" />'
				:'')+
				(DrawLib?
					'<img class="icon png24 draw_link bordered draw-listener'+(DrawLib.geomID!=N?'':' disabled')+'" title="Link to Current Drawing" src="'+
						(DrawLib.geomID!=N?
							'images/icons/link_drawing.png':'images/icons/link_drawing_grayed.png'
						)+'" />'
				:'')+
			'</p>'+
		'</div>');
	
	$("img.icon",oInput).tooltip({track:T,delay:0,showURL:F,extraClass:"tooltip"})
		/*.hover(mccl.MsgIconHoverIn,mccl.MsgIconHoverOut)*/.data("link_opts",link_opts)
		.data("ta",$('textarea',oInput));
	
	$("img.icon.extent_unlink",oInput).bind("click",{map:map},mccl.ExtentUnlinkClick);
	$("img.icon.select_unlink",oInput).bind("click",{map:map},mccl.SelectUnlinkClick);
	$("img.icon.draw_unlink",oInput).bind("click",{map:map},mccl.DrawUnlinkClick);
	$("img.icon.extent_link",oInput).bind("click",{map:map},mccl.ExtentLinkClick);
	$("img.icon.select_link",oInput).bind("click",{SelectLib:SelectLib},mccl.SelectLinkClick)
		.bind("UpdateSelection",{SelectLib:SelectLib},mccl.SelectLinkUpdate).trigger("UpdateSelection");
	$("img.icon.draw_link",oInput).bind("click",{DrawLib:DrawLib},mccl.DrawLinkClick)
		.bind("DrawUpdate",{DrawLib:DrawLib},mccl.DrawLinkUpdate)
		.bind("DrawingCleared",{DrawLib:DrawLib},mccl.DrawLinkUpdate).trigger("DrawUpdate");
	
	// Don't include these objects with the message data (plus json-encoding them will cause errors):
	if (msg_opts.extent) delete msg_opts.extent;
	if (msg_opts.selection) delete msg_opts.selection;
	if (msg_opts.drawing) delete msg_opts.drawing;
	if (msg_opts.map) delete msg_opts.map;
	if (msg_opts.SelectLib) delete msg_opts.SelectLib;
	if (msg_opts.DrawLib) delete msg_opts.DrawLib;
	
	MapChat.KeyProtect($("textarea",oInput));
	
	if (opts.cls) oInput.addClass(opts.cls);
	if (opts.css) oInput.css(opts.css);
	
	obj.text("").hide().after(oInput);
	
	return $('textarea',oInput).blur(function(){
		var ta = $(this);
		ta.data("leaveTO",setTimeout(function(){
			if (ta.val()!="") return; // Don't hide the dialog if there's text in it.
			oInput.remove();
			if (opts.close) opts.close(F);
		},200));
	}).focus(function(){
		var ta = $(this);
		if (ta.data("leaveTO"))
		{
			clearTimeout(ta.data("leaveTO"));
			ta.removeData("leaveTO");
		}
	}).keypress(function(e){
		switch(e.keyCode)
		{
			case 13:
				if (e.ctrlKey||e.metaKey||e.shiftKey) return;
				// TODO: send the message;
				msg_opts.text = $("textarea",oInput).val();
				
				$.extend(msg_opts,link_opts);
				
				// If set, the 'extent' should be a pointer to a map object - we just want the current extents as an array:
				if (msg_opts.extent) msg_opts.extent = msg_opts.extent.getGeoExtents();
				
				// Flag current selection/drawing as having been linked.
				if (msg_opts.selection && SelectLib) SelectLib.hasLinked = T;
				if (msg_opts.drawing && DrawLib) DrawLib.hasLinked = T;
				
				// Add the list of maps that will have events triggered on them if changes to user drawings occur.
				var maps = [];
				$.each(t.maps,function(mi,map){
					maps.push(map.currentMap);
				});
				msg_opts.maps = maps;
				msg_opts.timeDelta = MapChat.timeDelta;
				
				var msg = $.parseJSON($("chat_edit:first",MapChat.RunFunction("Chat_Edit",{args:{opts:$.toJSON(msg_opts)}})).text());
				
				// The above is rendered moot by this, but keeping selections/drawings after linking them can
				// be problematic, where people don't notice that they already have a selection or drawing, and
				// they keep adding to the selection/drawing without paying attention.
				if (msg_opts.selection && SelectLib) SelectLib.GetSelection({mode:'clear'});
				if (msg_opts.drawing && DrawLib) DrawLib.Clear();
				
				t.msgData.children('.chat_id-'+msg_opts.id).data("select_svg_new",T).data("draw_svg_new",T);
				
				t.context.filter('.edit-listener').trigger("EditedMessages",[[md.id]]);
				
				oInput.remove();
				obj.show();
				if (opts.close) opts.close(T,msg);
				return F;
				break;
			
			case 27:
				mcqe(9,"ChatLib",["CancelEditMessage",msg_opts.id]);
				oInput.remove();
				obj.text(md.text).show();
				if (opts.close) opts.close(F);
				break;
		}
		e.stopPropagation();
	}).focus();
}

/**
 * Produces a jQuery element or HTML text displaying the contents/controls for
 *   a chat message, which can be appended to any element that has been added
 *   to the current instance of ChatLib's context (see mccl.prototype.AddContext)
 *
 * @param mixed m - either the ID of a message in the msgData object,
 *                  or an actual jQuery object of a message selected from
 *                  the msgData object.
 * @param opts - options that control the output of this function.  Possible values include:
 *   boolean showReplies - indicates whether to add a replies section.
 *   boolean html - if true, the element is returned as only html
 *   boolean recurse - if true, when building html, recursively add html for any replies.
 *   boolean collapsed - if true, the message will have the chat_message_collapsed class added.
 */
mccl.prototype.DisplayMessage = function(m,opts)
{
	var t = this,
		opts = opts||{},
		showReplies = opts.showReplies||F,
		html = opts.html||F,
		recurse = opts.recurse||F,
		collapsed = opts.collapsed||F,
		rhtml = '';
	
	if (typeof(m)!="object") m = t.msgData.children(".chat_id-"+m);
	
	if (m.length>0)
	{
		md = m.data("md");
		
		// Make the HTML content for the message:
		var dm_html = '<div chat_id="'+md.id+'" class="chat_message chat_new_elem '+
				(showReplies?'chat_show_replies ':'')+
				((showReplies && collapsed)?'chat_message_collapsed ':'')+
				m.attr("class")+
			'">'+
				'<p class="chat_user">'+
					(showReplies?'<button class="chat_threadbutton">&nbsp;</button>':'')+
	// 				'<img class="chat_user_icon icon" alt="'+MapChat.UserInfo(md.username).data("u").username+
	// 					'" src="images/icons/user_'+(MapChat.UserInfo(md.username).data("u").online?'online':'offline')+'.png" />'+
					'<a class="chat_user_icon icon '+(MapChat.UserInfo(md.username).data("u").online?'':'offline')+'" name="'+md.username+'" role="button" unselectable="on" style="-moz-user-select: none;">'+
						'<span class="ui-icon">&nbsp;</span>'+
					'</a>'+
					'<strong>'+MapChat.UserInfo(md.username).data("u").display_name+'</strong>: '+
					'<span class="display_date">'+md.display_date+'</span>'+
				'</p>'+
				'<p class="chat_message_options">'+
					(md.extent?
						'<img class="chat_link_icon icon png24 link_extent bordered add_tooltip" src="images/icons/link_extent.png" svg_mode="extent" title="Zoom to Linked Extent" />'
					:'')+
					(md.selection?
						'<img class="chat_link_icon icon png24 link_select bordered add_tooltip" src="images/icons/link_selection.png" svg_mode="select" title="Zoom to Selected Features" />'
					:'')+
					(md.drawing?
						'<img class="chat_link_icon icon png24 link_draw bordered add_tooltip" src="images/icons/link_drawing.png" svg_mode="draw" title="Zoom to Drawing" />'
					:'')+
					(md.username!=MapChat.session.user_identity?"":
						'<button class="ui-state-default ui-corner-all" name="chat_edit_button">Edit</button>'
					)+
					'<button class="ui-state-default ui-corner-all" name="chat_reply_button">Reply</button><img class="icon_height_preserve" src="images/a_pixel.gif" />'+
				'</p>'+
				'<p class="chat_text'+(md.parent_id!=N?'':' chat_text_newtopic')+'">'+
					md.text.replace(/</g,'&lt;').replace(/>/g,'&gt;')+
				'</p>'+
				(showReplies?
					'<p class="chat_replies hidden"><button class="chat_repliesbutton">&nbsp;</button>'+
						'Show<span class="reply_ing">ing</span> <span class="chat_replies_shown">0</span> of <span class="chat_replies_total">'+md.replies.length+'</span> replies '+
						'<span class="more_replies_links">(get <span class="more_replies link">more</span>|<span class="all_replies link">all</span>)</span>'+
					'</p>'+
					((html&&recurse&&($.each(md.replies,function(i,mi){rhtml+=t.DisplayMessage(mi,{showReplies:T,html:T,recurse:T})||'';})))?rhtml:'')
				:"")+
			'</div>';
		if (!html)
		{
			return $(dm_html).data("md",md);
		}
		else return dm_html;
	}
}

/**
 * Binds events to message elements
 *
 * @param string f - a class to limit the binds to (e.g., new elements only)
 * @param boolean r - if f is specified, and r is true, the class name f will be removed from the elements
 * @param object libs - should contain members named map, SelectLib, and DrawLib, that point to the 
 *                      corresponding objects.
 */
mccl.prototype.BindMessageElems = function(f,r,libs)
{
	var t = this, elems = N;
	
	if (f)
		elems = $('div.chat_message.'+f,t.context);
	else
		elems = $('div.chat_message',t.context);
	
	// Add the 'md' and 'libs' data properties, then find any messages with the 
	// 'chat_show_replies' class that have replies, and trigger the NewReplies event.
	elems.each(function(){
			$(this).data("md",t.msgData.children('.chat_id-'+$(this).attr('chat_id')).data("md"))
				.data("libs",libs);
		}).filter(function(){
			var dm = $(this);
			if ((dm.hasClass("chat_show_replies") && 
				dm.data("md").replies.length>0))
			{
				dm.children('p.chat_replies').removeClass('hidden');
				return T;
			}
		}).trigger("NewReplies");
	
	if (f && r) elems.removeClass(f);
}

/**
 * Adds a map that will be updated when map-related events trigger changes (e.g., new user drawings):
 *
 * @param kaMap m - a kaMap map object.
 */
mccl.prototype.AddMap = function(m)
{
	var t = this;
	if ($.inArray(m,t.maps)==-1) t.maps.push(m);
	if (!t.defaultMap) t.defaultMap = m;
}

/**
 * Drops a map from the list of map objects to update for map-related events:
 *
 * @param kaMap m - the kaMap object that should be dropped.
 */
mccl.prototype.DropMap = function(m)
{
	var t = this, ia = $.inArray(m,t.maps);
	if (ia!=-1) t.maps.splice(ia,1);
	if (t.defaultMap == m) t.defaultMap = t.maps[0];
}

/**
 * Adds contexts within which elements will be search when triggering events:
 *
 * @param jQuery c - a jQuery object with the desired context selected.
 * @param object o - the class object that the elements belong to (e.g., an instance of MapChat.Chat.Dialog)
 * @param string cl - additional classes to add to the context object (e.g., message-listener, private_listener)
 * @param string rcl - classes to remove (i.e., one or more of the default draw-listener, select-listener, or chat-svg-listener classes);
 */
mccl.prototype.AddContext = function(c,o,cl,rcl)
{
	var t = this;
	
	// Make sure the object listens to desired events:
	c.addClass("draw-listener select-listener chat-svg-listener"+" "+(cl||''));
	if (rcl) c.removeClass(rcl);
	
	t.SetupDelegation(c);
	
	if (o)
	{
		// If a parent object is provided, we'll assume it has a displayCanvas and extentOpts/extentFill, etc. (needed by ShowSVG):
		c.bind("ShowSVG.ChatLib",{t:o},MapChat.ChatLib.ShowSVG)
			.bind("HideSVG.ChatLib",{t:o},MapChat.ChatLib.HideSVG);
	}
	
	t.context = $($.unique($.merge(t.context.get(),c.get())));
	
}

/**
 * Drops element contexts from the chatlib:
 *
 * @param jQuery c - the jQuery object that should be dropped.
 */
mccl.prototype.DropContext = function(c)
{
	var t = this, ia = $.inArray(c.get(0),t.context.get());
	if (ia!=-1)
	{
		t.RemoveDelegation(c);
		t.context = $(t.context.get().splice(ia,1));
	}
}

/**
 * Adds flags to the globalFlags array.
 *
 * @param string|array f - the flag or flags to add
 */
mccl.prototype.AddGlobalFlag = function(f)
{
	var t = this;
	if (typeof(f)=="string") f = [f];
	$.each(f,function(i,v){
		if ($.inArray(f,t.globalFlags)==-1 && v!=N) t.globalFlags.push(v);
	});
}

/**
 * Removes flags from the globalFlags array.
 *
 * @param string|array f - the flag or flags to drop
 */
mccl.prototype.DropGlobalFlag = function(f)
{
	var t = this;
	if (typeof(f)=="string") f = [f];
	$.each(f,function(i,v){
		var idx = $.inArray(v,t.globalFlags);
		if (idx!=-1) t.globalFlags.splice(idx,1);
	});
}

/**
 * Returns user information - if not downloaded previously for *this* discussion, then
 *  re-download it (even if the main MapChat.UserInfo has already been called for this user
 *  outside the context of the current discussion)
 *
 * @param string u - the username to get information for
 */
mccl.prototype.UserInfo = function(u)
{
	var t = this;
	if (t.uInfo[u]==N)
	{
		t.uInfo[u] = MapChat.UserInfo(u,T);
	}
	return t.uInfo[u];
}

/**
 * Binds events to message element containers, making container's events preform event delegation
 *  for the child messages.  This should remove the need to bind events to individual message elements.
 *
 * @param jQuery o - the container object to apply the event delegation to.
 */
mccl.prototype.SetupDelegation = function(o)
{
	var t = this;
	
	// Delegation for the click event:
	o.bind("click.ChatLib",{t:t},function(e,opts){
		var et = $(e.target);
		switch (true)
		{
			// Delegation for reply buttons.
			case (et.attr("name")=="chat_reply_button"):
				mccl.ReplyButtonClick.apply(et,[e]);
			break;
			
			// Delegation for edit buttons.
			case (et.attr("name")=="chat_edit_button"):
				mccl.EditButtonClick.apply(et,[e]);
			break;
			
			// Delegation for the more_replies link:
			case (et.hasClass('more_replies')):
				// Limit to the first 25 non-cached messages:
				var dm = et.parents('.chat_message:first'), 
					md = dm.data('md'),getids=[];
				$.each(md.replies,function(i,id){
					if (dm.children('div.chat_message.chat_id-'+id).length==0)
						getids.push(id);
					if (getids.length==25) return F;
				});
				t.Get({describe:getids});
			break;
			
			// Delecation for the all_replies link:
			case (et.hasClass('all_replies')):
				// Get all non-cached messages:
				var dm = et.parents('.chat_message:first'), 
					md = dm.data('md'),getids=[];
				$.each(md.replies,function(i,id){
					if (dm.children('div.chat_message.chat_id-'+id).length==0)
						getids.push(id);
				});
				t.Get({describe:getids});
			break;
			
			// Delegations for spatial link icons
			case (et.hasClass('chat_link_icon')):
				switch (true)
				{
					case (et.hasClass('link_extent')):
						mccl.ZoomExtentLink.apply(et,[e,opts]);
					break;
					
					case (et.hasClass('link_select')):
						mccl.ZoomSelectionLink.apply(et,[e,opts]);
					break;
					
					case (et.hasClass('link_draw')):
						mccl.ZoomDrawingLink.apply(et,[e,opts]);
					break;
				}
			break;
			
			case (et.hasClass('fake_parent_link')):
				var id = et.attr("chat_id");
				mcqe(9,"ChatDialog",["FakeParentClicked",id]);
				if (t.msgData.children('.chat_id-'+id).length==0)
				{
					t.Get({describe:[id]},{ignore:T});
				}
				// This is for chat_messages elements that load only parts of the locally cached discussion (e.g., the map icon callouts):
				else
				{
					t.context.filter('.message-listener').trigger("FakeMessagesReceived",[[id]]);
				}
			break;
			
			case (et.hasClass('chat_threadbutton')):
				et.parents('.chat_message:first').toggleClass('chat_message_collapsed');
			break;
			
			case (et.hasClass('chat_repliesbutton')):
				et.parents('.chat_message:first').toggleClass('chat_replies_collapsed');
			break;
			
			default:
				MapChat.Debug("Unmatched Click Target",et);
			break;
		}
		e.stopPropagation();
	});
	
	// Delegation for the UserOnline event:
	o.bind("UserOnline.ChatLib",{t:t},function(e){
		var et = $(e.target);
		switch (true)
		{
			case (et.hasClass('chat_user_icon')):
				mccl.UserOnline.apply(et,[e]);
			break;
		}
		e.stopPropagation();
	});
	
	// Delegation for the UserOffline event:
	o.bind("UserOffline.ChatLib",{t:t},function(e){
		var et = $(e.target);
		switch (true)
		{
			case (et.hasClass('chat_user_icon')):
				mccl.UserOffline.apply(et,[e]);
			break;
		}
		e.stopPropagation();
	});
	
	// Delegation for the MessageChanged event:
	o.bind("MessageChanged.ChatLib",{t:t},function(e){
		var et = $(e.target);
		if (et.hasClass("chat_message")) mccl.MessageChanged.apply(et,[e]);
		e.stopPropagation();
	});
	
	// Delegation for NewReplies event:
	o.bind("NewReplies.ChatLib",function(e,r){
		var et = $(e.target);
		if (et.hasClass("chat_message")) mccl.NewReplies.apply(et,[e,r]);
		e.stopPropagation();
	});
	
	// Delegation for mouseover event:
	o.bind("mouseover.ChatLib",function(e,r){
		var et = $(e.target);
		switch (true)
		{
			// Bind stuff to a freshly loaded icon (generated by the DisplayMessage function).  This
			// avoids having to load all of these bindings on initialization.
			case (et.hasClass('add_tooltip')):
				et.tooltip({track:T,delay:0,showURL:F,extraClass:"tooltip"})
					.bind('mouseenter',{t:t},mccl.MsgLinkIconHoverIn)
					.bind('mouseleave',{t:t},mccl.MsgLinkIconHoverOut)
					.removeClass('add_tooltip');
				try { et.mouseenter().mouseover(); }
				catch (err) { /* IE fun */ }
			break;
			
			case (et.hasClass('ui-state-default')):
				et.addClass('ui-state-hover');
			break;
		}
	});
	
	// Delegation for mouseout event:
	o.bind("mouseout.ChatLib",function(e,r){
		var et = $(e.target);
		switch (true)
		{
			case (et.hasClass('ui-state-hover')):
				et.removeClass('ui-state-hover');
			break;
		}
	});
	
	// Delegation for the scrollto event:
	o.bind("ScrollTo.ChatLib",function(e,s){
		var et = $(e.target);
		if (et.hasClass("chat_message")) mccl.ScrollTo(et,s);
	});
}

/**
 * Unbinds event delegations that were bound to the container in SetupDelegation.
 *
 * @param jQuery o - the container object to remove the event delegation from.
 */
mccl.prototype.RemoveDelegation = function(o)
{
	o.unbind(".ChatLib")
		.unbind("UserOnline.ChatLib")
		.unbind("UserOffline.ChatLib")
		.unbind("MessageChanged.ChatLib")
		.unbind("NewReplies.ChatLib")
		.unbind("mouseover.ChatLib")
		.unbind("mouseout.ChatLib");
}

/**
 * Loads all messages currently stored in the chatlib at once.
 *
 * @param boolean thr - if true, the html is built in threaded form recursively
 * @param boolean fakes - if true, include replies to parents that are not loaded yet...
 * @param jQuery msgData - if set, this will be used as the source of message data, in place of t.msgData.children.
 */
mccl.prototype.DisplayAllMessages = function(thr,fakes,msgData)
{
	var t = this,
		html = '',
		afakes = [],
		appended = [],
		msgs = msgData||t.msgData.children(':not(.chat_private)');
	
	if (thr)
	{
		var elems = msgs.filter(function(){
			var msg = $(this);
			return (msg.hasClass('chat_pid-null') || msg.hasClass('chat_pid-') || (fakes && msgs.filter('.chat_id-'+msg.data("md").parent_id).length==0));
		});
		elems.each(function(){
			var msg = $(this), md = msg.data("md");
			if ($.inArray(md.parent_id,afakes)==-1 && !(msg.hasClass('chat_pid-') || msg.hasClass('chat_pid-undefined') || msg.hasClass('chat_pid-null')))
			{
				afakes.push(md.parent_id);
				html += '<div chat_id="'+md.parent_id+'" class="chat_fake_message chat_fake_id-'+md.parent_id+'">'+
					'<p class="chat_fake_link">Replies to previous message '+
						'<strong><span class="fake_parent_link" chat_id="'+md.parent_id+'" style="cursor:pointer">#'+md.parent_id+'</span></strong>'+
					'</p>';
					
				elems.filter('.chat_pid-'+md.parent_id).each(function(){
					appended.push($(this).data("md").id);
					html += t.DisplayMessage($(this),{showReplies:T,html:T,recurse:T});
				});
				
				html += '</div>';
			}
			else if ($.inArray(md.id,appended)==-1)
			{
				html += t.DisplayMessage(msg,{showReplies:T,html:T,recurse:T});
			}
		});
	}
	else
	{
		var elems = msgs;
		elems.each(function(){
			html += t.DisplayMessage($(this),{showReplies:F,html:T});
		});
	}
	
	return html;
}

/**
 * Returns a private chat dialog for conversing with specific target users/groups.
 *
 * @param object targets - an associative array with 
 * @param object map - a kaMap object that will be used for displaying linked shapes, 
 *                     etc.  If null/not specified, the ChatLib's defaultMap will be used.
 */
mccl.prototype.GetPrivateDialog = function(targets,map,opts)
{
	if (targets.target_users) targets.target_users.sort();
	if (targets.target_groups) targets.target_groups.sort();
	
	var t = this, tid = $.toJSON(targets);
	
	if (t.privateDialogs[tid])
		t.privateDialogs[tid];
	else
		t.privateDialogs[tid] =
			MapChat.Class(["MapChat.Chat.PrivateDialog",["ui.dialog"]],(map||t.defaultMap),t,{targets:targets});
	
	return t.privateDialogs[tid];
}

// Functions that bound to various events...defined here so that it doesn't happen every time
//  an element is bound to the same function, rather than defining a new anonymous function
//  each time:

// For the chat user icons:
mccl.UserOnline = function(){
	$(this).filter('img').attr("src","images/icons/user_online.png");
	$(this).filter('a').removeClass('offline');
};
mccl.UserOffline = function(){
	$(this).filter('img').attr("src","images/icons/user_offline.png");
	$(this).filter('a').addClass("offline");
};

// For the message div elements:
mccl.MessageChanged = function(e)
{
	var t = e.data.t,
		dm = $(this),
		md = dm.data("md"),
		cmo = dm.children('p.chat_message_options'),
		b = cmo.children('button:first');
	
	cmo.children('img.icon').remove();
	
	dm.children("p.chat_text").text(md.text);
	
	if (md.extent)
		b.before('<img class="chat_link_icon icon png24 link_extent bordered" src="images/icons/link_extent.png" svg_mode="extent" title="Zoom to Linked Extent" />');
	
	if (md.selection)
		b.before('<img class="chat_link_icon icon png24 link_select bordered" src="images/icons/link_selection.png" svg_mode="select" title="Zoom to Selected Features" />');
	
	if (md.drawing)
		b.before('<img class="chat_link_icon icon png24 link_draw bordered" src="images/icons/link_drawing.png" svg_mode="draw" title="Zoom to Drawing" />');
	
	cmo.children('img.icon').unbind()
		.tooltip({track:T,delay:0,showURL:F,extraClass:"tooltip"})
		.bind('mouseenter',{t:t},mccl.MsgLinkIconHoverIn)
		.bind('mouseleave',{t:t},mccl.MsgLinkIconHoverOut);
	
}

mccl.NewReplies = function(e,r)
{
	var dm = $(this);
	if (r) $.each(r,function(i,v){ if ($.inArray(v,dm.data("md").replies)==-1) dm.data("md").replies.push(v); });
	$(".chat_replies_shown",dm.children(".chat_replies")).text(dm.children("div.chat_message").length);
	$(".chat_replies_total",dm.children(".chat_replies")).text(dm.data("md").replies.length);
	if (dm.data("md").replies.length>0) dm.children(".chat_replies").removeClass('hidden');
	if (dm.data("md").replies.length<=dm.children('div.chat_message').length) 
		dm.children('.chat_replies').children('.more_replies_links').addClass('hidden');
	else
		dm.children('.chat_replies').children('.more_replies_links').removeClass('hidden');
}

mccl.EditButtonClick = function(e)
{
	var eb = $(this),
		dm = eb.parents('div.chat_message:first'),
		md = dm.data('md'),
		context = e.data.t.context;
	
	if (!eb.data("EditMessage"))
	{
		mcqe(9,"ChatLib",["EditMessage",md.id]);
		
		eb.data("EditMessage",
		e.data.t.EditMessage({
			obj:dm.children("p.chat_text"),
			md:md,
			map:dm.data('libs').map,
			SelectLib:dm.data('libs').SelectLib,
			DrawLib:dm.data('libs').DrawLib,
			close:function(e,msg)
			{
				eb.removeData("EditMessage");
				if (msg)
				{
					$.extend(md,msg[md.id]);
					
					// Fix any array attributes that are still encoded as strings...
					if (!md.replies||md.replies=="") md.replies = [];
					else md.replies = $.parseJSON(md.replies);
					$.each(md.replies,function(i){md.replies[i] = parseInt(md.replies[i]);});
					
					mccl.FixMsgExtents(md);
					
					$(".chat_id-"+md.id,context).trigger("MessageChanged");
				}
			}
		}));
	}
	dm.trigger("ScrollTo",[md.id]);
}

mccl.ReplyButtonClick = function(e)
{
	// Display an input area above the new topic button:
	var rb = $(this), dm = rb.parents('div.chat_message:first'),
		md = dm.data("md");
	
	if (!rb.data("NewMessage"))
	{
		mcqe(9,"ChatLib",["ReplyToMessage",md.id]);
		
		var chat_opts = {parentID:md.id};
		if (md.private)
		{
			chat_opts.private = {};
			var tu = md.target_users,
				tg = md.target_groups;
			
			// Build the targetid for the dialog, then trigger the private listener event (so this message appears in the dialog)
			if (tu) chat_opts.private.target_users = (typeof(tu)=="string"?$.parseJSON(tu):tu);
			if (tg) chat_opts.private.target_groups = (typeof(tg)=="string"?$.parseJSON(tg):tg);
		}
		
		rb.data("NewMessage",
		e.data.t.NewMessage({
			chat_opts:chat_opts,
			obj:rb.parent().next(),
			eflags:MapChat.UniqueArray([].concat(e.data.t.globalFlags,dm.data("eflags"))),
			map:dm.data('libs').map,
			SelectLib:dm.data('libs').SelectLib,
			DrawLib:dm.data('libs').DrawLib,
			append:'sibling',
			above:F,
			close:function(e,msg)
			{
				rb.removeData("NewMessage");
				if (msg) dm.parents('.chat_messages:first').trigger("ScrollTo",[msg.data("md").id]);
			}
		}));
	}
	
	var cm = dm.parents('.chat_messages:first'), ci = dm.children('.chat_input');
	if (ci.offset().top<cm.offset().top)
		cm.scrollTo(ci,300);
	else if ((ci.offset().top+ci.outerHeight())>(cm.offset().top+cm.height()))
		cm.scrollTo(ci,300,{offset:{left:0,top:ci.outerHeight()-cm.height()}});
}

mccl.MsgLinkIconHoverIn = function(e)
{
	var t = e.data.t,
		ic = $(this),
		msg = ic.parents('div.chat_message:first'),
		c = msg.parents('.chat-svg-listener:first'),
		mode = ic.attr("svg_mode"),
		HideSVGTO = msg.data("HideSVGTO");
	msg.removeData("HideSVGTO");
	if (HideSVGTO) clearTimeout(HideSVGTO);
	if (msg.data(mode+"_svg") || mode=="extent")
		c.trigger("ShowSVG",[msg,mode,T,T]);
}

mccl.MsgLinkIconHoverOut = function(e)
{
	var t = e.data.t,
		ic = $(this),
		msg = ic.parents('div.chat_message:first'),
		c = msg.parents('.chat-svg-listener:first'),
		mode = ic.attr("svg_mode");
	if (msg.data(mode+"_svg") || mode=="extent") 
		c.trigger("HideSVG",[msg,mode,T]);
}

mccl.ExtentUnlinkClick = function(e)
{
	var ic = $(this);
	if (ic.data("checked")) ic.data("checked",F).removeClass('active').data("link_opts").unlink_extent = N;
	else ic.data("checked",T).addClass('active').data("link_opts").unlink_extent = T;
	ic.data("ta").focus();
	return F;
}

mccl.SelectUnlinkClick = function(e)
{
	var ic = $(this);
	if (ic.data("checked")) ic.data("checked",F).removeClass('active').data("link_opts").unlink_selection = N;
	else ic.data("checked",T).addClass('active').data("link_opts").unlink_selection = T;
	ic.data("ta").focus();;
	return F;
}

mccl.DrawUnlinkClick = function(e)
{
	var ic = $(this);
	if (ic.data("checked")) ic.data("checked",F).removeClass('active').data("link_opts").unlink_drawing = N;
	else ic.data("checked",T).addClass('active').data("link_opts").unlink_drawing = T;
	ic.data("ta").focus();
	return F;
}

mccl.ExtentLinkClick = function(e)
{
	var ic = $(this);
	if (ic.data("checked")) ic.data("checked",F).removeClass('active').data("link_opts").extent = F;
	else if (!ic.hasClass("disabled")) ic.data("checked",T).addClass('active').data("link_opts").extent = e.data.map;
	ic.data("ta").focus();
	return F;
}

mccl.SelectLinkClick = function(e)
{
	var ic = $(this);
	if (ic.data("checked")) ic.data("checked",F).removeClass('active').data("link_opts").selection = N;
	else if (!ic.hasClass("disabled")) ic.data("checked",T).addClass('active').data("link_opts").selection = e.data.SelectLib.selectIDs;
	ic.data("ta").focus();;
	return F;
}

mccl.DrawLinkClick = function(e)
{
	var ic = $(this);
	if (ic.data("checked")) ic.data("checked",F).removeClass('active').data("link_opts").drawing = N;
	else if (!ic.hasClass("disabled")) ic.data("checked",T).addClass('active').data("link_opts").drawing = e.data.DrawLib.geomID;
	ic.data("ta").focus();
	return F;
}

mccl.SelectLinkUpdate = function(e)
{
	var ic = $(this), SelectLib = e.data.SelectLib;
	if (SelectLib.count>0)
	{
		// Disabling the default behaviour of automatically linking new selections:
		SelectLib.hasLinked = T;
		
		ic.removeClass("disabled").attr("src","images/icons/link_selection.png");
		if (SelectLib.hasLinked && !ic.data("checked"))
			ic.removeClass('active')
				.data("link_opts").selection = N;
		else
			ic.addClass('active')
				.data("checked",T).data("link_opts").selection = SelectLib.selectIDs;
	}
	else
		ic.addClass("disabled").attr("src","images/icons/link_selection_grayed.png").data("checked",F)
			.removeClass('active')
			.data("link_opts").selection = N;
}

mccl.DrawLinkUpdate = function(e)
{
	var ic = $(this), DrawLib = e.data.DrawLib;
	if (DrawLib.geomID!=N)
	{
		// Disabling the default behaviour of automatically linking new selections:
		DrawLib.hasLinked = T;
		
		ic.removeClass("disabled").attr("src","images/icons/link_drawing.png");
		if (DrawLib.hasLinked && !ic.data("checked"))
			ic.removeClass('active')
				.data("link_opts").drawing = N;
		else
			ic.addClass('active')
				.data("checked",T).data("link_opts").drawing = DrawLib.geomID;
	}
	else
		ic.addClass("disabled").attr("src","images/icons/link_drawing_grayed.png").data("checked",F)
			.removeClass('active')
			.data("link_opts").drawing = N;;
}

mccl.ZoomExtentLink = function(e,nz)
{
	var msg = $(this).parents('div.chat_message:first'),
		md = msg.data("md"), map = msg.data('libs').map;
	
	mcqe(9,"ChatLib",["ZoomExtentLink",md.id]);
	
	if (!nz && map && md.extent)
	{
		var geoxy = map.getGeoExtents();
		if ((md.extent[2]-md.extent[0])>(geoxy[2]-geoxy[0])||(md.extent[3]-md.extent[1])>(geoxy[3]-geoxy[1]))
			map.zoomToExtents.apply(map,md.extent);
		else
			map.zoomTo((md.extent[0]+md.extent[2])/2,(md.extent[1]+md.extent[3])/2);
	}
}

mccl.ZoomSelectionLink = function(e,nz)
{
	var msg = $(this).parents('div.chat_message:first'),
		md = msg.data("md"), map = msg.data('libs').map;
	
	mcqe(9,"ChatLib",["ZoomSelectionLink",md.id]);
	
	if (!nz && map && md.selection)
	{
		var geoxy = map.getGeoExtents();
		if (md.selection.extent.length>2 && (md.selection.extent[2]-md.selection.extent[0])>(geoxy[2]-geoxy[0])||
			(md.selection.extent[3]-md.selection.extent[1])>(geoxy[3]-geoxy[1]))
			map.zoomToExtents.apply(map,md.selection.extent);
		else
			map.zoomTo((md.selection.extent[0]+md.selection.extent[2])/2,(md.selection.extent[1]+md.selection.extent[3])/2);
	}
	
	// If there is an 'new' flag, this means something has updated the message...so the geom could be obsolete:
	if (msg.data("select_svg_new")==T || !msg.data("select_svg"))
	{
		msg.data("select_svg_new",F);
		if (msg.data("select_geom")) msg.data("select_geom").drop();
		if (msg.data("select_line")) msg.data("select_line").drop();
		if (msg.data("select_point")) msg.data("select_point").drop();
		if (msg.data("select_simple_geom")) msg.data("select_simple_geom").drop();
		msg.removeData("select_svg").removeData("select_geom").removeData("select_line")
			.removeData("select_point").removeData("select_simple_geom");
		MapChat.RunFunction("Chat_Svg",{args:{opts:$.toJSON({selection:md.id})}},{
			async:T,
			complete: function(xhr){
				msg.data("select_svg",$.parseJSON($("chat_svg:first",xhr.responseXML).text()).select_svg[md.id]);
				e.data.t.context.filter('.chat-svg-listener:not(.map-context)').trigger("ShowSVG",[msg,"select",T]);
			},
			error: function(xhr,sStatus,nError){
				MapChat.Debug(["error","Chat_Svg failed: ",xhr.responseText]);
			}
		});
	}
	else
	{
		e.data.t.context.filter('.chat-svg-listener:not(.map-context)').trigger("ShowSVG",[msg,"select",T]);
	}
}

mccl.ZoomDrawingLink = function(e,nz)
{
	var msg = $(this).parents('div.chat_message:first'),
		md = msg.data("md"), map = msg.data('libs').map;
	
	mcqe(9,"ChatLib",["ZoomDrawingLink",md.id]);
	
	if (!nz && map && md.drawing && md.drawing.extent)
	{
		var geoxy = map.getGeoExtents();
		if (md.drawing.extent.length>2 && (md.drawing.extent[2]-md.drawing.extent[0])>(geoxy[2]-geoxy[0])||
			(md.drawing.extent[3]-md.drawing.extent[1])>(geoxy[3]-geoxy[1]))
			map.zoomToExtents.apply(map,md.drawing.extent);
		else
			map.zoomTo((md.drawing.extent[0]+md.drawing.extent[2])/2,(md.drawing.extent[1]+md.drawing.extent[3])/2);
	}
	
	// If there is an 'new' flag, this means something has updated the message...so the geom could be obsolete:
	if (msg.data("draw_svg_new")==T || !msg.data("draw_svg"))
	{
		msg.data("draw_svg_new",F);
		if (msg.data("draw_geom")) msg.data("draw_geom").drop();
		if (msg.data("draw_line")) msg.data("draw_line").drop();
		if (msg.data("draw_point")) msg.data("draw_point").drop();
		if (msg.data("draw_simple_geom")) msg.data("draw_simple_geom").drop();
		msg.removeData("draw_svg").removeData("draw_geom").removeData("draw_line")
			.removeData("draw_point").removeData("draw_simple_geom");
		MapChat.RunFunction("Chat_Svg",{args:{opts:$.toJSON({drawing:md.id})}},{
			async:T,
			complete: function(xhr){
				msg.data("draw_svg",$.parseJSON($("chat_svg:first",xhr.responseXML).text()).draw_svg[md.id]);
				e.data.t.context.filter(".chat-svg-listener:not(.map-context)").trigger("ShowSVG",[msg,"draw",T]);
			},
			error: function(xhr,sStatus,nError){
				MapChat.Debug(["error","Chat_Svg failed: ",xhr.responseText]);
			}
		});
	}
	else
	{
		e.data.t.context.filter(".chat-svg-listener:not(.map-context)").trigger("ShowSVG",[msg,"draw",T]);
	}
}


mccl.ShowSVG = function(e,msg,mode,timeout,showsimple)
{
	if (!(msg.data(mode+"_geom")||msg.data(mode+"_point")||msg.data(mode+"_line")) && msg.data(mode+"_svg_new")!=T)
	{
		if (mode=="extent")
		{
			geom = e.data.t.displayCanvas.newGeom("path",e.data.t.extentFill,T);
			ex = msg.data("md").extent;
			geom.addPoint([ex[0],ex[1]]);
			geom.addPoint([ex[2],ex[1]]);
			geom.addPoint([ex[2],ex[3]]);
			geom.addPoint([ex[0],ex[3]]);
			geom.addPoint([ex[0],ex[1]]);
			geom.draw();
			msg.data("extent_geom",geom);
		}
		else
		{
			var svg = msg.data(mode+"_svg");
			// Need to deal with complex geometries:
			if (svg.type==3)
			{
				var svgs = svg.shape.split(";"),
					points = [], lines = [], polys = [];
				$.each(svgs,function(i,s){
					if (s.indexOf("cx")!=-1) points.push(s);
					else if (s.indexOf("Z")!=-1) polys.push(s);
					else lines.push(s);
				});
				
				var simple=F, geom = F, line = F, point = F;
				
				if (polys.length>0) geom = e.data.t.displayCanvas.addGeom(polys.join(";"),e.data.t[mode+'Fill']);
				if (lines.length>0) line = e.data.t.displayCanvas.addGeom(lines.join(";"),e.data.t[mode+'Opts']);
				if (points.length>0) point = e.data.t.displayCanvas.addGeom(points.join(","),e.data.t[mode+'Fill']);
				
				msg.data(mode+"_geom",geom).data(mode+"_line",line)
					.data(mode+"_point",point).data(mode+"_simple_geom",simple);
			}
			else
			{
				var simple = (
					svg.simple?
					e.data.t.displayCanvas.addGeom(svg.simple,$.extend((showsimple?{display:'none'}:{}),e.data.t[mode+'Fill']))
					:F
				),
				geom = e.data.t.displayCanvas.addGeom(svg.shape,svg.type!=1?e.data.t[mode+'Fill']:e.data.t[mode+'Opts']);
				msg.data(mode+"_geom",geom).data(mode+"_simple_geom",simple);
			}
		}
	}
	else if (mode=="extent" || msg.data(mode+"_svg_new")!=T)
	{
		var geom = msg.data(mode+"_geom"),
			line = msg.data(mode+"_line"),
			point = msg.data(mode+"_point"),
			simple = msg.data(mode+"_simple_geom");
		
		if (mode=="extent")
		{
			geom.xy = [];
			ex = msg.data("md").extent;
			geom.addPoint([ex[0],ex[1]]);
			geom.addPoint([ex[2],ex[1]]);
			geom.addPoint([ex[2],ex[3]]);
			geom.addPoint([ex[0],ex[3]]);
			geom.addPoint([ex[0],ex[1]]);
			geom.setStyle(e.data.t.extentFill);
		}
		else
		{
			if (msg.data(mode+"_svg").type==3)
			{
				if (geom) geom.setStyle(e.data.t[mode+'Fill']);
				if (line) line.setStyle(e.data.t[mode+'Opts']);
				if (point) point.setStyle(e.data.t[mode+'Fill']);
			}
			else geom.setStyle(msg.data(mode+"_svg").type!=1?e.data.t[mode+'Fill']:e.data.t[mode+'Opts']);
		}
		
		if (simple)
		{
			if (showsimple) simple.setStyle({display:'visible'});
			else simple.setStyle({display:'none'});
			simple.draw();
		}
		
		if (geom) geom.draw();
		if (line) line.draw();
		if (point) point.draw();
	}
	
	if (geom)
	{
		$(geom.o).hover(
			function(){
				if (simple) simple.setStyle({display:'visible'});
				$('div.chat_id-'+msg.data("md").id+' p.chat_text',e.data.t.tabChat).addClass("chat_highlight");
			},
			function(){
				if (simple) simple.setStyle({display:'none'});
				$('div.chat_id-'+msg.data("md").id+' p.chat_text',e.data.t.tabChat).removeClass("chat_highlight");
			}
		);
		
		if (timeout) msg.data("HideSVGTO",setTimeout(function(){mccl.HideSVG(e,msg,mode);},5000));
	}
	
	if (line)
	{
		$(line.o).hover(
			function(){
				$('div.chat_id-'+msg.data("md").id+' p.chat_text',e.data.t.tabChat).addClass("chat_highlight");
			},
			function(){
				$('div.chat_id-'+msg.data("md").id+' p.chat_text',e.data.t.tabChat).removeClass("chat_highlight");
			}
		);
		
		if (timeout) msg.data("HideSVGTO",setTimeout(function(){mccl.HideSVG(e,msg,mode);},5000));
	}
	
	if (point)
	{
		$(point.o).hover(
			function(){
				$('div.chat_id-'+msg.data("md").id+' p.chat_text',e.data.t.tabChat).addClass("chat_highlight");
			},
			function(){
				$('div.chat_id-'+msg.data("md").id+' p.chat_text',e.data.t.tabChat).removeClass("chat_highlight");
			}
		);
		
		if (timeout) msg.data("HideSVGTO",setTimeout(function(){mccl.HideSVG(e,msg,mode);},5000));
	}
}

mccl.HideSVG = function(e,msg,mode)
{
	var geom = msg.data(mode+"_geom"),
		line = msg.data(mode+"_line"),
		point = msg.data(mode+"_point"),
		simple = msg.data(mode+"_simple_geom");
	
	if (geom) geom.drop();
	if (line) line.drop();
	if (point) point.drop();
	if (simple) simple.drop();
}


/**
 * Sets the scroll height of the chat lists to display a particular message
 *
 * @param object dm - a jQuery element of the message (or the chat_message container itself).
 * @param mixed opts - either the id of a message to scroll to, or boolean True to scroll to the end (which
 *                     is for legacy compability with the original form of this function), or an object
 *                     array of parameters with the following possible values:
 *   mixed msgID - same as above, but must be either a valid message ID or a boolean value.
 *   boolean stick - if true, it will only scroll to the identified message if it appears that the
 *                   scrollable parent is at or near the bottom already.
 *   boolean threaded - if true, the message elements will be treated as though they are
 *                      threaded (regardless of the presence of the chat_show_replies class)
 */
mccl.ScrollTo = function(dm,opts)
{
	MapChat.Debug('ST',dm,opts);
	var t = this,
		o = (dm.hasClass("chat_message")?dm.parents(".chat_messages:first"):dm),
		msgID = (typeof(opts)=="object"?opts.msgID:opts)||(dm?dm.attr("chat_id"):N);
	
	if (dm.hasClass("chat_show_replies") || opts.threaded)
	{
		if (msgID===T) setTimeout(function(){o.scrollTo(o.children(':last'),300,{offset:{left:0,top:o.children(':last').outerHeight()}});},10);
		else if (msgID)
		{
			// Threaded message elements:
			var ctm = (dm.hasClass("chat_message")?dm:$(".chat_id-"+msgID,o));
				ctm_parent = ctm.parent(),
				ctm_prev = ctm.prev(),
				ctmpos = ctm.offset().top-o.children(":first").offset().top;
			
			// If the top of the message is out of view, then reposition the scroll positions of the threaded/list divs:
			if (ctmpos<o.scrollTop() || ctmpos>(o.scrollTop()+o.height()) || ctm.height()>o.height())
			{
				// If there is a previous sibling message, or the parent has already been loaded, go to the top of the current message
				if (ctm_prev.length>0 && ctm_prev.hasClass("chat_message") || !ctm_parent.hasClass("chat_fake_message"))
					setTimeout(function(){o.scrollTo(ctm,300);},10);
				
				// Otherwise, position relative to the top of the fake placeholder for the parent message:
				else setTimeout(function(){o.scrollTo(ctm_parent);},10);
			}
			
			// If the bottom of the message is out of view, then reposition so the bottom of the message lines up with the
			else if ((ctmpos+ctm.outerHeight())>(o.scrollTop()+o.height()))
			{
				if (!opts.stick || opts.stick && Math.abs((o.children(":last").offset().top+o.children(":last").outerHeight())-(o.scrollTop()+o.height())>5))
				{
					var ctm_bottom = ctmpos+ctm.outerHeight()+ctm.children('div.chat_message:first').offset().top;
					setTimeout(function(){o.scrollTo(ctm,300,{offset:{left:0,top:ctm.outerHeight()-o.height()}});},10);
				}
			}
		}
	}
	else
	{
		//if (cl.data("stick")===T) cl.scrollTop(cl.children(":last").offset().top+cl.children(":last").outerHeight(T));
		if (msgID===T)  setTimeout(function(){o.scrollTo(o.children(':last'),300,{offset:{left:0,top:o.children(':last').outerHeight()}});},10);
		else if (msgID)
		{
			// List message elements:
			var clm = (dm.hasClass("chat_message")?dm:$(".chat_id-"+msgID,o)),
				clmpos = clm.offset().top-o.children(":first").offset().top;
			
			if (clmpos<o.scrollTop())
				setTimeout(function(){o.scrollTo(clm);},10);
			else if ((clmpos+clm.outerHeight())>(o.scrollTop()+o.height()))
			{
				if (!opts.stick || opts.stick && Math.abs((o.children(":last").offset().top+o.children(":last").outerHeight())-(o.scrollTop()+o.height())>5))
				{
					setTimeout(function(){o.scrollTo(clm,300,{offset:{left:0,top:clm.outerHeight()-o.height()}});},10);
				}
			}
		}
	}
}

/**
 * Appends a message to a threaded view (also checks for fake placeholders, and/or newer messages)
 *
 * @param object msg - a jQuery object containing the displayed message element.
 * @param object o - a jQuery object containing the parent element that the message will be appended to.
 */
mccl.AppendThread = function(dm,o)
{
	var md = dm.data("md"),
		p = $("div.chat_id-"+md.parent_id, o);
	
	if (p.length==0) p = o;
	
	var c = (o.hasClass('chat_messages')?o:o.parents('.chat_messages:first')),
	
		// Make sure there isn't an existing 'fake' placeholder for this message:
		fm = $("div.chat_fake_id-"+md.id,c);
	
	if (fm.length!=0)
	{
		$("p.chat_fake_link",fm).remove();
		dm.append(fm.children());
		fm.remove();
 	}
		
	// Get the first message in 'p' that is newer than 'm':
	var newM = p.children().filter(function(){
		return ($(this).data("md") && $(this).data("md").id>md.id);
	}).filter(":first");
	
	// If there is a newer message, append this one before it...
	if (newM.length!=0)
		newM.before(dm);
	else
		p.append(dm);
	
	if (p.hasClass('chat_message')) p.trigger("NewReplies");
	
	// Append any existing replies (or fake placeholders) - this must be done in sequential order:
	if (md.replies && md.replies.length>0)
	{
		md.replies.sort();
		var replyfilter = [];
		
		$.each(md.replies,function(i,id){
			replyfilter.push('.chat_id-'+id+',.chat_fake_id-'+id);
		});
		
		dm.append($(replyfilter.join(","),p)).trigger("NewReplies");
	}
	
	// If this is a thread-child, but it's parent is the messages container (and not the actual 
	//   parent message), add a fake place holder for the parent message:
	if (md.parent_id!=N && md.parent_id>=0 && dm.parent().hasClass("chat_messages"))
	{
		var fp = $('<div chat_id="'+md.parent_id+'" class="chat_fake_message chat_fake_id-'+md.parent_id+'">'+
			'<p class="chat_fake_link">'+
				'Replies to previous message '+
				'<strong><span class="fake_parent_link" chat_id="'+md.parent_id+'" style="cursor:pointer">#'+
					md.parent_id+
				'</span></strong>:'+
			'</p>'+
		'</div>').data("md",{id:md.parent_id,depth:md.depth-1});
		
		newM = p.children().filter(function(){
			return ($(this).data("md") && $(this).data("md").id>md.parent_id);
		}).filter(":first");
	
		if (newM.length!=0)
			newM.before(fp);
		else
			p.append(fp);
		
		fp.append(dm);
	}
}

/**
 * Appends a message to a sequential view
 *
 * @param object msg - a jQuery object containing the displayed message element.
 * @param object o - a jQuery object containing the parent element that the message will be appended to.
 */
mccl.AppendList = function(dm,o)
{
	// Get the first message in 'p' that is newer than 'm'
	newM = o.children().filter(function(){
		return $(this).data("md").id>dm.data("md").id;
	}).filter(":first");
	
	// If there is a newer message, append this one before it...
	if (newM.length!=0)
		newM.before(dm);
	
	// Otherwise, add it to the end of the parent element:
	else
		o.append(dm);
}

/**
 * Get full user/group names for any specified targets.
 */
mccl.TargetNames = function(targets)
{
	var tgNames = {};
	if (targets.target_users)
	{
		if (!tgNames.target_users) tgNames.target_users = [];
		$.each(targets.target_users.sort(),function(i,u){
			if (u!=MapChat.session.user_identity)
				tgNames.target_users.push(MapChat.UserInfo(u).data("u").display_name);
		});
	}
	
	if (targets.target_groups)
	{
		if (!tgNames.target_groups) tgNames.target_groups = [];
		$.each(targets.target_groups.sort(),function(i,g){
			MapChat.Debug("warn","MapChat.GroupInfo not yet implemented");
			tgNames.target_groups.push(g);
			//tgNames.groups.push(MapChat.GroupInfo(g).data("g").description);
		});
	}
	
	return tgNames;
}


mccl.FixMsgExtents = function(md)
{
	md.extent = mccl.FixExtent(md.extent);
	md.selection = mccl.FixExtent(md.selection);
	md.drawing = mccl.FixExtent(md.drawing);
}

mccl.FixExtent = function(ext)
{
	if (typeof(ext)!="string") return ext;
	var s = $.parseJSON(ext);
	if (s.length>4)
		return {extent:[s[0],s[1],s[2],s[3]],centroid:s[4],point:s[5]};
	else
		return $.parseJSON(ext);
}