/**
 * A library for handling the features selected in a map object
 *
 * @class MapChat.DrawLib
 *
 * @param object map - the kaMap map object that features will be selected in.
 * @param object opts - options
 *
 * Depends:
 *	kamap/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 */
MapChat.DrawLib = function(map)
{
	var t = this,
		map = (t.map = map||MapChat.map),
		drawList = (t.drawList = {}),
		curGeom = (t.curGeom = N),
		geomType = (t.geomType = N),
		geomProps = (t.geomProps = N),
		geomID = (t.geomID = N),
		geomHistory = (t.geomHistory = []),
		count = (t.count = 0)
		hasLinked = F;
		
	// Check for any existing drawings to start off:
	t.GetDrawing({mode:'list'},T);
}

//shortcut to the MapChat.DrawLib class:
var mcdl = MapChat.DrawLib;

/**
 * Sends drawing options to the server, and parses the results.
 *
 * @param object opts - options for the Select_Feature API
 * @param boolean quiet - if true, don't trigger the listeners (i.e., when the lib is first loaded).
 *
 * Possible values that can be included in the opts parameter:
 *  mode: one of 'add', 'cut', 'open', 'list', 'save' - defines wither to add new coordinates to the current (or new) drawing,
 *        cut areas out of a drawing, or list the current selection
 *  style: one of 'point', 'circle', 'line', 'polygon' - defines how the drawing will be performed
 *  coords: the coordinates that are to be used to create/update a drawing
 *  props: properties to assign to a drawing
 *  id: this should be the id of the drawing to modify (if 'add' or 'cut') or return/set data for (if 'open' or 'save')
 */
mcdl.prototype.GetDrawing = function(opts,quiet)
{
	var t = this;
	$.extend(opts,{map:t.map.currentMap,cellSize:t.map.cellSize});
	
	// If we have a current ID set, then make sure it gets passed.  Otherwise either a new geometry will be 
	// created (if opts.id is null), or new properties/coordinates will be applied to another existing 
	// geometry at the specified ID - in the latter case, any current geometry will be replaced by data from
	// the one identified by opts.id.
	if (t.geomID && !opts.id) opts.id = t.geomID;
	
	MapChat.RunFunction("Draw_Features",{args:{opts:$.toJSON(opts)}},{
		async:T,
		complete:function(xhr){
			try
			{
				t.ParseDrawing(xhr,quiet);
			}
			catch (e)
			{
				MapChat.Debug(["error","Draw_Features: Error parsing AJAX response: ",e.message,e,xhr.responseText]);
				// not sure what else to do here:
				//alert("Sorry - a problem was encountered processing your drawing.");
				MapChat.QueueEvent(8,"SpatialTools",["DrawingError"]);
				//t.Clear();
				$(".draw-listener",t.map.context).trigger("DrawError");
				$.unblockUI();
			}
		},
		error: function(xhr,sStatus,nError){
			MapChat.Debug(["warn","Draw_Features: Error on AJAX request. Status: %s Error: %s ",sStatus,nError]);
			// not sure what else to do here:
			//alert("Sorry - a problem was encountered processing your drawing.");
			MapChat.QueueEvent(8,"SpatialTools",["DrawingError"]);
			//t.Clear();
			$(".draw-listener",t.map.context).trigger("DrawError");
			$.unblockUI();
		}
	});
}

/**
 * Parses a drawing result, and triggers events on the map's context elements
 * that are classed as draw-listener:
 */
mcdl.prototype.ParseDrawing = function(xhr,quiet)
{
	var t = this,
		sel = $("draw_results:first",xhr.responseXML),
		newProps = N,
		geomData = $.parseJSON(sel.text()),
		mode = sel.attr("mode");
	
	
	switch (mode)
	{
		case "list":
			t.drawList = geomData||{};
			$.each(t.drawList,function(i,d){
				if (d.label && typeof(d.label)=="string" && d.label.length>0)
					d.displayText = d.label+' ('+d.geom_type+' ID# '+d.id+')';
				else
					d.displayText = '[Unlabelled] ('+d.geom_type+' ID# '+d.id+')';
			});
			$(".draw-listener",t.map.context).trigger("DrawListUpdate");
		break;
		
		case "delete":
			delete t.drawList[geomData];
			t.hasLinked = F;
			if (geomData==t.geomID) t.Clear();
			$(".draw-listener",t.map.context).trigger("DrawListUpdate");
		break;
		
		default:
			
			if (geomData)
			{
				$.each(geomData,function(i,v){
					switch (i)
					{
						case "id":
							t.geomID = v;
						break;
						
						case "shape":
							if (t.curGeom && mode!="save") t.geomHistory.push(t.curGeom);
							t.curGeom = v;
						break;
						
						case "geom_type":
							t.geomType = v;
						break;
						
						case "isnew":
							t.geomHistory = [];
						break;
						
						default:
							if (!newProps) newProps = {};
							newProps[i] = v;
						break;
					}
				});
				
				// calculate the display text for this object:
				if (geomData.label && typeof(geomData.label)=="string" && geomData.label.length>0)
					geomData.displayText = geomData.label+' ('+geomData.geom_type+' ID# '+geomData.id+')';
				else
					geomData.displayText = '[Unlabelled] ('+geomData.geom_type+' ID# '+geomData.id+')';
					
				newProps.displayText = geomData.displayText;
			
				// If new properties have been returned, set them as the properties for the current geometry.
				if (newProps) t.geomProps = newProps;
				
				// Check if the label has changed for a current drawing:
				var triggerLU = F;
				if (t.drawList[geomData.id] && t.drawList[geomData.id].label != geomData.label) triggerLU = T;
				
				// Update the drawings list...
				t.drawList[geomData.id] = geomData;
				
				// If it's a new geometry, or if the label has changed, trigger the list update on any listeners:
				if (geomData.isnew || triggerLU)
				{
					$(".draw-listener",t.map.context).trigger("DrawListUpdate");
					geomData.isnew = N;
				}
				
				t.hasLinked = F;
			
			}
			
			if (!quiet) $(".draw-listener",t.map.context).trigger("DrawUpdate");
		break;
	}
	
}

mcdl.prototype.Clear = function()
{
	var t = this;
	t.curGeom = N;
	t.geomType = N;
	t.geomHistory = [];
	t.geomProps = N;
	t.geomID = N;
	$(".draw-listener",t.map.context).trigger("DrawingCleared",[t]);
	MapChat.QueueEvent(8,"SpatialTools",["DrawingCleared"]);
}
