/**
 * An object that presents a GUI for putting together Irrigation plans for farms in MapChat.
 *
 * @class MapChat.Irrigation.Planner
 *
 * @param kamap - the ka-Map API object that this tool relates to
 * @param options - an object listing options for the tool
 *
 * Depends:
 *	kamap/htdocs/kaMap.js
 *	kamap/htdocs/kaTool.js
 *	MapChat.js
 *	MapChat/application.js
 *	MapChat/MapLayers.js
 *	jquery/ui/ui.dialog.js
 *	jquery/jquery.autocomplete.js
 */

// Get an alias for the the MapChat.Irrigation namespace (to make the code shorter).
var mci = MapChat.UseNS("MapChat.Irrigation"),

// Alias the disable/enable button functions and the QueueEvent function:
	dbtn = MapChat.DisableButton, ebtn = MapChat.EnableButton, mcqe = MapChat.QueueEvent;

/**
 * The class object for the Irrigation Planner dialog.
 *
 * @param object kamap - the map object that this dialog is associated with
 * @param object options - an array of options for creating this object.
 */
mci.Planner = function(kamap,options)
{
	if (!options) options = {};
	var t = this,
		mapLayers = (t.mapLayers = MapChat.MapSingleton("MapChat.MapLayers",MapChat.map)),
		options = (t.options = options),
		map = (t.map = kamap),
		cursor = (t.cursor = "crosshair"),
		name = (t.name = "MapChat.Irrigation.Planner"),
		zIndex = (t.zIndex = options.zIndex||300),
		currentPlan = (t.currentPlan = null),
		
		lineStyle = (t.lineStyle = {stroke:
			{stroke:options.lineColor||"#007",'stroke-width':options.stroke||2,'stroke-opacity':options.opacity||.65}}),
		closeStyle = (t.closeStyle = {stroke:
			{stroke:options.closeColor||"#C73",'stroke-width':options.stroke||2,'stroke-opacity':options.opacity||1.0}}),
		fillStyle = (t.fillStyle = {fill:
			{fill:options.fillColor||"#00F",'fill-opacity':options.opacity||.35},r:5}),
			
		displayCanvas = (t.displayCanvas = MapChat.Class("MapChat.Canvas",{map:map})),
		displayGeom = (t.displayGeom = null),
		closeGeom = (t.closeGeom = null),
		
		divTabs = (t.divTabs = null),
		ulTabs = (t.ulTabs = null),
		planTabs = (t.planTabs = {}),
		curTab = (t.curTab = null),
		panelIDs = (t.panelIDs = []),
		panelHeight = (t.panelHeight = null),
		planInfo = (t.planInfo = null),
		featureCounts = (t.featureCounts = null),
		featureInfo = (t.featureInfo = {"paddocks":[],"networks":[],"equipment":[],"storage":[],"takes":[],"other":[]}),
		drawing = (t.drawing = F),
		drawMode = (t.drawMode = null),
		lastMapTool = (t.lastMapTool = null),
		
		content = (t.content = 
			$('<div class="autoscroll" style="font-size:.9em;line-height:16px"><p>Enter/select an irrigation plan title:</p></div>')
			.dialog({title:options.title||"Irrigation Planner",width:450,height:300,modal:F,autoOpen:F,closeOnEscape:F,
				containment:options.containment||document,
				resize:function()
				{
					// Make the continer is fit to the dialog's height (so we get scrolling):
					t.content.css({height:$(".ui-resizable-s", t.dialog).offset().top-t.content.offset().top});
					if (t.divTabs)
					{
						t.panelHeight = t.content.height() - 
							((t.divTabs.children('div[id^="mcip_"]:visible').offset().top) - t.content.offset().top)-8;
						t.divTabs.children(".ui-tabs-panel").css({height:t.panelHeight});
					}
				},
				open:function()
				{
					// Setup the autocomplete/bind events for the planTitle object.
					t.BindTitle();
					
					t.content.css({height:$(".ui-resizable-s", t.dialog).offset().top-t.content.offset().top});
					
					if (!t.divTabs)
					{
						// Check to see if there is already a plan open in the map - if so, load it in the dialog.
						$.each(t.displayGroups,function(n,dg){
							if (dg.plan_type=="Irrigation")
							{
								t.currentPlan = n;
								t.planTitle.val(n);
								var pi = $(MapChat.RunFunction("Plan_Open",{args:{plan_id:t.plansLUT[n].id,mapfile:t.map.currentMap}}));
								t.LoadPlan(pi);
								t.CheckTitle();
								return F;
							}
						});
					}
					
					// For now, this is needed to ensure the arrow keys continue to work.
					if ($.browser.msie) document.onkeydown = kaTool_redirect_onkeypress;
					else document.onkeypress = kaTool_redirect_onkeypress;
					
				},
				close:function()
				{
					// Just to make sure things are cleaned up if a user closes the dialog while the
					// autocomplete list is open:
					t.planTitle.acHide().unautocomplete();
					t.mainP.hide("");
					t.deactivate();
					t.displayCanvas.deactivate();
					t.displayCanvas.clear();
					mcqe(3,"Irrigation.Planner",["Close"]);
					
					// For now, this is needed to ensure the arrow keys continue to work.
					if ($.browser.msie) document.onkeydown = kaTool_redirect_onkeypress;
					else document.onkeypress = kaTool_redirect_onkeypress;
				}
			})
		),
		
		titleP = (t.titleP = $('<p />').appendTo(content)),
		planTitle = (t.planTitle = $('<input class="dropDown" type="text" style="width:250px;" />').appendTo(titleP)),
		planTitleDD = (t.planTitleDD = $('<div class="dropDownButton" />').appendTo(titleP).click(function(){
			planTitle.trigger("ddToggle");
			t.CheckTitle();
		})),
		buttonPanel = (t.buttonPanel = $('<p class="buttonpanel" style="clear:left;" />').appendTo(content)),
		openPlan = (t.openPlan = dbtn($('<button class="ui-state-default ui-corner-all" />')).text(MapChat.discussion.ro?"Open":"Open/Create")
			.click(function(){t.OpenPlan();}).appendTo(buttonPanel).after("&nbsp;")),
		copyPlan = (t.copyPlan = dbtn($('<button class="ui-state-default ui-corner-all" />')).text("Copy").click(function(){t.CopyPlan();}).appendTo(buttonPanel)),
		delPlan = (t.delPlan = dbtn($('<button class="ui-state-default ui-corner-all" />')).text("Delete").click(function(){t.DeletePlan();}).appendTo(buttonPanel).before("&nbsp;")),
		mainP = (t.mainP = $('<p />').appendTo(content)),
		displayCanvas = (t.displayCanvas = MapChat.Class("MapChat.Canvas",{map:this.map})),
		displayGeom = (t.displayGeom = null),
		plans = (t.plans = {}),
		plansLUT = (t.plansLUT = []),
		planList = (t.planList = []),
		dialog = (t.dialog = content.parents(".ui-dialog:first"));
		
	//content.data("dialog").uiDialog.data("draggable").setData("containment","document");
	
	MapChat.KeyProtect(t.planTitle);
	
	kaTool.apply( t, [map] );
	
	$.extend(fillStyle,lineStyle);
	
	/** get all the prototype functions from the kaTool object */
	for (var p in kaTool.prototype)
	{
		if (!mcip.prototype[p])
			mcip.prototype[p]= kaTool.prototype[p];
	}
	
}

// Make an alias for the MapChat.Irrigation.Planner tool:
var mcip = mci.Planner;

/**
 * Opens the dialog.
 */
mcip.prototype.Open = function()
{
	
	$.blockUI();
	
	var t = this,
		groups = (t.groups = t.mapLayers.getGroups()),
		displayGroups = (t.displayGroups = t.mapLayers.getDisplayGroups());
	
	t.planList = [];
	t.plans = $.parseJSON($(MapChat.RunFunction("Plan_Listing",{args:{type:"Irrigation"}})).find("plans").text());
	$.each(t.plans,function(){ t.plansLUT[this.title] = this; t.planList.push(this.title); });
	
	t.displayCanvas.activate();
	
	if (t.drawing) t.activate();
	
	t.content.dialog("open").dialog("moveToTop");
	
	mcqe(3,"Irrigation.Planner",["Open"]);
	
	$.unblockUI();
	
	return t;
}

/**
 * Activate this tool.  Deactivate any store a reference to any currently active map tool.
 */
mcip.prototype.activate = function()
{
	var t = this;
	
	// Deactivate any current 'nav' tool:
	if (MapChat.map.toolGroups["nav"] && (ai = MapChat.map.toolGroups["nav"].activeItem)) t.lastMapTool = ai;
	else if (t.map.currentTool) t.lastMapTool = t.map.currentTool;
	else t.lastMapTool = t.map.defaultTool;
	
	if (t.lastMapTool) t.lastMapTool.deactivate();
	
	MapChat.QueueEvent(3,"Irrigation.Planner",["DrawStart",(t.curTab?t.curTab.data("layer"):''),(t.curTab?t.curTab.data("shape"):'')]);
	
	kaTool.prototype.activate.apply(t);
	
	if ($.browser.msie) document.onkeydown = kaTool_redirect_onkeypress;
	else document.onkeypress = kaTool_redirect_onkeypress;
};

/**
 * Deactivate this tool in the map interface...restore any other map tool that was active.
 */
mcip.prototype.deactivate = function()
{
	var t = this;
	
	mcqe(3,"Irrigation.Planner",["DrawEnd",(t.curTab?t.curTab.data("layer"):''),(t.curTab?t.curTab.data("shape"):'')]);
	
	if (t.displayGeom)
	{
		t.displayGeom.draw();
		if (t.closeGeom)
		{
			t.closeGeom.xy = [[[]]];
			t.closeGeom.draw();
		}
		t.ShowCoordinates(t.displayGeom,$("[name='coordinates']",t.curTab));
		$("[name='coordinates']",t.curTab).removeAttr(D);
		$("[name='draw']",t.curTab).text("Draw Feature");
		t.drawing = F;
	}
	
	kaTool.prototype.deactivate.apply(t);
	
	if (t.lastMapTool) t.lastMapTool.activate(T);
};

/**
 * MapChat.Irrigation.Planner.onmouseup( e )
 *
 * called when a mouse button is released over theInsideLayer.
 *
 * e - object, the event object or null (in ie)
 */
mcip.prototype.onmouseup = function(e)
{
	var t = this;
	e = $.event.fix(e?e:event);
	if (e.button==2||e.metaKey||e.shiftKey)
	{
		return t.cancelEvent(e);
	}
	
	// Only add points for line/polygon shapes - points get
	else if (t.drawing && t.drawMode!="point")
	{
		var geoms = t.displayGeom.xy.length,
			curGeomPts = (geoms>0?t.displayGeom.xy[geoms-1].length:0),
			curPt = t.adjustPixPosition( e.pageX, e.pageY );
			
		var geoPt = MapChat.Round(t.map.pixToGeo(curPt[0],curPt[1]));
		
		// Ignore points equal to the last one (e.g., on a double-click event):
		if (curGeomPts>0 && geoPt[0] == t.displayGeom.lastPoint()[0] && geoPt[1] == t.displayGeom.lastPoint()[1]) return;
		
		t.displayGeom.addPoint(geoPt);
		
		mcqe(3,"Irrigation.Planner",["AddPoint",geoPt[0],geoPt[1]]);
		
		t.SetMinMax(t.curTab,geoPt);
		
		ebtn($('[name^="undo_"]',t.curTab));
		
		t.displayGeom.draw();
	}
};


/**
 * MapChat.Irrigation.Planner.onmousemove( e )
 *
 * called when a mouse button is moved over theInsideLayer.
 *
 * e - object, the event object or null (in ie)
 */
mcip.prototype.onmousemove = function(e)
{
	var t = this;
	
	// Only add points for line/polygon shapes - points get
	if (t.drawing && t.drawMode!="point")
	{
		e = $.event.fix(e?e:event);
	
		var geoms = t.displayGeom.xy.length,
			curGeomPts = (geoms>0?t.displayGeom.xy[geoms-1].length:0),
			curPt = t.adjustPixPosition( e.pageX, e.pageY );
		
		var geoPt = MapChat.Round(t.map.pixToGeo(curPt[0],curPt[1]));
		
		// Ignore points equal to the last one (e.g., on a double-click event):
		if (curGeomPts>0 && geoPt[0] == t.displayGeom.lastPoint()[0] && geoPt[1] == t.displayGeom.lastPoint()[1]) return;
		
		if (t.drawMode=="polygon" && geoms>0 && t.displayGeom.xy[geoms-1].length>0)
		{
			t.closeGeom.xy[0] = [t.displayGeom.xy[geoms-1][0],geoPt];
			t.closeGeom.draw();
		}
		
		t.displayGeom.addPoint(geoPt);
		if (geoms>0 && t.displayGeom.xy[geoms-1].length>1) t.displayGeom.draw();
		t.displayGeom.dropPoint();
	}
};


/**
 * MapChat.Irrigation.Planner.ondblclick( e )
 *
 * called when a double-click event is triggered on theInsideLayer.
 *
 * e - object, the event object or null (in ie)
 */
mcip.prototype.ondblclick = function(e)
{
	var t = this;
	
	if (t.drawing)
	{
		e = $.event.fix(e?e:event);
		
		if (t.drawMode=="polygon" && !t.displayCanvas.validateGeom(t.displayGeom)) return F;
		
		if (t.closeGeom)
		{
			t.closeGeom.xy = [[[]]];
			t.closeGeom.draw();
		}
		
		if (t.drawMode!="point") t.displayGeom.addPart();
		
		else
		{
			var curPt = t.adjustPixPosition( e.pageX, e.pageY ),
				geoPt = MapChat.Round(t.map.pixToGeo(curPt[0],curPt[1]));
			t.displayGeom.addPoint(geoPt);
			t.SetMinMax(t.curTab,geoPt);
			mcqe(3,"Irrigation.Planner",["AddPoint",geoPt[0],geoPt[1]]);
		}
		
		t.displayGeom.draw();
		
		ebtn($('[name^="undo_"]',t.curTab));
		
	}
};




/**
 * Loads information about a selected plan for the server, or creates one if the plan title entered is new.
 */
mcip.prototype.OpenPlan = function()
{
	var t = this, pt = t.planTitle.val();
	if (pt != t.currentPlan)
	{
		// Close any previously opened plan.
		if (t.currentPlan && t.currentPlan.length>0)
		{
			if (!t.ClosePlan(t.currentPlan,T)) return;
		}
		
		// Reset the selected plan title.
		t.planTitle.val(pt);
		
		if ($.inArray(pt,t.planList)>=0)
		{
			var pi = $(MapChat.RunFunction("Plan_Open",{args:{plan_id:t.plansLUT[pt].id,mapfile:t.map.currentMap}}));
			t.LoadPlan(pi);
			t.CheckTitle();
		}
		else
		{
			var pc = $(MapChat.RunFunction("Plan_Create",{args:{title:pt,type:"Irrigation",mapfile:t.map.currentMap}}));
			if (pc.find("plan_id:first").text().length>0)
			{
				t.plansLUT[pt] = $.parseJSON(pc.find("plan_info:first").text());
				t.planList.push(pt);
				t.LoadPlan(pc);
				t.planTitle.unautocomplete();
				t.BindTitle();
				t.CheckTitle();
			}
		}
	}
	else if (pt.length>0 && pt == t.currentPlan)
	{
		t.ClosePlan(t.planTitle.val());
	}
	else
	{
		ebtn(t.openPlan).text("Open/Create");
	}
}

/**
 * Checks whether a title refers to an existing plan - sets the open/create button's text/state accordingly.
 */
mcip.prototype.CheckTitle = function()
{
	var t = this, op = t.openPlan, cp = t.copyPlan, dp = t.delPlan;
	if (t.planTitle.val().length>0 && t.planTitle.val()!=t.currentPlan)
	{
		if ($.inArray(t.planTitle.val(),t.planList)>=0)
		{
			ebtn(op).text("Open");
			ebtn(cp);
			ebtn(dp);
		}
		else if (!MapChat.discussion.ro && !t.displayGroups[t.planTitle.val()])
		{
			ebtn(op).text("Create");
			dbtn(cp);
			dbtn(dp);
		}
		else
		{
			ebtn(op).text(MapChat.discussion.ro?"Open":"Open/Create");
			dbtn(cp);
			dbtn(dp);
		}
	}
	else if (t.planTitle.val().length>=0 && t.planTitle.val()==t.currentPlan)
	{
		ebtn(op).text("Close");
		ebtn(cp);
		ebtn(dp);
	}
	else if (t.planTitle.val().length==0)
	{
		dbtn(op).text(MapChat.discussion.ro?"Open":"Open/Create");
		dbtn(cp);
		dbtn(dp);
	}
}

/**
 * Loads the plan currently selected - assumes it has already been created.
 */
mcip.prototype.LoadPlan = function(p)
{
	var t = this, pt = t.planTitle.val();
	t.currentPlan = pt;
	
	if (!$.inArray(pt,t.displayGroups)>=0)
	{
		var mlayers = $.parseJSON(p.find("plan_maplayers:first").text()),
			ilayers = [], nLayers = [];
		
		$.each(mlayers,function(n,v){ilayers.push(v);nLayers.push(n);});
		ilayers.sort();
		
		var group = {
			name:"USER_LAYERS",
			visible:T,
			opacity:parseInt(p.find("opacity:first").text()),
			imageformat:p.find("imageformat:first").text(),
			queryable:F,
			tileSource:"redraw",
			redrawInterval:1,
			refreshInterval:-1,
			layers:ilayers,
			userLayer:T
		};
		
		t.mapLayers.addDisplayGroup({name:pt,status:T,partial_status:F,removeable:T,plan_type:"Irrigation"},group,t.map.currentMap);
	}
	
	if (!t.divTabs)
	{
		// If the tabs haven't been created yet, then create them.
		t.divTabs = $('<div />');
		
		// Get several unique IDs that will be used to refer to the div elements - for ui.tabs to work.
		t.panelIDs = MapChat.UID("mcip",6);
		
		var ulTabs = $('<ul><li><a href="#'+this.panelIDs[0]+'"><span>Paddocks</span></a></li><li><a href="#'+this.panelIDs[1]+'"><span>Networks</span></a></li><li><a href="#'+this.panelIDs[2]+'"><span>Equipment</span></a></li><li><a href="#'+this.panelIDs[3]+'"><span>Storage</span></a></li><li><a href="#'+this.panelIDs[4]+'"><span>Takes</span></a></li><li><a href="#'+this.panelIDs[5]+'"><span>Other</span></a></li></ul><div id="'+this.panelIDs[0]+'" /><div id="'+this.panelIDs[1]+'" /><div id="'+this.panelIDs[2]+'" /><div id="'+this.panelIDs[3]+'" /><div id="'+this.panelIDs[4]+'" /><div id="'+this.panelIDs[5]+'" />').appendTo(t.divTabs);
		
		
		t.divTabs.appendTo(t.content);
		
		t.divTabs.find("div").addClass("autoscroll").addClass("planner_panel");
		t.divTabs.tabs({show:function(){t.ShowTab();}}).bind("tabsselect",function(){return t.SelectTab();});
		t.ulTabs = t.divTabs.find("ul:first");
		
		// Setup the panel height property for the panels.
		t.panelHeight = t.content.height() - 
			(t.divTabs.children('div[id^="mcip_"]:visible').offset().top - t.content.offset().top)-8;
		t.divTabs.children(".ui-tabs-panel").css({height:t.panelHeight});
		
	}
	else
	{
		// otherwise, make sure they are visible.
		t.divTabs.show();
	}
	
	t.featureCounts = $.parseJSON(p.find("plan_feature_counts:first").text());
	t.planInfo = $.parseJSON(p.find("plan_info:first").text());
	
	t.InitializeTabs();
	
}

/**
 * Sets up the UI controls for each of the tabs for the irrigation planner.
 */
mcip.prototype.InitializeTabs = function()
{
	var t = this;
	
	// Get the panels:
	var divPaddocks = $("#"+t.panelIDs[0]);
	var divNetworks = $("#"+t.panelIDs[1]);
	var divEquipment = $("#"+t.panelIDs[2]);
	var divStorage = $("#"+t.panelIDs[3]);
	var divTakes = $("#"+t.panelIDs[4]);
	var divOther = $("#"+t.panelIDs[5]);
	
	// Make the content for each of the tabs.
	t.BuildTabContent(
		divPaddocks,
		{title:"Paddock",feature_count:t.featureCounts.paddocks,fields:[
			{n:"Primary Irrigation Type",ac:["Contour Flood","Wild Flood","Border Dyke","Centre Pivot","K-Line","Spray (stationary)","Spray (travelling)","Other","None"]},
			{n:"Additional Irrigation Type",ac:["Contour Flood","Wild Flood","Border Dyke","Centre Pivot","K-Line","Spray (stationary)","Spray (travelling)","Other","None"]},
			{n:"Size of Irrigated Area (ha)"},{n:"Max. Rate (litres/sec.)"},{n:"Normal Rate (litres/sec.)"},{n:"Low Rate (litres/sec.)"},
			{n:"Start of Irrigation"},{n:"End of Irrigation"},{n:"Frequency of Irrigation"},{n:"Crop Type"},{n:"Annual Crop Yield"},{n:"Soil Type"}
		]}).data("shape","polygon").data("layer","paddocks").data("cur_feature","##");
	
	t.BuildTabContent(
		divNetworks,
		{title:"Network",feature_count:t.featureCounts.networks,fields:[
			{n:"Feature Type",ac:["Race","Creek/Stream","Pipeline","Other"]},
			{n:"Capacity (litres/sec.)"}
		]}).data("shape","line").data("layer","networks").data("cur_feature","##");
		
		
	t.BuildTabContent(
		divEquipment,
		{title:"Equipment",feature_count:t.featureCounts.equipment,fields:[
				{n:"Feature Type",ac:["Centre Pivot","K-Line","Pump","Generator","Wier","Other"]},
				{n:"Radius of Centre Pivot (metres)"},
				{n:"Number of Segments (C-Pivot) or Pods (K-Line)"},
				{n:"Power Input/Output"}
		]}).data("shape","point").data("layer","equipment").data("cur_feature","##");
		
		
	t.BuildTabContent(
		divStorage,
		{title:"Storage",feature_count:t.featureCounts.storage,fields:[
				{n:"Volume (tonnes/cubic metres)"},
				{n:"Purpose"}
		]}).data("shape","polygon").data("layer","storage").data("cur_feature","##");
	
	t.BuildTabContent(
		divTakes,
		{title:"Take",feature_count:t.featureCounts.takes,fields:[
				{n:"Type of Permit",ac:["Mining Right/Deemed Permit","RMA","Other"]},
				{n:"Maximum Rate (litres/sec.)"},
				{n:"Maximum Daily (cubic metres)"},
				{n:"Maximum Weekly (cubic metres)"},
				{n:"Maximum Monthly (cubic metres)"},
				{n:"Area of use (ha)"},
				{n:"Purpose"}
		]}).data("shape","point").data("layer","takes").data("cur_feature","##");
	
	t.BuildTabContent(
		divOther,
		{title:"Feature",feature_count:t.featureCounts.other,fields:[
				{n:"Description"}
		],geomType:"geometry"}).data("shape","point").data("layer","other").data("cur_feature","##");
	
	
	$("[name='draw']",t.divTabs).click(function(){
		
		if (!t.drawing)
		{
			var p = t.curTab;
			
			t.SetChanged(p)
			t.drawMode = p.data("shape");
			
			$("[name='coordinates']",p).attr(D,T);
			$(this).text("Finish Drawing");
			
			t.drawing = T;
			t.activate();
		}
		else
		{
			if (t.displayGeom.numPoints()==0 || t.displayCanvas.validateGeom(t.displayGeom))
			{
				t.drawing = F;
				if (t.drawMode=="polygon" && t.displayGeom.numPoints()>0) t.displayGeom.addPart();
				t.deactivate();
			}
			else
			{
				alert("Error - Polygons may not have self-intersecting sides.\n\nPlease finish or undo your drawing.");
			}
		}
	});
	
	$('[name="coordinates"]',t.divTabs).blur(function(){t.ApplyCoordinates(t.displayGeom,$(this));})
	
	$('.dropDownButton',t.divTabs).click(function(){ $(this).prev().trigger("ddToggle"); });
	
	$('[name="save"]',t.divTabs).click(function(){ t.SaveChanges($(this)); });
	
	// The paddocks tab is selected by default - so we'll initialize the geometry for it.
	t.InitGeometry(divPaddocks);
	
	MapChat.KeyProtect($("input,textarea",t.divTabs));
}

/**
 * Builds the content for a tab in the planner interface.
 *
 * @param object el - the div element for the tab panel
 * @param object opts - the data associated with this tab, including the title, fields, feature_count.  The
 *  fields property should be an array of objects with a property n (the name of the field) and optionally
 *  a property named ac which would contain a list of possible values for an auto-complete dropdown.
 */
mcip.prototype.BuildTabContent = function(el,opts)
{
	var t = this, c = "";
	c += '<p><button name="previous" class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">&lt;&lt;</button> <input name="feature_idx" value="##" style="width:30px;" /> of <span class="feature_count">'+opts.feature_count+' </span>&nbsp;<button name="next" class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">&gt;&gt;</button>&nbsp;|&nbsp;<button class="ui-state-default ui-corner-all" name="new">New</button>&nbsp;<button name="save" class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">Save</button>&nbsp;<button name="reset" class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">Reset</button>&nbsp;<button name="delete" class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">Delete</button></p><p><button class="ui-state-default ui-corner-all" name="draw">Draw Feature</button>&nbsp;|&nbsp;<button name="undo_coord" class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">Undo Coord.</button>&nbsp;<button name="undo_all_coords"  class="ui-state-default ui-corner-all ui-state-disabled" disabled="true">Undo All Coords.</button></p>';
	if (opts.geomType=="geometry") c += '<p>Shape Type: <input type="radio" name="geom_point" value="point" checked="true" /> Point&nbsp;&nbsp;&nbsp;<input type="radio" name="geom_line" value="line" /> Line&nbsp;&nbsp;&nbsp;<input type="radio" name="geom_polygon" value="polygon" /> Polygon</p>';
	c += '<p style="font-weight:bold;">'+opts.title+' Details:</p><p style="margin-left:16px;"><table>';
	$.each(opts.fields,function(){
		c += '<tr><td>'+this.n+':&nbsp;</td><td valign="bottom" width="180"><input type="text" name="'+MapChat.NormalizeString(this.n)+'" class="attribute_field" />';
		if (this.ac)
		{
			c += '<div class="dropDownButton" />';
		}
		c += '</td></tr>';
	});
	c += '<tr><td colspan="2">Notes:<br /><textarea name="notes" style="width:320px;height:50px;" class="attribute_field" /></td></tr></table></p><tr><td colspan="2">Shape Coordinates:<br /><textarea name="coordinates" style="width:320px;height:100px;" /></td></tr></table></p>';
	
	el.append(c);
	$.each(opts.fields,function(){
		if (this.ac)
		{
			$("[name='"+MapChat.NormalizeString(this.n)+"']", el).css({float:"left"})
				.autocomplete(this.ac,
					{minChars:0,selectFirst:F,highlight:F,mustMatch:F,matchContains:T,scrollHeight:220});
		}
	});
	
	if (opts.feature_count>0)
	{
		$(".feature_count",el).text(opts.feature_count);
		ebtn($('[name="next"],[name="previous"]',el));
	}
	else
	{
		dbtn($('[name="next"],[name="previous"]',el));
	}
	
	var chk = function(){
		var ti = $(this), cf = t.featureInfo[el.data("layer")][el.data("cur_feature")];
		if ((!cf && ti.val()!="") || (cf && cf.attr && cf.attr[ti.attr("name")]!=ti.val()))
		{
			t.SetChanged(el);
		}
	};

	if (opts.geomType=="geometry")
	{
		$('[type="radio"][name^="geom_"]',el).click(function(){
			if ($(this).val()!=el.data("shape") && t.displayGeom && t.displayGeom.numPoints()>0)
				var doChange = confirm("Discard currently drawn features?");
			else
				var doChange = T;
			
			if (doChange)
			{
				$('[type="radio"][name^="geom_"]',el).removeAttr("checked");
				$(this).attr("checked",T);
				el.data("shape",$(this).val());
				t.InitGeometry(el);
				t.drawMode = el.data("shape");
			}
			else
			{
				$('[type="radio"][name^="geom_"]',el).removeAttr("checked");
				$('[type="radio"][name^="geom_"][value="'+el.data("shape")+'"]',el).attr("checked",T);
			}
		});
	}
	
	$('[name="next"],[name="previous"]',el).click(function(){
		
		if (el.data("changed") && !confirm("Discard changes to the current feature?")) return;
		
		var idx = $('[name="feature_idx"]',el), ly = el.data("layer"), nextIdx;
		
		switch ($(this).attr("name"))
		{
			case "next":
				if (idx.val()=="##") nextIdx = 1;
				else if (parseInt(idx.val())<t.featureCounts[ly]) nextIdx = 1+parseInt(idx.val());
				else nextIdx = 1;
				break;
			case "previous":
				if (idx.val()=="##") nextIdx = parseInt(t.featureCounts[ly]);
				else if (parseInt(idx.val())>1) nextIdx = parseInt(idx.val())-1;
				else nextIdx = parseInt(t.featureCounts[ly]);
				break;
		}
		
		t.SelectFeature(nextIdx,el);
	});
	
	$('[name="feature_idx"]',el).keyup(function(e){
		if (e.keyCode==13)
		{
			var idx = $(this).val();
			if (idx>=1 && idx<=t.featureCounts[el.data("layer")])
			{
				el.data("cur_feature",$(this).val());
				t.SelectFeature($(this).val(),el);
			}
			else
			{
				$(this).val(el.data("cur_feature"));
			}
		}
	}).blur(function(){
		$(this).val(el.data("cur_feature"));
	});
	
	$('[name="new"]',el).click(function(){
		if (!el.data("changed")||confirm("Discard current changes?")) t.SelectFeature("##",el);
	});
	
	// This is how we'll keep track of whether anything has changed.
	el.data("changed",F);
	$("textarea,input[name!='feature_idx']",el).click(chk).keypress(chk).keyup(chk).focus(function(){
		var ti = $(this); 
		if ((!ti.data("lastval") && ti.val().length>0) || (ti.data("lastval") && ti.data("lastval")!=ti.val())) chk.apply(this);
		ti.data("lastval",ti.val());
	});
	
	$('[name="reset"]',el).click(function(){t.DiscardChanges(el,T);});
	
	$('[name="undo_coord"]',el).click(function(){
		t.UndoCoord(t.displayGeom,$(this));
		t.SetChanged(el);
	});
	
	$('[name="undo_all_coords"]',el).click(function(){
		t.UndoAllCoords($(this));
		t.SetChanged(el);
	});
	
	$('[name="delete"]',el).click(function(){
		if (confirm("Delete the current feaure?")) t.DeleteFeature(el);
	});
	
	return el;
}

/**
 * Handles what happens when tabs are switched - first load should create controls, after that
 * switching should handle anything related to leaving the previous tab, and opening the new tab.
 */
mcip.prototype.ShowTab = function()
{
	// the ui.tabs script currently has a feature that clears any previously set style properties, so we'll have to
	// reset these each time a tab is displayed...just to be sure.
	this.divTabs.children(".ui-tabs-panel").css({height:this.panelHeight});
	
	var t = this,
		curTabID = (t.divTabs?t.divTabs.tabs('option', 'selected'):0),
		panelID = t.panelIDs[curTabID],
		panel = t.divTabs.find("#"+panelID);
	
	t.lastPanelID = panelID;
	t.lastTabID = curTabID;
	
	t.curTab = panel;
	
	mcqe(3,"Irrigation.Planner",["ShowTab",(t.curTab?t.curTab.data("layer"):'')]);
	
	t.InitGeometry(panel);
	
}

/**
 * Handles what happens when tabs are switched - first load should create controls, after that
 * switching should handle anything related to leaving the previous tab, and opening the new tab.
 */
mcip.prototype.SelectTab = function()
{
	var t = this;
	if (t.lastPanelID)
	{
		if ($("#"+t.lastPanelID,t.divTabs).data("changed"))
		{
			if (t.lastTabID!=null && !confirm("Discard unsaved changes?"))
			{
				// Only way to make this work is to make the ShowTab callback switch back.
				return F;
			}
			else
			{
				t.DiscardChanges($("#"+t.lastPanelID,t.divTabs));
			}
		}
	}
}


/**
 * Displays an editable list of coordinates in the tab.
 *
 * @param object g - the canvas geometry object with coordinate to display
 * @param object ta - the text area that the coordinates will be displayed in.
 */
mcip.prototype.ShowCoordinates = function(g,ta)
{
	var gt = "";
	if (g.type=="path")
	{
		$.each(g.xy,function(i){
			if (i>0) gt += " | ";
			$.each(this,function(j){
				if (j>0) gt += " , ";
				gt += this[0]+" "+this[1];
			});
		});
	}
	else
	{
		$.each(g.xy,function(i){
			if (i>0) gt += " | ";
			gt += this[0]+" "+this[1];
		});
	}
	ta.val(gt);
}

/**
 * Updates the displayGeometry with the coordinates in a textarea object.
 *
 * @param object g - the canvas geometry object to display apply the coordinates to.
 * @param object ta - the text area that the coordinates are displayed in.
 */
mcip.prototype.ApplyCoordinates = function(g,ta)
{
	var t = this, el = ta.parents(".ui-tabs-panel:first"),
		xy = ta.val().replace(/\ ,\ /g,",").replace(/\ \|\ /g,"|").replace(/\|$/g,"").split("|");
	
	if (g.type=="path")
	{
		g.xy = [];
		$.each(xy,function(i){
			if (this.length>0)
			{
				g.addPart();
				var cxy = this.split(",");
				$.each(cxy,function(){
					var c = this.split(" ");
					g.addPoint([1*c[0],1*c[1]]);
					t.SetMinMax(el,[1*c[0],1*c[1]]);
				});
			}
		});
	}
	else
	{
		while (g.numPoints()>0)
		{
			g.dropPoint();
		}
		if (xy.length>0)
		$.each(xy,function(){
			c = this.split(" ");
			if (c.length>1)
			{
				g.addPoint([1*c[0],1*c[1]]);
				t.SetMinMax(el,[1*c[0],1*c[1]]);
			}
		});
	}
	if (g.numPoints()>0) ebtn($('[name^="undo_"]',el));
	g.draw();
}


/**
 * Deletes the last drawn coordinate.
 *
 * @param object g - the canvas geometry object to display apply the coordinates to.
 * @param object ud - the undo button that was clicked.
 */
mcip.prototype.UndoCoord = function(g,ud)
{
	var el = ud.parents(".ui-tabs-panel:first"),
		xy = g.dropPoint();
	g.draw();
	
	mcqe(3,"Irrigation.Planner",["UndoCoord",xy[0],xy[1]]);
	
	if (g.numPoints()==0)
	{
		dbtn(ud);
		dbtn($('[name="undo_all_coords"]',el));
	}
	if (!this.drawing)
	{
		this.ShowCoordinates(this.displayGeom,$('[name="coordinates"]',el));
	}
}


/**
 * Deletes all geometry data from the current feature.
 *
 * @param object ud - the undo button that was clicked.
 */
mcip.prototype.UndoAllCoords = function(ud)
{
	var el = ud.parents(".ui-tabs-panel:first");
	mcqe(3,"Irrigation.Planner",["UndoAllCoords"]);
	this.InitGeometry(el);
}


/**
 * Initializes the displayGeometry based on the current panel's state.
 *
 * @param object el - the div panel element associated with the current UI tab.
 */
mcip.prototype.InitGeometry = function(el)
{
	var t = this;
	// Get rid of any existing geometries.
	if (t.displayGeom)
	{
		t.displayCanvas.dropGeom(t.displayGeom);
		t.displayGeom = null;
	}
	if (t.closeGeom)
	{
		t.displayCanvas.dropGeom(t.closeGeom);
		t.closeGeom = null;
	}
	
	switch (el.data("shape"))
	{
		case "point":
			t.displayGeom = t.displayCanvas.newGeom("point",t.fillStyle);
			break;
			
		case "line":
			t.displayGeom = t.displayCanvas.newGeom("path",t.lineStyle);
			break;
		
		case "polygon":
			t.displayGeom = t.displayCanvas.newGeom("path",t.fillStyle,T);
			t.closeGeom = t.displayCanvas.newGeom("path",t.closeStyle);
			t.closeGeom.xy = [[]];
			break;
	}
	
	if (!t.drawing)
	{
		$('[name="coordinates"]',el).val("");
	}
	
	dbtn($('[name^="undo_"]',el));
	
}


/**
 * Discards any changes to the tab.
 *
 * @param object el - the div panel element associated with the current UI tab.
 * @param boolean conf - if this is true, the user will be asked before discarding changes.
 */
mcip.prototype.DiscardChanges = function(el,conf)
{
	var t = this;
	
	if (conf && !confirm("Discard current changes?")) return;
	
	t.InitGeometry(el);
	
	// Only log this if the user was asked for confirmation - if not, then it's a side-effect of a different event.
	if (conf) mcqe(3,"Irrigation.Planner",["DiscardChanges",el.data("layer")]);
	
	if (t.drawing)
	{
		// Call the click function which should end the drawing.
		$("[name='draw']",el).click();
	}
	
	var idx = $('[name="feature_idx"]',el);
	if (idx.val()!="##")
	{
		var f = t.featureInfo[el.data("layer")][parseInt(idx.val())];
		$('.attribute_field[name!="coordinates"]',el).each(function(){
			$(this).val(f.attr[$(this).attr("name")]);
		});
		
		if ($('input[type="radio"][name^="geom_"]',el).length>0)
		{
			$('input[type="radio"][name^="geom_"]',el).removeAttr("checked");
			if (f.shape)
			{
				$('input[type="radio"][value="'+f.shape+'"]',el).attr("checked",T);
				el.data("shape",f.shape);
			}
			else
			{
				$('input[type="radio"][value="point"]',el).attr("checked",T);
				el.data("shape","point");
			}
			
			// Re-initialize the geometry to make sure it's the correct type.
			t.InitGeometry(el);
		}
		
		if (f.coords)
			$('[name="coordinates"]',el).val(f.coords);
		else
			$('[name="coordinates"]',el).val("");
			
	}
	else
	{
		$('textarea,input[type="text"][name!="feature_idx"]',el).val("");
		if ($('input[type="radio"][name^="geom_"]',el).length>0)
		{
			$('input[type="radio"][name^="geom_"]',el).removeAttr("checked");
			$('input[type="radio"][value="point"]',el).attr("checked",T);
			el.data("shape","point");
		}
	}
	
	dbtn($('[name="save"],[name="reset"]',el));
	
	el.data("changed",F);
	t.ApplyCoordinates(t.displayGeom,$('[name="coordinates"]',el));
	
	// Add a part to non-point geometries, just to make sure the next
	// time draw feature is selected, the new point isn't attached to the 
	// previous shape in a multi-point geometry:
	if (el.data("shape")!="point") t.displayGeom.addPart();
}


/**
 * Saves changes that have been input for a new/existing feature in the tab.
 *
 * @param object s - the save button element that was clicked.
 */
mcip.prototype.SaveChanges = function(s)
{
	var t = this, el = s.parents(".ui-tabs-panel:first");
	
	if (t.drawing) $('[name="draw"]',el).click();
	
	if (!t.drawing)
	{
		var ly = el.data("layer"), f = {layer:ly,attr:{}};
		
		$('.attribute_field[name!="coordinates"]',el).each(function(){
			f.attr[$(this).attr("name")] = $(this).val();
		})
		
		var idx = $('[name="feature_idx"]',el);
		if (idx.val()!="##") f.idx = parseInt(idx.val());
		
		var coords = $('[name="coordinates"]',el);
		if (coords.val()!="")
		{
			f.coords = coords.val().replace(/\ ,\ /g,",").replace(/\ \|\ /g,"|").replace(/\|$/g,"");
			f.shape = el.data("shape");
		}
		
		var n = $.parseJSON($(MapChat.RunFunction("Plan_SaveFeature",
			{args:{plan_id:t.plansLUT[t.planTitle.val()].id,feature:$.toJSON(f)}})).find("plan_feature_id").text());
		
		if (parseInt(n)>0)
		{
			f.id = n;
			if (f.idx)
			{
				t.featureInfo[ly][f.idx] = f;
			}
			else
			{
				t.featureCounts[ly]++;
				f.idx = t.featureCounts[ly];
				$(".feature_count",el).text(f.idx);
				$('[name="feature_idx"]',el).val(f.idx);
				el.data("cur_feature",f.idx);
				t.featureInfo[ly][f.idx] = f;
			}
			t.DiscardChanges(el,F);
			ebtn($('[name="next"],[name="previous"],[name="delete"]',el));
		}
	}
	
	t.ResetTiles(el);
}

/**
 * Should handle moving to previous/next features in a tab within the planner.
 *
 * @param string idx - the index of the feautre.
 * @param object el - the panel that the feature belongs to.
 */
mcip.prototype.SelectFeature = function(idx,el)
{
	el.removeData("minx");
	el.removeData("maxx");
	el.removeData("miny");
	el.removeData("maxy");
	el.data("cur_feature",idx);
	
	var t = this, ly = el.data("layer");
	
	mcqe(3,"Irrigation.Planner",["SelectFeature",el.data("layer"),idx]);
	
	if (idx!="##" && !t.featureInfo[ly][parseInt(idx)])
	{
		var f = $.parseJSON($(MapChat.RunFunction("Plan_GetFeature",
			{args:{plan_id:t.plansLUT[t.planTitle.val()].id,layer:ly,idx:idx}})).find("plan_feature_info").text());
		
		t.featureInfo[ly][parseInt(idx)] = f;
		
		ebtn($('[name="delete"]',el));
		
	}
	else if (idx=="##")
		dbtn($('[name="delete"]',el));
	else
		ebtn($('[name="delete"]',el));
		
	$('[name="feature_idx"]',el).val(idx);
	t.DiscardChanges(el,F);
}

/**
 * Stores the min/max coordinates so we can search for tiles later.
 *
 * @param object el - the panel that coords are associated with.
 * @param object xy - the next set of xy coordinates.
 */
mcip.prototype.SetMinMax = function(el,xy)
{
	if (el.data("minx")==null)
		el.data("minx",xy[0]);
	else
		el.data("minx",Math.min(el.data("minx"),xy[0]));
	
	if (el.data("maxx")==null)
		el.data("maxx",xy[0]);
	else
		el.data("maxx",Math.max(el.data("maxx"),xy[0]));
	
	if (el.data("miny")==null)
		el.data("miny",xy[1]);
	else
		el.data("miny",Math.min(el.data("miny"),xy[1]));
	
	if (el.data("maxy")==null)
		el.data("maxy",xy[1]);
	else
		el.data("maxy",Math.max(el.data("maxy"),xy[1]));
}


/**
 * Sets the state of the layer to 'changed' - updating the corresponding controls as well.
 *
 * @param el - the panel with the feature data that has been 'changed'
 */
mcip.prototype.SetChanged = function(el)
{
	el.data("changed",T);
	ebtn($('[name="save"],[name="reset"]',el));
}

/**
 * Deletes the currently selected feature.
 *
 * @param el - the panel with the feature data that is being deleted.
 */
mcip.prototype.DeleteFeature = function(el)
{
	var t = this, idx = $('[name="feature_idx"]',el), ly = el.data("layer");
	if (idx.val()=="##") return;
	else fidx = parseInt(idx.val());
	
	var f = t.featureInfo[ly][fidx];
	
	var del = parseInt($(MapChat.RunFunction("Plan_DeleteFeature",
		{args:{plan_id:t.plansLUT[t.planTitle.val()].id,layer:ly,fid:f.id}})).find("plan_feature_deleted").text());
	
	if (del>0)
	{
		t.ResetTiles(el);
		t.ResetTiles(el);
		t.featureCounts[ly]--;
		t.featureInfo[ly].splice(fidx,1);
		if (t.featureCounts[ly]==0) t.SelectFeature("##",el);
		else if (t.featureCounts[ly]>=fidx) t.SelectFeature(fidx,el);
		else t.SelectFeature(fidx-1,el);
		$('.feature_count',el).text(t.featureCounts[ly]);
	}
}

/**
 * Reloads any tiles that are within 5 pixels of the bounding rectangle for the feature data in the current panel.
 *
 * @param el - the panel with the feature data that is being deleted.
 */
mcip.prototype.ResetTiles = function(el)
{
	var t = this;
	$.each(MapChat.FindTilesByExtent(
			t.map.unadjustPixPosition(el.data("minx"),el.data("miny"),T),
			t.map.unadjustPixPosition(el.data("maxx"),el.data("maxy"),T),
			t.map,"USER_LAYERS",5),
		function(){ t.map.resetTile(""+this,F,T);}
	);
}

/**
 * Closes the current plan.
 *
 * @param string pt - the plan title (needs to be passed here, because this.planTitle() may have changed)
 * @param boolean skipRefresh - if true, the map is not refreshed after the layer is dripped.
 * @param boolean skipConf - if true, the user is not asked to close the plan.
 */
mcip.prototype.ClosePlan = function(pt,skipRefresh,skipConf)
{
	var t = this, conf = (t.curTab.data("changed")?"Discard changes and close the current plan?":"Close the current plan?");
	if (skipConf||confirm(conf))
	{
		mcqe(3,"Irrigation.Planner",["ClosePlan",t.planTitle.val()]);
		
		// This assumes that the map's display group associated with this plan only contains layers from the plan.
		t.mapLayers.dropDisplayGroup(pt,((!!skipRefresh)?"USER_LAYERS":F),t.map.currentMap);
		
		// Clear the content from the tabbed section of the planner, and delete any other references to variables and such.
		t.divTabs.remove();
		if (t.displayGeom) t.displayCanvas.dropGeom(t.displayGeom);
		if (t.closeGeom) { t.closeGeom.xy=[[[]]]; t.closeGeom.draw() };
		t.planTabs = {};
		t.panelIDs = [];
		
		t.displayGeom = t.divTabs = t.currentPlan = t.ulTabs = t.curTab =
			t.panelHeight = t.planInfo = t.featureCounts = t.drawMode = null;
		
		t.featureInfo = {"paddocks":[],"networks":[],"equipment":[],"storage":[],"takes":[],"other":[]};
		t.drawing = F;
		
		t.planTitle.val("");
		t.CheckTitle();
		return T;
	}
	else
	{
		return F;
	}
}

/**
 * Copies the entire contents of the current plan to a new plan title that the user will be prompted for.
 */
mcip.prototype.CopyPlan = function()
{
	
	var t = this, copyTitle = t.planTitle.val(), np = prompt("Create a copy of this plan under a new name:",copyTitle+" (copy)");
	if (!!np && np.length>0)
	{
		if ($.inArray(np,t.planList)>=0 || t.displayGroups[np])
		{
			alert("Error: The name '"+np+"' is already used in this map.");
			return F;
		}
		else
		{
			if (t.currentPlan && !t.ClosePlan(t.currentPlan,T)) return F;
			
			var pc = $(MapChat.RunFunction("Plan_Copy",{args:{plan_id:t.plansLUT[copyTitle].id,
				newTitle:np,mapfile:t.map.currentMap}}));
			
			if (pc.find("plan_id:first").text().length>0)
			{
				t.plansLUT[np] = $.parseJSON(pc.find("plan_info:first").text());
				t.planList.push(np);
				t.currentPlan = np;
				t.planTitle.val(np);
				t.LoadPlan(pc);
				t.planTitle.unautocomplete();
				t.BindTitle();
				t.CheckTitle();
			}
		}
	}
}

/**
 * Deletes a plan
 */
mcip.prototype.DeletePlan = function()
{
	
	var t = this, dt = t.planTitle.val(), cp = F, dp;
	
	if (dt==t.currentPlan)
	{
		if (this.curTab.data("changed")) cp = dp = confirm("Discard changes, close, and delete the plan '"+dt+"'?");
		else  cp = dp = confirm("Close and delete the plan '"+dt+"'?");
	}
	else dp = confirm("Delete the plan '"+dt+"'?");
	
	if (dp)
	{
		if (cp) t.ClosePlan(dt,F,T);
		pd = $.parseJSON($(MapChat.RunFunction("Plan_Delete",{args:{plan_id:t.plansLUT[dt].id}})).find("plan_deleted:first").text());
		if (pd)
		{
			t.plansLUT[dt] = null;
			delete t.plansLUT[dt];
			t.planList.splice($.inArray(dt,t.planList),1);
			t.planTitle.val(cp?"":t.currentPlan||"");
			t.planTitle.unautocomplete();
			t.BindTitle();
			t.CheckTitle();
		}
		else alert("Error deleting the plan on the server.");
	}
}

/**
 * Binds events to the planTitle input that makes sure the open/create/close/copy/etc... buttons are in the correct state.
 */
mcip.prototype.BindTitle = function()
{
	var t = this;
	t.planTitle.autocomplete(t.planList,{minChars:0,selectFirst:F,
		max:999,highlight:F,mustMatch:F,matchContains:T,scrollHeight:220});
	t.planTitle.keyup(function(){ t.CheckTitle(); }).keydown(function(){ t.CheckTitle(); })
		.keypress(function(){ t.CheckTitle(); }).click(function(){ t.CheckTitle(); })
		.blur(function(){ t.CheckTitle(); }).focus(function(){ t.CheckTitle(); });
}