/**
 * An object that produces a GUI for controlling map layers in MapChat.
 *
 * @class MapChat.LayerControl
 *
 * @param kamap - the ka-Map API object that this LayerControl object relates to.
 * @param options - an object listing options for the dialog (e.g., autoOpen:true)
 *
 * Depends:
 *	kamap/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 *	MapChat/MapLayers.js
 *	jquery/ui/ui.dialog.js
 */
MapChat.LayerControl = function(kamap,opts)
{
	if (!opts) opts = {};
	var t = this,
		map = (t.map = kamap),
		mapLayers = (t.mapLayers = MapChat.getMapClassObject("MapChat.MapLayers",kamap)),
		container = (t.container = $('<div class="mcTools-mapLayerContainer autoscroll layerstate-listener display-group-listener"><ul></ul></div>'))
		options = (t.options = $.extend({},opts)),
		dialog = (t.dialog = null),
		groups = (t.groups = null),
		displaygroups = (t.displaygroups = null),
		cur_ly = (t.cur_ly = null),
		dgs = (t.dgs = {}),
		ly_lists = (t.ly_lists = {});
	
	container.bind("LayerStatesChanged",function(e,tl,dgs){t.LayerStatesChanged(tl,dgs);});
	
	container.bind("ResetDisplayGroup",function(e,dgn){t.ResetDisplayGroup(dgn);});
	
	if (options.autoOpen)
	{
		t.Open();
	}
}

/**
 * Opens the layer control object, and builds all the stuff that goes into it.
 */
MapChat.LayerControl.prototype.Open = function()
{
	$.blockUI();
	
	var self = this,
		g = (self.groups = this.mapLayers.getGroups()),
		dg = (self.displaygroups = this.mapLayers.getDisplayGroups());
	
	MapChat.QueueEvent(3,"LayerControl",["Open"]);
	
	if (this.dialog==null)
	{
		this.container.dialog({title:"Map Layers",width:200,height:200,modal:false,autoOpen:false,closeOnEscape:false,
			resizeStart:function()
			{
				if (self.cur_ly) self.cur_ly.hide();
			},
			dragStart:function()
			{
				if (self.cur_ly) self.cur_ly.hide();
			},
			resizeStop:function()
			{
				// Make the continer is fit to the dialog's height (so we get scrolling):
				self.container.css({height:$(".ui-dialog-buttonpane", self.dialog).offset().top-self.container.offset().top});
			},
			buttons:{"Update Map":function(){
				var toggleLayers;
				$.each(self.ly_lists,function(dgn){
					$(this).find(".mcTools-mapLayerCheckbox").each(function(){
						var lycb = $(this), cbStatus = (lycb.val()==1), oLayer = dg[dgn].Layers[lycb.attr("name")];
						if (cbStatus!=oLayer.status)
						{
							if (!toggleLayers) toggleLayers = {};
							if (!toggleLayers[oLayer.mapfile.name]) toggleLayers[oLayer.mapfile.name] = {};
							toggleLayers[oLayer.mapfile.name][oLayer.name] = cbStatus;
						}
					});
				});
				if (toggleLayers)
				{
					self.mapLayers.SubmitLayerStates(toggleLayers);
				}
			}},
			close:function()
			{
				if (self.cur_ly) self.cur_ly.hide();
				$.each(self.ly_lists,function(){ $(this).hide();} );
				MapChat.QueueEvent(3,"LayerControl",["Close"]);
			}
		});
		
		//this.container.data("dialog").uiDialog.data("draggable").setData("containment","document");
		
		$.each(dg,function(){
			self.AddNewDisplayGroup(this);
		});
		
		MapChat.AddEventCallback(this.mapLayers.getMapItem(),
			"DisplayGroupAdded",function(mf,dg){
				self.displaygroups[dg.name] = dg;
				self.AddNewDisplayGroup(dg,true);
			});
		
		MapChat.AddEventCallback(this.mapLayers.getMapItem(),
			"DisplayGroupDropped",function(mf,dgn){
				self.RemoveDisplayGroup(dgn);
			});
		
		this.dialog = this.container.parents(".ui-dialog:first");
		
		MapChat.AddContext(this.map,this.dialog);
		
	}
	
	this.container.dialog("open").dialog("moveToTop");
	
	this.container.css({height:$(".ui-dialog-buttonpane", this.dialog).offset().top-this.container.offset().top});
	
	$.unblockUI();
	
	return this;
	
}

/**
 * Handles removal of display groups.
 *
 * @param object dgn - the name of the display group to remove.
 */
MapChat.LayerControl.prototype.RemoveDisplayGroup = function(dgn)
{
	var t = this;
	t.dgs[dgn].remove();
	t.ly_lists[dgn].remove();
	delete t.dgs[dgn];
	delete t.ly_lists[dgn];
}

/**
 * Resets a layer list after classes in one of the layers has been changed.
 *
 * @param object dgn - the name of the display group to reset.
 */
MapChat.LayerControl.prototype.ResetDisplayGroup = function(dgn)
{
	var t = this;
	if (t.ly_lists[dgn])
	{
		t.ly_lists[dgn].hide().children("ul.mcTools-mapLayerList").children().remove();
		t.displaygroups[dgn].displayed = F;
		t.displaygroups[dgn].classes = N;
		
		var updateLayerClasses = [];
		$.each(t.displaygroups[dgn].Layers,function(ln,l){
			updateLayerClasses.push(ln);
		});
		
		var newLayerClasses = $.parseJSON($(MapChat.RunFunction("Map_Info",{args:{mapfile:t.map.currentMap,
			info:"layer_classes",layer:$.toJSON(updateLayerClasses)}})).find("layer_classes:first").text());
		
		$.each(newLayerClasses,function(ln,cl){
			if (t.displaygroups[dgn] && t.displaygroups[dgn].Layers
				&& t.displaygroups[dgn].Layers[ln])
				t.displaygroups[dgn].Layers[ln].classes = cl;
		});
		
	}
}

/**
 * Handles new display groups - adds them to the top of the layer control
 * based on the assumption that newly added layers will appear on top.
 *
 * @param object dg - an object of properties for the display group.
 * @param boolean prepend - if true, the layer is added to the top.
 */
MapChat.LayerControl.prototype.AddNewDisplayGroup = function(dg,prepend)
{
	var self = this,
		dg_li = $('<li class="mcTools-mapLayerGroup"><a style="cursor:pointer;"><span> '+dg.name+'</span></a></li>');
	
	self.dgs[dg.name] = dg_li;
	
	if (prepend && self.container.find("ul:first").children(":first").length>0)
	{
		self.container.find("ul:first").children(":first").before(dg_li);
	}
	else
	{
		self.container.find("ul:first").append(dg_li);
	}
	
	var dg_ly_div = $('<div class="mcTools-mapLayerList" />').appendTo(document.body)
		.bind("mouseleave",function(){
			//dg.timeout = setTimeout(function(){dg_ly_div.hide();},1000);
		})
		.bind("mouseenter",function(){
			//clearTimeout(dg.timeout);
		}),
		dg_ly = $('<ul class="mcTools-mapLayerList" />').appendTo(dg_ly_div),
		dg_cb = $('<img class="mcTools-mapLayerCheckbox" src="'+
			(dg.status?(dg.partial_status?'images/checkbox_partial.png':'images/checkbox_checked.png'):'images/checkbox_unchecked.png')+'" />')
			.val(dg.status?(dg.partial_status?2:1):0)
			
			// This is only for ie6's benefit, but it blurts a bunch of 'stack overflow at line: 0' error messages, so too bad for ie6 users.
			// .load(function(){ MapChat.applyPNGFilter(this); })
			
			.click(function(e){
				var cb = $(this), setStatus = (cb.val()==1?0:1);
				cb.val(setStatus).attr("src",(setStatus==1?'images/checkbox_checked.png':'images/checkbox_unchecked.png'));
				
				if (!dg.displayed) dg_li.find("span").trigger("click",{stayHidden:true});
				MapChat.QueueEvent(3,"LayerControl",["ToggleDisplayGroup",dg_li.find("span").text(),setStatus]);
				
				dg_ly.find(".mcTools-mapLayerCheckbox").each(function(){
					$(this).val(setStatus).attr("src",(setStatus==1?'images/checkbox_checked.png':'images/checkbox_unchecked.png'));
				});
			})
			
			.bind("checkStatus",function(e){
				var status, partial;
				dg_ly.find(".mcTools-mapLayerCheckbox").each(function(){
					partial = (partial||(status===false && $(this).val()==1)||(status && $(this).val()!=1));
					status = (status||($(this).val()==1));
				});
				setStatus = (partial?2:(status?1:0));
				$(this).val(setStatus).attr("src",
					(setStatus==2?'images/checkbox_partial.png':
						(setStatus==1?'images/checkbox_checked.png':'images/checkbox_unchecked.png')
					));	
			});
	
	dg.checkbox = dg_cb;
	
	dg_li.find("a:first").prepend(dg_cb);
	
	dg.displayed = false;
	
	self.ly_lists[dg.name] = dg_ly_div;
			
	dg_li.find("span").click(function(e,opts){
		$.blockUI();
		
		if (dg.timeout) clearTimeout(dg.timeout);
		
		if (!dg.displayed)
		{
			var ly = dg.getLayers();
			
			$.each(ly,function(){
				
				var cly = this, ly_li = $('<li></li>').appendTo(dg_ly);
						
				ly_li.append('<a class="mcTools-mapLayerItem"><span> '+this.name+'</span></a>');
				
				if (cly.classes.length==1)
				{
					ly_li.find("a:first").prepend('<img class="mcTools-mapLayerIcon" src="api?function=Map_ClassIcon&args[map]='+
						dg.mapfile.name+'&args[layer]='+cly.index+'&args[class]='+0+(MapChat.useCookies?'':'&sid='+MapChat.session.sid)+
						'&rand='+Math.random()+'" />');
				}
				
				if (self.groups.tile_source!="cache")
				{
					ly_li_cb = $('<img class="mcTools-mapLayerCheckbox" src="'+
						(cly.status?'images/checkbox_checked.png':'images/checkbox_unchecked.png')+'" />').val(cly.status?1:0)
						
						// This is only for ie6's benefit, but it blurts a bunch of 'stack overflow at line: 0' error messages, so too bad for ie6 users.
						// .load(function(){ MapChat.applyPNGFilter(this); })
						
						.click(function(e){
							var tcb = $(this);
							tcb.val((tcb.val()==1?0:1));
							tcb.attr("src",(tcb.val()==1?'images/checkbox_checked.png':'images/checkbox_unchecked.png'));
							dg_cb.trigger("checkStatus");
							MapChat.QueueEvent(3,"LayerControl",["ToggleLayer",tcb.attr("name"),tcb.val()]);
						})
						
						.attr("name",this.name)
						
						.prependTo(ly_li.find("a:first"));
					
					cly.checkbox = ly_li_cb;
					
				};
				
				if (cly.classes.length>1)
				{
					ly_li.find("span:first").addClass("mcTools-mapLayerMultiClass");
					var ly_li_cl = $('<ul class="mcTools-mapLayerIconList"></ul>').appendTo(ly_li);
					$.each(cly.classes,function(ci,cv){
						ly_li_cl.append('<li><img class="mcTools-mapLayerIcon" src="api?function=Map_ClassIcon&args[map]='+
							dg.mapfile.name+'&args[layer]='+cly.index+'&args[class]='+
							ci+(MapChat.useCookies?'':'&sid='+MapChat.session.sid)+'&rand='+Math.random()+'" /> '
						+cv+'</li>');
					});
					ly_li.find("span:first").click(function(){
						$(this).toggleClass("mcTools-mapLayerExpanded");
						ly_li_cl.toggle();
						if ((dg_ly_div.height()+dg_ly_div.offset().top)>$(window).height())
						{
							MapChat.LayerControl.MakeListScrollable(dg_ly_div,$(window).height()-dg_ly_div.offset().top);
						}
						if (dg_ly.height()+6<dg_ly_div.height() && dg_ly.height()+6<$(window).height()-dg_ly_div.offset().top)
						{
							MapChat.LayerControl.MakeListUnscrollable(dg_ly_div,dg_ly.height()+6);
						}
					});
				}
			});
			dg.displayed = true;
		}
		
		if (opts && opts.stayHidden)
		{
			dg_ly_div.hide();
		}
		else
		{
			// Determine the optimal placement for the list - seems:
			dg_os = dg_li.offset();
			c_w = self.container.width();
			
			ww = $(window).width();
			wh = $(window).height();
			
			ly_w = dg_ly_div.width();
			ly_h = dg_ly_div.height();
			
			ly_left = (((ww-dg_os.left-c_w)<ly_w)?dg_os.left-ly_w:dg_os.left+c_w);
			
			ly_top = (((wh-dg_os.top)<ly_h)?(ly_h<=wh?wh-ly_h:0):dg_os.top);
			
			dg_ly_div.css({left:ly_left,top:ly_top,position:'absolute','z-index':self.dialog.css('z-index')+1});
			
			if (ly_h>wh)
			{
				MapChat.LayerControl.MakeListScrollable(dg_ly_div,wh);
			}
			
			if (self.cur_ly && self.cur_ly!=dg_ly_div) self.cur_ly.hide();
			
			self.cur_ly = dg_ly_div;
			
			dg_ly_div.toggle();
		}
		$.unblockUI();
	});
}

/**
 * This should callback after the mapLayers' SubmitLayerStates method is called.  This will ensure
 *  consistency of the checkbox states in the layer control.
 *
 * @param object tl - the toggleLayers object that was supplied to the SubmitLayerStates method.
 * @param object dgs - array of displaygroup state information
 */
MapChat.LayerControl.prototype.LayerStatesChanged = function(tl,dgs)
{
	var t = this,
		ml = t.mapLayers;
	
	$.each(tl,function(mapname,ly_list){
		$.each(ly_list,function(ly_name,ly_state){
			var oLayer = ml.mapfiles[mapname].Layers[ly_name];
			
			if (oLayer)
			{
				// update the layer's checkbox state
				if (oLayer.checkbox) oLayer.checkbox.attr('src','images/checkbox_'+(oLayer.status?'':'un')+'checked.png');
			}
		});
	});
	
	// Now update all affected display groups:
	$.each(dgs[t.map.currentMap],function(dgn){
		var oDGroup = t.displaygroups[dgn];
		if (oDGroup && oDGroup.checkbox) oDGroup.checkbox.attr('src','images/checkbox_'+(oDGroup.partial_status?'partial':(oDGroup.status?'checked':'unchecked'))+'.png');
	});
	
}

/**
 * Makes an object scrollable by setting overflow to auto, specifying a fixed height, and
 * expanding the width of the element to account for the scrollbar.
 */
MapChat.LayerControl.MakeListScrollable = function(div,maxHeight)
{
	div.css({width:div.width()+18,overflow:'auto',height:maxHeight});
}

/**
 * Resets an object to have an automatic width/height.
 */
MapChat.LayerControl.MakeListUnscrollable = function(div)
{
	div.css({width:'auto',height:'auto'});
}

