/**
 * An object for accessing/managing mapfile layers in MapChat.  The idea is to manage mapfile layers
 * from a single object - any other objects that need to interact or modify the mapfile's layers
 * should use a common instance of this object for a given ka-Map API object.  Maps, layers, 
 * and layer groups might be extended with events that other objects can listen to (e.g., when
 * layers are changed in some way).
 *
 * @class MapChat.MapLayers
 *
 * @param kamap - the ka-Map API object that this MapLayers object relates to.
 *
 * Depends:
 *	kamap/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 */
MapChat.MapLayers = function(kamap)
{
	var self = this,
		kamap = (this.kamap = kamap),
		mapfiles = (this.mapfiles = {});
};

// an alias for the MapChat.MapLayers object:
var mcml = MapChat.MapLayers;

/**
 * This holds the information for each mapfile in the MapLayers' ka-Map object.
 *
 * @param string mapname - the name of the mapfile in the ka-Map object.
 */
mcml.MapFile = function(mapname)
{
	var name = (this.name = mapname),
		Layers = (this.Layers = {}),
		eventCallbacks = (this.eventCallbacks = {
			// these arrays should contain callback functions that should be
			// executed if/when groups are added/removed from the mapfile...
			GroupsUpdated:[],
			DisplayGroupsUpdated:[],
			DisplayGroupDropped:[],
			GeneralUpdate:[],
			GroupAdded:[],
			DisplayGroupAdded:[],
			LayerListUpdate:[]
		});
};

/**
 * Return a specified object from the mapfiles property.
 *
 * @param string mapname - the name of the mapfile to list groups
 *  from. If omitted, this.kamap.currentMap is used.
 */
mcml.prototype.getMapItem = function(mapname)
{
	// if the mapfile layer info object for the current hasn't been created yet, make a new one...
	if (!this.mapfiles[mapname||this.kamap.currentMap])
	{
		return this.mapfiles[mapname||this.kamap.currentMap] = new mcml.MapFile(mapname||this.kamap.currentMap);
	}
	else
	{
		return this.mapfiles[mapname||this.kamap.currentMap];
	}
};

// The following functions get lists of layer groups (either mapfile layer groups,
// or MapChat display groups).  They should only update the actual lists in the object
// if something has changed in the organization of layer groups - if this happens, any
// references to existing layer objects will be broken.

/**
 * Update the mapfile's layer group information from the server.
 *
 * @param string mapname - the name of the mapfile to list groups
 *  from. If omitted, this.kamap.currentMap is used.
 * @param object groups - allows the group information to be retrieved from
 *  another function that can call this script.
 */
mcml.prototype.getGroups = function(mapname,groups)
{
	var mf = this.getMapItem(mapname);
	
	// If any group information is missing, or is flagged as changed...get it from the server.
	if (!mf.Groups || mf.groupsChanged)
	{
		var groupInfo = groups||MapChat.RunFunction("Map_Info",{"args[mapfile]":mf.name,"args[info]":"groups"});
		
		mf.Groups = {};
		$.each($.parseJSON($(groupInfo).find("groups").text()),function(){
			mf.Groups[this.name] = new mcml.Group(mf,this);
		});
		mf.groupsChanged = F;
		
	}
	
	return mf.Groups;
	
};

/**
 * Update the mapfile's layer display group information from the server.
 *
 * @param string mapname - the name of the mapfile to list groups
 *  from. If omitted, this.kamap.currentMap is used.
 * @param object groups - allows the group information to be retrieved from
 *  another function that can call this script.
 */
mcml.prototype.getDisplayGroups = function(mapname,groups)
{
	var mf = this.getMapItem(mapname);
	
	// If any group information is missing, or is flagged as changed...get it from the server.
	if (!mf.DisplayGroups || mf.displayGroupsChanged)
	{
		var groupInfo = groups||MapChat.RunFunction("Map_Info",{"args[mapfile]":mf.name,"args[info]":"display_groups"});
		
		mf.DisplayGroups = {};
		$.each($.parseJSON($(groupInfo).find("display_groups").text()),function(){
			mf.DisplayGroups[this.name] = new mcml.DisplayGroup(mf,this);
		});
		mf.displayGroupsChanged = F;
		
	}
	
	return mf.DisplayGroups;
	
};

/**
 * Update the mapfile's layer group/display group information from the server.
 *
 * @param string mapname - the name of the mapfile to list groups
 *  from. If omitted, this.kamap.currentMap is used.
 */
mcml.prototype.getAllGroups = function(mapname)
{
	var mf = this.getMapItem(mapname);
	
	// If any group information is missing, or is flagged as changed...get it from the server.
	if (!mf.DisplayGroups || !mf.Groups || mf.groupsChanged || mf.displayGroupsChanged)
	{
		var groupInfo = MapChat.RunFunction("Map_Info",{"args[mapfile]":mf.name,"args[info]":"all_groups"});
		
		this.getGroups(mapname,groupInfo);
		
		this.getDisplayGroups(mapname,groupInfo);
	}
	
	return mf;
	
};

/**
 * Load the entire list of layers...instead of just the group names.
 *
 * @param string mapname - the name of the mapfile to list layers
 *  from. If omitted, this.kamap.currentMap is used.
 */
mcml.prototype.getAllLayers = function(mapname)
{
	var mf = this.getMapItem(mapname);
	
	var layerInfo = $.parseJSON($(MapChat.RunFunction("Map_Info",{"args[mapfile]":mf.name,"args[info]":"all_layers"}))
		.find("all_layers").text());
	
	if (!mf.Groups || mf.groupsChanged) mf.Groups = {};
	if (!mf.DisplayGroups || mf.displayGroupsChanged) mf.DisplayGroups = {};
	
	$.each(layerInfo,function(){
		
		if (!mf.Groups[this.group])
			mf.Groups[this.group] = new mcml.Group(mf,this.group);
		if (!mf.DisplayGroups[this.display_group])
			mf.DisplayGroups[this.display_group] = new mcml.DisplayGroup(mf,this.display_group);
		
		var oLayer = new mcml.Layer(mf,this);
		
		if (!mf.Groups[this.group].Layers[this.name])
			mf.Groups[this.group].Layers[this.name] = oLayer;
		if (!mf.DisplayGroups[this.display_group].Layers[this.name])
			mf.DisplayGroups[this.display_group].Layers[this.name] = oLayer;
			
	});
	
	return mf;
	
};

/**
 * This just gets a list of layer names, without any extra layer properties
 * (e.g., for tools like Identify, that just need a dropdown list).
 *
 * @param string mapname - the name of the mapfile to list layers
 *  from. If omitted, this.kamap.currentMap is used.
 * @param object propsFilter - an associative array of property names/values that
 *  must exist in the layer object on the server (e.g., {status:1}).
 * @param object metaFilter - an associative array of metadata names/values that
 *  must exist in the layer object on the server (e.g., {mapchat_display_group:"Cadastral Layers"}).
 */
mcml.prototype.getLayerNames = function(mapname,propsFilter,metaFilter)
{
	if (!propsFilter) propsFilter = {};
	if (!metaFilter) metaFilter = {};
	
	var mf = this.getMapItem(mapname),
		layerNameListID = $.toJSON([propsFilter,metaFilter]);
	
	// layer name lists will be cached based on the filters applied - if anything has changed, this cache will be reset
	if (!mf.layerNameLists || mf.groupsChanged || mf.displayGroupsChanged || mf.layersChanged) mf.layerNameLists = {};
	
	// If the current filter isn't in the cache, then get the corresponding list of layer names.
	if (!mf.layerNameLists[layerNameListID])
	{
		mf.layerNameLists[layerNameListID] = $.parseJSON($(MapChat.RunFunction("Map_Info",{"args[mapfile]":mf.name,"args[info]":"layer_names",
		"args[props]":$.toJSON(propsFilter),"args[meta]":$.toJSON(metaFilter)})).find("layer_names").text());
	}
	
	return mf.layerNameLists[layerNameListID];
};

/**
 * This just gets a list of layer names, without any extra layer properties
 * (e.g., for tools like Identify, that just need a dropdown list).
 *
 * @param object toggleLayers - an array of mapfile/layer names with true/false states to apply
 *  to the map states on the server.  If the response from the server is successuful, then apply
 *  those to the corresponding layer ojbects in this map.
 */
mcml.prototype.SubmitLayerStates = function(toggleLayers)
{
	var t = this;
	$.blockUI();
	
	var result = $(MapChat.RunFunction("Map_LayerStates",{"args[0]":$.toJSON(toggleLayers)}));
	
	if (result.find("success").length);
	{
		var drawGroups = {};
		
		$.each(toggleLayers,function(mapname,ly_list){
			$.each(ly_list,function(ly_name,ly_state){
				var oLayer = t.mapfiles[mapname].Layers[ly_name];
				if (oLayer)
				{
					if (oLayer.status != ly_state) drawGroups[oLayer.group] = T;
					oLayer.Update({"status":ly_state});
				}
			});
		});
		
		var groupStates = $.parseJSON(result.find("group_states").text());
		var displayGroupStates = $.parseJSON(result.find("display_group_states").text());
		
		$.each(groupStates,function(mapname,g_list){
			$.each(g_list,function(){
				var group = this;
					kaGroup = t.kamap.aMaps[mapname].getLayer(group.name),
					mcGroup = t.getGroups()[group.name];
					
				mcGroup.status = group.status;
				mcGroup.partial_status = group.partial_status;
				
				if (kaGroup.visible!=group.status) kaGroup.setVisibility(group.status);
				
				kaGroup.aLayers = group.layers;
				
				// Hack to make groups that started out in an off state to display - the group's visible
				// scales should actually be determined by the Init.php script, but ATM we're not concerned
				// with scale visibility for layer groups, so we'll look the other way for now:
				
				if (!kaGroup.isVisible() && group.status)
				{
					$.each(kaGroup.scales,function(i){ kaGroup.scales[i] = 1; });
					if (t.kamap.currentMap==mapname)
					{
						t.kamap.setMapLayers();
						kaGroup.setTileLayer();
					}
				}
				
				// Only redraw the group if it's in the currently displayed mapfile:
				else if (drawGroups[group.name] && kaGroup.visible && t.kamap.currentMap==mapname && kaGroup.aLayers.length>0)
				{
					kaGroup.tileSource = 'redraw';  // Force this layer to redraw...
					kaGroup.redrawInterval = 1;  // Give it a redraw interval of 1 - otherwise it doesn't actually redraw.
					kaGroup.setTileLayer();
				}
				
			});
		});
		
		$.each(displayGroupStates,function(mapname,g_list){
			$.each(g_list,function(dgn,states){
				var dg = t.getDisplayGroups()[dgn];
				dg.status = states.status;
				dg.partial_status = states.partial_status;
			});
		});
		$(".layerstate-listener").trigger("LayerStatesChanged",[toggleLayers,displayGroupStates]);
		
	}
	$.unblockUI();
};

/**
 * Adds a displaygroup of layers to the map.
 *
 * @param object dg - the properties of the display group that the
 *  layers belong to.
 * @param object g - the properties of the group that should be added
 *  to the kamap object.
 * @param string m - the name of the mapfile to add the display
 *  group to. If omitted, this.kamap.currentMap is used.
 */
mcml.prototype.addDisplayGroup = function(dg,g,m)
{
	var mf = this.getMapItem(m);
	
	// Only do stuff if there isn't already a display group by this name...
	if (!mf.DisplayGroups[dg.name])
	{
		// If the layer group isn't already loaded, add it to the groups list,
		// and add a corresponding kamap _layer object current map in the kamap object:
		if (!mf.Groups[g.name])
		{
			var aScales = [];
			$.each(this.kamap.aMaps[mf.name].aScales,function(){aScales.push(1);});
			g.scales = aScales;
			mf.Groups[g.name] = new mcml.Group(mf,g);
			var kg = new _layer(g);
			this.kamap.aMaps[mf.name].addLayer(kg);
			this.kamap.setMapLayers();
			kg.setTileLayer(T);
			
			MapChat.TriggerEvent(mf,'GroupAdded',mf.Groups[g.name]);
			
			
			// Using the new event approach - anything that should respond to this should be added to the maps context
			// property, have the 'layerlist-listener' class added to it, and a binding for an event named 'LayerListUpdate':
			$(".layergroup-listener",this.kamap.context).trigger("GroupAdded",[mf.Groups[g.name]]);
			
		}
		
		// Otherwise, we only need to reset the layers in this group to include those
		// that belong to the new displaygroup:
		else
		{
			var kg = this.kamap.aMaps[mf.name].getLayer(g.name);
			$.each(g.layers,function(){kg.aLayers.push(this);});
			kg.aLayers.sort();
			if (!kg.visible) kg.setVisibility(T);
			if (kg.isVisible()) kg.setTileLayer(T);
		}
		
		oNewDG = {};
		oNewDG[dg.name] = new mcml.DisplayGroup(mf,dg);
		mf.DisplayGroups = $.extend(oNewDG,mf.DisplayGroups);
		MapChat.TriggerEvent(mf,'DisplayGroupAdded',mf.DisplayGroups[dg.name]);
		
		// Using the new event approach - anything that should respond to this should be added to the maps context
		// property, have the 'layerlist-listener' class added to it, and a binding for an event named 'LayerListUpdate':
		$(".displaygroup-listener",this.kamap.context).trigger("DisplayGroupAdded",[mf.DisplayGroups[dg.name]]);
		
		mf.layerNameLists = F; //forces layer lists to be re-requested - not terribly efficient, but works for now.
		MapChat.TriggerEvent(mf,'LayerListUpdate');
		
		// Using the new event approach - anything that should respond to this should be added to the maps context
		// property, have the 'layerlist-listener' class added to it, and a binding for an event named 'LayerListUpdate':
		$('.layerlist-listener',this.kamap.context).trigger("LayerListUpdate");
	}
};

/**
 * Drops a displaygroup of layers to the map.
 *
 * @param string dgn - the name of the display group is to be dropped
 * @param string skipRefresh - if set, the group (not display group) of this name is not refreshed.
 * @param string m - the name of the mapfile to drop the display group
 *  from. If omitted, this.kamap.currentMap is used.
 */
mcml.prototype.dropDisplayGroup = function(dgn,skipRefresh,m)
{
	var t = this, mf = this.getMapItem(m);
	
	// Only do stuff if there is a display group by this name...
	if (mf.DisplayGroups[dgn])
	{
		var dg = mf.DisplayGroups[dgn];
		
		// Get the server to delete the layers from the mapfile...which should return the adjustments to indexes of affected layers,
		// and updated group layer index arrays - we'll need to use this information to refresh the layer indexes for any layers
		// that were above the dropped layers.
		var a = $.parseJSON($(MapChat.RunFunction("Map_DropDisplayGroup",{args:{dg:dgn,mapfile:mf.name}})).find("affected:first").text());
		
		$.each(a.layers,function(n,s){ mf.Layers[n].index = mf.Layers[n].index-s; });
		$.each(a.groups,function(n,l){
			kg = t.kamap.aMaps[mf.name].getLayer(n)
			kg.aLayers = l;
			kg.userLayer = T;
			if (skipRefresh!=n) kg.setTileLayer();
		});
		
		// be sure to remove any lingering references to the display group's layers...
		$.each(dg.Layers,function(){
			var ly = this;
			if (ly.mapfile.Groups[ly.group]) delete ly.mapfile.Groups[ly.group].Layers[ly.name];
			delete (ly.mapfile.Layers[ly.name]);
		});
		
		MapChat.TriggerEvent(mf,'DisplayGroupDropped',dgn);
		
		// Using the new event approach - anything that should respond to this should be added to the maps context
		// property, have the 'layerlist-listener' class added to it, and a binding for an event named 'LayerListUpdate':
		$(".displaygroup-listener",this.kamap.context).trigger("DisplayGroupDropped",[dgn]);
		
		delete mf.DisplayGroups[dgn];
		
		mf.layerNameLists = F; //forces layer lists to be re-requested - not terribly efficient, but works for now.
		MapChat.TriggerEvent(mf,'LayerListUpdate');
		
		// Using the new event approach - anything that should respond to this should be added to the maps context
		// property, have the 'layerlist-listener' class added to it, and a binding for an event named 'LayerListUpdate':
		$('.layerlist-listener',this.kamap.context).trigger("LayerListUpdate");
	}
};

/**
 * An object for containing information about single layer group (not display group)
 *
 * @param mf - the corresponding mapfile object from the MapChat.MapLayers object's mapfiles property.
 * @param g - array of properties for the group.
 */
mcml.Group = function(mf,g)
{
	$.extend(this,g);
	var Layers = (this.Layers = {}),
		mapfile = (this.mapfile = mf),
		eventCallbacks = (this.eventCallbacks = {
			statusChanged:[]
		}),
		Status = (this.Status = mcml.Status),
		Update = (this.Update = mcml.Update);
};

/**
 * Returns an array of layer objects that correspond to this group.  If the
 * list hasn't been loaded before, then it is requested from the server.
 */
mcml.Group.prototype.getLayers = function()
{
	var self = this, oLayer,
		layerInfo = $.parseJSON($(MapChat.RunFunction("Map_Info",{"args[mapfile]":this.mapfile.name,
			"args[info]":"group_layers","args[group]":this.name})).find("group_layers").text());
	$.each(layerInfo,function(){
		
		// Re-use existing layer objects if they have already been created
		if (self.mapfile.Layers[this.name])
		{
			oLayer = self.mapfile.Layers[this.name];
			// TODO: oLayer.update(this); - maybe implement this to update new properties from the server, but at the moment it's not needed
		}
		else
		{
			oLayer = new mcml.Layer(self.mapfile,this);
		}
		
		self.Layers[this.name] = oLayer;
		
	});
	
	return this.Layers;
	
};

/**
 * An object for containing information about single layer display group (not regular group)
 *
 * @param mf - the corresponding mapfile object from the MapChat.MapLayers object's mapfiles property.
 * @param dg - array of properties for the displaygroup
 */
mcml.DisplayGroup = function(mf,dg)
{
	$.extend(this,dg);
	var Layers = (this.Layers = {}),
		mapfile = (this.mapfile = mf),
		eventCallbacks = (this.eventCallbacks = {
			statusChanged:[]
		}),
		Status = (this.Status = mcml.Status),
		Update = (this.Update = mcml.Update);
};

/**
 * Returns an array of layer objects that correspond to this display group.  If the
 * list hasn't been loaded before, then it is requested from the server.
 */
mcml.DisplayGroup.prototype.getLayers = function()
{
	var self = this, oLayer,
		layerInfo = $.parseJSON($(MapChat.RunFunction("Map_Info",{"args[mapfile]":this.mapfile.name,
			"args[info]":"display_group_layers","args[group]":this.name})).find("display_group_layers").text());
	$.each(layerInfo,function(){
		
		// Re-use existing layer objects if they have already been created
		if (self.mapfile.Layers[this.name])
		{
			oLayer = self.mapfile.Layers[this.name];
			// TODO: oLayer.update(this); - maybe implement this to update new properties from the server, but at the moment it's not needed
		}
		else
		{
			oLayer = new mcml.Layer(self.mapfile,this);
		}
		
		self.Layers[this.name] = oLayer;
		
	});
	
	return this.Layers
	
};

/**
 * Returns a single Layer object for the named layer
 */
mcml.prototype.getLayerInfo = function(layer,mapfile)
{
	var self = this, mf = this.getMapItem(mapfile), oLayer;
	
	if (mf.Layers[layer]) return mf.Layers[layer];
	
	var li = $.parseJSON($(MapChat.RunFunction("Map_Info",{args:{mapfile:mf.name,
		info:"layer_info",layer:layer}})).find("layer_info:first").text());
	
	// If the layer doesn't exist, the value of li should be false...
	if (li===F) return li;
	
	return new mcml.Layer(mf,li);
};

/**
 * An object for containing information and events relating to mapfile layers.
 * 
 * @param mf - the corresponding mapfile object from the MapChat.MapLayers object's mapfiles property.
 * @param layerInfo - the object array of the layer's information
 */
mcml.Layer = function(mf,layerInfo)
{
	$.extend(this,layerInfo);
	var t = this, mf = (t.mapfile = mf),
		eventCallbacks = (t.eventCallbacks = {
			statusChanged:[]
		}),
		Status = (t.Status = mcml.Status),
		Update = (t.Update = mcml.Update);
	
	// Add this layer to the mapfile's array of layer objects.
	mf.Layers[t.name] = t;
};

/**
 * Gets/sets the on/off status of the current object.  This is meant to be added as a method of individual Layer/Group objects.
 *
 * @param boolean status - if set, this will change the status of the mapfile, and trigger any statusChanged callbacks.
 */
mcml.Status = function(status)
{
	if (status==null)
	{
		return this.status;
	}
	else if (status!=this.status)
	{
		this.status = status;
		MapChat.TriggerEvent(this,'statusChanged');
	}
};

/**
 * Updates general properties of an object.  This is meant to be added as a method of individual Layer/Group objects.
 *
 * @param object props - an list of properties/values.
 */
mcml.Update = function(props)
{
	var self = this;
	$.each(props,function(n,v){
		if (v!=self[n])
		{
			self[n] = v;
			if (self.eventCallbacks[n+"Changed"])
			{
				MapChat.TriggerEvent(self,n+'Changed');
			}
		}
	});
};


