/**
 * Construct a menu object instance on a given kaMap instance
 *
 * @class
 * MapChat.mapMenu is a tool that handles right-clicks on the 
 * corresponding map object:
 *
 * @base kaTool
 * @param Object oKaMap - the kaMap object to display the menu on
 * @param string defaultMenu - the html for the default menu items.
 */
MapChat.MapMenu = function( oKaMap, defaultMenu ) {
	kaTool.apply( this, [oKaMap] );
	
	this.map = oKaMap;
	
	/** the name of this tool */
	this.name = 'MapChat.MapMenu';
	
	/** this is an info tool - it should always have events triggered on it */
	this.bInfoTool = true;
	
	/** get all the prototype functions from the kaTool object */
	for (var p in kaTool.prototype) {
		this[p] = kaTool.prototype[p];
	}
	
	/** store a list of menu objects */
	this.menus = [];
	
	/** load the defaultMenu */
	this.menus[0] = this.loadMenu(defaultMenu,this.map);
	
	/** set the current menu */
	this.currentMenu = this.menus[0];
	
	/** overload the mouseup function to do our stuff */
	this.onmouseup = function(e)
	{
		e = $.event.fix(e?e:event);
		if (e.button==2||e.metaKey||e.shiftKey) // Added ctrl/shift+click ability, for browsers like Epiphany & Opera
		{
			$(this.currentMenu)
				.css({top:e.pageY+1,left:e.pageX+1})
				.slideDown(500);
			this.map.theInsideLayer.focus();
			MapChat.QueueEvent(3,"MapMenu",["Show",e.pageX,e.pageY]);
		}
		return false;
	};
	
	/** overload the mousedown function - hide any open menus */
	this.onmousedown = function(e)
	{
		e = $.event.fix(e?e:event);
		cm = this.currentMenu;
		$(document).find(".mcTools-menu").each(function(){
			if (e.button==2||e.metaKey||e.shiftKey) // Added ctrl/+click ability, for browsers like Epiphany & Opera
			{
				// if it's a right-click, only hide non-current menus
				if (this != cm[0])
				{
					$(this).slideUp(500);
				}
			}
			else
			{
				$(this).slideUp(500);
			}
		});
	}
	
	/**
	 * overload the keypress function - if ESC is pressed, hide any 
	 * open menus, then defer to the kaTool keypress event - not the
	 * most logical way of dealing with this, but it seems to work for:
	 */
	this.onkeypress = function(e)
	{
		e = $.event.fix(e?e:event);
		if (e.keyCode==27) //escape key
		{
			$.each(this.menus,function(){ $(this).slideUp(500); });
		}
		kaTool.prototype.onkeypress.apply(this, [e]);
	}
	
	/** overload the mouseout function - so mousing over the menu doesn't mess-up the ESC key */
	this.onmouseout = function(e)
	{
		return false;
	}
};


/**
 * Generate a menu object from HTML text containing a <DIV>, and a <UL>
 * with a series of <LI> items.  Each item should have
 *
 * @param string html contains the DIV and UL/LI elements for the menu
 */
MapChat.MapMenu.prototype.loadMenu = function(html,map)
{
	var t = this, menu = $(html)
		.css({display:"none",position:"absolute","z-index":1})
		.addClass(".mcTools-menu");
	
	// Setup each menu item:
	menu.find("li").each(function(){
		
		var li = this, $li = $(li)
			.click(function(){ li.act=false; t.runMenuItem.apply(li,[map]); })
			.hover(
				function(){ $(li).css({"background-color": '#CCF',border:'1px solid #AAF'}); },
				function(){ $(li).css({"background-color": '#FFF',border:'1px solid #FFF'}); }
			);
		
		// save the LI's attributes to the jQuery object's data store.
		$li.data("object",$li.attr("object"));
		$li.data("otype",$li.attr("otype"));
		$li.data("deps",$.parseJSON($li.attr("deps")));
		$li.data("cargs",$.parseJSON($li.attr("cargs")));
		$li.data("action",$li.attr("action"));
		$li.data("args",$.parseJSON($li.attr("args")));
		
		// remove the attributes from the markup.
		$li.removeAttr("object").removeAttr("otype").
		removeAttr("deps").removeAttr("cargs").
		removeAttr("action").removeAttr("args");
		
	});
	$("#MapChat").append(menu);
	return menu;
}

/**
 * Run the action for a menu item - if an addtional script is required,
 * then load that first with the action as a callback.
 */
MapChat.MapMenu.prototype.runMenuItem = function(map)
{
	var t = this, $t = $(t), cl = null;
	if ($t.attr("require")) MapChat.Using($t.attr("require"));
	$t.parents("div.mcTools-menu:first").slideUp(500,function(){
		
		var o = $t.data("object"), otype = $t.data("otype"),
			deps = $t.data("deps"), cargs = $t.data("cargs"),
			act = $t.data("action"), args = $t.data("args"), a = [];
		
		// if there are dependencies, then merge the object name into an array with the dependencies:
		if (deps) o = [o,deps];
		else o = [o,null];
		
		// if there are additional arguments for creating the class object:
		if (cargs) o.push(cargs);
		
		// Get whatever class this menu item is configured for:
		if (otype=="MapSingleton")
			cl = MapChat.MapSingleton(o,map);
		else if (otype=="Singleton")
			cl = MapChat.Singleton(o);
		else
			cl = MapChat.Class(o);
		
		if (!args) args = [];
		
		cl[act].apply(cl,args);
		
	});
}