/**
 * An object for accessing/managing tool bars attached to maps in MapChat
 *
 * @class MapChat.ToolBars
 *
 * @param kamap - the ka-Map API object that this object relates to.
 *
 * Depends:
 *	kamap/kaMap.js
 *	MapChat.js
 *	MapChat/application.js
 *	jquery/ui/ui.core.js
 *	jquery/ui/ui.draggable.js
 */
MapChat.ToolBars = function(kamap)
{
	var map = (this.map = kamap),
		toolBars = (this.toolBars = {});
		//buttonGroups = (this.buttonGroups = {});
}

/**
 * Construct a tool bar identified by the name of its xml node in
 * the discussion config xml file.
 *
 * @param string configName - the node name in the discussion config xml file to get items from.
 */
MapChat.ToolBars.prototype.getToolBar = function(configName)
{
	if (!this.toolBars[configName])
	{
		this.toolBars[configName] = new MapChat.ToolBar({
			items:MapChat.discussion.config.find(configName+":first").children(),
			title:MapChat.discussion.config.find(configName+":first").attr("title"),
			parent:$('#MapChat'),
			containment:'document',
			top:0,
			left:22,
			map:this.map,
			tbs:this
		});
	}
	else
	{
		this.toolBars[configName].Show();
	}
	return this.toolBars[configName];
}


/**
 * Construct a menu object instance on a given kaMap instance
 *
 * @class MapChat.ToolBar a basic toolbar element.
 *
 */
MapChat.ToolBar = function(options)
{
	var self = this,
		
		options = (this.options = options),
		
		map = (this.map = options.map),
		
		tbs = (this.tbs = options.tbs),
		
		tbContainer = (this.tbContainer = $('<div style="display:none;" class="mcTools-mapToolBar"></div>').appendTo(options.parent)),
			
		tbHandle = (this.tbHandle = 
			$('<div class="mcTools-mapToolBarHandle"><span>'+
				(options.title||'Toolbar')+
				'</span></div>').appendTo(tbContainer)
				.append('<div class="mcTools-mapToolBarClose"><span>X</span></div>')),
	
		tbStrip = (this.tbStrip = $('<div />').addClass("mcTools-mapToolBarStrip").appendTo(tbContainer)),
	
		activeItem = (this.activeItem = {});
	
	$.each(options.items,function(i,v){ tbStrip.append(MapChat.xmlText(v)); });
	
	options.autoShow = (options.autoShow!=undefined?options.autoShow:true);
	
	// Assuming icon images are 16px wide, with a 1px border, and .spanner elements are 5px wide:
	tbContainer.css({
		width:Math.max(75,(tbStrip.find(".icon").length*18)+
			(tbStrip.find(".spanner").length*5)),
		top:(options.top||0),
		left:(options.left||20),
		"z-index":1000
	});
	
	tbClose = $(".mcTools-mapToolBarClose",tbHandle).hover(
		function()
		{
			$(this).addClass('mcTools-mapToolBarClose-hover');
		},
		function()
		{
			$(this).removeClass('mcTools-mapToolBarClose-hover');
		}
	).click(function(){
		self.Hide();
		MapChat.QueueEvent(3,"ToolBars",["Close",options.title]);
	});
	
	tbContainer.draggable({ handle:tbHandle,containment: (options.containment||"document")});
	
	$.each(tbStrip.find('.icon'),function(){
		
		var $this = $(this), item = new MapChat.ToolBarItem({
			el:$this,
			map:map,
			group:$this.attr("group"),
			toggle:$this.attr("toggle"),
			groupdefault:($this.attr("groupdefault")=="true"),
			toolName:$this.attr("tool"),
			title:$this.attr("title"),
			tbName:$this.attr("tbName"),
			tool:false,
			using:($this.attr("using")?$.parseJSON($this.attr("using")):null),
			toolOpts:$.extend(($this.attr("options")?$.parseJSON($this.attr("options")):{}),{map:self.map})
		});
		
		if (item.group)
		{
			if (!self.map.toolGroups[item.group]) self.map.toolGroups[item.group] = {itemsNamed:{},items:[],activeItem:false};
			self.map.toolGroups[item.group].items.push(item);
			if (item.tbName) self.map.toolGroups[item.group].itemsNamed[item.tbName] = item;
		}
		
		if (item.group && item.groupdefault)
		{
			self.map.toolGroups[item.group].activeItem = item;
			item.el.css({margin:"0px",border:"1px solid #F00"});
		}
		item.el.hover(
				function()
				{
					item.el.css({margin:"0px",border:"1px solid #00F"});
				},
				function()
				{
					if (((!item.group || !(item.el[0]==self.map.toolGroups[item.group].activeItem.el[0])) && !item.toggle) ||
						(item.toggle && !item.active))
					{
						item.el.css({margin:"1px",border:"0px"});
					}
					else
					{
						if (!item.toggle) item.el.css({margin:"0px",border:"1px solid #F00"});
						else item.el.css({margin:"0px",border:"1px solid #0F0"});
					}
				}
			)
			.click(function(){
				if (item.group)
				{
					if (ai = self.map.toolGroups[item.group].activeItem)
					{
						ai.deactivate();
						self.previousItem = ai;
					}
					self.map.toolGroups[item.group].activeItem = item;
					item.el.css({margin:"0px",border:"1px solid #F00"});
				}
				
				if (item.toggle)
				{
					if (!item.active)
					{
						item.el.css({margin:"1px",border:"0px"});
						item.activate();
					}
					else
					{
						item.el.css({margin:"0px",border:"1px solid #0F0"});
						item.deactivate();
					}
				}
				else
					item.activate();
				
			});
		if (item.title)
		{
			MapChat.Using("jquery.tooltip");
			item.el.tooltip({
				track: true,
				delay: 0,
				showURL: false,
				extraClass: "tooltip"
			});
		}
	});
	
	(options.autoShow && this.Show());
	
	return this;
}

/**
 * Shows the toolbar on the map.
 */
MapChat.ToolBar.prototype.Show = function()
{
	$.each(this.map.toolGroups,function(){
		if (this.activeItem)
		{
			this.activeItem.getTool();
			this.activeItem.tool.activate();
		}
	});
	this.tbContainer.show();
	MapChat.QueueEvent(3,"ToolBars",["Show",this.options.title]);
};

/**
 * Hides the toolbar from the map.
 */
MapChat.ToolBar.prototype.Hide = function()
{
	with (this)
	{
		$.each(map.toolGroups,function(){ if (this.activeItem && this.activeItem.tool) this.activeItem.tool.deactivate(); });
		tbContainer.hide();
		if (map.defaultTool) map.defaultTool.activate();
	}
};


/**
 * Toolbar item object.
 *
 * @param object options - the properties that should be assigned to the toolbar item.
 */
MapChat.ToolBarItem = function(options)
{
	$.extend(this,options);
};

/**
 * Loads/returns the actual tool associated with the button.
 */
MapChat.ToolBarItem.prototype.getTool = function(){
	with (this)
	{
		if (!tool)
		{
			$.blockUI();
			tool=MapChat.Class([toolName,using],toolOpts);
			$.unblockUI();
		}
	}
};

/**
 * Deactivates the tool and sets the button's style accordingly.
 */
MapChat.ToolBarItem.prototype.deactivate = function()
{
	this.active = F;
	with (this)
	{
		el.css({margin:"1px",border:"0px"});
		if (tool)
		{
			tool.deactivate();
		}
	}
}

/**
 * Activates the tool and sets the button's style accordingly.
 *
 * @param boolean quiet - allows a tool to be reactivated 'quietly'
 */
MapChat.ToolBarItem.prototype.activate = function(quiet)
{
	this.active = T;
	with (this)
	{
		el.css({margin:"0px",border:"1px solid #F00"});
		getTool();
		tool.activate(quiet);
	}
}