<?php 

class MapChat_API_Map_Info
{
	public function __construct()
	{
		// Load the Response Object from the Api Controller
		$this->xml = Zend_Registry::get("XMLResponse");
		$this->discussion = Zend_Registry::get("discussion");
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * Return a JSON-encoded array of map/group/layer information, depending on the input.
	 *
	 * @param string $mapfile - the name of the map that info is being requested for...
	 * @param string $info - the type of info to request (defaults to a list of groups)
	 * @param boolean $reverse - if true, reverses lists of groups/layers (e.g., for intuitive
	 *  display in the client interface, such as the layer control tool).
	 * @param string $group - if the information requested is layers or display_layers, then
	 *  setting this value will limit the returned layers to the corresponding group/display group.
	 * @param string $props - a JSON-encoded associative array of properties that can be used when
	 *  querying for a list of layers ($info="layer_names")
	 * @param string $meta - a JSON-encoded associated array of metadata properties that can be
	 *  used when querying a for a list of layers ($info="layer_names")
	 * @param string $layer - the name of a layer to get information for ($info="layer_info")
	 */
	public function Info($mapfile,$info="groups",$reverse=true,$group="",$props="[]",$meta="[]",$layer="")
	{
		$oMap = $this->discussion->Map_getMapObj($mapfile);
		switch ($info)
		{
			case "all_groups":
				$this->Info($mapfile,"groups",$reverse);
				$this->Info($mapfile,"display_groups",$reverse);
				break;
			
			case "groups":
				$aGroups = $oMap->getAllGroupNames();
				foreach ($aGroups as $groupIdx => $groupName)
				{
					$aLayersIdx = $oMap->getLayersIndexByGroup($groupName);
					$bGroupStatus = NULL;
					$oLayer = $oMap->getLayer($aLayersIdx[0]);
					$sTileSource = $oLayer->getMetaData("tile_source");
					$sImageFormat = $oLayer->getMetaData("imageformat");
					$sRedrawInterval = $oLayer->getMetaData("redraw_interval");
					$sRefreshInterval = $oLayer->getMetaData("refresh_interval");
					$bPartialStatus = false;
					
					foreach ($aLayersIdx as $i)
					{
						$oLayer = $oMap->getLayer($i);
						if (($bGroupStatus && $oLayer->status==MS_OFF) || ($bGroupStatus===false && $oLayer->status==MS_ON))
						{
							$bPartialStatus = true;
						}
						if ($oLayer->status==MS_ON)
						{
							$bGroupStatus = true;
						}
					}
					
					$aGroups[$groupIdx] = array(
						"name"=>$groupName,
						"status"=>$bGroupStatus,
						"tile_source"=>$sTileSource,
						"imageformat"=>$sImageFormat,
						"redraw_interval"=>$sRedrawInterval,
						"refresh_interval"=>$sRefreshInterval,
						"partial_status"=>$bPartialStatus,
					);
					
				}
				$this->xml->groups = Zend_Json::encode(($reverse?array_reverse($aGroups):$aGroups));
				break;
			
			case "display_groups":
				$aDisplayGroupInfo = array();
				$aDisplayGroups = array();
				for ($i=0;$i<$oMap->numlayers;$i++)
				{
					$oLayer = $oMap->getLayer($i);
					$sDisplayGroup = $oLayer->getMetaData("mapchat_display_group");
					if (!isset($aDisplayGroupInfo[$sDisplayGroup]))
					{
						$aDisplayGroupInfo[$sDisplayGroup] = array(
							"name"=>$sDisplayGroup,
							"status"=>($oLayer->status==MS_ON),
							"partial_status"=>false,
						);
					}
					else
					{
						if (($aDisplayGroupInfo[$sDisplayGroup]["status"] && $oLayer->status==MS_OFF) || 
							(!$aDisplayGroupInfo[$sDisplayGroup]["status"] && $oLayer->status==MS_ON))
						{
							$aDisplayGroupInfo[$sDisplayGroup]["partial_status"] = true;
						}
						$aDisplayGroupInfo[$sDisplayGroup]["status"] = $aDisplayGroupInfo[$sDisplayGroup]["status"]||($oLayer->status==MS_ON);
					}
					
					// This makes the assumption that a given display group contains only layers from only one plan
					if ($oLayer->getMetaData("mapchat_plan_type")!="")
						$aDisplayGroupInfo[$sDisplayGroup]["plan_type"] = $oLayer->getMetaData("mapchat_plan_type");
				}
				
				// Rewrite the array without group names as keys (which are redundant)...
				foreach ($aDisplayGroupInfo as $sGroup => $aInfo)
				{
					$aDisplayGroups[] = $aInfo;
				}
				
				$this->xml->display_groups = Zend_Json::encode(($reverse?array_reverse($aDisplayGroups):$aDisplayGroups));
				break;
			
			case "layer_classes":
				$aLayers = Zend_Json::decode($layer);
				if (!is_array($aLayers)) $aLayers = array($aLayers);
				$aReturn = array();
				foreach ($aLayers as $layer)
				{
					$aReturn[$layer] = $this->discussion->Layer_getClasses($oMap,$layer,$reverse);
				}
				$this->xml->layer_classes = Zend_Json::encode($aReturn);
				break;
			
			case "layer_info":
				$this->xml->layer_info = Zend_Json::encode($this->discussion->Layer_getProps($oMap,$layer,$reverse));
				break;
			
			case "all_layers":
				$aLayers = array();
				for ($i=0;$i<$oMap->numlayers;$i++)
				{
					$oLayer = $oMap->getLayer($i);
					$aLayers[] = $this->discussion->Layer_getProps($oMap,$oLayer,$reverse);
				}
				$this->xml->all_layers = Zend_Json::encode($reverse?array_reverse($aLayers):$aLayers);
				break;
			
			case "layer_names":
				$aLayers = array();
				for ($i=0;$i<$oMap->numlayers;$i++)
				{
					$oLayer = $oMap->getLayer($i);
					$returnName = true;
					foreach (Zend_Json::decode($props) as $key => $value)
					{
						if (is_array($value)) $returnName = ($returnName && in_array($oLayer->{$key},$value));
						else $returnName = ($returnName && ($oLayer->{$key} == $value));
						if (!$returnName) break;
					}
					if ($returnName)
					{
						foreach (Zend_Json::decode($meta) as $key => $value)
						{
							$returnName = ($returnName && ($oLayer->getMetaData($key) == $value));
							if (!$returnName) break;
						}
					}
					if ($returnName) $aLayers[] = $oLayer->name;
				}
				$this->xml->layer_names = Zend_Json::encode($reverse?array_reverse($aLayers):$aLayers);
				break;
			
			case "group_layers":
				$aLayers = array();
				$aLayersIdx = $oMap->getLayersIndexByGroup($group);
				foreach ($aLayersIdx as $i)
				{
					$oLayer = $oMap->getLayer($i);
					$aLayers[] = $this->discussion->Layer_getProps($oMap,$oLayer,$reverse);
				}
				$this->xml->group_layers = Zend_Json::encode($reverse?array_reverse($aLayers):$aLayers);
				break;
			
			case "display_group_layers":
				$aLayers = array();
				for ($i=0;$i<$oMap->numlayers;$i++)
				{
					$oLayer = $oMap->getLayer($i);
					$sDisplayGroup = $oLayer->getMetaData("mapchat_display_group");
					if ($sDisplayGroup==$group)
					{
						$aLayers[] = $this->discussion->Layer_getProps($oMap,$oLayer,$reverse);
					}
				}
				$this->xml->display_group_layers = Zend_Json::encode($reverse?array_reverse($aLayers):$aLayers);
				break;
		}
		$this->discussion->Event_log(array(
			"type"=>5,
			"text"=>"Map_Info",
			"data"=>array($mapfile,$info,$reverse,$group,$props,$meta,$layer),
			"username"=>$this->me->username,
		));
		return $this->xml;
	}
}

?>