<?php

class MapChat_Discussion
{

	private $_aDiscussionInfo = NULL;
	private $_bExists = false;
	private $_aSubClasses = array();
	
	/*
	 * Identify the properties required to restore this object
	 * (i.e., the properties that must be saved to the session).
	 */
	public function __sleep()
	{
		return array("_aDiscussionInfo","_bExists");
	}
	
	/*
	 * Identify the properties that must be reinitialized when this
	 * object is restored from a saved session:
	 */
	public function __wakeup()
	{
		$this->db = Zend_Registry::get("db");
		$this->config = Zend_Registry::get("config");
		
		if (isset($this->_aDiscussionInfo["config"]))
		{
			$this->discussion_config = simplexml_load_file($this->config->BaseDir."/config/".$this->_aDiscussionInfo["config"]);
		}
	}
	
	/*
	 * Allow this object to be cast as a string containing the discussion name.
	 */
	public function __toString() 
	{
		if (isset($this->_aDiscussionInfo["discussion"]))
		{
			return $this->_aDiscussionInfo["discussion"];
		}
		else
		{
			return "";
		}
	}
	
	/*
	 * Allow the discussion properties accessed as object members.
	 */
	public function __get($n) 
	{
		if (isset($this->_aDiscussionInfo[$n]))
		{
			return $this->_aDiscussionInfo[$n];
		}
		elseif ($n=="discussion_config" && isset($this->_aDiscussionInfo["config"]))
		{
			return $this->discussion_config->config;
		}
		elseif ($n=="mapfiles" && isset($this->_aDiscussionInfo["mapfile"]))
		{
			return explode(",",$this->_aDiscussionInfo["mapfile"]);
		}
		elseif ($n=="name")
		{
			// just a more intuitive alternative to the __toString, or $this->discussion overload:
			return $this->_aDiscussionInfo["discussion"];
		}
		else
		{
			return NULL;
		}
	}
	
//   // Disabled for now - this might make it too easy to mess up the database
//   // while developing/testing code.
//   /*
// 	 * Allow the values of discussion properties to be set as object members.
// 	 */
// 	public function __set($n,$v) 
// 	{
// 		$this->setInfo(array($n => $v));
// 	}
	
	/**
	 * Overflow calls to dynamically subclasses (e.g., $this->Event_log($params) should
	 * load a subclass of the type MapChat_Discussion_Event, and call the log() method
	 * of that class with the supplied arguments.
	 *
	 * @param string $f the name of the subclass and the method to execute, separated
	 *   by an underscore.
	 * @param array $args the arguments supplied to the function.
	 */
	public function __call($f,$args)
	{
		$base = substr($f,0,strpos($f,"_"));
		$func = substr($f,strpos($f,"_")+1);
		
		$r = new ReflectionClass("MapChat_Discussion_".$base);
		if (!isset($this->_aSubClasses[$base]))
		{
			$this->_aSubClasses[$base] = $r->newInstance($this);
		}
		return $r->getMethod($func)->invokeArgs($this->_aSubClasses[$base],$args);
	}

	
	/*
	 * Construct the object by querying for the record in the
	 * discussions table that matches the name supplied to 
	 * the object constructor:
	 */
	public function __construct($sDiscussion=false)
	{
		$this->db = Zend_Registry::get("db");
		$this->config = Zend_Registry::get("config");
		
		if ($sDiscussion!==false && strlen($sDiscussion)>0)
		{
			$discussion = $this->db->fetchRow('SELECT *,st_asewkt(extent) as extent_text FROM discussions WHERE discussion = ? LIMIT 1',$sDiscussion);
			if ($discussion)
			{
				$this->_aDiscussionInfo = $discussion;
				$this->_bExists = true;
				$this->discussion_config = simplexml_load_file($this->config->BaseDir."/config/".$this->_aDiscussionInfo["config"]);
			}
			
			// Even if the discussion does not exist, we should at least
			// store the name, so we know what name the object
			// was initialized with:
			else
			{
				$this->_aDiscussionInfo = array("discussion"=>$sDiscussion);
			}
		}
	}
	
	/*
	 * Returns true if there is an array saved in the aDiscussionInfo
	 * property (i.e., either the discussion existed when the object
	 * was instantiated, or it has since been created).
	 */
	public function exists()
	{
		return $this->_bExists;
	}
	
	/*
	 * This function returns info from the discussion's record in discussions.
	 */
	public function getInfo($sID=false)
	{
		// If an ID is provided that references a column in the discussions
		// table, then return only that particular attribute...
		if ($sID!=false)
		{
			if (isset($this->_aDiscussionInfo[$sID]))
				return $this->_aDiscussionInfo[$sID];
			
			// If an invalid column ID was supplied:
			else
				return false;
		}
		
		// If nothing is supplied for a column ID, then return the full
		// discussion's record as an array:
		else
		{
			return $this->_aDiscussionInfo;
		}
	}
	
	/*
	 * This function updates attributes of the discussion in the database.
	 * the $aInfo must be an array with values referenced by keys that
	 * correspond to the columns in discussions.  This only relates to the
	 * record in the discussions table - this doesn't modify the schema that
	 * is related to a given discussion.
	 */
	public function setInfo($aInfo)
	{
		if (!$this->exists())
		{
			throw new MapChat_Exception("Cannot update non-existent discussion."); 
		}
		
		if (is_array($aInfo)&&count($aInfo)>0)
		{
			// Get the discussions table metatada:
			$aTblGroups = $this->db->describeTable("discussions","public");
			
			// Loop through the supplied values...
			foreach ($aInfo as $sKey => $sValue)
			{
				// If one of the variables doesn't match, throw an exception...
				if (!isset($aTblGroups[$sKey])) 
				{
					throw new MapChat_Exception("Invalid column specified: ".$sKey);
				}
				elseif ($aTblGroups[$sKey]["PRIMARY"]) 
				{
					throw new MapChat_Exception("Altering the column '".$sKey."' is not allowed, as it is a primary key.");
				}
				elseif($sKey=="srid")
				{
					// TODO: add support for modifying the projection?  This would mean reprojecting all spatial geometries
					// stored in the discussion - doesn't seem worth the trouble at this point.
					throw new MapChat_Exception("Altering the spatial reference for this discussion (i.e., ".$sKey." = ".$this->srid.") is not currently supported.");
				}
				elseif($sKey=="mapfile")
				{
					$mapfiles = explode(",",$sValue);
					foreach ($mapfiles as $mapfile)
					{
						if  (!file_exists($this->config->BaseDir."/maps/".$mapfile))
						{
							throw new MapChat_Exception("The mapfile '".$mapfile."' cannot be found in the MapChat maps directory.");
						}
					}
				}
				elseif($sKey=="config" && !file_exists($this->config->BaseDir."/config/".$sValue))
				{
					throw new MapChat_Exception("The configuration file '".$sValue."' cannot be found in the MapChat config directory.");
				}
				elseif($sKey=="title" && strlen($sValue)==0)
				{
					throw new MapChat_Exception("The discussion title must be not be blank - it is needed for display purposes.");
				}
			}
			
			// Clear the $this->discussion_config property if a 'config' file has been specified:
			unset($this->discussion_config);
			
			// Execute the update:
			$this->db->update("discussions",$aInfo,$this->db->quoteInto("discussion = ?",$this->_aDiscussionInfo["discussion"]));
			
			$this->_aDiscussionInfo = $this->db->fetchRow('SELECT *,st_asewkt(extent) as extent_text FROM discussions WHERE discussion = ? LIMIT 1',$this->_aDiscussionInfo["discussion"]);
			
			$this->discussion_config = simplexml_load_file($this->config->BaseDir."/config/".$this->_aDiscussionInfo["config"]);
			
		}
		
		// The input wasn't an array, or was empty:
		else
		{
			throw new MapChat_Exception("Invalid input for MapChat_Group::setInfo() - must be an array.");
		}
		
	}
	
	/*
	 * This function returns either description or discussion (if description is not
	 * available) - useful for getting an appropriate display name.
	 */
	public function displayName()
	{
		return (strlen($this->_aDiscussionInfo["description"])>0?$this->_aDiscussionInfo["description"]:$this->_aDiscussionInfo["discussion"]);
	}
	
	/*
	 * This function can be used to create new discussions.  The $aDiscussionInfo 
	 * parameter should be an array containing values with keys that 
	 * match the columns of the discussions table in the MapChat2 database.
	 * 
	 * @param array $aDiscussionInfo an associative array where the keys correspond to columns
	 *   in the discussions table of the database.
	 * @param array $aMemberUsers - unused at this time
	 * @param array $aMemberGroups - unused at this time
	 */
	public function create($aDiscussionInfo,$aMemberUsers=array(),$aMemberGroups=array())
	{
		// The name is optional, provided that one was supplied when 
		// the current instance of the MapChat_Group object was created...
		if (!isset($aDiscussionInfo["discussion"]))
		{
			if ($this->_aDiscussionInfo["discussion"])
			{
				$aDiscussionInfo["discussion"] = $this->_aDiscussionInfo["discussion"];
			}
			else
			{
				throw new MapChat_Exception("A name must be specified in order to create a new discussion");
			}
		}
		
		$tmpDiscussion = new MapChat_Discussion($aDiscussionInfo["discussion"]);
		if ($tmpDiscussion->exists())
		{
			throw new MapChat_Exception("The discussion '".$aDiscussionInfo["discussion"]."' already exists.");
		}
		
		// Verify that the required parameters are all OK:
		if (strlen($aDiscussionInfo["discussion"])<3)
		{
			throw new MapChat_Exception("The name '".$aDiscussionInfo["discussion"]."' is too short.  Please use three or more characters.");
		}
		if (!isset($aDiscussionInfo["title"])||strlen($aDiscussionInfo["title"])==0)
		{
			throw new MapChat_Exception("A title must be provided for this discussion.");
		}
		if (!ctype_alnum(str_replace("_","",$aDiscussionInfo["discussion"])))
		{
			throw new MapChat_Exception("The name '".$aDiscussionInfo["discussion"]."' is not allowable.  Only alphanumeric characters and unerscores are allowed (i.e., A-Z, a-z, 0-9, and/or _).");
		}
		if (is_numeric($aDiscussionInfo["discussion"][0]))
		{
			throw new MapChat_Exception("The name '".$aDiscussionInfo["discussion"]."' is not allowable - the discussion name must not start with a number.");
		}
		$mapfiles = explode(",",$aDiscussionInfo["mapfile"]);
		foreach ($mapfiles as $mapfile)
		{
			if (!file_exists($this->config->BaseDir."/maps/".$mapfile))
			{
				throw new MapChat_Exception("The mapfile '".$mapfile."' cannot be found in the MapChat maps directory.");
			}
		}
		if (!file_exists($this->config->BaseDir."/config/".$aDiscussionInfo["config"]))
		{
			throw new MapChat_Exception("The configuration file '".$aDiscussionInfo["config"]."' cannot be found in the MapChat config directory.");
		}
		if (!(isset($aDiscussionInfo["srid"]) && is_numeric($aDiscussionInfo["srid"])))
		{
			throw new MapChat_Exception("A numeric spatial reference ID is required.");
		}
		else
		{
			$srid = $this->db->fetchRow("select * from spatial_ref_sys where srid = ?",$aDiscussionInfo["srid"]);
			if (!is_array($srid))
			{
				throw new MapChat_Exception("The spatial reference id (".$aDiscussionInfo["srid"].") is not available in the spatial_ref_sys table.");
			}
		}
		
		// TODO: could add some code to parse the extent field - right now, this will need to be
		// a full PostGIS EWKT representation, but we could optionally allow an array of x-y
		// coordinates, which would be rewritten into an expression that results in a geometry.
		
		$dmod = new MapChat_User((isset($aDiscussionInfo["moderator"])?$aDiscussionInfo["moderator"]:false));
		if (!$dmod->exists())
		{
			throw new MapChat_Exception("An existing user must be specified as the moderator for the discussion.");
		}
		
		// Get the discussions table metatada:
		$aTblGroups = $this->db->describeTable("discussions","public");
		
		// Loop through the supplied values...
		foreach ($aDiscussionInfo as $sKey => $sValue)
		{
			// If one of the variables doesn't match, throw an exception...
			if (!isset($aTblGroups[$sKey])) 
			{
				throw new MapChat_Exception("Invalid column specified: ".$sKey);
			}
		}
		
		// Before moving on, make sure there isn't an existing schema with the same name as the
		// one chosen for the discussion:
		if ($this->_hasSchema())
		{
			throw new MapChat_Exception("Cannot create a discussion named '".$this->discussion."' because the database already has a schema with this name.");
		}
		
		$this->db->beginTransaction();
		
		try
		{
			// Insert the record into the database.
			$this->db->insert('discussions', $aDiscussionInfo);
			
			$dbc = $this->db->getConnection();
			
			// Create the discussion schema:
			if (file_exists($this->config->BaseDir."/setup/db/discussion_schema.sql"))
			{
				// the discussion_schema.sql must have a '?' as a placeholder wherever the discussion
				// name needs to be substituted.  We're also going to replace 'english' with whatever
				// tsvector locale is specified in the config wherever there is an occurrance of 
				// "to_tsvector('english',":
				$dbc->exec(str_replace("to_tsvector('english',","to_tsvector('".$this->config->db_tsconfig."',",str_replace("?",$aDiscussionInfo["discussion"],
					file_get_contents($this->config->BaseDir."/setup/db/discussion_schema.sql"))));
			}
		}
		catch (Exception $e)
		{
			$this->db->rollBack();
			throw new MapChat_Exception($e->getMessage(),$e->getCode(),$e->getLine(),$e->getFile());
		}
		
		$this->_aDiscussionInfo = $this->db->fetchRow('SELECT *,st_asewkt(extent) as extent_text FROM discussions WHERE discussion = ? LIMIT 1',$aDiscussionInfo["discussion"]);
			
		$this->db->commit();
		
		$this->_bExists = true;
		
		$this->discussion_config = simplexml_load_file($this->config->BaseDir."/config/".$this->_aDiscussionInfo["config"]);
		
	}
	
	/*
	 * Determines if there is an existing schema that matches the discussion name:
	 */
	private function _hasSchema()
	{
		$schema = $this->db->fetchRow("select * from pg_namespace where nspname = ? limit 1",$this->discussion);
		return is_array($schema);
	}
	
	/*
	 * Adds members to the discussion.
	 * 
	 * @param MapChat_User|MapChat_Group $member the member to add to the discussion
	 * @param boolean $readonly the readonly state to assign to this member
	 */
	public function addMember($member,$readonly=false)
	{
		// Check the input for either a MapChat_User or MapChat_Group class:
		if (!(gettype($member)=="object" && (get_class($member)=="MapChat_User"||get_class($member)=="MapChat_Group")))
		{
			throw new MapChat_Exception("First argument must be a class of the type MapChat_User or MapChat_Group.");
		}
		
		$mtype = (get_class($member)=="MapChat_User"?"user":"group");
		
		if ($member->exists())
		{
			if (!$this->hasMember($member,false))
			{
				try
				{
					$this->db->insert($mtype."_discussion", array($mtype."name"=>"".$member."","discussion"=>$this->_aDiscussionInfo["discussion"],"readonly"=>(integer)$readonly));
				}
				catch (Exception $e)
				{
					throw new MapChat_Exception($e->getMessage(),$e->getCode(),$e->getLine(),$e->getFile());
				}
			}
// 			else
// 			{
// 				throw new MapChat_Exception("The user '".$member."' is already a member of this discussion");
// 			}
		}
		else
		{
			throw new MapChat_Exception("Cannot add '".$member."', because the user doesn't exist.");
		}
	}
	
	/*
	 * Checks to see if users or groups are members of a discussion:
	 * 
	 * @param MapChat_User|MapChat_Group $member the member to look for
	 * @param boolean $recurse determines whether to check for membership of users via
	 *   group membership (if true), or only direct membership as individuals.
	 * @param boolean $readonly if set, only returns true if the user is a member if their
	 *   readonly status matches this parameter (taking into consideration group permissions
	 *   if recurse is true)
	 */
	public function hasMember($member,$recurse=true,$readonly=NULL)
	{
		// Check the input for either a MapChat_User or MapChat_Group class:
		if (!(gettype($member)=="object" && (get_class($member)=="MapChat_User"||get_class($member)=="MapChat_Group")))
		{
			throw new MapChat_Exception("First argument must be a class of the type MapChat_User or MapChat_Group.");
		}
		
		if ($member->exists())
		{
			switch (get_class($member))
			{
				case "MapChat_User":
					
					$ud_select = $this->db->select()->from('user_discussion')->where("username = ?",$member->username)->where("discussion = ?",$this->discussion);
					if (!is_null($readonly))
					{
						$ud_select->where("readonly = ?::boolean",(integer)$readonly);
					}
					$ud = $this->db->fetchRow($ud_select->limit(1));
					
					
					if (!$ud && $recurse)
					{
						$ud_select = $this->db->select()->from('group_user')->joinUsing("group_discussion","groupname")->where("username = ?",$member->username)->where("discussion = ?",$this->discussion);
						if (!is_null($readonly))
						{
							$ud_select->where("readonly = ?::boolean",(integer)$readonly);
						}
						$ud = $this->db->fetchRow($ud_select->limit(1));
					}
					
					// If we found a result, we're recursing into groups, and looking at readonly users,
					// then make sure there isn't a conflicting non-readonly membership for the user:
					if (is_array($ud) && count($ud)>0 && $recurse && !is_null($readonly) && $readonly)
					{
						return !$this->hasMember($member,true,false);
					}
					else
					{
						return (is_array($ud) && count($ud)>0);
					}
					break;
					
				case "MapChat_Group":
					$gd_select = $this->db->select()->from("group_discussion")->where("groupname = ?",$member->groupname)->where("discussion = ?",$this->discussion);
					if (!is_null($readonly))
					{
						$gd_select->where("readonly = ?::boolean",(integer)$readonly);
					}
					$gd = $this->db->fetchRow($gd_select->limit(1));
					return (is_array($gd)&&count($gd)>0);
					break;
			}
		}
		else
		{
			// Should this cause an exception?  Just returning false for now - obviously the discussion won't have non-existent members.
			//throw new MapChat_Exception("The user or group '".$member."' doesn't exist.");
			return false;
		}
	}
	
	/*
	 * Drops members from the discussion - this doesn't deal with users that are members
	 * via group membership, only with direct membership between the user/group and the
	 * discussion itself, as listed in either user_discussion or group_discussion tables.
	 * 
	 * @param MapChat_User|MapChat_Group $member the member to drop from the discussion
	 */
	public function dropMember($member)
	{
		// Check the input for either a MapChat_User or MapChat_Group class:
		if (!(gettype($member)=="object" && (get_class($member)=="MapChat_User"||get_class($member)=="MapChat_Group")))
		{
			throw new MapChat_Exception("First argument must be a class of the type MapChat_User or MapChat_Group.");
		}
		
		$mtype = (get_class($member)=="MapChat_User"?"user":"group");
		
		// check for membership (without recursion):
		if ($this->hasMember($member,false))
		{
			$this->db->delete($mtype."_discussion",array(
				$this->db->quoteInto($mtype."name = ?",$member->{$mtype."name"}),
				$this->db->quoteInto("discussion = ?",$this->discussion)
			));
		}
// 		else
// 		{
// 			// Should this cause an exception?  If the discussion never had the member, dropping it 
// 			// shouldn't calse a problem since the end result is the same.
// 			throw new MapChat_Exception("The discussion doesn't have the ".$mtype." '".$member."'.");
// 		}
	}
	
	/*
	 * Lists users in the discussion.  If $readonly is not NULL, then
	 * returned users will be those with a readonly attribute that
	 * matches the value of $readonly.
	 * 
	 * @param boolean $recurse indicates whether to check for members via group membership, or only
	 *   consider membership assigned directly to individual users.
	 * @param boolean $readonly if this is assigned true/false, then only users that have a
	 *   matching readonly status are returned (if recurse is true, then group readonly status is
	 *   taken into consideration)
	 * @param boolean $fullRecord if true the function returns an associated array for the
	 *   entire record from the users table.
	 * @param boolean $clean - if true, the password is stripped from the records
	 */
	public function listUsers($recurse=true,$readonly=NULL,$fullRecord=false,$clean=false)
	{
		if ($fullRecord)
		{
			$lu_select = $this->db->select()->from("users")->joinUsing("user_discussion","username",array())->where("discussion = ?",$this->discussion);
			if (!is_null($readonly))
			{
				$lu_select->where("readonly = ?::boolean",(integer)$readonly);
			}
			$lu = $this->db->fetchAll($lu_select);
		}
		else
		{
			$lu_select = $this->db->select()->from("user_discussion",array("username"))->where("discussion = ?",$this->discussion);
			if (!is_null($readonly))
			{
				$lu_select->where("readonly = ?::boolean",(integer)$readonly);
			}
			$lu = $this->db->fetchCol($lu_select);
		}
		
		if ($recurse)
		{
			if ($fullRecord)
			{
				$lgu_select = $this->db->select()->distinct()->from(array("u"=>"users"))->join(array("gu"=>"group_user"),"gu.username=u.username",array())->join(array("gd"=>"group_discussion"),"gu.groupname=gd.groupname",array())->where("discussion = ?",$this->discussion);
				if (!is_null($readonly))
				{
					$lgu_select->where("readonly = ?::boolean",(integer)$readonly);
				}
				$lgu = $this->db->fetchAll($lgu_select);
			}
			else
			{
				$lgu_select = $this->db->select()->from("group_user",array("username"))->joinUsing("group_discussion","groupname")->where("discussion = ?",$this->discussion);
				if (!is_null($readonly))
				{
					$lgu_select->where("readonly = ?::boolean",(integer)$readonly);
				}
				$lgu = $this->db->fetchCol($lgu_select);
			}
			
			if (!$fullRecord) $lu = array_unique(array_merge($lu,$lgu));
			else
			{
				$tempusers = array();
				foreach ($lu as $u) $tempusers[] = $u["username"];
				foreach ($lgu as $i => $u) if (in_array($u["username"],$tempusers)) unset($lgu[$i]);
				$lu = array_merge($lu,$lgu);
			}
			
					
			// If we're looking for readonly users, then we want to exclude anyone that
			// is also a member via groups that are not readonly:
			if ($recurse && $readonly)
			{
				$lu = array_diff($lu,$this->listUsers($recurse,false,false));
			}
		}
		
		if ($clean && $fullRecord) foreach ($lu as $i => $u)
		{
			unset($lu[$i]["password"]);
		}
		
		// Sort the user list:
		if (!$fullRecord) sort($lu);
		else
		{
			$sortedUsers = array();
			foreach ($lu as $u) $sortedUsers[$u["username"]] = $u;
			ksort($sortedUsers);
			$lu = array();
			foreach ($sortedUsers as $u) $lu[] = $u;
			
		}
		
		return $lu;
	}
	
	/*
	 * Lists groups in the discussion.  If $readonly is not NULL, then
	 * returned groups will be those with a readonly attribute that
	 * matches the value of $readonly.
	 * 
	 * @param boolean $readonly if null, readonly status is ignored, otherwise, list only
	 *   the member groups that match the value of this parameter
	 * @param boolean $fullRecord if true, the function returns an assocative array of the
	 *   entire record from the groups table.
	 */
	public function listGroups($readonly=NULL,$fullRecord=false)
	{
		if ($fullRecord)
		{
			$lgu_select = $this->db->select()->from("groups")->joinUsing("group_discussion","groupname",array())->where("discussion = ?",$this->discussion);
			if (!is_null($readonly))
			{
				$lgu_select->where("readonly = ?::boolean",(integer)$readonly);
			}
			$lgu = $this->db->fetchAll($lgu_select);
		}
		else
		{
			$lgu_select = $this->db->select()->from("group_discussion",array("groupname"))->where("discussion = ?",$this->discussion);
			if (!is_null($readonly))
			{
				$lgu_select->where("readonly = ?::boolean",(integer)$readonly);
			}
			$lgu = $this->db->fetchCol($lgu_select);
		}
		return $lgu;
	}
	
	/*
	 * Allows members' readonly status to be toggled.
	 * 
	 * @param MapChat_User|MapChat_Group $member
	 * @param boolean $readonly the value to assign to the member's readonly state.
	 */
	public function setReadonly($member,$readonly=true)
	{
		// Check the input for either a MapChat_User or MapChat_Group class:
		if (!(gettype($member)=="object" && (get_class($member)=="MapChat_User"||get_class($member)=="MapChat_Group")))
		{
			throw new MapChat_Exception("First argument must be a class of the type MapChat_User or MapChat_Group.");
		}
		
		$mtype = (get_class($member)=="MapChat_User"?"user":"group");
		
		// check for membership (without recursion):
		if ($this->hasMember($member,false))
		{
			$this->db->update($mtype."_discussion",array("readonly"=>(integer)$readonly),array(
				$this->db->quoteInto($mtype."name = ?","".$member.""),
				$this->db->quoteInto("discussion = ?",$this->discussion),
			));
		}
	}
	
	/*
	 * Get a member's readonly status - raise an error if the member doesn't exist.
	 * 
	 * @param MapChat_User|MapChat_Group $member
	 * @param boolean $recurse determines whether to check group permissions (true), or just
	 *   permissions directly assigned to the individual user.
	 */
	public function getReadonly($member,$recurse=true)
	{
		if ($this->hasMember($member,$recurse))
		{
			return $this->hasMember($member,$recurse,true);
		}
		else
		{
			throw new MapChat_Exception("The user/group '".$member->displayName()."' is not member of this discussion.");
		}
	}
	
	/*
	 * List all discussions in the database.  This can be called without creating an
	 * instance of this object:  MapChat_Discussion::listAllDiscussions();
	 * 
	 * @param boolean $disabled if set, only discussions with matching disabled states
	 *   are returned by this function.
	 */
	public function listAllDiscussions($disabled=NULL)
	{
		if (!is_null($disabled))
		{
			return Zend_Registry::get("db")->fetchAll("select * from discussions where disabled = ? order by discussion;",(integer)$disabled);
		}
		else
		{
			return Zend_Registry::get("db")->fetchAll("select * from discussions order by discussion;");
		}
	}
	
	/*
	 * List all discussion moderator users in the database.  This can be called without creating an
	 * instance of this object:  MapChat_Discussion::listAllDiscussionModerators();
	 * 
	 * @param boolean $currentUser if set, the current user is included (i.e., so that someone
	 *   with can_create_discussion permissions can be the moderator for the groups they create.
	 */
	public function listDiscussionModerators($currentUser=false)
	{
		if ($currentUser===false)
		{
			return Zend_Registry::get("db")->fetchAll("select distinct users.* from users join group_user using (username) join groups using (groupname) where is_admin or can_manage_discussions;");
		}
		else
		{
			return Zend_Registry::get("db")->fetchAll("select distinct users.* from users join group_user using (username) join groups using (groupname) where is_admin or can_manage_discussions or username = ?;",$currentUser->username);
		}
	}
	
	/*
	 * Get the username of the current discussion's moderator:
	 */
	public function getModerator()
	{
		return new MapChat_User($this->moderator);
	}
	
	/**
	 * List user online states
	 *
	 * @param string $username - only return the record for the specified username.
	 * @param boolean $status - if set, only users of the requested online status is returned.
	 * @param string $filter - if set, only users that match this where condition will be returned.
	 * @param boolean $onlyUsername - if set, only usernames are returned in a single array
	 *
	 * @return array
	 */
	public function getOnlineUsers($username=false,$status=NULL,$filter=false,$onlyUsername=false)
	{
		$s = $this->db->select()->from("user_status",array(
			"username"=>"username",
			"online_time"=>new Zend_Db_Expr("round(date_part('epoch',online_time)*1000.0)"),
			"offline_time"=>new Zend_Db_Expr("round(date_part('epoch',offline_time)*1000.0)"),
		),$this->name);
		
		if ($username!==false) $s = $s->where("username = ?",$username);
		if ($status===true) $s = $s->where("offline_time is null or online_time > offline_time");
		elseif ($status===false) $s = $s->where("online_time is null or offline_time >= online_time");
		if ($filter!==false) $s = $s->where($filter);
		
		if ($onlyUsername) return $this->db->fetchCol($s);
		
		$aUsers = $this->db->fetchAssoc($s);
		
		// If only one user was requested, return a single array, otherwise 
		// return the array of arrays for multiple users.
		return (($username!==false)?($aUsers?array_shift($aUsers):false):$aUsers);
	}
	
}

?>