<?php
/**
 * class MapChat_Discussion_Chat
 * Provides the general utilities for managing chat messages in the
 * MapChat database.
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Chat
{
	const REGULAR_MESSAGE = 1;
	
	public function __construct($discussion)
	{
		$this->db = Zend_Registry::get("db");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * Logs a message in the database:
	 * 
	 * @param string $opts - an array of parameters that describe the message:
	 *
	 * Possible values that can be included in the opts parameter:
	 *  clientTime: the time reported by the client, adjusted for the time delta relative to the server (should be pretty much equal).
	 *  parentID: if specified, this message is in reply to a previous message.
	 *  from: if set, this is the user the message will appear to originate from (this can be ignored for now)
	 *  flags: an array of text flags - if defined, these will be saved in a space-separated string in the chat_log
	 *  selection: a list of integers that references a set of features selected by the user - if specified, these will be included
	 *             in the spatial links for the chat message.
	 *  drawing: an integer that references the id of a user-drawn feature - if specified, this will be included
	 *           in the spatial links for the chat message.
	 *  extent: an array of xmin,ymin,xmax,ymax coordinates to explicitly define a geographical extent for this message - if
	 *          the min/max coordinates are the same (or max is missing), the extent will be treated as a point.  This will be set as the
	 *          the extent for this message (ignored if a selection is supplied).
	 *  distance: (not yet implemented) if set, the geometry for the chat_log will be a buffer around the extent and/or selected
	 *            features included above
	 *  type: if set, this will determine the 'type' of message (e.g., if some level of argumentation structure is to be used in a discussion)
	 *  private: if set, this should be an array containing lists of usernames and/or groupnames, to whom the messages are targeted.
	 *  text: the text of the message
	 *
	 * @return id - the integer generated as the identifier for the new record in chat_log
	 */
	public function Send(array $opts)
	{
		$aMsg = array();
		$aReturn = array();
		
		// It's probably best to leave the timestamp up to the server.
		//if (isset($opts["clientTime"])) $aMsg["time"] = new Zend_Db_Expr("to_timestamp(".($opts["clientTime"]*1)."/1000.0)");
		
		if (isset($opts["from"])) $aMsg["username"] = $opts["from"];
		else $aMsg["username"] = $this->me->username;
		
		if (isset($opts["flags"]))
		{
			if (!is_array($opts["flags"])) $opts["flags"] = array($opts["flags"]);
			$aMsg["flags"] = implode(" ",$opts["flags"]);
		}
		
		if (isset($opts["extent"]) && is_array($opts["extent"]))
		{
			$aMsg["extent"] = 'SRID='.$this->discussion->srid.';POLYGON(('.$opts["extent"][0]." ".$opts["extent"][1].",".
				$opts["extent"][2]." ".$opts["extent"][1].",".$opts["extent"][2]." ".$opts["extent"][3].",".
				$opts["extent"][0]." ".$opts["extent"][3].",".$opts["extent"][0]." ".$opts["extent"][1].'))';
		}
		
		$aMsg["type"] = @$opts["type"];
		$aMsg["text"] = @$opts["text"];
		
		// If a parent message is named, add its depth+1 as the value for depth (otherwise, it defaults to zero).
		if (isset($opts["parentID"]))
		{
			$aMsg["parent_id"]=$opts["parentID"];
			$aMsg["depth"] = new Zend_Db_Expr($this->db->quoteInto("(select depth+1 as parent_depth from ".$this->discussion.".chat_log where id = ?)",$opts["parentID"]));
		}
		else
			$aMsg["depth"] = 0;
		
		$this->db->insert($this->discussion.".chat_log",$aMsg);
		$newID = $this->lastChatID();
		
		// If target users or groups are named, then add the corresponding links in the chat_target_users and chat_target_gropus tables.
		if (isset($opts["private"]))
		{
			if (isset($opts["private"]["target_users"]) && count($opts["private"]["target_users"]))
			{
				foreach ($opts["private"]["target_users"] as $pUser)
				{
					$this->db->insert($this->discussion.".chat_target_users",array("id"=>$newID,"username"=>$pUser));
				}
			}
			
			if (isset($opts["private"]["target_groups"]) && count($opts["private"]["target_groups"]))
			{
				foreach ($opts["private"]["target_groups"] as $pGroup)
				{
					$this->db->insert($this->discussion.".chat_target_groups",array("id"=>$newID,"groupname"=>$pGroup));
				}
				
			}
		}
		
		
		if (isset($opts["selection"]) && is_array($opts["selection"]))
		{
			/** TODO: check for $opts["distance"] - if present, the geometry for the record should be buffer of the selected geometries. **/
			foreach ($opts["selection"] as $selID)
			{
				$this->db->insert($this->discussion.".chat_selection_links",array("chat_id"=>$newID,"selection_id"=>$selID));
			}
		}
		
		if (isset($opts["drawing"]))
		{
			
			/** TODO: check for $opts["distance"] - if present, the geometry for the record should be buffer of the selected geometries. **/
			
			// Get the data for the drawing(s):
			$drawings = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
				"deleted, visible, case when st_geometrytype(geom) in ".
				"('ST_Polygon','ST_MultiPolygon') then '3' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
				"then '2' when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then '1' when geom is null then null else '4' end as geom_type, geom, ".
				"st_area(geom) as area, st_perimeter(geom) as perimeter, st_length(geom) as length, ".
				"st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from ".
				$this->discussion.".user_drawings where id in ".MapChat_Array::toIntPG($opts["drawing"])." and username = ".
				$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetchAll(PDO::FETCH_ASSOC);
			
			foreach ($drawings as $d)
			{
				// Make the equivalant information that would have been retrieved via mapscript using the select tool:
				$attr_json = array(
					array(
						"ti"=>-1,"si"=>$d["id"],"ci"=>-1,"area"=>$d["area"],"length"=>max($d["perimeter"],$d["length"]),
						"type"=>$d["geom_type"],"layertype"=>$d["geom_type"],"extent"=>array(array($d["xmin"],$d["ymin"]),array($d["xmax"],$d["ymax"])),
					),
					array(
						"id"=>$d["id"],
						"Label"=>$d["label"],
						"User"=>$this->me->displayName()
					)
				);
				
				// Insert the drawing into the user_selections, but with deleted = true (since this is not actually part of any current selection set)
				$this->db->getConnection()->query("insert into ".$this->discussion.
					".user_selections (username,layer_name,user_layer,user_drawing,deleted,feature_id,attr_json,geom) values (".
					$this->db->quote($this->me->username).",'[User Drawings]',true,true,true,".$this->db->quote($d["id"]).",".
					$this->db->quote(Zend_Json::encode($attr_json)).",".$this->db->quote($d["geom"])."::geometry);");
				
				$this->db->insert($this->discussion.".chat_selection_links",array("chat_id"=>$newID,"selection_id"=>$this->db->lastSequenceId($this->discussion.".user_selections_id_seq")));
			}
		}
		
		return array_pop($this->Get(array("describe"=>array($newID))));
	}

	/**
	 * Updates a message in the database with new properties:
	 * 
	 * @param string $opts - an array of parameters that describe the message:
	 *
	 * Possible values that can be included in the opts parameter:
	 *  id: the ID of the message to edit (required)
	 *  clientTime: the time reported by the client, adjusted for the time delta relative to the server (should be pretty much equal).
	 *  parent_id: if specified, this message is in reply to a previous message.
	 *  flags: an array of text flags - if defined, these will be saved in a space-separated string in the chat_log
	 *  selection: an integer that references the id of a set of features selected by the user - if specified, these will be included
	 *             in the spatial links for the chat message.
	 *  drawing: an integer that references the id of a user-drawn feature - if specified, this will be included
	 *           in the spatial links for the chat message.
	 *  extent: an array of xmin,ymin,xmax,ymax coordinates to explicitly define a geographical extent for this message - if
	 *          the min/max coordinates are the same (or max is missing), the extent will be treated as a point.  This will be set as the
	 *          the extent for this message (ignored if a selection is supplied).
	 *  distance: (not yet implemented) if set, the geometry for the chat_log will be a buffer around the extent and/or selected
	 *            features included above
	 *  type: if set, this will determine the 'type' of message (e.g., if some level of argumentation structure is to be used in a discussion)
	 *  text: the text of the message
	 *
	 * @return array - the message queried from $this->Get(id)
	 */
	public function Edit(array $opts)
	{
		// Get a copy of the old message so we can log it in the events table when we're done:
		$aOldMsg = array_shift($this->Get(array("describe"=>array($opts["id"]),"maps"=>$opts["maps"],"timeDelta"=>$opts["timeDelta"])));
		// Ensure a copy of the ID is the first element in the oldmsg data:
		array_unshift($aOldMsg,$opts["id"]);
		$aEvent = array("data"=>$aOldMsg,"type"=>MapChat_EventTypes::MESSAGE_EDIT);
		$aMsg = array();
		
		// It's probably best to leave the timestamp up to the server.  This will prevent confusion about messages that might have
		// small margins of error with the difference between server and client time.
		// If the clientTime is provided, then pass it to the event - the timestamp in the chat_log should not be altered:
		//if (isset($opts["clientTime"])) $aEvent["time"] = new Zend_Db_Expr("to_timestamp(".($opts["clientTime"]*1)."/1000.0)");
		
		if (isset($opts["flags"]))
		{
			if (!is_array($opts["flags"])) $opts["flags"] = explode(" ",$opts["flags"]);
			if (!in_array("edited",$opts["flags"])) $opts["flags"][] = "edited";
			$aMsg["flags"] = implode(" ",$opts["flags"]);
		}
		
		if (isset($opts["extent"]) && is_array($opts["extent"]))
		{
			$aMsg["extent"] = 'SRID='.$this->discussion->srid.';POLYGON(('.$opts["extent"][0]." ".$opts["extent"][1].",".
				$opts["extent"][2]." ".$opts["extent"][1].",".$opts["extent"][2]." ".$opts["extent"][3].",".
				$opts["extent"][0]." ".$opts["extent"][3].",".$opts["extent"][0]." ".$opts["extent"][1].'))';
			
			// If there was a previous extent, and there wasn't a request to unlink it, then expand the extent to include both:
			if (!(isset($opts["unlink_extent"]) && $opts["unlink_extent"]) && isset($aOldMsg["extent"]) && strlen($aOldMsg["extent"])>0)
			{
				$aMsg["extent"] = new Zend_Db_Expr('st_envelope(st_collect('.$this->db->quote($aMsg["extent"]).
					',extent))');
			}
		}
		
		// Otherwise, set the extent to null if requested by the user.
		else if (isset($opts["unlink_extent"]) && $opts["unlink_extent"])
		{
			$aMsg["extent"] = new Zend_Db_Expr('null');
		}
		
		$aMsg["type"] = @$opts["type"];
		$aMsg["text"] = @$opts["text"];
		
		// If a new parent message is named, add its depth+1 as the value for depth (otherwise, it defaults to zero).
		if (isset($opts["parent_id"]) && $opts["parent_id"]!=$aOldMsg["parent_id"])
		{
			$aMsg["parent_id"]=$opts["parent_id"];
			$aMsg["depth"] = new Zend_Db_Expr($this->db->quoteInto("(select depth+1 as parent_depth from ".$this->discussion.".chat_log where id = ?)",$opts["parent_id"]));
		}
		
		// If there was previously a parent ID, and the new message data says to delete the current one, then set it to null in chat_log:
		elseif (isset($opts["remove_parent"]) && $opts["remove_parent"] && $aOldMsg["parent_id"]!=NULL)
		{
			$aMsg["parent_id"] = NULL;
			$aMsg["depth"] = 0;
		}
		
		$this->db->update($this->discussion.".chat_log",$aMsg,$this->db->quoteInto("id = ?",$opts["id"]));
		
		// Unlink any previously linked selections if requested:
		if (isset($opts["unlink_selection"]) && $opts["unlink_selection"])
		{
			$this->db->query("update ".$this->discussion.
				".chat_selection_links csl set deleted = true, deletion_date = current_timestamp from ".
				$this->discussion.".user_selections us where us.id = csl.selection_id and ".
				"not us.user_drawing and not csl.deleted and csl.chat_id = ".$this->db->quote($opts["id"]));
		}
		
		// Link any new selections if requested:
		if (isset($opts["selection"]) && is_array($opts["selection"]))
		{
			/** TODO: check for $opts["distance"] - if present, the geometry for the record should be buffer of the selected geometries. **/
			foreach ($opts["selection"] as $selID)
			{
				$this->db->insert($this->discussion.".chat_selection_links",array("chat_id"=>$opts["id"],"selection_id"=>$selID));
			}
		}
		
		// Unlink any previously linked drawings if requested:
		if (isset($opts["unlink_drawing"]) && $opts["unlink_drawing"])
		{
			$this->db->query("update ".$this->discussion.
				".chat_selection_links csl set deleted = true, deletion_date = current_timestamp from ".
				$this->discussion.".user_selections us where us.id = csl.selection_id and ".
				"us.user_drawing and not csl.deleted and csl.chat_id = ".$this->db->quote($opts["id"]));
		}
		
		// Link any new drawings if requested:
		if (isset($opts["drawing"]))
		{
			/** TODO: check for $opts["distance"] - if present, the geometry for the record should be buffer of the selected geometries. **/
			
			// Get the data for the drawing(s):
			$drawings = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
				"deleted, visible, case when st_geometrytype(geom) in ".
				"('ST_Polygon','ST_MultiPolygon') then '3' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
				"then '2' when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then '1' when geom is null then null else '4' end as geom_type, geom, ".
				"st_area(geom) as area, st_perimeter(geom) as perimeter, st_length(geom) as length, ".
				"st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from ".
				$this->discussion.".user_drawings where id in ".MapChat_Array::toIntPG($opts["drawing"])." and username = ".
				$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetchAll(PDO::FETCH_ASSOC);
			
			foreach ($drawings as $d)
			{
				// Make the equivalant information that would have been retrieved via mapscript using the select tool:
				$attr_json = array(
					array(
						"ti"=>-1,"si"=>$d["id"],"ci"=>-1,"area"=>$d["area"],"length"=>max($d["perimeter"],$d["length"]),
						"type"=>$d["geom_type"],"layertype"=>$d["geom_type"],"extent"=>array(array($d["xmin"],$d["ymin"]),array($d["xmax"],$d["ymax"])),
					),
					array(
						"id"=>$d["id"],
						"Label"=>$d["label"],
						"User"=>$this->me->displayName()
					)
				);
				
				// Insert the drawing into the user_selections, but with deleted = true (since this is not actually part of any current selection set)
				$this->db->getConnection()->query("insert into ".$this->discussion.
					".user_selections (username,layer_name,user_layer,user_drawing,deleted,feature_id,attr_json,geom) values (".
					$this->db->quote($this->me->username).",'[User Drawings]',true,true,true,".$this->db->quote($d["id"]).",".
					$this->db->quote(Zend_Json::encode($attr_json)).",".$this->db->quote($d["geom"])."::geometry);");
				
				$this->db->insert($this->discussion.".chat_selection_links",array("chat_id"=>$opts["id"],"selection_id"=>$this->db->lastSequenceId($this->discussion.".user_selections_id_seq")));
			}
		}
		
		//Magic
		
		$this->discussion->Event_log($aEvent);
		
		return $this->Get(array("describe"=>array($opts["id"]),"maps"=>$opts["maps"],"timeDelta"=>$opts["timeDelta"]));
	}

	/**
	 * Gets messages from the database.
	 * 
	 * @param string $opts - only messages that match the options specified in this array will be returned.
	 *
	 * Possible values that can be included in the opts parameter:
	 *  maps: an array of map names that should be checked for user-visibility states.
	 *  sinceTime: if specified, only messages recorded after this number of milliseconds since the UNIX 'epoch' time will be returned.
	 *  sinceID: if specified, -2 gets the most recent batch of messages; greater than or equal to -1 gets the next batch
	 *           of in sequential order messages with IDs greater than sinceID.
	 *  limit: defaults to 25; 0 will retrieve all messages; values greater than zero will limit the number of messages returned to
	 *         the specified value.
	 *  from: if set, only return messages from the specified username
	 *  reverse: if true, then get messages *before* sinceID
	 *  depth: the thread-depth to get messages at; -1/null is any depth, 0 is the top-level message that starts a thread, 1, etc...
	 *  flags: an array of text flags - if defined, only messages that satisfy these flags will be returned.
	 *  selection: an integer that references the id of a set of features selected by the user - only messages explicitly linked
	 *             to these features will be returned.
	 *  intersects: the same as 'selection', but will return any messages that are referenced to geographic features that intersect
	 *              the selection set identified.
	 *  extent: an array of xmin,ymin,xmax,ymax coordinates to explicitly define a geographical extent for this message - if
	 *          the min/max coordinates are the same (or max is missing), the extent will be treated as a point.
	 *  distance: the distance to search for nearby messages (works with intersects or extent).
	 *  centroid: if true (default), spatial filters are evaluated using the centroid of spatial messages.
	 *  list: if true, only IDs will be returned for matching messages
	 *  listgeo: if true, only IDs and centre-coordinates will be returned for matching messages
	 *  describe: if an array of IDs is supplied, details will be returned for the matching messages - NOTE: all above parameters
	 *            will be ignored in the request if this is used
	 *  children: if true, then the messages returned will be replies to sinceID or IDs listed in 'describe' (ignored if sinceID or
	 *            describe are not set)
	 *
	 * @return array - a list of records from chat_log
	 */
	public function Get($opts)
	{
		// This is the last time this function ran during this particular PHP session, for this particular discussion:
		if (is_array(MapChat_Session::ns()->discussions) && isset(MapChat_Session::ns()->discussions[$this->discussion->name]) &&
			is_array(MapChat_Session::ns()->discussions[$this->discussion->name]))
		{
			if (!isset(MapChat_Session::ns()->discussions[$this->discussion->name]["lastChatGet"]))
				MapChat_Session::ns()->discussions[$this->discussion->name]["lastChatGet"] = null;
			$lastChatGet = MapChat_Session::ns()->discussions[$this->discussion->name]["lastChatGet"];
			MapChat_Session::ns()->discussions[$this->discussion->name]["lastChatGet"] = $this->db->fetchOne("select now()::timestamptz");
		}
		else
		{
			$lastChatGet = null;
			if (!is_array(MapChat_Session::ns()->discussions))
			{
				MapChat_Session::ns()->discussions = array($this->discussion->name=>array("lastChatGet"=>$this->db->fetchOne("select now()::timestamptz")));
			}
			else
			{
				MapChat_Session::ns()->discussions[$this->discussion->name] = array("lastChatGet"=>$this->db->fetchOne("select now()::timestamptz"));
			}
		}
		
		// This is the last time this user has polled the chat in this discussion *ever* (including previous sessions):
		$lastChatPoll = $this->db->fetchOne($this->db->select()->from("user_status","last_poll",$this->discussion->name)
			->where("username = ?",$this->me->username));
		
		// Now save lastChatGet in the user_status last_poll column.
		$this->db->update($this->discussion->name.".user_status",
			array("last_poll"=>new Zend_Db_Expr($this->db->quote(MapChat_Session::ns()->discussions[$this->discussion->name]["lastChatGet"])."::timestamptz")),
			$this->db->quoteInto("username = ?",$this->me->username));
		
		// Determine whether to select just the IDs, or full details of the messages:
		if (@($opts["list"] && !$opts["describe"] &&  !$opts["listgeo"])) $cols = "id";
		else if (@($opts["listgeo"] && !$opts["describe"])) $cols = "cl.id, ".
			"'['||st_xmin(cl.extent)||','||st_ymin(cl.extent)||','||st_xmax(cl.extent)||','||st_ymax(cl.extent)||']' as extent, ".
			"csls.selection, csld.drawing";
		else $cols = "cl.id, cl.parent_id, cl.username, cl.type, cl.text, cl.time, cl.flags, cl.depth, ".
			"'['||st_xmin(cl.extent)||','||st_ymin(cl.extent)||','||st_xmax(cl.extent)||','||st_ymax(cl.extent)||']' as extent, ".
			"csls.selection, csld.drawing, round(date_part('epoch',time)*1000.0) as epoch_ms";
		
		// Begin the select statement...
		$s = $this->db->select()->from(array("cl"=>"chat_log"),$cols,$this->discussion->name)
		
			// Join info about linked selections:
			->joinLeft(array("csls"=>$this->db->select()->from(array("us1"=>"user_selections"),
				array("selection"=>new Zend_Db_Expr("'['||min(st_xmin(geom))||','||min(st_ymin(geom))||','||".
					"max(st_xmax(geom))||','||max(st_ymax(geom))||','''||st_assvg(st_centroid(st_union(geom)),0,".
						$this->db->quote($this->discussion->discussion_config->precision).")||''',['||st_x(".
							"case when st_geometrytype(st_union(geom)) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(st_union(geom)) when st_geometrytype(st_union(geom)) in ('ST_LineString') ".
							"then st_line_interpolate_point(st_union(geom),0.5) else st_centroid(st_union(geom)) end)||','||st_y(".
							"case when st_geometrytype(st_union(geom)) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(st_union(geom)) when st_geometrytype(st_union(geom)) in ('ST_LineString') ".
							"then st_line_interpolate_point(st_union(geom),0.5) else st_centroid(st_union(geom)) end)||']]'"),
					"ugeom"=>new Zend_Db_Expr("st_union(geom)")
				),$this->discussion->name)->join(array("csl1"=>$this->discussion.".chat_selection_links"),"us1.id = csl1.selection_id",array("csl1.chat_id"))
				->where("not us1.user_drawing and not csl1.deleted")->group("chat_id")),"csls.chat_id = cl.id",array("selection"))
			
			// Join info about linked drawings:
			->joinLeft(array("csld"=>$this->db->select()->from(array("us2"=>"user_selections"),
				array("drawing"=>new Zend_Db_Expr("'['||min(st_xmin(geom))||','||min(st_ymin(geom))||','||".
					"max(st_xmax(geom))||','||max(st_ymax(geom))||','''||st_assvg(st_centroid(st_union(geom)),0,".
						$this->db->quote($this->discussion->discussion_config->precision).")||''',['||st_x(".
							"case when st_geometrytype(st_union(geom)) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(st_union(geom)) when st_geometrytype(st_union(geom)) in ('ST_LineString') ".
							"then st_line_interpolate_point(st_union(geom),0.5) else st_centroid(st_union(geom)) end)||','||st_y(".
							"case when st_geometrytype(st_union(geom)) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(st_union(geom)) when st_geometrytype(st_union(geom)) in ('ST_LineString') ".
							"then st_line_interpolate_point(st_union(geom),0.5) else st_centroid(st_union(geom)) end)||']]'"),
					"ugeom"=>new Zend_Db_Expr("st_union(geom)")
				),$this->discussion->name)->join(array("csl2"=>$this->discussion.".chat_selection_links"),"us2.id = csl2.selection_id",array("csl2.chat_id"))
				->where("us2.user_drawing and us2.deleted and not csl2.deleted")->group("chat_id")),"csld.chat_id = cl.id",array("drawing"))
			
			// Join counts from the target users table:
			->joinLeft(array("ctu"=>$this->db->select()
			->from("chat_target_users",array("id"=>"id","num_target_users"=>"count(*)",
			"target_users"=>"'['||mapchat_concat_list(username)||']'",
			"target_users_array"=>"mapchat_concat_array(ARRAY[username])"),
			$this->discussion->name)->group("id")),"cl.id = ctu.id",array("target_users"))
			
			// Join counts from the target groups table:
			->joinLeft(array("ctg"=>$this->db->select()
			->from("chat_target_groups",array("id"=>"id","num_target_groups"=>"count(*)",
			"target_groups"=>"'['||mapchat_concat_list(groupname)||']'",
			"target_groups_array"=>"mapchat_concat_array(ARRAY[groupname])"),$this->discussion->name)
			->group("id")),"cl.id = ctg.id",array("target_groups"))
			
			// Get the number of replies per message:
			->joinLeft(array("clr"=>$this->db->select()
			->from("chat_log",array("id"=>"parent_id","replies"=>"'['||mapchat_concat_list(id::varchar)||']'"),$this->discussion->name)
			->group("parent_id")),"cl.id = clr.id",array("replies"));
			
			if (@$opts["sinceID"]==-2)
			{
				// Get the last time a corresponding private dialog was closed.
				$s = $s->joinLeft(array("elpd"=>$this->db->select()
				->from("event_log",array("private_dialog_closed"=>"max(time)","tid"=>new Zend_Db_Expr("data[2]")),$this->discussion->name)
				->where("text = 'ChatPrivateDialog' and username = ?",$this->me->username)->group("data[2]")),
				"elpd.tid = mapchat_targets_json(target_users_array,target_groups_array)",
				array("private_dialog_closed"=>"private_dialog_closed"));
			}
			
			// Add a column indicating if the message is private or not:
			$s = $s->columns(array(
				"private"=>"case when (num_target_users>0 or num_target_groups>0) then true else false end",
			))
		
			// Add conditions that limit results to public messages, the users' own messages, or private messages that 
			// target the current user:
			->where("(num_target_users is null and num_target_groups is null) or (num_target_users+num_target_groups)=0 or ((cl.username = ".$this->db->quote($this->me->username).
			" or ".$this->db->quote($this->me->username)." = any(target_users_array) or ".MapChat_Array::toPG($this->me->listGroups()).
			" && target_groups_array) and (".
				(@($opts["describe"] && !$opts["listgeo"])?"true":
					(@($opts["sinceID"]==-2)?
						"private_dialog_closed is null or time>private_dialog_closed":
						($lastChatPoll&&strlen($lastChatPoll)>0?"time>".$this->db->quote($lastChatPoll)."::timestamptz":"false")).
					(@is_array($opts["private"])?" or (".
						(@is_array($opts["private"]["target_users"])?"target_users_array = ".MapChat_Array::toPG($opts["private"]["target_users"]):"").
						((@is_array($opts["private"]["target_users"]) && @is_array($opts["private"]["target_groups"]))?" and ":"").
						(@is_array($opts["private"]["target_groups"])?"target_groups_array = ".MapChat_Array::toPG($opts["private"]["target_groups"]):"").
					")":"")
				).
			"))")
			->where("not deleted");
		
		/** TODO: need some way to filter results to those linked to a set of selected features **/
		/** TODO: add a filter for the 'intersects' option **/
		
		if (isset($opts["extent"]))
		{
			if (isset($opts["centroid"]) && !$opts["centroid"])
			{
				// Results are based on their full geometries matching in this case - need to fix.
				$s = $s->where("(not cl.extent is null and cl.extent && ".$extentWKT." and st_intersects(cl.extent,".$extentWKT.")) or ".
					"(not csls.ugeom is null and st_intersects(csls.ugeom,".$extentWKT.")) or ".
					"(not csld.ugeom is null and st_intersects(csld.ugeom,".$extentWKT."))");
			}
			else
			{
				// Results must have their centroids match the spatial filter.
				$extentWKT = $this->db->quote("SRID=".$this->discussion->srid.";POLYGON((".($opts["extent"][0]*1.0)." ".($opts["extent"][1]*1.0).
					",".($opts["extent"][0]*1.0)." ".($opts["extent"][3]*1.0).
					",".($opts["extent"][2]*1.0)." ".($opts["extent"][3]*1.0).
					",".($opts["extent"][2]*1.0)." ".($opts["extent"][1]*1.0).
					",".($opts["extent"][0]*1.0)." ".($opts["extent"][1]*1.0)."))");
				
				$cache = false;
				
				if (@$opts["geocache"])
				{
					if (!isset(MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]))
						MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"] = array("extent"=>$extentWKT);
					else if (!isset(MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]["extent"]))
						MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]["extent"] = $extentWKT;
					else
					{
						$cache = MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]["extent"];
						
						$aCache = $this->db->getConnection()->query("select st_asewkt(st_difference(".$extentWKT.",".
							MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]["extent"].
							")) as difference, st_asewkt(st_union(".$extentWKT.",".
							MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]["extent"].
							")) as cache;")->fetch(PDO::FETCH_ASSOC);
						
						MapChat_Session::ns()->discussions[$this->discussion->name]["geocache"]["extent"] = $this->db->quote($aCache["cache"]);
						
						$extentWKT = $this->db->quote($aCache["difference"]);
					}
				}
				
				$s = $s->where("(not cl.extent is null and cl.extent && ".$extentWKT.
					" and st_intersects(st_centroid(cl.extent),".$extentWKT.")".
						($cache?
							" and not st_intersects(st_centroid(cl.extent),".$cache.")":
						"").") or ".
					"(not csls.ugeom is null and st_intersects(case when st_geometrytype(csls.ugeom) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(csls.ugeom) when st_geometrytype(csls.ugeom) in ('ST_LineString') then ".
							"st_line_interpolate_point(csls.ugeom,0.5) else ".
							"st_centroid(csls.ugeom) end,".$extentWKT.")".
						($cache?
							" and not st_intersects(case when st_geometrytype(csls.ugeom) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(csls.ugeom) when st_geometrytype(csls.ugeom) in ('ST_LineString') then ".
							"st_line_interpolate_point(csls.ugeom,0.5) else ".
							"st_centroid(csls.ugeom) end,".$cache.")"
						:"").") or ".
					"(not csld.ugeom is null and st_intersects(case when st_geometrytype(csld.ugeom) in ('ST_Polygon','ST_MultiPolygon') then ".
						"st_pointonsurface(csld.ugeom) when st_geometrytype(csld.ugeom) in ('ST_LineString') then ".
						"st_line_interpolate_point(csld.ugeom,0.5) else ".
						"st_centroid(csld.ugeom) end,".$extentWKT.")".
						($cache?
							" and not st_intersects(case when st_geometrytype(csld.ugeom) in ('ST_Polygon','ST_MultiPolygon') then ".
							"st_pointonsurface(csld.ugeom) when st_geometrytype(csld.ugeom) in ('ST_LineString') then ".
							"st_line_interpolate_point(csld.ugeom,0.5) else ".
							"st_centroid(csld.ugeom) end,".$cache.")"
						:"").")");
			}
		}
		
		// If the 'from' option is set:
		if (isset($opts["from"])) $s = $s->where("cl.username = ?",$opts["from"]);
		
		// If a thread-depth is specified:
		if (isset($opts["depth"]) && $opts["depth"]>=0) $s = $s->where("cl.depth = ?",$opts["depth"]);
		
		// If any flags are requested:
		if (@(is_array($opts["flags"]) && count($opts["flags"])>0))
			$s = $s->where("cl.flags::tsvector @@ ?::tsquery",implode(" & ",$opts["flags"]));
		
		// If children are requested, then add the appropriate conditions to the statement:
		if (@($opts["children"] && isset($opts["sinceID"]))) $s = $s->where("cl.parent_id = ?",$opts["sinceID"]);
		elseif (isset($opts["describe"]) && is_array($opts["describe"]))
		{
			// Make sure all IDs supplied are integers:
			$aIDlist = array();
			foreach (array_unique($opts["describe"]) as $nID)
			{
				if (is_numeric($nID)) $aIDlist[] = intval($nID);
			}
			if (@($opts["children"]))
				$s = $s->where("cl.parent_id in (".implode(",",$aIDlist).")");
			else
				$s = $s->where("cl.id in (".implode(",",$aIDlist).")");
		}
		
		// The default option is to find times/ids greater than sinceTime/sinceID, in chronological order from a given
		// point in time (or from the beginning if unspecified).  If sinceID is -2, this is effectively the same as as
		// requesting items in reverse with no sinceID.
		if (@!($opts["reverse"] || $opts["sinceID"]==-2))
		{
			if (isset($opts["sinceTime"]))
				$s = $s->where("cl.time > to_timestamp(?::bigint/1000.0)",$opts["sinceTime"]);
			
			if (isset($opts["sinceID"]))
				$s = $s->where("cl.id > ?",$opts["sinceID"]);
			
			
			// Make sure messages returned are in chronological order
			$s = $s->order("cl.id asc");
		}
		
		// Otherwise, if $opts["reverse"] is true or $opts["sinceID"]==-2, search backwards from any supplied time/id, in
		// reverse chronological order...
		else
		{
			if (isset($opts["sinceTime"]))
				$s = $s->where("cl.time < to_timestamp(?::bigint/1000.0)",$opts["sinceTime"]);
			
			if (isset($opts["sinceID"]) && @$opts["sinceID"]>=0)  // Ignore sinceID values of -1 or -2
				$s = $s->where("cl.id < ?",$opts["sinceID"]);
			
			// Make sure messages returned are in chronological order
			$s = $s->order("cl.id desc");
		}
		
		// set the limit, if needed.
		switch (true)
		{
			case (!isset($opts["limit"])):
				$s = $s->limit(25); // Default to 25, to prevent entire discussions being downloaded.
				break;
			case ($opts["limit"]>0):
				$s = $s->limit($opts["limit"]); // If any other limit is specified, use that
				break;
			
			// The remaining case of $opts["limit"]<=0 means no limit will be used on the results returned:
		}
		
		// if the selection is in reverse order, then re-sort it in chronological order:
		if (@($opts["reverse"]||$opts["sinceID"]==-2)) $s = $this->db->select()->from(array("tdesc"=>$s))->order("tdesc.id asc");
		
		// For some reason, the query planner is deciding to use nested loops, causing major performance issues...until that
		// is fixed, the enable_nestloop option will be set to 'off' while performing the chat messages query:
		$nestloop = $this->db->fetchOne('show enable_nestloop;');
		$this->db->exec('set enable_nestloop = off;');
		
		$aNewMessages = $this->db->fetchAssoc($s);
		
		// Resetting the enable_nestloop option to whatever it's original value was:
		$this->db->exec('set enable_nestloop = '.$nestloop.';');
		
		if ($lastChatGet)
		{
			// Look for any messages that have changed since the lastChatGet:
			$aEdited = $this->db->fetchCol($this->db->select()
				->from("event_log",array("id"=>new Zend_Db_Expr("data[1]")),$this->discussion->name)
				->where("time >= ?::timestamptz",$lastChatGet)
				->where("type = ?",MapChat_EventTypes::MESSAGE_EDIT)
				->where("not username = ?",$this->me->username));
			
			// If there are any messages that have been changed, send the id list to the client - the client
			// will decide if it needs to download any (i.e., if it has a stale copy of each message):
			if (count($aEdited)>0) $this->discussion->Event_ReturnEvent("EditedMessages",$aEdited);
			
			// Return the extents of any recent changes:
			$aNewMapExts = array();
			if (@is_array($opts["maps"]))
			{
				foreach ($opts["maps"] as $mapfile)
				{
					
					// Get a lookup of which drawings are currently displayed in the user's map:
					$aDisplayed = $this->discussion->User_ListVisible($mapfile);
					
					// Look for any users' drawings that have been updated, and send the coordinates to the client:
					$select = $this->db->select()->from("user_drawings",
						array(
							"username"=>"username",
							"xmin"=>new Zend_Db_Expr("st_xmin(st_extent(geom))"),
							"ymin"=>new Zend_Db_Expr("st_ymin(st_extent(geom))"),
							"xmax"=>new Zend_Db_Expr("st_xmax(st_extent(geom))"),
							"ymax"=>new Zend_Db_Expr("st_ymax(st_extent(geom))"),
						),$this->discussion->name)
					->where(
						$this->db->quoteInto("(visible and (creation_date >= ?::timestamptz",$lastChatGet).
						$this->db->quoteInto(" or modification_date >= ?::timestamptz",$lastChatGet).
						$this->db->quoteInto(" or deletion_date >= ?::timestamptz",$lastChatGet).
						$this->db->quoteInto(")) or (not visible and visibility_change >= ?::timestamptz)",$lastChatGet)
					);
					
					switch (true)
					{
						// Return all by the user's own drawings:
						case (!$aDisplayed[0] && $aDisplayed[2]):
							$select = $select->where("not username = ?",$this->me->username);
						break;
						
						// Return only the user's own drawings (all users is off, and all individual users are off)
						case ($aDisplayed[0] && !$aDisplayed[2] && count($aDisplayed[1])==0):
							$select = $select->where("username = ?",$this->me->username);
						break;
						
						// Return the user's own drawings, and any other individual users' drawings
						case ($aDisplayed[0] && !$aDisplayed[2] && count($aDisplayed[1])>0):
							$select = $select->where("username = ".$this->db->quote($this->me->username).
								" or username = any(".MapChat_Array::toPG($aDisplayed[1]).")");
						break;
						
						// Return only other individual users' drawings
						case (!$aDisplayed[0] && !$aDisplayed[2] && count($aDisplayed[1])>0):
							$select = $select->where("username = any(".MapChat_Array::toPG($aDisplayed[1]).")");
						break;
						
						// Default is to return everything - i.e., current users and all users are on, so the
						// individuals display states don't a difference in terms of refreshing map tiles on the client.
					}
					
					$aNewExts = $this->db->fetchAssoc($select->group("username"));
					
					if (is_array($aNewExts) && count($aNewExts)>0)
						$aNewMapExts[$mapfile] = $aNewExts;
				}
			}
			
			if (count($aNewMapExts)>0)
					$this->discussion->Event_ReturnEvent("NewUserDrawings",$aNewMapExts);
		}
		
		return $aNewMessages;
	}

	/**
	 * Returns the id of the last logged chat message:
	 */
	public function lastChatID()
	{
		return $this->db->lastSequenceId($this->discussion.".chat_log_id_seq");
	}

	/**
	 * Returns the thread-depth of a given message.
	 *
	 * @param integer $id - the id of the message to query the depth for.
	 * @param boolean $calculate - if true, the depth is determined by the mapchat_thread_depth SQL procedure
	 */
	public function ThreadDepth($id,$calculate)
	{
		if ($calculate) return $this->db->fetchOne($this->db->quoteInto("select mapchat_thread_depth(?,'".$this->discussion."');",$id));
		else return $this->db->fetchOne($this->db->select()->from("chat_log","depth",$this->discussion->name)->where("id = ?",$id));
	}
	
	/**
	 * Returns the total number of messages
	 *
	 * @param array $targets - an array with lists of users and/or groups; if set the count
	 *                         returned will indicate the total number of messages matching
	 *                         the selected combination.
	 */
	public function MessageCount($targets=array())
	{
		$s = $this->db->select()->from(array("cl"=>"chat_log"),"count(*)",$this->discussion->name);
		
		if (isset($targets["users"]) && is_array($targets["users"]) && count($targets["users"])>0)
		{
			$sUsers = MapChat_Array::toPG($targets["users"]);
			$s = $s->where("(select mapchat_concat_array(ARRAY[username]) from ".$this->discussion.".chat_target_users ctu where ctu.id = cl.id) @> ".
				$sUsers." and (select mapchat_concat_array(ARRAY[username]) from ".$this->discussion.".chat_target_users ctu where ctu.id = cl.id) <@ ".
				$sUsers);
		}
		
		if (isset($targets["groups"]) && is_array($targets["groups"]) && count($targets["groups"])>0)
		{
			$sGroups = MapChat_Array::toPG($targets["groups"]);
			$s = $s->where("(select mapchat_concat_array(ARRAY[groupname]) from ".$this->discussion.".chat_target_groups ctu where ctu.id = cl.id) @> ".
				$sGroups." and (select mapchat_concat_array(ARRAY[groupname]) from ".$this->discussion.".chat_target_groups ctu where ctu.id = cl.id) <@ ".
				$sGroups);
		}
		
		if (count($targets)==0)
		{
			$s = $s->where("cl.id not in (select id from ".$this->discussion.".chat_target_users) and cl.id not in (select id from ".$this->discussion.".chat_target_groups)");
		}
		
		$s->where("not deleted");
			
		return $this->db->fetchOne($s);
		
	}
	
	/**
	 * Returns svg geometry data for selections/drawings linked to specified chat messages:
	 *
	 * @param array $opts - defines what to return
	 *
	 * Possible values that can be included in the opts parameter:
	 *  selection: if specified, ths should be an array of IDs for chat messages to return selections svg shapes
	 *  drawing: if specified, this should be an array of IDs for chat messages to return drawings as svg shapes
	 */
	public function GetMessageSVG($opts)
	{
		$aReturn = array();
		
		if (isset($opts["selection"]))
		{
		
			$aSelection = $this->db->fetchAssoc($this->db->select()->from(($this->db->select()->from(array("us"=>"user_selections"),
				array("geom"=>"st_union(geom)"),$this->discussion->name)
				->join(array("csl"=>$this->discussion.".chat_selection_links"),"us.id = csl.selection_id",array("chat_id"))
				->where("csl.chat_id in ".MapChat_Array::toIntPG($opts["selection"]))->where("not us.user_drawing and not csl.deleted")
				->group("csl.chat_id")),
					array(
						"id"=>"chat_id",
						"shape"=>new Zend_Db_Expr("st_assvg(geom,0,".
							$this->db->quote($this->discussion->discussion_config->precision).")"),
						"centroid"=>new Zend_Db_Expr("case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
							"then st_assvg(st_pointonsurface(case when st_isvalid(geom) then geom else st_buffer(geom,0) end)) when st_geometrytype(geom) in ('ST_LineString') ".
							"then st_assvg(ST_line_interpolate_point(geom,0.5)) else st_assvg(st_centroid(geom)) end"),
						"type"=>new Zend_Db_Expr("case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
							"then 2 when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
							"then 1 when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 0 else 3 end"),
						"simple"=>new Zend_Db_Expr("st_assvg(st_envelope(geom))")
					)));
			
			$aReturn["select_svg"] = array();
			foreach ($aSelection as $s)
			{
				if (strlen($s["shape"])>$this->discussion->discussion_config->svg_limit)
				{
					$s["shape"] = $s["centroid"];
					$s["type"] = 0;
					unset($s["centroid"]);
				}
				else
				{
					unset($s["centroid"]);
					unset($s["simple"]);
				}
				$aReturn["select_svg"][$s["id"]] = $s;
			}
			
		}
		
		if (isset($opts["drawing"]))
		{
			$aSelection = $this->db->fetchAssoc($this->db->select()->from(($this->db->select()->from(array("us"=>"user_selections"),
				array("geom"=>"st_union(geom)"),$this->discussion->name)
				->join(array("csl"=>$this->discussion.".chat_selection_links"),"us.id = csl.selection_id",array("chat_id"))
				->where("csl.chat_id in ".MapChat_Array::toIntPG($opts["drawing"]))->where("us.user_drawing and us.deleted and not csl.deleted")
				->group("csl.chat_id")),
					array(
						"id"=>"chat_id",
						"shape"=>new Zend_Db_Expr("st_assvg(geom,0,".
							$this->db->quote($this->discussion->discussion_config->precision).")"),
						"centroid"=>new Zend_Db_Expr("case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
							"then st_assvg(st_pointonsurface(case when st_isvalid(geom) then geom else st_buffer(geom,0) end)) when st_geometrytype(geom) in ('ST_LineString') ".
							"then st_assvg(ST_line_interpolate_point(geom,0.5)) else st_assvg(st_centroid(geom)) end"),
						"type"=>new Zend_Db_Expr("case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
							"then 2 when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
							"then 1 when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 0 else 3 end"),
						"simple"=>new Zend_Db_Expr("st_assvg(st_envelope(geom))")
					)));
			
			$aReturn["draw_svg"] = array();
			foreach ($aSelection as $s)
			{
				if (strlen($s["shape"])>$this->discussion->discussion_config->svg_limit)
				{
					$s["shape"] = $s["centroid"];
					$s["type"] = 0;
					unset($s["centroid"]);
				}
				else
				{
					unset($s["centroid"]);
					unset($s["simple"]);
				}
				$aReturn["draw_svg"][$s["id"]] = $s;
			}
			
		}
		
		return $aReturn;
	}
}

?>
