<?php
/**
 * class MapChat_Discussion_Draw
 * Provides the general utilities for managing drawings in the
 * MapChat database.
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Draw
{
	public function __construct($discussion)
	{
		$this->db = Zend_Registry::get("db");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * Modifies the current user's drawn features:
	 * 
	 * @param string $opts - an array of parameters that control how the selection is updated.
	 *
	 * Possible values that can be included in the opts parameter:
	 *  map: the name of the mapfile to query (required)
	 *  cellSize: the client pixel size (required - for snapping...if it is ever implemented)
	 *  mode: one of 'add', 'cut', 'open', 'list' - defines wither to add new coordinates to the current (or new) drawing,
	 *        cut areas out of a drawing, or list the current selection
	 *  style: one of 'point', 'circle', 'line', 'polygon' - defines how the drawing will be performed
	 *  coords: the coordinates that are to be used to create/update a drawing
	 *  props: properties to assign to a drawing
	 *  id: this should be the id of the drawing to modify (if 'add' or 'cut') or return data for (if 'open')
	 *
	 * @return array - the draw results
	 */
	public function Features(array $opts)
	{
		$aResults = array("mode"=>$opts["mode"],"results"=>array($opts["mode"]=>array()));
		$oMap = $this->discussion->Map_getMapObj($opts["map"]);
		
		$sGeom = false;
		
		try
		{
		
			if (isset($opts["coords"]) && is_array($opts["coords"]) && isset($opts["style"]))
			{
				switch ($opts["style"])
				{
					case "polygon":
						foreach ($opts["coords"] as $aShape)
						{
							$aPoints = array();
							
							foreach ($aShape as $aCoord)
							{
								$aPoints[] = implode(" ",$aCoord);
							}
							$aPoints[] = implode(" ",$aShape[0]);
							
							$sPolyWKT = $this->db->quote("SRID=".$this->discussion->srid.";POLYGON((".implode(",",$aPoints)."))");
							
							if (!$sGeom) $sGeom = $sPolyWKT;
							else
							{
								$sGeom = $this->db->quote($this->db->getConnection()->query(
									"select case when st_contains(".$sGeom.",".$sPolyWKT.") then st_difference(".
									$sGeom.",".$sPolyWKT.") else st_union(".$sGeom.",".$sPolyWKT.") end as geom;")->fetchColumn());
							}
						}
					break;
					
					case "circle":
						foreach ($opts["coords"] as $aCircle)
						{
							$sCircle = "st_buffer(".$this->db->quote("SRID=".$this->discussion->srid.";POINT(".$aCircle[0]." ".$aCircle[1].")").
								",".($aCircle[2]*1.0).")";
							
							if (!$sGeom) $sGeom = $sCircle;
							else
							{
								$sGeom = $this->db->quote($this->db->getConnection()->query(
									"select case when st_contains(".$sGeom.",".$sCircle.") then st_difference(".
									$sGeom.",".$sCircle.") else st_union(".$sGeom.",".$sCircle.") end as geom;")->fetchColumn());
							}
						}
					break;
					
					case "line":
						foreach ($opts["coords"] as $aShape)
						{
							$aPoints = array();
							
							foreach ($aShape as $aCoord)
							{
								$aPoints[] = implode(" ",$aCoord);
							}
							
							$sLineWKT = $this->db->quote("SRID=".$this->discussion->srid.";LINESTRING(".implode(",",$aPoints).")");
							
							if (!$sGeom) $sGeom = $sLineWKT;
							else
							{
								$sGeom = $this->db->quote($this->db->getConnection()->query(
									"select st_union(".$sGeom.",".$sLineWKT.") as geom;")->fetchColumn());
							}
						}
					break;
					
					case "point":
						foreach ($opts["coords"] as $aPoint)
						{
							$sPointWKT = $this->db->quote("SRID=".$this->discussion->srid.";POINT(".$aPoint[0]." ".$aPoint[1].")");
							if (!$sGeom) $sGeom = $sPointWKT;
							else
							{
								$sGeom = $this->db->quote($this->db->getConnection()->query(
									"select st_union(".$sGeom.",".$sPointWKT.") as geom;")->fetchColumn());
							}
						}
					break;
				}
			}
			
			$curDrawing = false;
			// Check to see if we are working on an existing geometry:
			if (isset($opts["id"]))
			{
				$curDrawing = $this->db->getConnection()->query(
					"select id, layer_id, username, attributes[1] as label, ".
					"deleted, visible, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then 'polygon' ".
					"when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then 'line' when ".
					"st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end as geom_type from ".$this->discussion.
					".user_drawings where id = ".$this->db->quote($opts["id"])." and username = ".
					$this->db->quote($this->me->username)." and not deleted and layer_id is null;"
					)->fetch(PDO::FETCH_ASSOC);
			}
			
			// Now that we have the geometry, what to do with it:
			
			switch (@$opts["mode"])
			{
				case "add":
					if ($sGeom)
					{
						if ($curDrawing && count($curDrawing)>0) // Add to the existing geometry
						{
							$this->db->getConnection()->query(
								"update ".$this->discussion.".user_drawings set modification_date = current_timestamp, geom = st_union(geom,".
								$sGeom.") where id = ".$this->db->quote($opts["id"])." and username = ".$this->db->quote($this->me->username).
								" and not deleted and layer_id is null;");
							
							$aResults["results"] = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
								"deleted, visible, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_area(geom) ".
								"else null end as area, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_perimeter(geom) ".
								"when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then st_length(geom) else null end as len, ".
								"case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
								"then 'polygon' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then 'line' when ".
								"st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end as geom_type, st_assvg(geom,0,".
								$this->db->quote($this->discussion->discussion_config->precision).") as shape from ".$this->discussion.
								".user_drawings where id = ".$this->db->quote($opts["id"])." and username = ".
								$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetch(PDO::FETCH_ASSOC);
						}
						else // or create a new one
						{
							$this->db->getConnection()->query(
								"insert into ".$this->discussion.".user_drawings (geom,username) values (".$sGeom.",".
								$this->db->quote($this->me->username).");");
							
							$id = $this->db->lastSequenceId($this->discussion.".user_drawings_id_seq");
							
							$aResults["results"] = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
								"deleted, visible, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_area(geom) ".
								"else null end as area, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_perimeter(geom) ".
								"when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then st_length(geom) else null end as len, ".
								"case when st_geometrytype(geom) in ".
								"('ST_Polygon','ST_MultiPolygon') then 'polygon' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
								"then 'line' when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end as geom_type, ".
								"st_assvg(geom,0,".$this->db->quote($this->discussion->discussion_config->precision).
								") as shape from ".$this->discussion.".user_drawings where id = ".$this->db->quote($id)." and username = ".
								$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetch(PDO::FETCH_ASSOC);
							
							$aResults["results"]["isnew"] = true;
						}
					}
					
				break;
				
				case "cut":
					if ($sGeom)
					{
						if ($curDrawing && count($curDrawing)>0) // Cut from an existing geometry
						{
							$this->db->getConnection()->query(
								"update ".$this->discussion.".user_drawings set modification_date = current_timestamp, geom = st_difference(geom,".
								$sGeom.") where id = ".$this->db->quote($opts["id"])." and username = ".$this->db->quote($this->me->username).
								" and not deleted and layer_id is null;");
							
							$aResults["results"] = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
								"deleted, visible, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_area(geom) ".
								"else null end as area, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_perimeter(geom) ".
								"when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then st_length(geom) else null end as len, ".
								"case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
								"then 'polygon' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then 'line' when ".
								"st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end as geom_type, st_assvg(geom,0,".
								$this->db->quote($this->discussion->discussion_config->precision).") as shape from ".$this->discussion.
								".user_drawings where id = ".$this->db->quote($opts["id"])." and username = ".
								$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetch(PDO::FETCH_ASSOC);
						}
					}
				break;
				
				case "open":
					$aResults["results"] = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
						"deleted, visible, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_area(geom) ".
						"else null end as area, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_perimeter(geom) ".
						"when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then st_length(geom) else null end as len, ".
						"case when st_geometrytype(geom) in ".
						"('ST_Polygon','ST_MultiPolygon') then 'polygon' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
						"then 'line' when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end as geom_type, ".
						"st_assvg(geom,0,".$this->db->quote($this->discussion->discussion_config->precision).
						") as shape from ".$this->discussion.".user_drawings where id = ".$this->db->quote($opts["id"])." and username = ".
						$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetch(PDO::FETCH_ASSOC);
				break;
				
				case "list":
					$aResults["results"] = $this->db->fetchAssoc($this->db->select()->from("user_drawings",array(
							"id"=>"id","label"=>new Zend_Db_Expr("attributes[1]"),"deleted"=>"deleted",
							"visible"=>"visible",
							"geom_type"=>"case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') ".
								"then 'polygon' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then 'line' when ".
								"st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end"
						),$this->discussion->name)->where("username = ?",$this->me->username)->where("not deleted")
						->where("layer_id is null"));
					if (count($aResults["results"])==0) $aResults["results"] = null;
				break;
				
				case "save":
				
					if ($sGeom) $this->db->getConnection()->query(
						"update ".$this->discussion.".user_drawings set modification_date = current_timestamp, geom = ".
						$sGeom." where id = ".$this->db->quote($opts["id"])." and username = ".$this->db->quote($this->me->username).
						" and not deleted and layer_id is null;");
						
					if (isset($opts["props"]))
					{
						if (isset($opts["props"]["label"]))
						{
							if (strlen($opts["props"]["label"])>0)
								$opts["props"]["attributes"] = new Zend_Db_Expr('ARRAY['.$this->db->quote($opts["props"]["label"]).']');
							else
								$opts["props"]["attributes"] = '{}';
							unset($opts["props"]["label"]);
						}
						
						if (isset($opts["props"]["visible"])) $opts["props"]["visible"] = (integer)$opts["props"]["visible"];
						
						$opts["props"]["modification_date"] = new Zend_Db_Expr("current_timestamp");
						$opts["props"]["visibility_change"] = new Zend_Db_Expr("current_timestamp");
						
						$this->db->update($this->discussion.".user_drawings",$opts["props"],
							"id = ".$this->db->quote($opts["id"])." and username = ".$this->db->quote($this->me->username).
							" and not deleted and layer_id is null");
					}
					
					$aResults["results"] = $this->db->getConnection()->query("select id, layer_id, username, attributes[1] as label, ".
						"deleted, visible, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_area(geom) ".
						"else null end as area, case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_perimeter(geom) ".
						"when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') then st_length(geom) else null end as len, ".
						"case when st_geometrytype(geom) in ".
						"('ST_Polygon','ST_MultiPolygon') then 'polygon' when st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') ".
						"then 'line' when st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') then 'point' else 'geom' end as geom_type, ".
						"st_assvg(geom,0,".$this->db->quote($this->discussion->discussion_config->precision).
						") as shape from ".$this->discussion.".user_drawings where id = ".$this->db->quote($opts["id"])." and username = ".
						$this->db->quote($this->me->username)." and not deleted and layer_id is null;")->fetch(PDO::FETCH_ASSOC);
						
				break;
				
				case "delete":
					$this->db->getConnection()->query(
						"update ".$this->discussion.".user_drawings set deleted = true, deletion_date = current_timestamp".
						" where id = ".$this->db->quote($opts["id"])." and username = ".$this->db->quote($this->me->username).
						" and not deleted and layer_id is null;");
					$aResults["results"] = $opts["id"];
				break;
			}
			
			$this->discussion->Event_log(array(
				"type"=>8,
				"text"=>"Draw_Features",
				"data"=>array(
						$opts["mode"],
						(isset($opts["style"])?$opts["style"]:"noStyle"),
						(isset($opts["cellSize"])?$opts["cellSize"]:"noCellSize"),
						(isset($opts["map"])?$opts["map"]:"noMap"),
						(isset($opts["coords"])?$opts["coords"]:"noCoords"),
						(isset($opts["props"])?$opts["props"]:"noProps"),
						(isset($opts["id"])?$opts["id"]:"noID"),
					),
				"geom"=>($sGeom?new Zend_Db_Expr($sGeom):NULL),
				"username"=>$this->me->username,
			));
		}
		catch (Exception $e)
		{
			throw new MapChat_Exception("Draw Feature Error: unable to process input drawing with self-intersecting sides.");
		}
		
		return $aResults;
		
		
	}

}

?>