<?php
/**
 * class MapChat_Discussion_Event
 * Provides the general utilities for managing events in the
 * MapChat database.
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Event
{
	
	public function __construct($discussion)
	{
		$this->db = Zend_Registry::get("db");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * Logs an event in the database:
	 * 
	 * @param array an associative array properties that correspond to
	 *   the columns in the [discussion].event_log table.
	 * @return integer the id of the event that was created
	 */
	public function log(array $aEvent)
	{
		if (isset($aEvent["data"]) && is_array($aEvent["data"]) && count($aEvent["data"])>0)
		{
			$data = array();
			foreach ($aEvent["data"] as $dv)
			{
				$data[] = $this->db->quote("".$dv);
			}
			$aEvent["data"] = new Zend_Db_Expr("ARRAY[".implode(",",$data)."]");
		}
		
		// If a timestamp is provided (assumed to be in milliseconds since the UNIX epoch), and it 
		// is not already converted to a Zend_Db_Expr object:
		if (isset($aEvent["time"]) && gettype($aEvent["time"])!="object")
		{
			$aEvent["time"] = new Zend_Db_Expr("to_timestamp(".($aEvent["time"]*1)."/1000.0)");
		}
		
		// If no username is provided, default to the current user:
		if (!isset($aEvent["username"])) $aEvent["username"] = $this->me->username;
		
		$this->db->insert($this->discussion.".event_log",$aEvent);
		return $this->lastEventID();
	}

	/**
	 * Logs an event in the database:
	 * 
	 * @param array an array of events that were queued by the client.
	 * @return array - integers of the id of the event that was created
	 */
	public function logQueue(array $aEvents)
	{
		$aLoggedEvents = array();
		foreach ($aEvents as $i => $sEvent)
		{
			$aEvent = Zend_Json::decode($sEvent);
			
			if (isset($aEvent["extent"]))
			{
				$ex = $aEvent["extent"];
				$sGeom = "SRID=".$this->discussion->srid.";POLYGON((".$ex[0]." ".$ex[1].",".$ex[0]." ".$ex[3].",".$ex[2]." ".$ex[3].",".$ex[2]." ".$ex[1].",".$ex[0]." ".$ex[1]."))";
			}
			else
			{
				$sGeom = NULL;
			}
			
			$aEventLog = array(
				"parent_id"=>(isset($aEvent["event_parent"])?$aEvent["event_parent"]:
					(isset($aEvent["queue_parent"])?$aLoggedEvents[$aEvent["queue_parent"]]["id"]:NULL)),
				"username"=>$this->me->username,
				"type"=>(isset($aEvent["type"])?$aEvent["type"]:NULL),
				"text"=>(isset($aEvent["text"])?$aEvent["text"]:NULL),
				"data"=>(isset($aEvent["data"])?$aEvent["data"]:NULL),
				"extent"=>$sGeom,
			);
			
			// Assuming that time is set in milliseconds since UNIX 'epoch' time
			if (isset($aEvent["time"]))
			{
				$aEventLog["time"] = new Zend_Db_Expr("to_timestamp(".($aEvent["time"]*1)."/1000.0)");
			}
			$aLoggedEvents[] = $this->log($aEventLog);
		}
		return $aLoggedEvents;
	}

	/**
	 * Logs a user's entry into the discussion interface:
	 */
	public function logEntry($data=array())
	{
		$data = array_merge(array("login",$_SERVER['REMOTE_ADDR']),$data);
		$aEvent = array(
			"username" => $this->me->username,
			"type" => MapChat_EventTypes::ENTER_DISCUSSION,
			"data"=> MapChat_Array::toPG($data,true,true),
			// Need to get the extents somehow.
		);
		return $this->log($aEvent);
	}
	
	/**
	 * Returns the id of the last logged event:
	 */
	public function lastEventID()
	{
		return $this->db->lastSequenceId($this->discussion.".event_log_id_seq");
	}
	
	/**
	 * Returns an array of values for the last event.
	 *
	 * @param integer $type - if specified, the last event of this type will be returned.
	 *
	 * @return array - an associative array of values.
	 */
	public function lastEvent($type=false,$username=false)
	{
		$s = $this->db->select()->from("event_log","*, round(date_part('epoch',time)*1000.0) as epoch_ms",$this->discussion->name);
		if ($type!==false)
		{
			$s = $s->where("type = ?",$type);
		}
		if ($username!==false)
		{
			$s = $s->where("username = ?",$username);
		}
		return $this->db->fetchRow($s->order("time desc")->limit(1));
	}
	
	/**
	 * Updates a user's last online time:
	 * 
	 * @param array $username - set the online state for someone other than
	 *  the current user.
	 * @return integer the time that the user was updated to...
	 */
	public function userOnline($username=false)
	{
		$droppedUsers = array();
		$onlineUsers = array();
		
		if ($username==false) $username = $this->me->username;
		
		// Get the current time as an integer...
		$now = $this->db->fetchOne("select round(date_part('epoch',now())*1000.0) as epoch_ms;");
		
		// Make sure the user is in the user_status table...if not, then insert, otherwise update.
		if ($this->discussion->getOnlineUsers($username))
		{
			// Before updating our current time, look for anyone who has gone offline since the last time this method was executed:
			$droppedUsers = $this->db->fetchCol($this->db->select()
				->from(array("us"=>"user_status"),"username",$this->discussion->name)
				->join(array("my"=>$this->discussion.".user_status"),
					"not us.username = ".$this->db->quote($this->me->username).
						" and my.username = ".$this->db->quote($this->me->username).
						" and us.offline_time > us.online_time and us.offline_time > my.online_time",
					array())
			);
			
			// Check for anyone who has signed on since our last online time
			$onlineUsers = $this->db->fetchCol($this->db->select()
				->from(array("el"=>"event_log"),"username",$this->discussion->name)
				->join(array("my"=>$this->discussion.".user_status"),
					"not el.username = ".$this->db->quote($this->me->username).
						" and my.username = ".$this->db->quote($this->me->username).
						" and el.time > my.online_time and el.type = ".MapChat_EventTypes::ENTER_DISCUSSION,
					array())
			);
			
			// Check to see if the current user has been timed-out:
			$AmIOffline = $this->discussion->getOnlineUsers($username,false);
			
			$this->db->update($this->discussion.".user_status",array(
				"online_time"=>new Zend_Db_Expr("to_timestamp(".($now*1)."/1000.0)"),
			),$this->db->quoteInto("username = ?",$username));
			
			if ($AmIOffline && count($AmIOffline)>0)
			{
				$lastEvent = $this->lastEvent(false,$username);
				
				if (@$lastEvent["type"]!=MapChat_EventTypes::ENTER_DISCUSSION)
				{
					// Since we appeared to be offline before, let's log the re-entry
					$this->logEntry(array("User Un-Timed-Out"));
				}
			}
		}
		else
		{
			$this->db->insert($this->discussion.".user_status",array(
				"username"=>$username,
				"online_time"=>new Zend_Db_Expr("to_timestamp(".($now*1)."/1000.0)"),
			));
		}
		
		// Update any other users whose online state has expired
		$this->db->update($this->discussion->name.".user_status",array("offline_time"=>new Zend_Db_Expr("to_timestamp(".($now*1)."/1000.0)")),"(to_timestamp(".($now*1)."/1000.0) - online_time)>'30 seconds'::interval and (online_time > offline_time or offline_time is null)");
		
		// List the users that got timed-out.
		$offlineUsers = $this->discussion->getOnlineUsers(false,false,"round(date_part('epoch',offline_time)*1000.0) = ".($now*1));
		foreach ($offlineUsers as $aUser)
		{
			$this->log(array(
				"type"=>MapChat_EventTypes::LEAVE_DISCUSSION,
				"text"=>"User Timed-Out (3rd-party session)",
				"data"=>$aUser,
				"username"=>$aUser["username"],
				"time"=>new Zend_Db_Expr("to_timestamp(".($now*1)."/1000.0)"),
			));
			// Add this username to the list of dropped users
			$droppedUsers[] = $aUser["username"];
		}
		
		// Any situation that calls the userOnline() method should also send this information back to the
		// client within the same AJAX request.
		if (count($droppedUsers)>0) $this->ReturnEvent("DroppedUsers",$droppedUsers);
		if (count($onlineUsers)>0) $this->ReturnEvent("AddedUsers",$onlineUsers);
		
		return $now;
	}
	
	/**
	 * Adds event trigger names and data to a registry variable, which is checked when AJAX requests are completed,
	 * and sent to the client in an extra events tag.
	 *
	 * @param string $name - the name of the event that will be triggered by the client
	 * @param mixed $data - the data that will be sent to this event's listeners.
	 */
	public function ReturnEvent($name,$data=NULL)
	{
		MapChat_Session::ns()->ClientEvents[] = array($name,$data);
	}
	
	
}

?>