<?php
/**
 * class MapChat_Discussion_Layer
 * Provides the general utilities for managing user's mapfiles when
 * interacting with the discussion:
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Layer
{
	private $_aMapText = false;
	private $_aTmpMaps = false;
	private $_bWriteToDB = false;
	
	/**
	 * Get the MapChat config/session/discussion environment, and
	 * attempt to load the most recent maps from the database.
	 */
	public function __construct($discussion)
	{
		$this->db = Zend_Registry::get("db");
		$this->config = Zend_Registry::get("config");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * Return a list layers organized by mapfile/group, and with their
	 * type, status, enabled properties as attributes.
	 *
	 * @param string $map if specified, only layers from the corresponding
	 *   mapfile will be returned:
	 * @param boolean $reverse if true, return groups/layers in reverse
	 *   order (i.e., so they can be listed from top to bottom relative
	 *   to the user's perspective, rather than the bottom-to-top order
	 *   that mapserver draws them in)
	 */
	public function listLayers($map=false,$reverse=true)
	{
		$aMapLayers = array();
		if ($map!==false&&strlen($map)>0)
		{
			$aMaps = array($map);
		}
		else
		{
			$aMaps = $this->discussion->Map_getMapFiles();
		}
		
		foreach ($aMaps as $aMap)
		{
			$aMapLayers[$aMap["title"]] = array("groups" => array() );
			$oMap = ms_newMapObj( $aMap['text'] , $this->config->BaseDir."/maps" );
			$aMapGroups = array();
			$aGroups = $oMap->getAllGroupNames();
			foreach ($aGroups as $sGroup)
			{
				$aGroup = array("layers" => array());
				$aLayers = array();
				$aLayersIdx = $oMap->getLayersIndexByGroup($sGroup);
				foreach ($aLayersIdx as $iLayerIdx)
				{
					$oLayer = $oMap->getLayer($iLayerIdx);
					
					// Save the current status, then set the status to 'MS_ON', to
					// allow us to test for scale visibility:
					$currentStatus = $oLayer->status;
					$oLayer->set("status", MS_ON);
					
					$aLayer = array(
						"index" => $oLayer->index,
						"status" => $currentStatus,
						"type" => $oLayer->type,
						"classes" => array(),
						"enabled" => $oLayer->isVisible(),
						"group" => $oLayer->group,
						"display_group" => $oLayer->getMetaData("mapchat_display_group"),
						"option_type" => $oLayer->getMetaData("mapchat_option_type"),
					);
					
					if ($oLayer->getMetaData("mapchat_plan_type")!="") $aLayer["plan_type"] = $oLayer->getMetaData("mapchat_plan_type");
					if ($oLayer->getMetaData("mapchat_userlayer")=="true") $aLayer["userlayer"] = true;
					
					// Set the status back to whatever it was before:
					$oLayer->set("status",$currentStatus);
					
					$aClasses = array();
					for ($i=0;$i<$oLayer->numclasses;$i++)
					{
						$aClasses[] = $oLayer->getClass($i)->name;
					}
					
					// With the way classes work, it's more appropriate to do the opposite wrt the listing order...
					$aLayer["classes"] = ($reverse?$aClasses:array_reverse($aClasses));
					
					$aLayers[$oLayer->name] = $aLayer;
				}
				$aGroup["layers"] = ($reverse?array_reverse($aLayers):$aLayers);
				$aMapGroups[$sGroup] = $reverse?array_reverse($aGroup):$aGroup;
			}
			$aMapLayers[$aMap["title"]]["groups"] = $reverse?array_reverse($aMapGroups):$aMapGroups;
		}
		return $reverse?array_reverse($aMapLayers):$aMapLayers;
	}
	
	public function setProps($map,$layer,$props)
	{
		$oMap = $this->discussion->Map_getMapObj($map);
		$oLayer = $oMap->getLayerByName($layer);
		foreach($props as $n => $v)
		{
			$oLayer->set($n,(defined($v)?constant($v):$v));
		}
		$this->discussion->Map_saveMapObj($map);
	}
	
	public function getProps($map,$layer,$reverse=true)
	{
		$oMap = (gettype($map)!="string"?$map:$this->discussion->Map_getMapObj($map));
		$oLayer = @(gettype($layer)!="string"?$layer:$oMap->getLayerByName($layer));
		
		// If the layer doesn't exist, return false
		if (!$oLayer) return false;
		
		$currentStatus = $oLayer->status;
		$oLayer->set("status",MS_ON);
		$aReturn = array(
			"name" => $oLayer->name,
			"index" => $oLayer->index,
			"group" => $oLayer->group,
			"display_group" => $oLayer->getMetaData("mapchat_display_group"),
			"option_type" => $oLayer->getMetaData("mapchat_option_type"),
			"status" => $currentStatus,
			"type" => $oLayer->type,
			"classes" => array(),
			"enabled" => $oLayer->isVisible(),
			"maxscaledenom" => $oLayer->maxscaledenom,
			"minscaledenom" => $oLayer->minscaledenom,
		);
		$oLayer->set("status",$currentStatus);
		
		$aClasses = array();
		for ($i=0;$i<$oLayer->numclasses;$i++)
		{
			$oClass = $oLayer->getClass($i);
			if ($oClass->status==MS_ON) $aClasses[] = $oLayer->getClass($i)->name;
		}
		
		// With the way classes work, it's more appropriate to do the opposite wrt the listing order...
		$aReturn["classes"] = ($reverse?$aClasses:array_reverse($aClasses));
		
		if ($oLayer->getMetaData("mapchat_aliases")!="" && $oLayer->getMetaData("mapchat_fields")!="")
		{
			// The number and order of fields and aliases are assumed to match here.
			$aFields = explode("\t",$oLayer->getMetaData("mapchat_fields"));
			$aAliases = explode("\t",$oLayer->getMetaData("mapchat_aliases"));
			$aFieldAliases = array();
			foreach ($aFields as $i => $field)
			{
				$aFieldAliases[$field]=$aAliases[$i];
			}
			$aReturn["field_aliases"] = $aFieldAliases;
		}
		
		if ($oLayer->getMetaData("mapchat_option_group")!="") $aReturn["option_group"]=$oLayer->getMetaData("mapchat_option_group");
		
		return $aReturn;
	}
	
	public function getClasses($map,$layer,$reverse=true)
	{
		$oMap = (gettype($map)!="string"?$map:$this->discussion->Map_getMapObj($map));
		$oLayer = @(gettype($layer)!="string"?$layer:$oMap->getLayerByName($layer));
		
		// If the layer doesn't exist, return false
		if (!$oLayer) return false;
		
		$currentStatus = $oLayer->status;
		$aReturn = array();
		$aClasses = array();
		for ($i=0;$i<$oLayer->numclasses;$i++)
		{
			$oClass = $oLayer->getClass($i);
			if ($oClass->status==MS_ON) $aClasses[] = $oLayer->getClass($i)->name;
		}
		
		// With the way classes work, it's more appropriate to do the opposite wrt the listing order...
		return ($reverse?$aClasses:array_reverse($aClasses));
	}
	
	/**
	 * Return an imageObj with a legend icon for specified layer to the browser
	 *
	 * @param string $map the mapfile that contains the desired layer.
	 * @param integer $layer the index of the layer containing the desired class.
	 * @param integer $class the index of the class to create the icon.
	 */
	public function getClassIcon($map,$layer,$class)
	{
		$oMap = $this->discussion->Map_getMapObj($map);
		$oLayer = $oMap->getLayer($layer);
		$oClass = $oLayer->getClass($class);
		$size = 12;
		
		// May need better logic to determine the optimal image size:
		for ($i=0;$i<$oClass->numstyles;$i++)
		{
			$oStyle = $oClass->getStyle($i);
			$size = max($size,$oStyle->size,$oStyle->width);
		}
		
		$img = $oClass->createLegendIcon($size,$size);
		$tImg = tempnam($this->config->tmpDir,'mc_');
		$sImg = $img->saveImage($tImg);

		header("Content-type: ".$oMap->outputformat->mimetype);
		header("Content-length: ".$tImg);

		// open the file in a binary mode
		$p = fopen($tImg, 'rb');
		fpassthru($p);
	}
	
	/**
	 * Query a map layer at the specified coordinates, and return an
	 * array of fields/values for the returned features.
	 *
	 * @param string $map the mapfile that contains the desired layer.
	 * @param string $layer the name of the layer to identify features.
	 * @param numeric $x,$y the coordinates to query on the layer.
	 * @param numeric $cellSize the current geographic size of a pixel in
	 *   the client display - so we can set an appropriate tolerance for
	 *   querying features in vector layers.
	 * @return array the list of field/value pairs for each identified feature.
	 */
	public function Identify($map,$layer,$x,$y,$cellSize)
	{
		$oMap = $this->discussion->Map_getMapObj($map);
		$oLayer = $oMap->getLayerByName($layer);
		if (in_array($oLayer->type,array(MS_LAYER_POINT, MS_LAYER_LINE, MS_LAYER_POLYGON)))
		{
			$oLayer->set("toleranceunits",$oMap->units);
			$oLayer->set("tolerance",$cellSize*3);
		}
		else
		{
			$oLayer->set("tolerance",0);
		}
		
		$oLayer->set("template","dummy.html");
		
		$oTestShape = ms_shapeObjFromWkt('POLYGON((0 0,0 1,1 1,1 0,0 0))');
// 		$sWKT = $oTestShape->toWkt();
		
		$oQPoint = ms_newPointObj();
		$oQPoint->setXY($x,$y);
		
		$aResults = array();
		
		if (@$oLayer->queryByPoint($oQPoint,MS_MULTIPLE,$oLayer->tolerance) == MS_SUCCESS)
		{
			$oLayer->open();
			$nResults = $oLayer->getNumResults();
			
			$sLayerProj = false;//$oLayer->getProjection();
			$sMapProj = false;//$oMap->getProjection();
			$bProjShape = false;
			if ($sLayerProj && $sMapProj && $sLayerProj!=$sMapProj)
			{
				$bProjShape = true;
				$oLayerProj = ms_newProjectionObj($sLayerProj);
				$oMapProj = ms_newProjectionObj($sMapProj);
			}
			
			for ($i=0;$i<$nResults;$i++)
			{
				$oResult = $oLayer->getResult($i);
				if (ms_GetVersionInt()<50600)
				{
					$oShape = $oLayer->getShape($oResult->tileindex, $oResult->shapeindex);
				}
				else
				{
					$oShape = $oLayer->resultsGetShape($oResult->shapeindex, $oResult->tileindex);
				}
				if ($bProjShape) $oShape->project($oLayerProj,$oMapProj);
				
				$wkt = MapChat_ShapeObj::toWkt($oShape);
				
				// Convert the wkt to SVG - precision defined in the xml config should be appropriate
				// for the scales and coordinate system used in the map - for UTM, zero is probably fine. Fewer
				// decimals means less data being transferred to the client:
				$sSVG = $this->db->getConnection()->query("select st_assvg(st_multi(st_forcerhr('".$wkt."'::geometry)),0,".
					$this->db->quote($this->discussion->discussion_config->precision).");")->fetchColumn();
				
				$simplified = false;
				
				// Return a single point + max extents if a geometry is particularly large.
				if (strlen($sSVG) > $this->discussion->discussion_config->svg_limit)
				{
					$simplified = true;
					
					if (@$oShape->type==MS_SHAPE_POLYGON) $shapeInfo = $this->db->getConnection()->query(
							"select st_assvg(st_pointonsurface(geom),0,".$this->db->quote($this->discussion->discussion_config->precision).
							") as svg, st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from (select '"
							.$wkt."'::geometry as geom) as g;"
						)->fetch(PDO::FETCH_ASSOC);
					
					elseif (@$oShape->type==MS_SHAPE_LINE) $shapeInfo = $this->db->getConnection()->query(
							"select case when st_geometrytype(geom) in ('ST_LineString') then st_assvg(ST_line_interpolate_point(geom),0.5) else st_centroid(geom) end as svg, st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from (select '".$wkt."'::geometry as geom) as g;"
						)->fetch(PDO::FETCH_ASSOC);
					
					else $shapeInfo = $this->db->getConnection()->query(
							"select st_assvg(st_centroid(geom)) as svg, st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from (select '".$wkt."'::geometry as geom) as g;"
						)->fetch(PDO::FETCH_ASSOC);
					
					$sSVG = $shapeInfo["svg"];
					$aExtent = array(array($shapeInfo["xmin"],$shapeInfo["ymin"]),array($shapeInfo["xmax"],$shapeInfo["ymax"]));
				}
				
				$aResults[] = array(array("ti"=>$oResult->tileindex,"si"=>$oResult->shapeindex,"ci"=>$oResult->classindex,"area"=>@$oShape->getArea(),"length"=>@$oShape->getLength(),"shape"=>$sSVG,"type"=>($simplified?MS_SHAPE_POINT:@$oShape->type),"layertype"=>@$oLayer->type,"extent"=>(isset($aExtent)?$aExtent:false)),$oShape->values);
			}
		}
		
		return $aResults;
	}
	
	/**
	 * Creates a new user-defined layer in the discussion.
	 *
	 * @param string $name - the name of the layer to create
	 * @param integer $shape_type - the shape type that this layer will be treated as (1=POINT, 2=LINE, 3=POLYGON)
	 * @param array $attributes - an array of attributes (keys are names, values are types)
	 * @param string $username - optionally specify a different username than the current user.
	 *
	 * @return integer - the id of the newly created layer.
	 */
	public function Create($name,$description,$shape_type,$attributes,$username=false)
	{
		$attrNames = array();
		$attrTypes = array();
		
		foreach ($attributes as $attr => $type)
		{
			$attrNames[] = $this->db->quoteInto("?",$attr);
			$attrTypes[] = $this->db->quoteInto("?",$type);
		}
		
		$oAttr = new Zend_Db_Expr("array[array[".implode(",",$attrNames)."],array[".implode(",",$attrTypes)."]]");
		
		$this->db->insert($this->discussion.".user_layers",array(
			"name"=>$name,
			"description"=>$description,
			"username"=>$username?$username:$this->me->username,
			"attributes"=>$oAttr,
			"shape_type"=>$shape_type,
		));
		
		return $this->db->lastSequenceId($this->discussion.".user_layers_id_seq");
	}
}

?>
