<?php
/**
 * class MapChat_Discussion_Map
 * Provides the general utilities for managing user's mapfiles when
 * interacting with the discussion:
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Map
{
	private $_aMapText = false;
	private $_aTmpMaps = false;
	private $_bWriteToDB = false;
	private $_bCheckOF = false;
	
	/**
	 * Get the MapChat config/session/discussion environment, and
	 * attempt to load the most recent maps from the database.
	 *
	 * @param boolean $init if set to true, the user's maps will
	 *   default to the original discussion maps.
	 */
	public function __construct($discussion,$init=false)
	{
		$this->db = Zend_Registry::get("db");
		$this->config = Zend_Registry::get("config");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
		if ($init)
		{
			$this->_initMaps();
		}
		else
		{
			$this->_loadMaps();
		}
	}
	
	/**
	 * Until PHP/MapScript is setup to load map text from strings instead
	 * of *.map files, we're going to have to manage loading/saving maps
	 * by reading/writing to temporary files.  This function should cleanup
	 * any mapfiles that are lingering in the temp folder:
	 */
	public function __destruct()
	{
		foreach ($this->_aTmpMaps as $name => $aMap)
		{
			if ($aMap["save"])
			{
				$tname = tempnam( $this->config->tmpDir, "mc_".$name ).".map";
				$aMap["object"]->save($tname);
				$newMapText = file_get_contents($tname);
				unlink($tname);
				
				if ($this->_aMapText[$name]["text"]!=$newMapText)
				{
					$this->_aMapText[$name]["text"] = $newMapText;
					$this->_bWriteToDB = true;
				}
			}
		}
		
		if ($this->_bWriteToDB)
		{
			$this->_writeToDB();
		}
	}
	
	/**
	 * Read the text from the default mapfiles for the current discussion
	 * into an array:
	 */
	private function _initMaps()
	{
		$this->_aMapText = array();
		foreach ($this->discussion->mapfiles as $mapfile)
		{
			// MapFiles are saved with certain elements in different orders than they are loaded (e.g., metadata), so we'll go through
			// the save rigamarole here so that the next time the mapfile's text is compared to itself, it won't be re-saved just because
			// the metadata elements have been re-arranged.
			$tname = tempnam( $this->config->tmpDir, "mc_init" ).".map";
			$oMapTemp = ms_newMapObj($this->config->BaseDir."/maps/".$mapfile,$this->config->BaseDir."/maps");
			
			// Now save the old map so we can get its actual written text
			$oMapTemp->save($tname);
			$mapText = file_get_contents($tname);
			unlink($tname);
			
			$this->_aMapText[basename($mapfile,".map")]["text"]=$mapText;
			$this->_aMapText[basename($mapfile,".map")]["time"]=MapChat_Time::stamp();
		}
		
		// Here, we add the requisite user drawings layer to the USER_LAYERS group:
		$this->_initChatLayers(basename($mapfile,".map"));
		
		// make sure when this object dies, it saves the mapfiles to the DB:
		$this->_bWriteToDB = true;
		$this->_bCheckOF = true;
		
	}
	
	/**
	 * Read the latest serialized mapfile array for this user from the current discussion
	 */
	private function _loadMaps()
	{
		$this->_aMapText = @unserialize($this->db->fetchOne($this->db->select()->from("user_maps","mapfile",$this->discussion)->where("username = ?",$this->me)->order("time desc")->limit(1)));
		if (!is_array($this->_aMapText))
		{
			$this->_initMaps();
		}
	}
	
	private function _writeToDB()
	{
		$this->db->insert($this->discussion.".user_maps",array(
			"username" => $this->me->username,
			"mapfile" => serialize($this->_aMapText),
		));
	}
	
	/**
	 * If PHP/MapScript can ever read mapfiles text strings instead of *.map files,
	 * then we could simply refer to $this->_aMapText, but until that happens we're
	 * stuck doing it this way.  The result from this function should be compatible
	 * with the '$aszMapFiles' config variable used for Ka-Map:
	 *
	 * TODO: do we care about different sets of scales for different maps?  I don't
	 *   think 'format' is an issue, since that can be defined in layer groups.  Not
	 *   sure if we're concerned with the Ka-Map Auth approach either.
	 */
	public function getMapfiles()
	{
		if (!is_array($this->_aTmpMaps))
		{
			$aScales = explode(",",$this->discussion->discussion_config->scales);
			$this->aTmpMaps = array();
			foreach ($this->_aMapText as $name => $aMapText)
			{
				$this->_aTmpMaps[$name] = array(
					'title' => $name,
					'text' => $aMapText["text"],
					'scales' => $aScales,
					'format' => 'AGG_PNG24', 'time' => $aMapText["time"],
					'save' => false,
					/* Sample authorized_users entry. See auth.php for more details:
					 * ,'authorized_users' => array('popplace' => array('user1', 'user2'),
					 *                              'park'     => array('user1')
					 */
				);
			}
		}
		return $this->_aTmpMaps;
	}

	/**
	 * Update the mapfile data - and set the flag to indicate that 
	 * the mapfiles should be saved to the DB when this object dies.
	 */
	public function saveMapfile($name,$oMap)
	{
		// MapScript still only reads from text strings - can't save to a text string,
		// so we still have to save to a file then read the text...but this won't be
		// as frequent as reading, so it shouldn't be too bad:
		$tname = tempnam( $this->config->tmpDir, "mc_".$name ).".map";
		
		$oMap->save($tname);
		$newMapText = file_get_contents($tname);
		unlink($tname);
		if ($this->_aMapText[$name]["text"]!=$newMapText)
		{
			$this->_aMapText[$name]["text"] = $newMapText;
			$this->_aMapText[$name]["time"] = MapChat_Time::stamp();
			$this->_bWriteToDB = true;
		}
	}

	/**
	 * Get a MapScript mapobject:
	 *
	 * @param string $name name of the mapfile/object to get:
	 */
	public function getMapObj($name)
	{
		$this->getMapFiles();
		if (!isset($this->_aTmpMaps[$name]["object"]))
		{
			// The exact approach that will ultimately be used for loading maps from text strings is not yet established:
			if (function_exists("ms_newMapObjFromString"))
			{
				$this->_aTmpMaps[$name]["object"] = ms_newMapObjFromString($this->_aTmpMaps[$name]['text'],$this->config->BaseDir."/maps");
			}
			else
			{
				try
				{
					$this->_aTmpMaps[$name]["object"] = @ms_newMapObj($this->_aTmpMaps[$name]['text'],$this->config->BaseDir."/maps");
					
					if (!$this->_aTmpMaps[$name]["object"]) throw new Exception();
				}
				catch (Exception $e)
				{
					// can't open directly from strings...do it the old-fashioned way.
					$tname = tempnam( $this->config->tmpDir, "mc_".$name ).".map";
					file_put_contents($tname,$this->_aTmpMaps[$name]['text']);
					$this->_aTmpMaps[$name]["object"] = ms_newMapObj($tname,$this->config->BaseDir."/maps");
					unlink($tname);
				}
			}
		}
		return $this->_aTmpMaps[$name]["object"];
	}
	
	/**
	 * Flag a map object for saving when this class is destroyed.
	 *
	 * @param string $name the name of the mapfile to be saved.
	 */
	public function saveMapObj($name)
	{
		if (strlen($name)>0 && isset($this->_aTmpMaps[$name]))
		{
			$this->_aTmpMaps[$name]["save"] = true;
			
			// Save the timestamp now - we can't do this in the __destroy() function, since
			// that class will no longer be available.
			$this->_aMapText[$name]["time"] = MapChat_Time::stamp();
		}
		else
		{
			throw new MapChat_Exception("Error - cannot save unidentified map object.");
		}
	}
	
	
	/**
	 * Add a layer to a mapfile (e.g., when creating/opening Irrigation Plans)
	 *
	 * @param string $mapfile - the name of the mapfile to add the layers to
	 * @param array $props - an associative array of properties that correspond to MapServer's mapfile syntax.
	 */
	public function addLayer($mapfile,$props)
	{
		$oMap = $this->getMapObj($mapfile);
		
		// For now, we're making the assumption that layers with the same name are equivalent.
		$oLayer = @$oMap->getLayerByName($props["name"]);
		if ($oLayer) return $oLayer->index;
		
		// If no existing layer is found, add it to the map.
		$oLayer = ms_newLayerObj($oMap);
		foreach ($props as $pname => $pval)
		{
			switch ($pname)
			{
				case "classes":
					foreach($pval as $class)
					{
						$oClass = ms_newClassObj($oLayer);
						if (!isset($class["styles"])) $oStyle = ms_newStyleObj($oClass);
						foreach ($class as $cname => $cval)
						{
							switch($cname)
							{
								case "label":
									foreach ($cval as $lname => $lval)
									{
										$oLabel = $oClass->label;
										switch ($lname)
										{
											case "color":
											case "outlinecolor":
											case "shadowcolor":
											case "backgroundcolor":
											case "backgroundshadowcolor":
												$oLabel->{$lname}->setRGB($lval[0],$lval[1],$lval[2]);
												break;
											
											default:
												$oLabel->set($lname,$sval);
												break;
										}
									}
									break;
							
								case "styles":
									foreach($cval as $style)
									{
										$oStyle = ms_newStyleObj($oClass);
										foreach ($style as $sname => $sval)
										{
											switch ($sname)
											{
												case "color":
												case "outlinecolor":
												case "backgroundcolor":
													$oStyle->{$sname}->setRGB($sval[0],$sval[1],$sval[2]);
													break;
												
												default:
													$oStyle->set($sname,$sval);
													break;
											}
										}
									}
									break;
								
								case "expression":
									$oClass->setExpression($cval);
									break;
								
								default:
									$oClass->set($cname,$cval);
									break;
							}
						}
					}
					break;
				
				case "projection":
					$oLayer->setProjection($pval);
					break;
				
				case "connectiontype":
					$oLayer->setConnectionType($pval);
					break;
				
				case "metadata":
					foreach ($pval as $mname => $mval)
					{
						$oLayer->setMetaData($mname,$mval);
					}
					break;
				
				default:
					$oLayer->set($pname,$pval);
					break;
			}
		}
		$this->saveMapfile($mapfile,$oMap);
		return $oLayer->index;
	}
	
	/**
	 *	Initializes the mapfile with chat layers with 'my shapes' and 'other users shapes'
	 */
	private function _initChatLayers($mapfile)
	{
		
		$pg_conn=Zend_Registry::get("pg_conn");
		$d = Zend_Registry::get("discussion");
		$me = Zend_Registry::get("me");
		$db = Zend_Registry::get("db");
		
		$imageformat = $d->discussion_config->user_imageformat;
		$opacity = $d->discussion_config->user_opacity;
		
		if ($imageformat == "GD_GIF")
		{
			$hatchStyle = array(
				array(
					"symbolname"=>"hatch",
					"size"=>9,
					"width"=>1,
					"angle"=>45,
				),
				array(
					"symbolname"=>"hatch",
					"size"=>9,
					"width"=>1,
					"angle"=>135,
				),
			);
		}
		else
		{
			$hatchStyle=array(array(),array());
		}
		
		$aUserPoly = array(
			"type"=>MS_LAYER_POLYGON,
			"connectiontype"=>MS_POSTGIS,
			"connection"=>$pg_conn,
			"data"=>"geom from (select geom, id, username, case when length(full_name)>0 then full_name else username end as display_name, ".
				"attributes[1] as label from ".$d->name.".user_drawings ud join users u using (username) where ".
				"not deleted and visible and st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') and ".
				"layer_id is null order by id) as foo using unique id using srid=".$d->srid,
			"name"=>"User Drawings (polygon)",
			"group"=>"USER_LAYERS",
			"status"=>MS_ON,
			"projection"=>"init=epsg:".$d->srid,
			"classitem"=>"username",
			
			"metadata"=>array(
				"mapchat_description"=>"User Drawings (polygon)",
				"mapchat_display_group"=>"User Drawings",
				"mapchat_option_type"=>"checkbox",
				"mapchat_aliases"=>"Shape ID\tLabel\tUser",
				"mapchat_fields"=>"id\tlabel\tdisplay_name",
				"mapchat_userlayer"=>"true",
				"mapchat_userdrawing"=>"true",
				"imageformat"=>$imageformat,
				"opacity"=>$opacity,
			),
			
			"classes"=>array(
				
				array(
					"name"=>"My Drawings",
					"expression"=>$me->username,
					"status"=>MS_ON,
					"styles"=>array(
						array_merge(array(
							"color"=>array(124,255,175),
							"opacity"=>60,
						),$hatchStyle[0]),
						array_merge(array(
							"color"=>array(124,255,175),
							"opacity"=>60,
						),$hatchStyle[1]),
						array(
							"outlinecolor"=>array(124,255,175),
							"width"=>2,
							"opacity"=>100,
						),
					),
				),
				
				array(
					"name"=>"Other People's Drawings",
					"status"=>MS_ON,
					"styles"=>array(
						array_merge(array(
							"color"=>array(255,175,124),
							"opacity"=>60,
						),$hatchStyle[0]),
						array_merge(array(
							"color"=>array(255,175,124),
							"opacity"=>60,
						),$hatchStyle[1]),
						array(
							"outlinecolor"=>array(255,175,124),
							"width"=>2,
							"opacity"=>100,
						),
					),
				),
				
			),
		);
		
		$this->addLayer($mapfile,$aUserPoly);
		
		$aUserLine = array(
			"type"=>MS_LAYER_LINE,
			"connectiontype"=>MS_POSTGIS,
			"connection"=>$pg_conn,
			"data"=>"geom from (select geom, id, username, case when length(full_name)>0 then full_name else username end as display_name, ".
				"attributes[1] as label from ".$d->name.".user_drawings ud join users u using (username) where ".
				"not deleted and visible and st_geometrytype(geom) in ('ST_LineString','ST_MultiLineString') and ".
				"layer_id is null order by id) as foo using unique id using srid=".$d->srid,
			"name"=>"User Drawings (line)",
			"group"=>"USER_LAYERS",
			"status"=>MS_ON,
			"projection"=>"init=epsg:".$d->srid,
			"classitem"=>"username",
			
			"metadata"=>array(
				"mapchat_description"=>"User Drawings (line)",
				"mapchat_display_group"=>"User Drawings",
				"mapchat_option_type"=>"checkbox",
				"mapchat_aliases"=>"Shape ID\tLabel\tUser",
				"mapchat_fields"=>"id\tlabel\tdisplay_name",
				"mapchat_userlayer"=>"true",
				"mapchat_userdrawing"=>"true",
				"imageformat"=>$imageformat,
				"opacity"=>$opacity,
			),
			
			"classes"=>array(
				
				array(
					"name"=>"My Drawings",
					"expression"=>$me->username,
					"status"=>MS_ON,
					"styles"=>array(
						array(
							"color"=>array(124,255,175),
							"width"=>2,
							"opacity"=>100,
						),
					),
				),
				
				array(
					"name"=>"Other People's Drawings",
					"status"=>MS_ON,
					"styles"=>array(
						array(
							"color"=>array(255,175,124),
							"width"=>2,
							"opacity"=>100,
						),
					),
				),
			),
		);
		
		$this->addLayer($mapfile,$aUserLine);
		
		$aUserPoint = array(
			"type"=>MS_LAYER_POINT,
			"connectiontype"=>MS_POSTGIS,
			"connection"=>$pg_conn,
			"data"=>"geom from (select geom, id, username, case when length(full_name)>0 then full_name else username end as display_name, ".
				"attributes[1] as label from ".$d->name.".user_drawings ud join users u using (username) where ".
				"not deleted and visible and st_geometrytype(geom) in ('ST_Point','ST_MultiPoint') and ".
				"layer_id is null order by id) as foo using unique id using srid=".$d->srid,
			"name"=>"User Drawings (point)",
			"group"=>"USER_LAYERS",
			"status"=>MS_ON,
			"projection"=>"init=epsg:".$d->srid,
			"classitem"=>"username",
			
			"metadata"=>array(
				"mapchat_description"=>"User Drawings (point)",
				"mapchat_display_group"=>"User Drawings",
				"mapchat_option_type"=>"checkbox",
				"mapchat_aliases"=>"Shape ID\tLabel\tUser",
				"mapchat_fields"=>"id\tlabel\tdisplay_name",
				"mapchat_userlayer"=>"true",
				"mapchat_userdrawing"=>"true",
				"imageformat"=>$imageformat,
				"opacity"=>$opacity,
			),
			
			"classes"=>array(
				
				array(
					"name"=>"My Drawings",
					"expression"=>$me->username,
					"status"=>MS_ON,
					"styles"=>array(
						array(
							"symbolname"=>"circle",
							"color"=>array(124,255,175),
							"outlinecolor"=>array(0,0,0),
							"size"=>9,
						),
					),
				),
				
				array(
					"name"=>"Other People's Drawings",
					"status"=>MS_ON,
					"styles"=>array(
						array(
							"symbolname"=>"circle",
							"color"=>array(255,175,124),
							"outlinecolor"=>array(0,0,0),
							"size"=>9,
						),
					),
				),
				
			),
		);
		
		$this->addLayer($mapfile,$aUserPoint);
		
	}
}
?>