<?php
/**
 * class MapChat_Discussion_Plan
 * Provides the general utilities for managing users' plans in the discussion:
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Plan
{
	
	/**
	 * Get the MapChat config/session/discussion environment.
	 *
	 * @param MapChat_Discussion $discussion the discussion class that should
	 *  be the parent to an instance of this class.
	 */
	public function __construct($discussion)
	{
		$this->db = Zend_Registry::get("db");
		$this->config = Zend_Registry::get("config");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * List plans of a certain type for this user.
	 *
	 * @param mixed $type the type of plan to list (either an integer value, or the
	 *  description value listed in the plan_types lookup table);
	 *
	 * @return array
	 */
	public function Listing($type)
	{
		$select = $this->db->select()->from(array("up" => $this->discussion.'.user_plans'))->join(array("pt"=>"plan_types"),"up.type=pt.id",array("description"))->where("not deleted and username = ?",$this->me->username);
		
		if (is_numeric($type))
		{
			$select = $select->where("type = ?",$type);
		}
		else
		{
			$select = $select->where("description = ?",$type);
		}
		
		return $this->db->fetchAssoc($select);
	}
	
	/**
	 * Creates a plan
	 *
	 * @param string $title - the title of the plan being created (must be unique per username)
	 * @param mixed $type - the type of plan to list (either an integer value, or the
	 *  description value listed in the plan_types lookup table);
	 * @param string $username - optionally specify a different username than the current user.
	 *
	 * @return integer - the id of the newly created plan
	 */
	public function Create($title,$type,$username=false)
	{
		if (strlen($title)<1)
		{
			throw new MapChat_Exception("Plan title must be specified");
		}
		
		if (!is_numeric($type))
		{
			$type = $this->db->fetchOne($this->db->quoteInto("select id from plan_types where description = ?",$type));
		}
		
		$this->db->insert($this->discussion.".user_plans",array(
			"title"=>$title,
			"type"=>$type,
			"username"=>$username?$username:$this->me->username,
		));
		
		return $this->db->lastSequenceId($this->discussion.".user_plans_id_seq");
		
	}
	
	/**
	 * Returns descriptive information about a given plan.
	 *
	 * @param integer $plan_id - the identifier of the plan in the user_plans table.
	 *
	 * @return array
	 */
	public function Info($plan_id)
	{
		
		$planInfo = $this->db->fetchRow($this->db->select()->from("user_plans","*",$this->discussion->name)->where("id = ?",$plan_id));
		
		return $planInfo;
		
	}
	
	/**
	 * Returns descriptive information about a given plan's layers.
	 *
	 * @param integer $plan_id - the identifier of the plan in the user_plans table.
	 *
	 * @return array
	 */
	public function LayerInfo($plan_id)
	{
		$aPlanLayers = $this->db->fetchAssoc($this->db->select()->from("user_layers","*",$this->discussion->name)->join($this->discussion->name.".user_plans_layers","id = layer_id")->where("plan_id = ?",$plan_id));
		$aNamedLayers = array(); // I want to add layers in a specific order, so I'll make sure they can be referenced by name.
		$aAddedLayers = array(); // This will store the layer information after it has been added to the mapfile.
		foreach ($aPlanLayers as $aPlanLayer)
		{
			$aNamedLayers[$aPlanLayer["name"]] = $aPlanLayer;
		}
		
		return $aNamedLayers;
	}
	
	/**
	 * Links plans to layers by ID in the user_plans_layers table.
	 *
	 * @param integer $plan_id - the identifier for the plan.
	 * @param integer $layer_id - the identifier for the layer.
	 *
	 * @return array
	 */
	public function linkToLayer($plan_id,$layer_id)
	{
		$this->db->insert($this->discussion.".user_plans_layers",array(
			"plan_id"=>$plan_id,
			"layer_id"=>$layer_id,
		));
	}
	
	/**
	 * Does the mapfile setup for a given type of plan - based on the plan type (listed the plan_types table),
	 * various layers are added to the named mapfile.
	 *
	 * @param integer $plan_id - the identifier for the plan to be setup.
	 * @param string $mapfile - if specified, the setup layers are added to the specified mapfile.
	 *
	 * @return array
	 */
	public function Setup($plan_id,$mapfile=false)
	{
		// Call the AddMapLayers function from the plan methods for the corresponding plan type...
		$r = $this->PlanMethods($plan_id);
		
		// Load/invoke the setup function for the class associated with the specified type of plan.
		$aUserLayers = $r->getMethod("Setup")->invoke(NULL,$plan_id);
		
		if ($mapfile)
		{
			return $r->getMethod("AddMapLayers")->invoke(NULL,$plan_id,$mapfile);
		}
	}
	
	/**
	 * Opens a plan in the specified mapfile.
	 *
	 * @param integer $plan_id - the identifier of the plan in the user_plans table.
	 * @param string $mapfile - the name of the mapfile to open the plan in.
	 *
	 * @return array
	 */
	public function Open($plan_id,$mapfile)
	{
		
		// Call the AddMapLayers function from the plan methods for the corresponding plan type...
		return $this->PlanMethods($plan_id)->getMethod("AddMapLayers")->invoke(NULL,$plan_id,$mapfile);
		
	}
	
	/**
	 * Returns a reflection class for the object containing static functions for a given plan type.
	 *
	 * @param integer $plan_id - the identifier for the plan to be setup.
	 *
	 * @return array
	 */
	public function PlanMethods($plan_id)
	{
		$type = $this->db->fetchOne($this->db->select()->from(array("pt"=>"plan_types"),"description")->
			join(array("up"=>$this->discussion->name.".user_plans",array()),"pt.id = up.type")->where("up.id = ?",$plan_id));
		
		$type = ucwords(implode("_",explode(" ",$type)));
		
		// Load/invoke the setup function for the class associated with the specified type of plan.
		return new ReflectionClass("MapChat_Discussion_Plan_".$type);
	}
	
	/**
	 * Gets the feature counts indexed by the layer names in a given plan:
	 *
	 * @param integer $plan_id - the identifier for the plan to be setup.
	 *
	 * @return array
	 */
	public function FeatureCounts($plan_id)
	{
		// Call the FeatureCounts function from the plan methods for the corresponding plan type...
// 		return $this->PlanMethods($plan_id)->getMethod("FeatureCounts")->invoke(NULL,$plan_id);
		
		$aNamedLayers = $this->LayerInfo($plan_id);
		
		$aReturn = array();
		foreach($aNamedLayers as $name => $aLayer)
		{
			$numFeatures = $this->db->fetchOne("select count(*) from ".$this->discussion->name.
				".user_drawings where not deleted and layer_id = ".$aLayer["id"].
				" and username = ".$this->db->quote($this->me->username).";");
			$aReturn[$name] = $numFeatures;
		}
		
		return $aReturn;
	}
	
	/**
	 * Inserts/updates a feature in a specific plan.
	 *
	 * @param integer $plan_id - the identifier for the plan to be setup.
	 * @param array $feature - the identifier for the plan to be setup.
	 *
	 * @return integer
	 */
	public function SaveFeature($plan_id,$feature)
	{
		// Call the SaveFeature function from the plan methods for the corresponding plan type...
// 		return $this->PlanMethods($plan_id)->getMethod("SaveFeature")->invoke(NULL,$plan_id,$feature);
		$aNamedLayers = $this->LayerInfo($plan_id);
		
		$aLayerAttr = MapChat_Layer::UserLayerAttrInfo($aNamedLayers[$feature["layer"]]["id"]);
		
		$aLayerFields = explode("\t",$aLayerAttr["fields"]);
		
		$aFeatureProps = array("layer_id"=>$aNamedLayers[$feature["layer"]]["id"],"username"=>$this->me->username);
		
		$aAttr = array();
		foreach ($aLayerFields as $attr)
		{
			$aAttr[] = $this->db->quote($feature["attr"][$attr]);
		}
		$aFeatureProps["attributes"] = new Zend_Db_Expr("ARRAY[".implode(",",$aAttr)."]");
		$aFeatureProps["modification_date"] = new Zend_Db_Expr("current_timestamp");
		
		if (isset($feature["coords"]))
		{
			switch ($feature["shape"])
			{
				case "polygon":
					$geometries = explode("|",$feature["coords"]);
					foreach ($geometries as $i => $coords)
					{
						$aCoords = explode(",",$coords);
						$aCoords[] = $aCoords[0];
						$geometries[$i] = $this->db->quote("SRID=".$this->discussion->srid.";POLYGON((".implode(",",$aCoords)."))");
					}
					$aFeatureProps["geom"] = new Zend_Db_Expr("st_multi(st_forcerhr(mapchat_append_polygon(ARRAY[".implode(",",$geometries)."])))");
					break;
				case "line":
					$geometries = explode("|",$feature["coords"]);
					foreach ($geometries as $i => $coords)
					{
						$geometries[$i] = $this->db->quote("SRID=".$this->discussion->srid.";LINESTRING(".$coords.")");
					}
					$aFeatureProps["geom"] = new Zend_Db_Expr("st_multi(mapchat_union_geometries(ARRAY[".implode(",",$geometries)."]))");
					break;
				case "point":
					$geometries = explode("|",$feature["coords"]);
					foreach ($geometries as $i => $coords)
					{
						$geometries[$i] = $this->db->quote("SRID=".$this->discussion->srid.";POINT(".$coords.")");
					}
					$aFeatureProps["geom"] = new Zend_Db_Expr("st_multi(mapchat_union_geometries(ARRAY[".implode(",",$geometries)."]))");
					break;
			}
		}
		else
		{
			$aFeatureProps["geom"] = new Zend_Db_Expr("NULL");
		}
		
		if (isset($feature["idx"]))
		{
			$fid = $this->db->fetchOne($this->db->select()->from("user_drawings","id",$this->discussion->name)
				->where("layer_id = ?",$aFeatureProps["layer_id"])->where("not deleted")
				->where("username = ?",$this->me->username)->order("id ASC")->limit(1,(1*$feature["idx"])-1));
			$this->db->update($this->discussion->name.".user_drawings",$aFeatureProps,$this->db->quoteInto("id = ?",$fid));
			return $fid;
		}
		else
		{
			$this->db->insert($this->discussion->name.".user_drawings", $aFeatureProps);
			return $this->db->lastSequenceId($this->discussion->name.".user_drawings_id_seq");
		}
	}
	
	/**
	 * Returns a feature in a specific plan.
	 *
	 * @param integer $plan_id - the id of the plan that hte feature will be saved to.
	 * @param string $layer - the name of the plan layer that the feature is in
	 * @param integer $idx - the feature's index (in sequential order by id) in the layer.
	 *
	 * @return array
	 */
	public function GetFeature($plan_id,$layer,$idx)
	{
		// Call the GetFeature function from the plan methods for the corresponding plan type...
// 		return $this->PlanMethods($plan_id)->getMethod("GetFeature")->invoke(NULL,$plan_id,$layer,$idx);
		$aNamedLayers = $this->LayerInfo($plan_id);
		
		$aLayerAttr = MapChat_Layer::UserLayerAttrInfo($aNamedLayers[$layer]["id"]);
		
		$aLayerFields = explode("\t",$aLayerAttr["fields"]);
		
		$aFeatureProps = array("layer_id"=>$aNamedLayers[$layer]["id"],"attr"=>array());
		
		$finfo = $this->db->fetchAssoc($this->db->select()->from("user_drawings","id, st_assvg(st_multi(st_forcerhr(geom)),0,".
				$this->db->quote($this->discussion->discussion_config->precision).") as svg".$aLayerAttr["sql"],$this->discussion->name)
			->where("layer_id = ?",$aFeatureProps["layer_id"])->where("not deleted")
			->where("username = ?",$this->me->username)->order("id ASC")->limit(1,(1*$idx)-1));
		
		$finfo = array_shift($finfo);
		
		foreach ($aLayerFields as $attr)
		{
			$aFeatureProps["attr"][$attr] = $finfo[$attr];
		}
		
		if (strlen($finfo["svg"])>0)
		{
			$aFeatureProps["coords"]="";
			if (strpos($finfo["svg"],"M")!==false)
			{
				$aFeatureProps["shape"]="line";
				$geoms = explode("M",$finfo["svg"]);
				foreach ($geoms as $g)
				{
					if (strlen($aFeatureProps["coords"])>0) $aFeatureProps["coords"] .= " | ";
					$c = explode(" ",trim($g));
					if (count($c)>1)
					{
						for ($i=0;$i<count($c);$i=$i+2)
						{
							if ($c[$i]!="Z")
							{
								if ($i>0) $aFeatureProps["coords"] .= " , ";
								$aFeatureProps["coords"] .= $c[$i]." ".(-1*$c[$i+1]);
							}
							else
							{
								$aFeatureProps["shape"] = "polygon";
							}
						}
					}
				}
			}
			else
			{
				$aFeatureProps["shape"] = "point";
				$geoms = explode(",",$finfo["svg"]);
				foreach ($geoms as $g)
				{
					if (strlen($aFeatureProps["coords"])>0) $aFeatureProps["coords"] .= " | ";
					$c = explode(" ",trim($g));
					$acx = explode("\"",$c[0]);
					$acy = explode("\"",$c[1]);
					$cx = $acx[1];
					$cy = (-1*$acy[1]);
					$aFeatureProps["coords"] .= $cx." ".$cy;
				}
			}
		}
		
		$aFeatureProps["id"] = $finfo["id"];
		
		return $aFeatureProps;
	}
	
	/**
	 * Deletes a feature from a specific plan.
	 *
	 * @param integer $plan_id - the id of the plan that the feature will be saved to.
	 * @param integer $layer - the name of the plan layer that the feature belongs to
	 * @param integer $fid - the feature's id in the user_drawings table (not the sequential index as in the Save/Get Feature APIs)
	 *
	 * @return integer - a value of one can be interpreted as the feature being deleted, zero being nothing deleted - any
	 *  other value is a problem.
	 */
	public function DeleteFeature($plan_id,$layer,$fid)
	{
		// Call the DeleteFeature function from the plan methods for the corresponding plan type...
// 		return $this->PlanMethods($plan_id)->getMethod("DeleteFeature")->invoke(NULL,$plan_id,$layer,$fid);
		$aNamedLayers = $this->LayerInfo($plan_id);
		
		$layer_id = $aNamedLayers[$layer]["id"];
		
		// Make sure that the feature actually exists and belongs to the specified plan and layer.
		$c = $this->db->fetchOne($this->db->select()->from(array("ud"=>"user_drawings"),"count(*)",$this->discussion->name)
			->joinUsing("user_plans_layers","layer_id",array(),$this->discussion->name)->where("plan_id = ?",$plan_id)
			->where("ud.layer_id = ?",$layer_id)->where("id = ?",$fid));
		
		// If the count is equal to 1, we should be safe to mark the feature as deleted:
		$d = $this->db->update($this->discussion->name.".user_drawings",array("deleted"=>true,"deletion_date"=>new Zend_Db_Expr("now()")),"id = ".$this->db->quote($fid)." and layer_id = ".$this->db->quote($layer_id));
		
		return $d;
	}
	
	
	/**
	 * Copies the contents of a specific plan to a new plan with a different title.
	 *
	 * @param integer $old_id - the id of the existing plan to copy.
	 * @param integer $newTitle - the title for the new copy of the plan.
	 * @param string $username - optionally specify a different username than the current user.
	 *
	 * @return array
	 */
	public function Copy($old_id,$newTitle,$username=false)
	{
		if (strlen($newTitle)<1)
		{
			throw new MapChat_Exception("Plan title must be specified");
		}
		
		$type = $this->db->fetchOne($this->db->select()->from("user_plans",array("type"),$this->discussion->name)
			->where("username = ?",$this->me->username)->where("id = ?",$old_id)->where("not deleted"));
		
		if (!is_numeric($type))
		{
			throw new MapChat_Exception("Error - could not access existing plan record.");
		}
		
		$this->db->insert($this->discussion.".user_plans",array(
			"title"=>$newTitle,
			"type"=>$type,
			"username"=>$username?$username:$this->me->username,
		));
		
		$new_id = $this->db->lastSequenceId($this->discussion.".user_plans_id_seq");
		
		// At this point, we've created the record in user_plans...now lets copy the related layers in user_layers:
		$oldLayers = $this->db->fetchAssoc($this->db->select()->from(array("ul"=>"user_layers"),
			array("id","name","description","attributes","shape_type","shape_attribute"),$this->discussion->name)
			->join(array("upl"=>"user_plans_layers"),"ul.id = upl.layer_id",array(),$this->discussion->name)
			->where("username = ?",$username?$username:$this->me->username)->where("plan_id = ?",$old_id)
			->where("not deleted")->order("id ASC"));
		
		foreach ($oldLayers as $oldLayer)
		{
			// Copy the layer:
			$oldLayerID = $oldLayer["id"];
			unset($oldLayer["id"]);
			$oldLayer["username"] = $username?$username:$this->me->username;
			$this->db->insert($this->discussion.".user_layers",$oldLayer);
			$newLayerID = $this->db->lastSequenceId($this->discussion.".user_layers_id_seq");
			$this->linkToLayer($new_id,$newLayerID);
			
			// Copy the layer's features:
			$oldDrawings = $this->db->fetchAssoc($this->db->select()->from(array("ud"=>"user_drawings"),
				array("geom","attributes"),$this->discussion->name)->where("username = ?",$username?$username:$this->me->username)
				->where("layer_id = ?",$oldLayerID)->where("not deleted")->order("id ASC"));
			
			foreach ($oldDrawings as $oldDrawing)
			{
				$oldDrawing["layer_id"] = $newLayerID;
				$oldDrawing["username"] = $username?$username:$this->me->username;
				$this->db->insert($this->discussion.".user_drawings",$oldDrawing);
			}
		}
		
		return $new_id;
	}
	
	
	/**
	 * Deletes the contents of a specific plan for a user.
	 *
	 * @param integer $plan_id - the id of the existing plan to copy.
	 * @param string $username - optionally specify a different username than the current user.
	 *
	 * @return boolean - success?
	 */
	public function Delete($plan_id,$username=false)
	{
		// This should be fairly simple - set deleted = true and deletion_time = now() for the plan,
		// its layers, and all of the features in it.  Only do this for records that are not already 
		// deleted though, since we don't want to change the deletion_date for any previous records.
		
		$username = ($username?$username:$this->me->username);
		
		// First, delete the plan...
		$dp = $this->db->update($this->discussion->name.".user_plans",
			array("deleted"=>true,"deletion_date"=>new Zend_Db_Expr("now()")),"not deleted and username = ".
			$this->db->quote($username)." and id = ".$this->db->quote($plan_id));
			
		if ($dp != 1) throw new MapChat_Exception("Something went wrong deleting the plan record in the database (result = ".$dp.")");
		
		// Now delete the layers...
		$dl = $this->db->update($this->discussion->name.".user_layers",
			array("deleted"=>true,"deletion_date"=>new Zend_Db_Expr("now()")),"not deleted and username = ".
			$this->db->quote($username)." and id in (select layer_id from ".$this->discussion->name.
			".user_plans_layers where plan_id = ".$this->db->quote($plan_id).")");
		
		// If no layers were deleted, that's unexpected...unless we implement plans that can have custom layers in it...in any case,
		// if there were no layers, then there were no drawings either, so this may not be an exepction - but it is for now.
		if ($dl<=0) throw new MapChat_Exception("Something went wrong deleting the plan layers in the database (result = ".$dl.")");
		
		// Now delete the drawings...
		$dd = $this->db->update($this->discussion->name.".user_drawings",
			array("deleted"=>true,"deletion_date"=>new Zend_Db_Expr("now()")),"not deleted and username = ".
			$this->db->quote($username)." and layer_id in (select layer_id from ".$this->discussion->name.
			".user_plans_layers where plan_id = ".$this->db->quote($plan_id).")");
		
		// If no features were deleted, that doesn't matter, since there may actually have been no current drawings.
		return true;
		
	}
	
}

?>
