<?php
/**
 * class MapChat_Discussion_Select
 * Provides the general utilities for managing selections in the
 * MapChat database.
 *
 * @category MapChat
 * @package MapChat
 */
class MapChat_Discussion_Select
{
	public function __construct($discussion)
	{
		$this->db = Zend_Registry::get("db");
		$this->discussion = $discussion;
		$this->me = Zend_Registry::get("me");
	}
	
	/**
	 * Modifies the current user's selected features:
	 * 
	 * @param string $opts - an array of parameters that control how the selection is updated.
	 *
	 * Possible values that can be included in the opts parameter:
	 *  map: the name of the mapfile to query (required)
	 *  cellSize: the client pixel size (required)
	 *  mode: one of 'add', 'remove', 'clear', 'list' - defines wither to add new features, remove features, clear, or list the current selection
	 *  style: one of 'point', 'circle', 'rectangle', 'by_id' - defines how the selection will be performed
	 *  layer: either the name of a vector layer in the map, or '[All Visible Layers]'
	 *  coords: the coordinates that are to be used to search for features
	 *  ids: if style is 'by_id', and mode is 'remove', this should be a list of ids from the user_selection table to remove
	 *
	 * @return array - the selection results
	 */
	public function Features(array $opts)
	{
		$aResults = array("mode"=>$opts["mode"],"results"=>array($opts["mode"]=>array()));
		$oMap = $this->discussion->Map_getMapObj($opts["map"]);
		
		switch (@$opts["style"])
		{
			case "point":
				$oPoint = ms_newPointObj();
				$oPoint->setXY($opts["coords"][0],$opts["coords"][1]);
			break;
			
			case "circle":
				$r = sqrt(pow($opts["coords"][0]-$opts["coords"][2],2)+pow($opts["coords"][1]-$opts["coords"][3],2));
				$sWKT = $this->db->fetchOne("select st_astext(st_buffer('POINT(".
					($opts["coords"][0]*1.0)." ".($opts["coords"][1]*1.0).
					")'::geometry,".$r."));");
				$aResults["mode"] = $r;
			break;
			
			case "rectangle":
				$sWKT = "POLYGON((".($opts["coords"][0]*1.0)." ".($opts["coords"][1]*1.0).
				",".($opts["coords"][0]*1.0)." ".($opts["coords"][3]*1.0).
				",".($opts["coords"][2]*1.0)." ".($opts["coords"][3]*1.0).
				",".($opts["coords"][2]*1.0)." ".($opts["coords"][1]*1.0).
				",".($opts["coords"][0]*1.0)." ".($opts["coords"][1]*1.0)."))";
			break;
		}
		
		$aLayers = array();
		if (@$oLayer = $oMap->getLayerByName($opts["layer"]))
		{
			$aLayers[] = $oLayer;
		}
		else
		{
			// Assuming the option to select from all visisble layers was chosen:
			for ($i=0;$i<$oMap->numlayers;$i++)
			{
				$oLayer = $oMap->getLayer($i);
				if ($oLayer->status==MS_ON && in_array($oLayer->type,array(MS_LAYER_POINT, MS_LAYER_LINE, MS_LAYER_POLYGON)))
				{
					$aLayers[] = $oLayer;
				}
			}
		}
		
		switch ($opts["mode"])
		{
			case "add":
				if (isset($oPoint))
				{
					foreach ($aLayers as $oLayer)
					{
						$oLayer->set("toleranceunits",$oMap->units);
						$oLayer->set("tolerance",$opts["cellSize"]*3);
						$oLayer->set("template","dummy.html");
						
						if (@$oLayer->queryByPoint($oPoint,MS_MULTIPLE,$oLayer->tolerance) == MS_SUCCESS)
						{
							$aRes = $this->GetResults($oLayer,$opts["map"]);
							$aResults["results"][$opts["mode"]][$oLayer->name] = $aRes;
						}
					}
					
				}
				else
				{
					$oShape = ms_shapeObjFromWkt($sWKT);
					foreach ($aLayers as $oLayer)
					{
						$oLayer->set("tolerance",0);
						$oLayer->set("template","dummy.html");
						if (@$oLayer->queryByShape($oShape) == MS_SUCCESS)
						{
							$aRes = $this->GetResults($oLayer,$opts["map"]);
							$aResults["results"][$opts["mode"]][$oLayer->name] = $aRes;
						}
					}
				}
			break;
			
			case "remove":
				if ($opts["style"]!="by_id")
				{
					if (isset($oPoint)) $sGeom = "'SRID=".$this->discussion->srid.";POINT(".$oPoint->x." ".$oPoint->y.")'::geometry";
					else $sGeom = "'SRID=".$this->discussion->srid.";".$sWKT."'::geometry";
					
					foreach ($aLayers as $oLayer)
					{
						// First get the IDs that are about to be deleted:
						$aDelSel = $this->db->getConnection()->query("select id, feature_id, layer_name from ".
							$this->discussion->name.".user_selections where username = ".$this->db->quote($this->me->username).
							" and mapfile = ".$this->db->quote($opts["map"]).
							" and not deleted and (layer_name = ".$this->db->quote($oLayer->name).
							" or layer_name = '' or layer_name is null) and st_intersects(geom,".
							$sGeom.");")->fetchAll(PDO::FETCH_ASSOC);
						
						$aDelIDs = array();
						foreach ($aDelSel as $aDel) $aDelIDs[] = $aDel["id"];
						
						if (count($aDelIDs)>0)
						{
							$sDelIDs = MapChat_Array::toPG($aDelIDs,true);
							
							// Now delete them from the users selection table:
							$this->db->getConnection()->query("update ".
								$this->discussion->name.".user_selections set deleted = true, deletion_date = now() where username = ".
								$this->db->quote($this->me->username)." and not deleted and id::text = any (".$sDelIDs."::text[]);");
							
							foreach ($aDelSel as $sel)
							{
								if (!isset($aResults["results"][$opts["mode"]][$sel["layer_name"]]))
									$aResults["results"][$opts["mode"]][$sel["layer_name"]] = array();
								$aResults["results"][$opts["mode"]][$sel["layer_name"]][$sel["id"]] = true;
							}
						}
					}
				}
				else
				{
					// First get the IDs that are about to be deleted:
					$aDelSel = $this->db->getConnection()->query("select id, feature_id, layer_name from ".
						$this->discussion->name.".user_selections where username = ".$this->db->quote($this->me->username).
						" and mapfile = ".$this->db->quote($opts["map"]).
						" and not deleted and id::text = any (".MapChat_Array::toPG($opts["ids"],true).
						"::text[]);")->fetchAll(PDO::FETCH_ASSOC);
						
					// Set the deleted status on the specified layers...
					$this->db->getConnection()->query("update ".
						$this->discussion->name.".user_selections set deleted = true, deletion_date = now() where username = ".
						$this->db->quote($this->me->username)." and mapfile = ".$this->db->quote($opts["map"]).
						" and not deleted and id::text = any (".MapChat_Array::toPG($opts["ids"],true)."::text[]);");
					
					foreach ($aDelSel as $sel)
					{
						if (!isset($aResults["results"][$opts["mode"]][$sel["layer_name"]]))
							$aResults["results"][$opts["mode"]][$sel["layer_name"]] = array();
						$aResults["results"][$opts["mode"]][$sel["layer_name"]][$sel["id"]] = true;
					}
				}
			break;
			
			case "clear":
				$this->db->update($this->discussion.".user_selections",array("deleted"=>(integer)true),
					$this->db->quoteInto("username = ?",$this->me->username));
				$aResults["results"] = true;
			break;
			
			case "list":
				$aSel = $this->db->getConnection()->query("select id, layer_name as ln, attr_json as aj, st_assvg(geom,0,".
					$this->db->quote($this->discussion->discussion_config->precision).") as svg, ".
					"case when st_geometrytype(geom) in ('ST_Polygon','ST_MultiPolygon') then st_assvg(st_pointonsurface(case when st_isvalid(geom) then geom else st_buffer(geom,0) end)) ".
					"when st_geometrytype(geom) in ('ST_LineString') then st_assvg(ST_line_interpolate_point(geom,0.5)) ".
					"else st_assvg(st_centroid(geom)) end as centroid, st_assvg(st_envelope(geom)) as simple from ".$this->discussion->name.
					".user_selections where username = ".$this->db->quote($this->me->username)." and mapfile = ".$this->db->quote($opts["map"]).
					" and not deleted;")->fetchAll(PDO::FETCH_ASSOC);
					
				foreach ($aSel as $sel)
				{
					$aj = Zend_Json::decode($sel["aj"]);
					
					if (strlen($sel["svg"]) <= $this->discussion->discussion_config->svg_limit)
					{
						$aj[0]["shape"] = $sel["svg"];
					}
					else
					{
						$aj[0]["shape"] = $sel["centroid"];
						$aj[0]["simple"] = $sel["simple"];
					}
					
					if (!isset($aResults["results"][$opts["mode"]][$sel["ln"]])) $aResults["results"][$opts["mode"]][$sel["ln"]] = array();
					$aResults["results"][$opts["mode"]][$sel["ln"]][$sel["id"]] = $aj;
				}
				if (count($aResults["results"][$opts["mode"]])==0) $aResults["results"][$opts["mode"]] = false;
			break;
		}
		
		$this->discussion->Event_log(array(
			"type"=>MapChat_EventTypes::SPATIAL_TOOLS,
			"text"=>"Select_Features",
			"data"=>array(
					$opts["mode"],
					(isset($opts["style"])?$opts["style"]:"noStyle"),
					(isset($opts["cellSize"])?$opts["cellSize"]:"noCellSize"),
					(isset($opts["map"])?$opts["map"]:"noMap"),
					(isset($opts["coords"])?$opts["coords"]:"noCoords"),
					(isset($opts["layer"])?$opts["layer"]:"noLayer"),
					(isset($opts["ids"])?$opts["ids"]:"noIDs"),
				),
			"geom"=>(isset($oPoint)?
				new Zend_Db_Expr("'SRID=".$this->discussion->srid.";POINT(".$oPoint->x." ".$oPoint->y.")'::geometry"):
					(isset($sWKT)?new Zend_Db_Expr("'SRID=".$this->discussion->srid.";".$sWKT."'::geometry"):NULL)),
			"username"=>$this->me->username,
		));
		
		return $aResults;
	}


	/**
	 * Modifies the current user's selected features:
	 * 
	 * @param layerObj $oLayer - a layer that has had some query results.
	 *
	 * @return array - the formatted results
	 */
	public function GetResults($oLayer,$sMapfile="")
	{
		$aResults = array();
		$oLayer->open();
		$nResults = $oLayer->getNumResults();
		
		$sLayerProj = false;//$oLayer->getProjection();
		$sMapProj = false;//$oMap->getProjection();
		$bProjShape = false;
		if ($sLayerProj && $sMapProj && $sLayerProj!=$sMapProj)
		{
			$bProjShape = true;
			$oLayerProj = ms_newProjectionObj($sLayerProj);
			$oMapProj = ms_newProjectionObj($sMapProj);
		}
		
		for ($i=0;$i<$nResults;$i++)
		{
			$oResult = $oLayer->getResult($i);
			if (ms_GetVersionInt()<50600)
			{
				$oShape = $oLayer->getShape($oResult->tileindex, $oResult->shapeindex);
			}
			else
			{
				$oShape = $oLayer->resultsGetShape($oResult->shapeindex,$oResult->tileindex);
			}
			if ($bProjShape) $oShape->project($oLayerProj,$oMapProj);
			
			$wkt = MapChat_ShapeObj::toWkt($oShape);
			
			// Convert the wkt to SVG - precision defined in the xml config should be appropriate
			// for the scales and coordinate system used in the map - for UTM, zero is probably fine. Fewer
			// decimals means less data being transferred to the client:
			$aSVG = $this->db->getConnection()->query("select st_assvg(st_multi(st_forcerhr('".$wkt."'::geometry)),0,".
				$this->db->quote($this->discussion->discussion_config->precision).") as svg, st_assvg(st_envelope('".$wkt."'::geometry)) as simple;")->fetch(PDO::FETCH_ASSOC);
			
			$sSVG = $aSVG["svg"];
			$sSimple = $aSVG["simple"];
			
			$simplified = false;
			
			// Return a single point + max extents if a geometry is particularly large.
			if (strlen($sSVG) > $this->discussion->discussion_config->svg_limit)
			{
				$simplified = true;
				
				if (@$oShape->type==MS_SHAPE_POLYGON) $shapeInfo = $this->db->getConnection()->query(
						"select st_assvg(st_pointonsurface(case when st_isvalid(geom) then geom else st_buffer(geom,0) end)) as svg, st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from (select '"
						.$wkt."'::geometry as geom) as g;"
					)->fetch(PDO::FETCH_ASSOC);
				
				elseif (@$oShape->type==MS_SHAPE_LINE) $shapeInfo = $this->db->getConnection()->query(
						"select case when st_geometrytype(geom) in ('ST_LineString') then st_assvg(ST_line_interpolate_point(geom,0.5)) else st_assvg(st_centroid(geom)) end as svg, st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from (select '".$wkt."'::geometry as geom) as g;"
					)->fetch(PDO::FETCH_ASSOC);
				
				else $shapeInfo = $this->db->getConnection()->query(
						"select st_assvg(st_centroid(geom)) as svg, st_xmin(geom) as xmin, st_ymin(geom) as ymin, st_xmax(geom) as xmax, st_ymax(geom) as ymax from (select '".$wkt."'::geometry as geom) as g;"
					)->fetch(PDO::FETCH_ASSOC);
				
				$sSVG = $shapeInfo["svg"];
				$aExtent = array(array($shapeInfo["xmin"],$shapeInfo["ymin"]),array($shapeInfo["xmax"],$shapeInfo["ymax"]));
			}
			
			// Make sure the result isn't already selected...skip the result if it is.
			if ($this->db->fetchOne($this->db->select()->from("user_selections",array("num"=>"count(*)"),$this->discussion->name)
				->where("username = ?",$this->me->username)->where("not deleted")->where("layer_name = ?",$oLayer->name)
				->where("feature_id = ?",$oResult->shapeindex))>0) continue;
			
			$aResult = array(
				array(
					"ti"=>$oResult->tileindex,
					"si"=>$oResult->shapeindex,
					"ci"=>$oResult->classindex,
					"area"=>@$oShape->getArea(),
					"length"=>@$oShape->getLength(),
					//"shape"=>$sSVG,
					"type"=>($simplified?MS_SHAPE_POINT:@$oShape->type),
					"layertype"=>@$oLayer->type,
					"extent"=>array(array(@$oShape->bounds->minx,@$oShape->bounds->miny),array(@$oShape->bounds->maxx,@$oShape->bounds->maxy))
				),
				$oShape->values
			);
			
			// Don't forget, Zend_Db causes php to segfault with statements that have large geometries:
			$this->db->getConnection()->query("insert into ".$this->discussion.
				".user_selections (username,mapfile,layer_name,user_layer,user_drawing,feature_id,attr_json,geom) values (".
				$this->db->quote($this->me->username).",".$this->db->quote($sMapfile).
				",".$this->db->quote($oLayer->name).",".($oLayer->group=="USER_LAYERS"?"true":"false").
				",".(strtoupper($oLayer->getMetaData("mapchat_user_drawings"))=="TRUE"?"true":"false").
				",".$this->db->quote($oResult->shapeindex).",".$this->db->quote(Zend_Json::encode($aResult)).
				",'SRID=".$this->discussion->srid.";".$wkt."'::geometry);");
			
			$id = $this->db->lastSequenceId($this->discussion.".user_selections_id_seq");
			
			$aResult[0]["shape"] = $sSVG;
			
			if ($simplified) $aResult[0]["simple"] = $sSimple;
			
			$aResults[$id]=$aResult;
			
		}
		
		return $aResults;
	}
}

?>
