<?php

class MapChat_Group
{

	private $_aGroupInfo = NULL;
	private $_bExists = false;
	private $_db = NULL;
	private $_config = NULL;
	
	/*
	 * Identify the properties required to restore this object
	 * (i.e., the properties that must be saved to the session).
	 */
	public function __sleep()
	{
		return array("_aGroupInfo","_bExists");
	}
	
	/*
	 * Identify the properties that must be reinitialized when this
	 * object is restored from a saved session:
	 */
	public function __wakeup()
	{
		$this->_db = Zend_Registry::get("db");
		$this->_config = Zend_Registry::get("config");
	}
	
	/*
	 * Allow this object to be cast as a string containing the groupname.
	 */
	public function __toString() 
	{
		if (isset($this->_aGroupInfo["groupname"]))
		{
			return $this->_aGroupInfo["groupname"];
		}
		else
		{
			return "";
		}
	}
	
	/*
	 * Allow the group properties accessed as object members.
	 */
	public function __get($n) 
	{
		if (isset($this->_aGroupInfo[$n]))
		{
			return $this->_aGroupInfo[$n];
		}
		else
		{
			return NULL;
		}
	}
	
//   // Disabled for now - this might make it too easy to mess up the database
//   // while developing/testing code.
//   /*
// 	 * Allow the values of group properties to be set as object members.
// 	 */
// 	public function __set($n,$v) 
// 	{
// 		$this->setInfo(array($n => $v));
// 	}
	
	/*
	 * Construct the object by querying for the record in the
	 * groups table that matches the groupname supplied to 
	 * the object constructor:
	 */
	public function __construct($sGroupname=false)
	{
		$this->_db = Zend_Registry::get("db");
		$this->_config = Zend_Registry::get("config");
		
		if ($sGroupname!==false && strlen($sGroupname)>0)
		{
			$group = $this->_db->fetchRow('SELECT * FROM groups WHERE groupname = ? LIMIT 1',$sGroupname);
			if ($group)
			{
				$this->_aGroupInfo = $group;
				$this->_bExists = true;
			}
			
			// Even if the group does not exist, we should at least
			// store the groupname, so we know what name the object
			// was initialized with:
			else
			{
				$this->_aGroupInfo = array("groupname"=>$sGroupname);
			}
		}
	}
	
	/*
	 * Returns true if there is an array saved in the aGroupInfo
	 * property (i.e., either the group existed when the object
	 * was instantiated, or it has since been created).
	 */
	public function exists()
	{
		return $this->_bExists;
	}
	
	/*
	 * Checks if this group has administrative rights.
	 */
	public function isAdmin()
	{
		// Return the is_admin property if it is available:
		if (isset($this->_aGroupInfo["is_admin"]))
		{
			return $this->_aGroupInfo["is_admin"];
		}
		
		// Either the group doesn't exist, or ths object
		// hasn't been instantiated with a groupname
		else
		{
			return false;
		}
	}
	
	public function hasPermission($sPermission)
	{
		if (isset($this->_aGroupInfo[$sPermission]))
		{
			return ($this->_aGroupInfo["is_admin"]||$this->_aGroupInfo[$sPermission]);
		}
		else
		{
			return false;
		}
	}
	
	/*
	 *	Checks if the group has permissions to create new groups.
	 */
	public function canCreateGroups()
	{
		return $this->hasPermission("can_create_groups");
	}
	
	/*
	 *	Checks if the group has permissions to create new users.
	 */
	public function canCreateUsers()
	{	
		return $this->hasPermission("can_create_users");
	}
	
	/*
	 *	Checks if the group has permissions to create new discussions.
	 */
	public function canCreateDiscussions()
	{
		return $this->hasPermission("can_create_discussions");
	}
	
	
	/*
	 *	Checks if the group has permissions to manage existing groups.
	 */
	public function canManageGroups()
	{
		return $this->hasPermission("can_manage_groups");
	}
	
	/*
	 *	Checks if the group has permissions to manage existing users
	 */
	public function canManageUsers()
	{
		return $this->hasPermission("can_manage_users");
	}
	
	/*
	 *	Checks if the group has permissions to manage existing discussions
	 */
	public function canManageDiscussions()
	{
		return $this->hasPermission("can_manage_discussions");
	}
	
	/*
	 * Checks if the group has any admin-related permissions:
	 */
	function hasSomeAdmin()
	{
		return ($this->isAdmin()||$this->canCreateDiscussions()||$this->canCreateGroups()||$this->canCreateUsers()||
			$this->canManageDiscussions()||$this->canManageGroups()||$this->canManageUsers());
	}
	
	/*
	 * This function returns info from the group's record in groups.
	 */
	public function getInfo($sID=false)
	{
		// If an ID is provided that references a column in the groups
		// table, then return only that particular attribute...
		if ($sID!=false)
		{
			if (isset($this->_aGroupInfo[$sID]))
				return $this->_aGroupInfo[$sID];
			
			// If an invalid column ID was supplied:
			else
				return false;
		}
		
		// If nothing is supplied for a column ID, then return the full
		// group's record as an array:
		else
		{
			return $this->_aGroupInfo;
		}
	}
	
	/*
	 * This function updates attributes of the group in the database.
	 * the $aInfo must be an array with values referenced by keys that
	 * correspond to the columns in groups.
	 */
	public function setInfo($aInfo)
	{
		if (!$this->exists())
		{
			throw new MapChat_Exception("Cannot update non-existent group."); 
		}
		
		if (is_array($aInfo)&&count($aInfo)>0)
		{
			// Get the groups table metatada:
			$aTblGroups = $this->_db->describeTable("groups","public");
			
			// Loop through the supplied values...
			foreach ($aInfo as $sKey => $sValue)
			{
				// If one of the variables doesn't match, throw an exception...
				if (!isset($aTblGroups[$sKey])) 
				{
					throw new MapChat_Exception("Invalid column specified: ".$sKey);
				}
				elseif ($aTblGroups[$sKey]["PRIMARY"]) 
				{
					throw new MapChat_Exception("Altering the column '".$sKey."' is not allowed, as it is a primary key.");
				}
			}
			
			// If one of the items is the password, make sure it gets encrypted:
			if (isset($aInfo["password"]))
			{
				if (strlen($aInfo["password"])>5)
				{
					$aInfo["password"] = new Zend_Db_Expr($this->_db->quoteInto('md5(?)',$aInfo["password"]));
				}
				else
				{
					throw new MapChat_Exception("The password must be at least 6 characters in length.");
				}
			}
			
			// Execute the update:
			$this->_db->update("groups",$aInfo,$this->_db->quoteInto("groupname = ?",$this->_aGroupInfo["groupname"]));
			
			$this->_aGroupInfo = $this->_db->fetchRow('SELECT * FROM groups WHERE groupname = ? LIMIT 1',$this->_aGroupInfo["groupname"]);
			
		}
		
		// The input wasn't an array, or was empty:
		else
		{
			throw new MapChat_Exception("Invalid input for MapChat_Group::setInfo() - must be an array.");
		}
		
	}
	
	/*
	 * This function returns either full_name or groupname (if full_name is not
	 * available) - useful for getting an appropriate display name.
	 */
	public function displayName()
	{
		return (strlen($this->_aGroupInfo["description"])>0?$this->_aGroupInfo["description"]:$this->_aGroupInfo["name"]);
	}
	
	/*
	 * This function can be used to create new groups.  The $aGroupInfo 
	 * parameter should be an array containing values with keys that 
	 * match the columns of the groups table in the MapChat2 database.
	 */
	public function create($aGroupInfo)
	{
		// The groupname is optional, provided that one was supplied when 
		// the current instance of the MapChat_Group object was created...
		if (!isset($aGroupInfo["groupname"]))
		{
			if ($this->_aGroupInfo["groupname"])
			{
				$aGroupInfo["groupname"] = $this->_aGroupInfo["groupname"];
			}
			else
			{
				throw new MapChat_Exception("A groupname must be specified in order to create a new group");
			}
		}
		
		$tmpGroup = new MapChat_Group($aGroupInfo["groupname"]);
		if ($tmpGroup->exists())
		{
			throw new MapChat_Exception("The group '".$aGroupInfo["groupname"]."' already exists.");
		}
		
		// Verify that the required parameters groupname, password, and moderator are all OK:
		if (strlen($aGroupInfo["groupname"])<3)
		{
			throw new MapChat_Exception("The groupname '".$aGroupInfo["groupname"]."' is too short.  Please use three or more characters.");
		}
		if (!isset($aGroupInfo["password"])||strlen($aGroupInfo["password"])<6)
		{
			throw new MapChat_Exception("A password of six or more characters must be provided.");
		}
		
		$gmod = new MapChat_User((isset($aGroupInfo["moderator"])?$aGroupInfo["moderator"]:false));
		if (!$gmod->exists())
		{
			throw new MapChat_Exception("An existing user must be specified as the moderator for the group.");
		}
		
		$aGroupInfo["password"] = new Zend_Db_Expr($this->_db->quoteInto('md5(?)',$aGroupInfo["password"]));
		
		// Get the groups table metatada:
		$aTblGroups = $this->_db->describeTable("groups","public");
		
		// Loop through the supplied values...
		foreach ($aGroupInfo as $sKey => $sValue)
		{
			// If one of the variables doesn't match, throw an exception...
			if (!isset($aTblGroups[$sKey])) 
			{
				throw new MapChat_Exception("Invalid column specified: ".$sKey);
			}
		}
		
		$this->_db->insert('groups', $aGroupInfo);
		
		$this->_aGroupInfo = $this->_db->fetchRow('SELECT * FROM groups WHERE groupname = ? LIMIT 1',$aGroupInfo["groupname"]);
		
		$this->_bExists = true;
	}
	
	/*
	 * This function checks whether a string matches the group's current
	 * password (e.g., for verifying actions such as editing a group's
	 * profile)
	 */
	public function checkPassword($sPassword)
	{
		return $this->_aGroupInfo["password"]==md5($sPassword);
	}
	
	/*
	 * Adds users to the group.
	 */
	public function addUser($user)
	{
		// Check the input for either a string, or a MapChat_User class:
		if (gettype($user)=="object" && get_class($user)=="MapChat_User")
		{
			$oUser = $user;
		}
		elseif (gettype($user)=="string")
		{
			$oUser = new MapChat_User($user);
		}
		else
		{
			throw new MapChat_Exception("Invalid input.");
		}
		
		if ($oUser->exists())
		{
			if (!$this->hasUser($oUser))
			{
				$this->_db->insert('group_user', array("username"=>$oUser->username,"groupname"=>$this->_aGroupInfo["groupname"]));
			}
// 			else
// 			{
// 				throw new MapChat_Exception("The user '".$oUser->username."' is already a member of this group.");
// 			}
		}
		else
		{
			throw new MapChat_Exception("Cannot add '".$oUser->username."', because the user doesn't exist.");
		}
	}
	
	/*
	 * Verifies that the user is a member of the group.
	 */
	public function hasUser($user)
	{
		// Check the input for either a string, or a MapChat_User class:
		if (gettype($user)=="object" && get_class($user)=="MapChat_User")
		{
			$oUser = $user;
		}
		elseif (gettype($user)=="string")
		{
			$oUser = new MapChat_User($user);
		}
		else
		{
			throw new MapChat_Exception("Invalid input.");
		}
		
		if ($oUser->exists())
		{
			$gu = $this->_db->fetchRow($this->_db->select()->from('group_user')->where("username = ?",$oUser->username)->where("groupname = ?",$this->groupname)->limit(1));
			return ($gu?true:false);
		}
		else
		{
			// Should this cause an exception?  Just returning false for now - obviously the group won't have non-existent users.
			//throw new MapChat_Exception("The user '".$oUser->username."' doesn't exist.");
			return false;
		}
	}
	
	/*
	 * Drops users from the group.
	 */
	public function dropUser($user)
	{
		// Check the input for either a string, or a MapChat_User class:
		if (gettype($user)=="object" && get_class($user)=="MapChat_User")
		{
			$oUser = $user;
		}
		elseif (gettype($user)=="string")
		{
			$oUser = new MapChat_User($user);
		}
		else
		{
			throw new MapChat_Exception("addUser: Invalid input.");
		}
		
		if ($this->hasUser($oUser))
		{
			$this->_db->delete('group_user',array(
				$this->_db->quoteInto("username = ?",$oUser->username),
				$this->_db->quoteInto("groupname = ?",$this->groupname)
			));
		}
// 		else
// 		{
// 			// Should this cause an exception?  If the group never had the user, dropping it 
// 			// shouldn't calse a problem since the end result is the same.
// 			throw new MapChat_Exception("The group doesn't have the user '".$oUser->username."'.");
// 		}
	}
	
	/*
	 * Lists users in the group.
	 */
	public function listUsers($fullRecord=false)
	{
		if (!$fullRecord)
		{
			return $this->_db->fetchCol("select username from group_user where groupname = ?",$this->groupname);
		}
		else
		{
			return $this->_db->fetchAll("select distinct users.* from users join group_user using (username) where groupname = ?",$this->groupname);
		}
	}
	
	public function listAllGroups()
	{
		return Zend_Registry::get("db")->fetchAll("select * from groups order by groupname;");
	}
	
	public function listGroupModerators($currentUser=false)
	{
		if ($currentUser===false)
		{
			return Zend_Registry::get("db")->fetchAll("select distinct users.* from users join group_user using (username) join groups using (groupname) where is_admin or can_manage_groups;");
		}
		else
		{
			return Zend_Registry::get("db")->fetchAll("select distinct users.* from users join group_user using (username) join groups using (groupname) where is_admin or can_manage_groups or username = ?;",$currentUser->username);
		}
	}
	
	public function getModerator()
	{
		return new MapChat_User($this->moderator);
	}
}

?>