<?php

/**
 * Imitates Ka-Map's tile.php for building tiles.
 */
class MapChat_KaMap_Tile
{
	public function getTile()
	{
		/**********************************************************************
		 *
		 * $Id: tile.php,v 1.41 2006/11/13 18:34:36 pspencer Exp $
		 *
		 * purpose: a simple phpmapscript-based tile renderer that implements
		 *          rudimentary caching for reasonable efficiency.  Note the
		 *          cache never shrinks in this version so your disk could
		 *          easily fill up!
		 *
		 * author: Paul Spencer (pspencer@dmsolutions.ca)
		 *
		 * modifications by Daniel Morissette (dmorissette@dmsolutions.ca)
		 *
		 * TODO:
		 *
		 *   - remove debugging stuff (not really needed now)
		 *
		 **********************************************************************
		 *
		 * Copyright (c) 2005, DM Solutions Group Inc.
		 *
		 * Permission is hereby granted, free of charge, to any person obtaining a
		 * copy of this software and associated documentation files (the "Software"),
		 * to deal in the Software without restriction, including without limitation
		 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
		 * and/or sell copies of the Software, and to permit persons to whom the
		 * Software is furnished to do so, subject to the following conditions:
		 *
		 * The above copyright notice and this permission notice shall be included
		 * in all copies or substantial portions of the Software.
		 *
		 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
		 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
		 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
		 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
		 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
		 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
		 * DEALINGS IN THE SOFTWARE.
		 *
		 **********************************************************************/
		
		/*
		 * the tile renderer accepts several parameters and returns a tile image from
		 * the cache, creating the tile only if necessary.
		 *
		 * all requests include the pixel location of the request at a certain scale
		 * and this script figures out the geographic location of the tile from the
		 * scale assuming that 0,0 in pixels is 0,0 in geographic units
		 *
		 * Request parameters are:
		 *
		 * map: the name of the map to use.  This is handled by config.php.
		 *
		 * t: top pixel position
		 * l: left pixel position
		 * s: scale
		 * g: (optional) comma-delimited list of group names to draw
		 * layers: (optional) comma-delimited list of layers to draw
		 * force: optional.  If set, force redraw of the meta tile.  This was added to
		 *        help with invalid images sometimes being generated.
		 * tileid: (optional) can be used instead of t+l to specify the tile coord.,
		 *         useful in regenerating the cache
		 */
		
		//include_once( '../include/config.php' );
		
		$kaConf = new MapChat_KaMap_Config();
		
		/* get the various request parameters
		 * also need to make sure inputs are clean, especially those used to
		 * build paths and filenames
		 */
		
		$top = isset( $_REQUEST['t'] ) ? intval($_REQUEST['t']) : 0;
		$left = isset( $_REQUEST['l'] ) ? intval($_REQUEST['l']) : 0;
		$scale = isset( $_REQUEST['s'] ) ? $_REQUEST['s'] : $kaConf->anScales[0];
		$bForce = isset($_REQUEST['force'])? true : false;
		$groups = isset( $_REQUEST['g'] ) ? $_REQUEST['g'] : "";
		$layers = isset( $_REQUEST['layers'] ) ? $_REQUEST['layers'] : "";
		$timestamp = isset($_REQUEST['ts']) ? $_REQUEST['ts'] : '';
		$interval = isset($_REQUEST['interval']) ? $_REQUEST['interval'] : 300;
		$version = isset($_REQUEST['version']) ? $_REQUEST['version'] : '';
		$userlayer = isset($_REQUEST['userlayer']) ? true : false;
		
		// dynamic imageformat ----------------------------------------------
		//use the function in config.php to set the output format
		$useAlphaCh = false;
		if (isset($_REQUEST['i'])) {
			$kaConf->setOutputFormat( $_REQUEST['i'] );
			if (strtoupper($_REQUEST['i']) == 'ALPHA') {
				$useAlphaCh = true;
			}
		}
		
		//----------------------------------------------------------------
		
		/* tileid=t#####l#### can be used instead of t+l parameters. Useful in
		 * regenerating the cache for instance.
		 */
		if (isset( $_REQUEST['tileid']) &&
				preg_match("/t(-?\d+)l(-?\d+)/", $_REQUEST['tileid'], $aMatch) )
		{
				$top = intval($aMatch[1]);
				$left = intval($aMatch[2]);
		}
		
		/* Check that user has the rights to see requested group(s).
		 */
		if (isset($kaConf->oAuth) && $groups)
		{
				$aszGroups = explode(",", $groups);
		
				foreach ($aszGroups as $thisGroup)
				{
						if (!$kaConf->oAuth->testPrivilege($thisGroup))
						{
								/* User is not authorized. */
								echo "You are not authorized to access group '$thisGroup'\n";
								exit;
						}
				}
		}
		
		/* bug 1253 - root permissions required to delete cached files */
		$orig_umask = umask(0);
		
		/* create the main cache directory if necessary */
		if (!@is_dir($kaConf->szMapCacheDir))
				$kaConf->makeDirs($kaConf->szMapCacheDir);
		
		/* Calculate the metatile's top-left corner coordinates.
		 * Include the $metaBuffer around the metatile to account for various
		 * rendering issues happening around the edge of a map
		 */
		$metaLeft = floor( ($left)/($kaConf->tileWidth*$kaConf->metaWidth) ) * $kaConf->tileWidth * $kaConf->metaWidth;
		$metaTop = floor( ($top)/($kaConf->tileHeight*$kaConf->metaHeight) ) * $kaConf->tileHeight * $kaConf->metaHeight;
		$szMetaTileId = "t".$metaTop."/l".$metaLeft;
		$metaLeft -= $kaConf->metaBuffer;
		$metaTop -= $kaConf->metaBuffer;
		
		/* caching is done by scale value, then groups and layers and finally metatile
		 * and tile id. Create a new directory if necessary
		 */
		$szGroupDir = $groups != "" ? $kaConf->normalizeString($groups) : "def";
		$szLayerDir = $layers != "" ? sha1($kaConf->normalizeString($layers)) : "def";
		
		/*
		 * MapChat has user-controlled layers - if the layer is flagged as user-modified, then
		 * a new cache will need to be started - an sha1 string from the username is appended
		 * to the path for this:
		 */
		$sUserDir = $userlayer ? sha1($kaConf->username)."/" : "";
		
		
		$szCacheDir = $kaConf->szMapCacheDir."/".$scale."/".$szGroupDir."/".$szLayerDir."/".$sUserDir.$szMetaTileId;
		if (!@is_dir($szCacheDir))
				$kaConf->makeDirs($szCacheDir);
		
		/* resolve cache hit - clear the os stat cache if necessary */
		$szTileId = "t".$top."l".$left;
		$szCacheFile = $szCacheDir."/".$szTileId.$kaConf->szImageExtension;
		clearstatcache();
		
		$szMetaDir = $szCacheDir."/meta";
		if (!@is_Dir($szMetaDir))
				$kaConf->makeDirs($szMetaDir);
				
		/* timestamp checking - ignore requests more than the interval in the future (bad client)*/
		$now = $kaConf->gmttime();
		if ($timestamp != '' && $now >= $timestamp - $interval) {
				//get the metatile cache timestamp
				$ts_file = $szMetaDir."/timestamp";
				clearstatcache();
				if (!file_exists($ts_file)) {
						$bForce = true;
				} else {
						$h = fopen($ts_file,"r");
						$meta_timestamp = fread($h,filesize($ts_file));
						fclose($h);
						if ($timestamp > $meta_timestamp + $interval) {
								$bForce = true;
						}    
				}
		}
		
		if ($version != '') {
				$version_file = $szMetaDir."/version";
				clearstatcache();
				if (!file_exists($version_file)) {
						$bForce = true;
				} else {
						$h = fopen($version_file, 'r');
						$meta_version = fread($h,filesize($version_file));
						fclose($h);
						if ($meta_version < $version) {
								$bForce = true;
						}
				}
		}
		
		/* simple locking in case there are several requests for the same meta
			tile at the same time - only draw it once to help with performance */
		$szLockFile = $szMetaDir."/lock_".$metaTop."_".$metaLeft;
		$fpLockFile = fopen($szLockFile, "a+");
		clearstatcache();
		if (!file_exists($szCacheFile) || $bForce || $userlayer)
		{
			flock($fpLockFile, LOCK_EX);
			fwrite($fpLockFile, ".");
			
			$cache_check = array();
			
			//check once more to see if the cache file was created while waiting for
			//the lock
			clearstatcache();
			if (!file_exists($szCacheFile) || $bForce || $userlayer)
			{
				if (!extension_loaded('MapScript'))
				{
						dl( $kaConf->szPHPMapScriptModule );
				}
				if (!extension_loaded('gd'))
				{
						dl( $kaConf->szPHPGDModule);
				}
				
				$oMap = $kaConf->discussion->Map_getMapObj($kaConf->szMap);
				
				/* Metatile width/height include 2x the metaBuffer value */
				$oMap->set('width', $kaConf->tileWidth * $kaConf->metaWidth + 2*$kaConf->metaBuffer);
				$oMap->set('height', $kaConf->tileHeight * $kaConf->metaHeight + 2*$kaConf->metaBuffer);

				/* Tell MapServer to not render labels inside the metaBuffer area
				* (new in 4.6)
				* TODO: Until MapServer bugs 1353/1355 are resolved, we need to
				* pass a negative value for "labelcache_map_edge_buffer"
				*/
				$oMap->setMetadata("labelcache_map_edge_buffer", -$kaConf->metaBuffer);

				$inchesPerUnit = array(1, 12, 63360.0, 39.3701, 39370.1, 4374754);
				$geoWidth = $scale/($oMap->resolution*$inchesPerUnit[$oMap->units]);
				$geoHeight = $scale/($oMap->resolution*$inchesPerUnit[$oMap->units]);

				/* draw the metatile */
				$minx = $metaLeft * $geoWidth;
				$maxx = $minx + $geoWidth * $oMap->width;
				$maxy = -1 * $metaTop * $geoHeight;
				$miny = $maxy - $geoHeight * $oMap->height;

				$nLayers = $oMap->numlayers;
				$oMap->setExtent($minx,$miny,$maxx,$maxy);
				$oMap->selectOutputFormat( $kaConf->szMapImageFormat );
				$aszLayers = array();
				if ($groups || $layers)
				{
						/* Draw only specified layers instead of default from mapfile*/
						if (strlen($layers)>0)
						{
								$aszLayers = explode(",", $layers);
						}

						if ($groups)
						{
								$aszGroups = explode(",", $groups);
						}

						for($i=0;$i<$nLayers;$i++)
						{
								$oLayer = $oMap->getLayer($i);
								switch (true)
								{
									//case ($aszGroups && in_array($oLayer->group,$aszGroups) && count($aszLayers)==0):
									case ($aszGroups && in_array($oLayer->group,$aszGroups) 
										&& $aszLayers && in_array($oLayer->index,$aszLayers)):
									//case ($aszGroups && $oLayer->group == '' && in_array( "__base__", $aszGroups) && count($aszLayers)==0):
										$oLayer->set("status", MS_ON );
										
										// The following is meant to handle caching for MapChat user drawing layers - so if the file exists, we're not forcing, but it's a user layer then check
										// the last modification time in the user_drawings table - if a layer_id is specified, then limit the check to just drawings in that layer...
										if (file_exists($szCacheFile) && !$bForce && $oLayer->getMetadata("mapchat_userlayer")=="true")
										{
											$db = Zend_Registry::get("db");
											if ($oLayer->getMetadata("mapchat_userlayer_id")!="")
											{
												$sCacheCheck = "select max(cache_date) as date from (select case when ud.deletion_date>ud.creation_date and ud.deletion_date>ud.modification_date then ud.deletion_date when ud.modification_date>ud.creation_date then ud.modification_date else ud.creation_date end as cache_date from ".
													$kaConf->discussion->name.".user_drawings ud join users u using (username) where layer_id = ".
													$oLayer->getMetadata("mapchat_userlayer_id")." and st_intersects(geom,'SRID=".
													$kaConf->discussion->srid.";POLYGON(($minx $miny, $minx $maxy, $maxx $maxy, $maxx $miny, $minx $miny))'::geometry)) as cache_check";
											}
											else
											{
												$sCacheCheck = "select max(cache_date) as date from (select case when ud.deletion_date>ud.creation_date and ud.deletion_date>ud.modification_date then ud.deletion_date when ud.modification_date>ud.creation_date then ud.modification_date else ud.creation_date end as cache_date from ".
													$kaConf->discussion->name.".user_drawings ud join users u using (username) where layer_id is null and st_intersects(geom,'SRID=".
													$kaConf->discussion->srid.";POLYGON(($minx $miny, $minx $maxy, $maxx $maxy, $maxx $miny, $minx $miny))'::geometry)) as cache_check";
											}
											if (!in_array($sCacheCheck,$cache_check))
											{
												$cache_check[] = $sCacheCheck;
											}
										}
										break;
										
									default:
										$oLayer->set("status", MS_OFF );
								}
						}
						//need transparency if groups or layers are used
						$oMap->outputformat->set("transparent", MS_ON );
				}
				else
				{
						$oMap->outputformat->set("transparent", MS_OFF );
				}
				
				$cache_time = 0;
				
				if (count($cache_check)>0)
				{
					$maxDate = $db->fetchOne("select max(timezone('GMT',date)) from (".implode(" union ",$cache_check).") as cache_dates;");
					if ($maxDate && strlen($maxDate)>0)
					{
						$cache_time = strtotime($maxDate);
						$dateTimeZoneGMT = new DateTimeZone("GMT");
						$dateTimeZoneLocal = new DateTimeZone(date_default_timezone_get());
						$tOff = $dateTimeZoneLocal->getOffset(new DateTime("now", $dateTimeZoneGMT));
					}
				}
				
				if ($cache_time>0) file_put_contents($szMetaDir.'/cache_time',array($tOff."\n",$cache_time."\n",(@filemtime($szCacheFile)-$tOff)."\n"));
				
				// Only create the tile if it does not exist, if it is force, or if it is a user layer and user drawings
				// are found that have a timestamp that is more recent than the last modified time of the tile.
				if (!file_exists($szCacheFile) || $bForce || (file_exists($szCacheFile) && $cache_time!==false && $cache_time>=(filemtime($szCacheFile)-$tOff)))
				{
					
					//record timestamp for this metatile
					$h = fopen($szMetaDir.'/timestamp', 'w+');
					fwrite($h, $now);
					fclose($h);
					
					//record mapfile version for this metatile
					$h = fopen($szMetaDir.'/version', 'w+');
					fwrite($h, $version);
					fclose($h);
					
					$szMetaImg = $szMetaDir."/t".$metaTop."l".$metaLeft.$kaConf->szImageExtension;
					$oImg = $oMap->draw();
					$oImg->saveImage($szMetaImg);
					$oImg->free();
					
					eval("\$oGDImg = ".$kaConf->szMapImageCreateFunction."('".$szMetaImg."');");
					if ($kaConf->bDebug)
					{
							$blue = imagecolorallocate($oGDImg, 0, 0, 255);
							imagerectangle($oGDImg, 0, 0, $kaConf->tileWidth * $kaConf->metaWidth - 1, $kaConf->tileHeight * $kaConf->metaHeight - 1, $blue );
					}
					for($i=0;$i<$kaConf->metaWidth;$i++)
					{
						for ($j=0;$j<$kaConf->metaHeight;$j++)
						{
							eval("\$oTile = ".$kaConf->szImageCreateFunction."( ".$kaConf->tileWidth.",".$kaConf->tileHeight." );");
							
							// GD's default background is black - make that transparent.  If the mapfile has transparency, then this
							// should carry over to the copied tile segments.
							if (!$useAlphaCh)
							{
								$nTransparent = imagecolorallocate($oTile, 0,0,0);
								if ($oMap->outputformat->transparent == MS_ON)
								{
									imagecolortransparent( $oTile,$nTransparent);
								}
							}
							else
							{
								ImageAlphaBlending($oTile, false);
								imagesavealpha($oTile, true);
							}
							
							$tileTop = $j*$kaConf->tileHeight + $kaConf->metaBuffer;
							$tileLeft = $i*$kaConf->tileWidth + $kaConf->metaBuffer;
							imagecopy( $oTile, $oGDImg, 0, 0, $tileLeft, $tileTop, $kaConf->tileWidth, $kaConf->tileHeight );
							/* debugging stuff */
							if ($kaConf->bDebug)
							{
									$black = imagecolorallocate($oTile, 1, 1, 1);
									$green = imagecolorallocate($oTile, 0, 128, 0 );
									$red = imagecolorallocate($oTile, 255, 0, 0);
									imagerectangle( $oTile, 1, 1, $kaConf->tileWidth-2, $kaConf->tileHeight-2, $green );
									imageline( $oTile, 0, $kaConf->tileHeight/2, $kaConf->tileWidth-1, $kaConf->tileHeight/2, $red);
									imageline( $oTile, $kaConf->tileWidth/2, 0, $kaConf->tileWidth/2, $kaConf->tileHeight-1, $red);
									imagestring ( $oTile, 3, 10, 10, ($metaLeft+$tileLeft)." x ".($metaTop+$tileTop), $black );
									imagestring ( $oTile, 3, 10, 30, ($minx+ ($kaConf->metaBuffer * $geoWidth) +
									($i * $kaConf->tileWidth *$geoWidth)) ." x ". (  $maxy - (  ($kaConf->metaBuffer *
									$geoWidth) + ($j * $kaConf->tileHeight * $geoHeight) )  ), $black );
							}
							
							$szTileImg = $szCacheDir."/t".($metaTop+$tileTop)."l".($metaLeft+$tileLeft).$kaConf->szImageExtension;
							eval("{$kaConf->szImageOutputFunction}( \$oTile, '".$szTileImg."' );");
							imagedestroy($oTile);
							$oTile = null;
						}
					}
					if ($oGDImg != null)
					{
							imagedestroy($oGDImg);
							$oGDImg = null;
					}
					if (!$kaConf->bDebug)
					{
							//unlink( $szMetaImg );
					}
				}
			}
			//release the exclusive lock
			flock($fpLockFile, LOCK_UN );
		}
		
		//flock($fpLockFile, LOCK_SH);
		
		//release lock
		fclose($fpLockFile);
		
		// check client cache using Minify's HTTP_ConditionalGet class
		$cgOptions = array(
				'lastModifiedTime' => filemtime($szCacheFile),
				'isPublic' => true,
		);
		require_once $kaConf->config->BaseDir.'/library/Minify/HTTP/ConditionalGet.php';
		$cg = new HTTP_ConditionalGet($cgOptions);
		
		if ($cg->cacheIsValid) {
			$cg->sendHeaders();
			
			// Not sure what, but something is sending default values for 
			// these two headers, so I'll blank them out here:
			header("Expires: "); // something is defaulting to 'Expires: Thu, 19 Nov 1981 08:52:00 GMT'
			header("Cache-Control: "); // defaults to 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0'
			
			exit();
		}
		// client will need output
		$headers = $cg->getHeaders();
		unset($cg);

		// add headers to those from ConditionalGet
		$headers["Expires"] = ""; // Same as above - prevent the browser from thinking the image is nearly 30 years old.
		$headers['Content-Type'] = $kaConf->szImageHeader;
		$headers['Content-Length'] = filesize($szCacheFile);
		
		// output headers & content
		foreach ($headers as $name => $val) {
				header($name . ': ' . $val);
		}
		
		$h = fopen($szCacheFile, "r");
		fpassthru($h);
		fclose($h);
		
		/* bug 1253 - root permissions required to delete cached files */
		umask($orig_umask);
	}
}

?>