<?php

class MapChat_Setup
{
	public function runSetup($params)
	{
		$dbconf = Zend_Registry::get("config")->database->toArray();
		
		$dbconf["dbname"] = $params["dbname"];
		$dbconf["username"] = $params["dbuser"];
		$dbconf["password"] = $params["dbpass"];
		
		// Try to connect to the maintenance DB:
		$setupdb = Zend_Db::factory('PDO_PgSQL',$dbconf);
		$sdc = $setupdb->getConnection();
		
		// Make sure we can find the PostGIS files...
		if (!((file_exists($params["postgis_path"]."/lwpostgis.sql")||file_exists($params["postgis_path"]."/postgis.sql")) &&
			file_exists($params["postgis_path"]."/spatial_ref_sys.sql")))
		{
			throw new MapChat_Exception("Problem locating lwpostgis.sql and spatial_ref_sys.sql in '".$params["postgis_path"]."'");
		}
		
		// Create the MapChat database:
		$sdc->exec("create database ".
			Zend_Registry::get("config")->database->dbname.
			" owner ".Zend_Registry::get("config")->database->username.
			" template template0 encoding '".Zend_Registry::get("config")->db_encoding."';"
		);
		
		// Let go of the maintenance DB - we don't want to mess with that too much.
		$setupdb->closeConnection();
		
		// Now substitute the name of the database that we just created, then we can run the setup in there...
		$dbconf["dbname"] = Zend_Registry::get("config")->database->dbname;
		
		// Now connect to the MapChat database:
		$db = Zend_Db::factory('PDO_PgSQL',$dbconf);
		$dbc = $db->getConnection();
		
		// Create the PLPGSQL language:
		$dbc->exec("create language plpgsql;");
		
		// Load PostgreSQL necessary contrib files (e.g., GiST used for PostGIS spatial indexing):
		$dbc->exec(MapChat_Setup::cleanPGContrib($params["contrib_path"]."/btree_gist.sql"));
		$dbc->exec(MapChat_Setup::cleanPGContrib($params["contrib_path"]."/tsearch2.sql"));
		
		// Load PostGIS:
		if (file_exists($params["postgis_path"]."/lwpostgis.sql")) // Load PostGIS < 1.4.x
			$dbc->exec(MapChat_Setup::cleanPGContrib($params["postgis_path"]."/lwpostgis.sql"));
		else // Load PostGIS >= 1.4.x
			$dbc->exec(MapChat_Setup::cleanPGContrib($params["postgis_path"]."/postgis.sql"));
		
		$dbc->exec(MapChat_Setup::cleanPGContrib($params["postgis_path"]."/spatial_ref_sys.sql"));
		
		// Give the MapChat db user ownership over the geometry_columns and spatial_ref_sys tables...
		$dbc->exec("ALTER TABLE geometry_columns OWNER to ".Zend_Registry::get("config")->database->username.";");
		try
		{
			$dbc->exec("ALTER TABLE geography_columns OWNER to ".Zend_Registry::get("config")->database->username.";");
		}
		catch (Exception $e)
		{
			// Do nothing - this is just to avoid problems with PostGIS versions prior to 1.5.x, which don't have the geography_columns table.
		}
		$dbc->exec("ALTER TABLE spatial_ref_sys OWNER to ".Zend_Registry::get("config")->database->username.";");
		
		// Load PostgreSQL contrib files:
		$setupdir = Zend_Registry::get("config")->BaseDir."/setup/db/";
		$load_order = require($setupdir."load_order.php");
		foreach ($load_order as $file)
		{
			if (file_exists($setupdir.$file))
			{
				try
				{
					$dbc->exec(MapChat_Setup::substituteParam("dbuser",Zend_Registry::get("config")->database->username,MapChat_Setup::cleanPGContrib($setupdir.$file)));
				}
				catch (Exception $e)
				{
					throw new MapChat_Exception("Error loading '".$setupdir.$file."': ".$e->getMessage());
				}
			}
			else
			{
				throw new MapChat_Exception("Error loading '".$setupdir.$file."'.");
			}
		}
		
		// Now add the default user/group records:
		$adminuser = array(
			"username" => "admin",
			"password" => new Zend_Db_Expr("md5('".$params["adminpass"]."')"),
			"full_name" => "Administrator",
			"email" => $params["email"],
			"organization" => "MapChat Admin",
			"confirmed" => true,
		);
		$db->insert('users', $adminuser);
		
		$admingroup = array(
			"groupname" => "admin",
			"password" => new Zend_Db_Expr("md5('".$params["admingpass"]."')"),
			"description" => "MapChat Administrators",
			"moderator" => 'admin',
			"is_admin" => true,
		);
		$db->insert('groups', $admingroup);
		
		$umodgroup = array(
			"groupname" => "umod",
			"password" => new Zend_Db_Expr("md5('".$params["umodgpass"]."')"),
			"description" => "MapChat User Moderators",
			"moderator" => 'admin',
			"can_create_users" => true,
			"can_manage_users" => true,
		);
		$db->insert('groups', $umodgroup);
		
		$gmodgroup = array(
			"groupname" => "gmod",
			"password" => new Zend_Db_Expr("md5('".$params["gmodgpass"]."')"),
			"description" => "MapChat Group Moderators",
			"moderator" => 'admin',
			"can_create_groups" => true,
			"can_manage_groups" => true,
		);
		$db->insert('groups', $gmodgroup);
		
		$dmodgroup = array(
			"groupname" => "dmod",
			"password" => new Zend_Db_Expr("md5('".$params["dmodgpass"]."')"),
			"description" => "MapChat Discussion Moderators",
			"moderator" => 'admin',
			"can_create_discussions" => true,
			"can_manage_discussions" => true,
		);
		$db->insert('groups', $dmodgroup);
		
		// This is the group for 'everyone' - password is irrelevant.
		$egroup = array(
			"groupname" => "everyone",
			"password" => new Zend_Db_Expr("md5('')"),
			"description" => "Everyone",
			"moderator" => 'admin',
		);
		$db->insert('groups', $egroup);
		
		$db->insert("group_user",array("username" => "admin", "groupname" => "admin"));
		$db->insert("group_user",array("username" => "admin", "groupname" => "everyone"));
		
		$dbc->exec("vacuum full analyze;");
	}
	
	public function cleanPGContrib($file)
	{
		if (file_exists($file))
		{
			// No vacuuming is allowed in this context - this assumes a 'vaccum' statement will
			// be on a separate line (i.e., the first word on the line is 'vacuum'):
			$aFile = file($file);
			for ($i=0; $i<count($aFile); $i++)
			{
				if (strpos(strtoupper(trim($aFile[$i])),"VACUUM")===0) $aFile[$i]="";
			}
			return implode("\n",$aFile);
		}
		else
		{
			throw new MapChat_Exception("Error cleaning PG Contrib file '".$file."'");
		}
	}
	
	// Substitute '?dbuser?' in a file with the default dbuser in the library config.
	public function substituteParam($paramName,$paramValue,$text)
	{
			$aText = explode("?".$paramName."?",$text);
			return implode($paramValue,$aText);
	}
	
	public function checkDB()
	{
		try
		{
			// Don't let anyone mess around with this utility if the database actually exists:
			$db = Zend_Db::factory('PDO_PgSQL',Zend_Registry::get("config")->database->toArray());
			$dbc = $db->getConnection();
		}
		catch (Exception $e)
		{
			// This is ok - a successful connection shouldn't be possible if the database
			// doesn't exist.
			
			// TODO: perhaps could do some detailed checking of the error message
			// to ensure that it is in fact one that says the database doesn't exist.
			return true;
		}
		
		// If we're still here, the connection was successful...that's bad, since we
		// don't want to be running setup on an existing database.
		throw new MapChat_Exception("The MapChat database already exists.");
		return false;
	}
}

?>
