-- This is meant to be called by the MapChat application, where all instances of '?'
-- will be substituted with the discussion name.  Could possibly divide this up like
-- the rest of the database setup.

-- Create the discussion schema:
create schema ?;

-- User status table (for keeping track of who's currently online):
create table ?.user_status (username varchar primary key references users on delete cascade on update cascade, online_time timestamp with time zone not null default current_timestamp, offline_time timestamp with time zone, last_poll timestamp with time zone);

-- Discussion event log - a generic table that all events in the discussion will be recorded in:
create table ?.event_log (id serial primary key, parent_id integer references ?.event_log, username varchar references users on delete set default on update cascade not null, type integer references event_types, text text, data text[], time timestamp with time zone not null default current_timestamp, flags varchar, depth integer not null default 0, deleted boolean not null default false);

-- Add an extent column for storing the current map extent when the event took place:
select addgeometrycolumn('?','event_log','extent',(select srid from discussions where discussion = '?' limit 1),'POLYGON',2);
CREATE INDEX event_log_extent_idx ON ?.event_log USING GIST ( extent );
CREATE INDEX event_log_parent_id_idx ON ?.event_log ( parent_id );
CREATE INDEX event_log_username_idx ON ?.event_log ( username );
CREATE INDEX event_log_text_idx ON ?.event_log using gin( to_tsvector('english',text) );
CREATE INDEX event_log_type_idx ON ?.event_log ( type );
CREATE INDEX event_log_time_idx ON ?.event_log ( time );
CREATE INDEX event_log_flags_idx ON ?.event_log using gin( (flags::tsvector) );
CREATE INDEX event_log_depth_idx ON ?.event_log ( depth );
CREATE INDEX event_log_deleted_idx ON ?.event_log ( deleted );

-- Build indexes for the first 10 data elements, just in case it's useful
CREATE INDEX event_log_data1_idx ON ?.event_log ((data[1]));
CREATE INDEX event_log_data2_idx ON ?.event_log ((data[2]));
CREATE INDEX event_log_data3_idx ON ?.event_log ((data[3]));
CREATE INDEX event_log_data4_idx ON ?.event_log ((data[4]));
CREATE INDEX event_log_data5_idx ON ?.event_log ((data[5]));
CREATE INDEX event_log_data6_idx ON ?.event_log ((data[6]));
CREATE INDEX event_log_data7_idx ON ?.event_log ((data[7]));
CREATE INDEX event_log_data8_idx ON ?.event_log ((data[8]));
CREATE INDEX event_log_data9_idx ON ?.event_log ((data[9]));
CREATE INDEX event_log_data10_idx ON ?.event_log ((data[10]));

-- Add column for storing arbitrary geometry data (maybe this isn't needed):
select addgeometrycolumn('?','event_log','geom',(select srid from discussions where discussion = '?' limit 1),'GEOMETRY',2);
CREATE INDEX event_log_geom_idx ON ?.event_log USING GIST ( geom ); 

-- Links between events - allows for multiple parent/child relationships:
create table ?.event_links (parent_id integer not null references ?.event_log, child_id integer not null references ?.event_log, primary key (parent_id,child_id), constraint ?_event_links_check_notself check (not parent_id = child_id));
CREATE INDEX event_links_child_id_idx ON ?.event_links ( child_id );
CREATE INDEX event_links_parent_id_idx ON ?.event_links ( parent_id );
CREATE INDEX event_links_child_id_parent_id_idx ON ?.event_links ( child_id,parent_id );

-- Target groups for a specific event (e.g., for interacting with a specific subset of users):
create table ?.event_target_groups (id integer references ?.event_log on delete cascade on update cascade, groupname varchar not null references groups on delete cascade on update cascade, primary key (id,groupname));

-- Target users for a specific event (e.g., for interacting with a specific subset of users):
create table ?.event_target_users (id integer references ?.event_log on delete cascade on update cascade, username varchar not null references users on delete set default on update cascade not null, primary key (id,username));

-- Would be ideal to allow the chat_log to inherit from event_log, but that won't maintain 
-- primary key structures very well. So chat and event records will be maintained separately.
create table ?.chat_log (id serial primary key, parent_id integer references ?.chat_log, username varchar references users on delete set default on update cascade not null, type integer references chat_types, text text not null, time timestamp with time zone not null default current_timestamp, flags varchar, depth integer not null default 0, deleted boolean not null default false);

-- Add an extent column for storing the current map extent when the chat message was recorded place:
select addgeometrycolumn('?','chat_log','extent',(select srid from discussions where discussion = '?' limit 1),'POLYGON',2);
CREATE INDEX chat_log_extent_idx ON ?.chat_log USING GIST ( extent );
CREATE INDEX chat_log_parent_id_idx ON ?.chat_log ( parent_id );
CREATE INDEX chat_log_username_idx ON ?.chat_log ( username );
CREATE INDEX chat_log_text_idx ON ?.chat_log using gin( to_tsvector('english',text) );
CREATE INDEX chat_log_type_idx ON ?.chat_log ( type );
CREATE INDEX chat_log_time_idx ON ?.chat_log ( time );
CREATE INDEX chat_log_flags_idx ON ?.chat_log USING gin( (flags::tsvector) );
CREATE INDEX chat_log_depth_idx ON ?.chat_log ( depth );
CREATE INDEX chat_log_deleted_idx ON ?.chat_log ( deleted );

-- Target groups for a specific chat message (e.g., for sending messages to a specific subset of users):
create table ?.chat_target_groups (id integer references ?.chat_log on delete cascade on update cascade, groupname varchar not null references groups on delete cascade on update cascade, primary key (id,groupname));

-- Target users for a specific chat message (e.g., for sending messages to a specific subset of users):
create table ?.chat_target_users (id integer references ?.chat_log on delete cascade on update cascade, username varchar not null references users on delete set default on update cascade not null, primary key (id,username));

-- Links between chat messages - allows for multiple parent/child relationships:
create table ?.chat_links (parent_id integer not null references ?.chat_log, child_id integer not null references ?.chat_log, primary key (parent_id,child_id), constraint ?_chat_links_check_notself check (not parent_id = child_id));
CREATE INDEX chat_links_child_id_idx ON ?.chat_links ( child_id );
CREATE INDEX chat_links_parent_id_idx ON ?.chat_links ( parent_id );
CREATE INDEX chat_links_child_id_parent_id_idx ON ?.chat_links ( child_id,parent_id );

-- Table listing user-drawn layers, categorized by username and layer name.  
-- Attributes for a given layer are described by the attributes column in this
-- table, which should list field names and their data types in a 2-dimensional
-- array.  The field shape_type determins if the layer is exclusively POINT (1),
-- LINE (2), POLYGON (3), or GEOMETRY (4) - in the case of GEOMETRY, the field
-- shape_attribute will be used to break the layer down into shape types (assuming
-- each the shape_type is the same for all features in the layer that have with the 
-- same value in the attribute that corresponds to shape_attribute)
CREATE TABLE ?.user_layers (id serial primary key, username varchar references users on delete set default on update cascade, name varchar not null, description varchar, attributes varchar[][], shape_type integer not null references shape_types, shape_attribute varchar, creation_date timestamp with time zone not null default current_timestamp, deleted boolean not null default false, deletion_date timestamp with time zone);

-- Build indexes for user_layers
CREATE INDEX user_layers_deleted_idx ON ?.user_layers ( deleted );
CREATE INDEX user_layers_id_deleted_idx ON ?.user_layers ( id,deleted );

-- Constraint ensures that shape_attribute is an actual attribute for this layer.
ALTER TABLE ?.user_layers ADD CONSTRAINT ?_user_layers_shape_attribute_exists CHECK (shape_attribute is null or (mapchat_in_2d_array(shape_attribute,attributes))[1] = 1);

-- Table containing user-drawn shapes - must correspond to the metadatata
-- stored in user_layers - attribute values are stored in an array, which
-- must correspond to the order of attribute names/types listed in the 
-- user_layers table.
CREATE TABLE ?.user_drawings (id serial primary key, layer_id integer references ?.user_layers, username varchar references users on update cascade not null, attributes text[], creation_date timestamp with time zone not null default current_timestamp, visible boolean not null default false, deleted boolean not null default false, deletion_date timestamp with time zone, modification_date timestamp with time zone, visibility_change timestamp with time zone);

-- Add a geometry column for storing the user-drawn geometries:
select addgeometrycolumn('?','user_drawings','geom',(select srid from discussions where discussion = '?' limit 1),'GEOMETRY',2);

-- Build indexes for user_drawings:
CREATE INDEX user_drawings_geom_idx ON ?.user_drawings USING GIST ( geom );
CREATE INDEX user_drawings_geom_geomtype_idx ON ?.user_drawings USING GIST ( geom, st_geometrytype(geom) );
CREATE INDEX user_drawings_geomtype_idx ON ?.user_drawings ( st_geometrytype(geom) );
CREATE INDEX user_drawings_deleted_idx ON ?.user_drawings ( deleted );
CREATE INDEX user_drawings_id_deleted_idx ON ?.user_drawings ( id,deleted );
CREATE INDEX user_drawings_username_deleted_idx ON ?.user_drawings ( username,deleted );
CREATE INDEX user_drawings_visible_idx ON ?.user_drawings ( visible );
CREATE INDEX user_drawings_id_visible_idx ON ?.user_drawings ( id,visible );
CREATE INDEX user_drawings_username_visible_idx ON ?.user_drawings ( username,visible );
CREATE INDEX user_drawings_visible_deleted_idx ON ?.user_drawings ( visible,deleted );
CREATE INDEX user_drawings_id_visible_deleted_idx ON ?.user_drawings ( id,visible,deleted );
CREATE INDEX user_drawings_username_visible_deleted_idx ON ?.user_drawings ( username,visible,deleted );
CREATE INDEX user_drawings_creation_date_idx on ?.user_drawings ( creation_date );
CREATE INDEX user_drawings_deletion_date_idx on ?.user_drawings ( deletion_date );
CREATE INDEX user_drawings_modification_date_idx on ?.user_drawings ( modification_date );
CREATE INDEX user_drawings_visibility_change_idx on ?.user_drawings ( visibility_change );

-- Build indexes for the first 10 attributes - we could do more, but is that necessary?
CREATE INDEX user_drawings_attributes1_idx ON ?.user_drawings ((attributes[1]));
CREATE INDEX user_drawings_attributes2_idx ON ?.user_drawings ((attributes[2]));
CREATE INDEX user_drawings_attributes3_idx ON ?.user_drawings ((attributes[3]));
CREATE INDEX user_drawings_attributes4_idx ON ?.user_drawings ((attributes[4]));
CREATE INDEX user_drawings_attributes5_idx ON ?.user_drawings ((attributes[5]));
CREATE INDEX user_drawings_attributes6_idx ON ?.user_drawings ((attributes[6]));
CREATE INDEX user_drawings_attributes7_idx ON ?.user_drawings ((attributes[7]));
CREATE INDEX user_drawings_attributes8_idx ON ?.user_drawings ((attributes[8]));
CREATE INDEX user_drawings_attributes9_idx ON ?.user_drawings ((attributes[9]));
CREATE INDEX user_drawings_attributes10_idx ON ?.user_drawings ((attributes[10]));

-- The user_maps table will store text representations of mapfiles - representing
-- the current state of a given user's map:
CREATE TABLE ?.user_maps (id serial primary key, username varchar not null references users, mapfile text, time timestamp with time zone not null default current_timestamp);

-- Build indexes for user_maps:
CREATE INDEX user_maps_username_time_idx ON ?.user_maps ( username,time );


-- Table containing plan-related information, which generally consists
-- Of a series of layers.
CREATE TABLE ?.user_plans (id serial primary key, type integer references plan_types, username varchar references users on delete set default on update cascade not null, title varchar, creation_date timestamp with time zone not null default current_timestamp, deleted boolean not null default false, deletion_date timestamp with time zone);

-- Prohibit more than one non-deleted plan under the same title for the same user:
CREATE UNIQUE INDEX user_plans_title_username_notdeleted_uq ON ?.user_plans (title,username) WHERE not deleted; 

CREATE INDEX user_plans_id_idx ON ?.user_plans ( id );
CREATE INDEX user_plans_username_idx ON ?.user_plans ( username );
CREATE INDEX user_plans_id_username_idx ON ?.user_plans ( id,username );
CREATE INDEX user_plans_id_username_type_idx ON ?.user_plans ( id,username,type );
CREATE INDEX user_plans_type_idx ON ?.user_plans ( type );
CREATE INDEX user_plans_id_type_idx ON ?.user_plans ( id,type );
CREATE INDEX user_plans_username_type_idx ON ?.user_plans ( username,type );

-- Link plans to layers - layer_id is unique here, because a single layer may not belong to more than one plan.
CREATE TABLE ?.user_plans_layers (plan_id integer not null references ?.user_plans, layer_id integer not null references ?.user_layers, unique (layer_id));
CREATE INDEX user_plans_layers_plan_id_layer_id_idx ON ?.user_plans_layers ( plan_id,layer_id );
CREATE INDEX user_plans_layers_plan_id_idx ON ?.user_plans_layers ( plan_id );
CREATE INDEX user_plans_layers_layer_id_idx ON ?.user_plans_layers ( layer_id );


-- This will be the table that stores user selections, which ultimately may be used for the chat_spatial_links.  User drawings that
-- are linked to chat messages (but are not part of a map-layer) will also be copied here - to be indicated by the 'user_drawing' field.
create table ?.user_selections (id serial primary key, username varchar references users on delete set default on update cascade not null, mapfile varchar, layer_name varchar not null, user_layer boolean not null default false, user_drawing boolean not null default false, feature_id integer not null, attr_json text not null, creation_date timestamp with time zone not null default current_timestamp, deleted boolean default false, deletion_date timestamp with time zone);

-- Add a geometry column for storing the geometry of the selected features:
select addgeometrycolumn('?','user_selections','geom',(select srid from discussions where discussion = '?' limit 1),'GEOMETRY',2);

-- Build indexes for user_selections:
CREATE INDEX user_selections_id_idx ON ?.user_selections ( id );
CREATE INDEX user_selections_username_idx ON ?.user_selections ( username );
CREATE INDEX user_selections_mapfile_idx ON ?.user_selections (mapfile);
CREATE INDEX user_selections_geom_idx ON ?.user_selections USING GIST ( geom );
CREATE INDEX user_selections_geom_geomtype_idx ON ?.user_selections USING GIST ( geom, st_geometrytype(geom) );
CREATE INDEX user_selections_geomtype_idx ON ?.user_selections ( st_geometrytype(geom) );
CREATE INDEX user_selections_feature_id_idx ON ?.user_selections ( feature_id );
CREATE INDEX user_selections_deleted_idx ON ?.user_selections ( deleted );
CREATE INDEX user_selections_feature_id_deleted_idx ON ?.user_selections ( feature_id,deleted );
CREATE INDEX user_selections_attr_json_idx ON ?.user_selections (attr_json);
CREATE INDEX user_selections_attr_json_ts_idx ON ?.user_selections using gin( to_tsvector('simple',attr_json) );
CREATE UNIQUE INDEX user_selections_username_feature_id_layer_name_deleted_uq ON ?.user_selections (username,feature_id,layer_name,deleted) WHERE not deleted;

-- Changing the way spatial links are done from the original MapChat version.  Selections are stored in the user_selections table,
-- and linking between chat messages and selected features will be stored in this table.
create table ?.chat_selection_links (id serial primary key, chat_id integer references ?.chat_log on delete cascade not null, selection_id integer references ?.user_selections on delete cascade not null, link_time timestamp with time zone not null default current_timestamp, deleted boolean default false, deletion_date timestamp with time zone);

-- Build indexes for chat_selection_links:
CREATE INDEX chat_selection_links_chat_id_idx ON ?.chat_selection_links ( chat_id );
CREATE INDEX chat_selection_links_selection_id_idx ON ?.chat_selection_links ( selection_id );
CREATE INDEX chat_selection_links_deleted_idx ON ?.chat_selection_links ( deleted );
CREATE INDEX chat_selection_links_chat_id_deleted_idx ON ?.chat_selection_links ( chat_id,deleted );
CREATE INDEX chat_selection_links_selection_id_deleted_idx ON ?.chat_selection_links ( selection_id,deleted );
CREATE INDEX chat_selection_links_chat_id_selection_id_deleted_idx ON ?.chat_selection_links ( chat_id,selection_id,deleted );


