--  Returns the index of the first match of the element (first argument) in the array (second argument)
--   Note: only works for one-dimensional arrays:
create or replace function mapchat_in_array(anyelement,anyarray) returns integer as
$$
	declare
		cur_idx integer;
		min_idx integer;
		max_idx integer;
		elem alias for $1;
		arr alias for $2;
	begin
		select into cur_idx array_lower(arr,1);
		select into min_idx array_lower(arr,1);
		select into max_idx array_upper(arr,1);
		while cur_idx>=min_idx and cur_idx<=max_idx and not arr[cur_idx] = elem
		loop
			cur_idx := cur_idx + 1;
		end loop;
		
		if cur_idx>=min_idx and cur_idx<=max_idx and arr[cur_idx] = elem then
			return cur_idx;
		else
			return null;
		end if;
	end;
$$ 
language 'plpgsql';

--  Returns the 2d index of the first match of the element (first argument) in the 2-dmensional
--   array (second argument)
create or replace function mapchat_in_2d_array(anyelement,anyarray) returns integer[] as
$$
	declare
		x_idx integer;
		y_idx integer;
		minx_idx integer;
		maxx_idx integer;
		miny_idx integer;
		maxy_idx integer;
		e alias for $1;
		a alias for $2;
	begin
		select into x_idx array_lower(a,1);
		select into y_idx array_lower(a,2);
		select into minx_idx array_lower(a,1);
		select into maxx_idx array_upper(a,1);
		select into miny_idx array_lower(a,2);
		select into maxy_idx array_upper(a,2);
		while x_idx>=minx_idx and x_idx<=maxx_idx
		loop
			while y_idx>=miny_idx and y_idx<=maxy_idx
			loop
				if e = a[x_idx][y_idx] then
					return array[x_idx,y_idx];
				end if;
				y_idx := y_idx +1;
			end loop;
			x_idx := x_idx + 1;
			y_idx := miny_idx;
		end loop;
		
		return null;
	end;
$$ 
language 'plpgsql';


-- Types and a procedure for parsing user_layer attributes:
create type mapchat_user_layer_attribute as (name varchar, type varchar, index integer);
create type mapchat_max_idx as (max integer);
create or replace function mapchat_user_layer_attributes(integer,varchar) returns setof mapchat_user_layer_attribute as
$$
	declare
		cur_idx integer;
		midx mapchat_max_idx;
		cur_attr mapchat_user_layer_attribute;
		layer_id alias for $1;
		discussion alias for $2;
	begin
		cur_idx := 1;
		execute 'select array_upper(attributes,2) as max from '||discussion||'.user_layers where id = '||layer_id||' limit 1' into midx;
		
		while cur_idx<=midx.max
		loop
			execute 'select attributes[1]['||cur_idx||'] as name, attributes[2]['||cur_idx||'] as type, '||cur_idx||' as index from '||discussion||'.user_layers where id = '||layer_id||' limit 1' into cur_attr;
			return next cur_attr;
			cur_idx := cur_idx + 1;
		end loop;
		return;
	end;
$$ 
language 'plpgsql';


-- Appends a polygon onto another - if the second polygon is contained by the first, it will be treated as a
-- hole, otherwise, it is unioned with the first geometry.
create or replace function mapchat_append_polygon(geometry,geometry) returns geometry as
$$
	declare
		poly_a alias for $1;
		poly_b alias for $2;
	begin
		if (st_geometrytype(poly_a) in ('ST_Polygon','ST_MultiPolygon') and st_geometrytype(poly_b) in ('ST_Polygon','ST_MultiPolygon') and st_contains(poly_a,poly_b) and not st_equals(poly_a,poly_b)) then
			return st_difference(poly_a,poly_b);
		else
			return st_union(poly_a,poly_b);
		end if;
	end;
$$
language 'plpgsql';

-- A shortcut for doing the above with a single argument being an array of geometries.
create or replace function mapchat_append_polygon(geometry[]) returns geometry as
$$
	declare
		polys alias for $1;
		cur_poly geometry;
		idx integer;
	begin
		if array_upper(polys,1)<=1 then
			return polys[1];
		else
			cur_poly = polys[1];
			idx := 2;
			while length(polys[idx]::text)>0
			loop
				cur_poly = mapchat_append_polygon(cur_poly,polys[idx]);
				idx := idx + 1;
			end loop;
			return cur_poly;
		end if;
	end;
$$
language 'plpgsql';

-- Unions many geometries - saves having to do it before inputting a bunch of geometries into a statement.
create or replace function mapchat_union_geometries(geometry[]) returns geometry as
$$
	declare
		geoms alias for $1;
		cur_geom geometry;
		idx integer;
	begin
		if array_upper(geoms,1)<=1 then
			return geoms[1];
		else
			cur_geom = geoms[1];
			idx := 2;
			while length(geoms[idx]::text)>0
			loop
				cur_geom = st_union(cur_geom,geoms[idx]);
				idx := idx + 1;
			end loop;
			return cur_geom;
		end if;
	end;
$$
language 'plpgsql';

-- Returns the depth of a given message in the chat_log table, calculated by its heirarchical position.
-- 	@param integer msg_id - the id of the message to calculate the depth for.
--  @param varchar discussion - the name of the discussion (i.e., the schema) that the message is located in.
create or replace function mapchat_thread_depth(integer,varchar) returns integer as
$$
	declare
		chat_id alias for $1;
		discussion alias for $2;
		thread_depth integer;
		parent_depth integer;
		parent_id integer;
		count_check integer;
	begin
		thread_depth := 0;
		execute 'select count(*) from '||discussion||'.chat_log where id = '||chat_id||';' into count_check;
		if count_check = 0 then
			return null;
		end if;
		execute 'select parent_id from '||discussion||'.chat_log where id = '||chat_id||';' into parent_id;
		if not parent_id is null then
			execute 'select mapchat_thread_depth('||parent_id::varchar||','''||discussion||''');' into parent_depth;
			thread_depth = parent_depth + 1;
		end if;
		return thread_depth;
	end;
$$
language 'plpgsql';

-- Returns true if the chat message ID appears in either chat_target_users or chat_target_groups
create or replace function mapchat_chat_private(integer,varchar) returns boolean as
$$
	declare
		chat_id alias for $1;
		discussion alias for $2;
		isprivate boolean;
	begin
		execute 'select count(*)>0 from '||discussion||'.chat_target_users where id = '||chat_id||';' into isprivate;
		if not isprivate then
			execute 'select count(*)>0 from '||discussion||'.chat_target_groups where id = '||chat_id||';' into isprivate;
		end if;
		return isprivate;
	end;
$$
language 'plpgsql';

-- Concatenates two strings - intended to work with the aggregate below (not sure how robust this 
-- will be with many special and/or escaped characters):
create or replace function mapchat_concat_list (text, text) RETURNS text AS
$$
	DECLARE
		t text;
	BEGIN
		IF  character_length($1) > 0 THEN
			t = $1||','''||replace(replace(replace($2,E'\n',E'\\n'),E'\\',E'\\\\'),'''',E'\\''')||'''';
		ELSE
			t = ''''||replace(replace(replace($2,E'\n',E'\\n'),E'\\',E'\\\\'),'''',E'\\''')||'''';
		END IF;
		RETURN t;
	END;
$$ LANGUAGE plpgsql;

-- Aggregates multiple strings into one long string, with apostrophes escaped and single 
-- quotes around each individual string, and separated by commas
create aggregate mapchat_concat_list (
	sfunc = mapchat_concat_list,
	basetype = text,
	stype = text,
	initcond = ''
);

-- Concatenates two text arrays - intended to work with the aggregate below:
create or replace function mapchat_concat_array (text[], text[]) RETURNS text[] AS
$$
	BEGIN
		RETURN $1||$2;
	END;
$$ LANGUAGE plpgsql;

-- Aggregates multiple strings into one long string, with apostrophes escaped and single 
-- quotes around each individual string, and separated by commas
create aggregate mapchat_concat_array (
	sfunc = mapchat_concat_array,
	basetype = text[],
	stype = text[],
	initcond = '{}'
);

-- Converts a text array into a JSON-compatible string:
create or replace function mapchat_concat_json (text[]) RETURNS text AS
$$
	DECLARE
		arr alias for $1;
		result text;
		glue varchar;
		i integer;
	BEGIN
		glue = '';
		result = '["';
		FOR i IN 1..array_upper(arr,1) LOOP
			result = result || glue || replace(arr[i],'"',E'\\"');
			glue = '","';
		END LOOP;
		RETURN result || '"]';
	END;
$$ LANGUAGE plpgsql;

-- Converts two arrays of target_users and target_groups into a JSON-compatible string,
-- which should be usable for looking up log entries that relate to private messages.
create or replace function mapchat_targets_json (text[],text[]) RETURNS text AS
$$
	DECLARE
		tu alias for $1;
		tg alias for $2;
		glue varchar;
		tustr text;
		tgstr text;
	BEGIN
		glue = '';
	
		IF NOT tu IS NULL THEN
			tustr = '"target_users":'||mapchat_concat_json(tu);
		ELSE
			tustr = '';
		END IF;
		
		IF NOT tg IS NULL THEN
			tgstr = '"target_users":'||mapchat_concat_json(tg);
			IF length(tustr) > 0 THEN
				glue = ',';
			END IF;
		ELSE
			tgstr = '';
		END IF;
		
		RETURN '{'||tustr||glue||tgstr||'}';
		
	END;
$$ LANGUAGE plpgsql;

